-- --8<--8<--8<--8<--
--
-- Copyright (C) 2007 Smithsonian Astrophysical Observatory
--
-- This file is part of rdb_lua
--
-- rdb_lua is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- rdb_lua is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the 
--       Free Software Foundation, Inc. 
--       51 Franklin Street, Fifth Floor
--       Boston, MA  02110-1301, USA
--
-- -->8-->8-->8-->8--



$ifnot RDB

-- base object
RDB = {
	Comments = {},
	Cols     = {},
	Defs     = {},
	KeyDefs  = {},
	Pos      = {},
	Read     = {},
	File     = {},
	opened   = nil,
	handle   = nil,
	line     = nil,
	Vars	 = {},
      }

-- open up an RDB file 

function RDB.split( line )
  local start, finish
  local field, fields, tab

  fields = {}
  fields.n = 0
  start = 1

  local len = strlen( line )
  while ( start ) do
    start, finish, field = strfind( line, "([^\t]*)", start) 

    fields.n = fields.n + 1
    fields[fields.n] = field

    if ( finish == len ) then
      start = nil
    else
      start = finish + 2
    end

  end

  return fields
end

function RDB:open( ... )
  
  if ( arg.n > 0 ) then

    self.File = arg[1]

    local oINPUT = _INPUT
    self.handle, errstr = readfrom( arg[1] )
    _INPUT = oINPUT

    if ( nil == self.handle ) then
      return nil, errstr .. ': ' .. self.File
    end

  else

     self.handle = _INPUT
     self.File = 'unknown file'

  end

  -- line counter
  self.line = 0

  -- now, read in header comments
  self.Comments = {}
  self.Comments.n = 0

  -- preload buffer
  local buf = read( self.handle )
  if ( nil ~= buf ) then
    self.line = self.line + 1
  end

  while buf and strfind( buf, "^%s*#" )  do
    self.Comments.n = self.Comments.n + 1
    self.Comments[self.Comments.n] = gsub( buf, '^%s*#(.*)', '%1' )

    local str_begin, str_end, var, value
    str_begin, str_end, var, value = strfind( buf, '^%s*#:%s*(%w*)%s*=%s*(.*)' )

    if ( str_begin ) then
      self.Vars[var] = value
    end

    -- reload buffer
    buf = read( self.handle )
    if ( nil ~= buf ) then
      self.line = self.line + 1
    end

  end

  -- check for EOF
  if nil == buf then
    return nil, 'unexpected EOF'
  end

  -- read in header
  self.Cols = RDB.split( buf )
  self.n = self.Cols.n

  -- read in column definitions
  buf = read( self.handle )
  if nil == buf then
    return nil, 'unexpected EOF'
  end
  self.line = self.line + 1

  self.Defs = RDB.split( buf )
  if self.Defs.n ~= self.Cols.n then
    return nil, self.File .. ': line ' .. self.line .. ": number of columns and number of defs don't match"
  end

  -- create table of positions and definitions keyed off of column names
  local tmp1 = {}
  local tmp2 = {}
  local n = 1
  while ( n <= self.n ) do
    tmp1[self.Cols[n]] = n 
    tmp2[self.Cols[n]] = self.Defs[n]
    n = n + 1
  end

  self.Pos = tmp1
  self.KeyDefs = tmp2

  self.opened = 1

  return 0
end

function RDB:close()

   local oINPUT = _INPUT
   _INPUT = self.handle
   readfrom( )
   _INPUT = oINPUT
end

function RDB:read( )

  local buf
  local rec
  local data = nil
  local i

  -- only read if there's been a successful open
  if self.opened then

    buf = read( self.handle )
    if buf then

      self.line = self.line + 1

      rec = RDB.split(buf)
      if ( rec.n ~= self.Cols.n ) then 
        error( self.File ..': line ' .. self.line .. 
			 ': error in RDB file: expected ' .. self.Cols.n .. ' columns, got ' .. rec.n );
      end
      data = {}
      i = 1
      while i <= self.n do
        if strfind( self.Defs[i], '^%d*N' ) then
          data[self.Cols[i]] = tonumber(rec[i])
        else
          data[self.Cols[i]] = rec[i]
        end
	i = i + 1
      end
    end  

  end

  return data
end

function RDB:new( ... )

  local rdb = RDB.clone( RDB )

  if ( arg.n > 0 ) then
    local err, errmsg
    err, errmsg = rdb:open( arg[1] )
    if nil == err then
      return nil, errmsg
    end
  end

  return rdb
end

function RDB.clone( what )
  local new = {}             -- create a new table
  local i, v = next(what, nil)  -- i is an index of t, v = t[i]

  while i do

    if type(v)=="table" then
      v = RDB.clone(v) 
    end
    new[i] = v
    i, v = next(what, i)        -- get next index
  end

  return new
end

function RDB:ncols( )
  return self.n
end

function RDB:cols( )
  return self.Cols
end

function RDB:pos( col )
  return self.Pos[col]
end

function RDB:defs( col )
  return self.KeyDefs[col]
end

function RDB:comments( )
  return self.Comments
end

function RDB:chk_cols( cols )

  return  foreach( cols, function( i, v )
		   if ( nil == %self:defs( v ) ) then
		      return v
                   end
                 end 
         )	

end

function RDB:var( varname )

  return self.Vars[varname]

end

function RDB:vars()

  return self.Vars

end

$end
