/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <string.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <linklist/linklist.h>
#include <luasup/luasup.h>

#include "init.h"

#include "aperture_queue.h"
#include "assembly_queue.h"
#include "assembly_check.h"
#include "xstack.h"

#include "common.h"


/*
   these keep track of the assembly and aperture id's.  aperture id's are
   unique within an assembly, not between.
*/
static unsigned long assembly_id;

/* the queue of assemblies */
static SLinkList assembly_queue = NULL;


/*---------------------------------------------------------------------*/

unsigned long get_assembly_id( void )
{
  return assembly_id;
}

/*---------------------------------------------------------------------*/
void
initialize_assembly_queue( void )
{
  tf_enter("initialize_assembly_queue");

  /* allocate assembly queue */
  if ( NULL == (assembly_queue = sll_new( NULL )) )
    tf_exit( ExitERR_alloc, "Unable to allocate assembly_queue" );

  assembly_id = 0;

  tf_leave();
}

void
cleanup_assembly_queue( void )
{
  if ( assembly_queue )
    sll_delete(assembly_queue, (SLL_DELETE) assembly_free);
}


/*--------------------------------------------------------------------*/
static int count_apertures( void *assembly, void *count )
{
  *((size_t*) count) += sll_count( ((Assembly*)assembly)->leaves );

  return 0;
}

size_t
assembly_queue_napertures( void )
{
  size_t count = 0;
  sll_utraverse(assembly_queue, count_apertures, &count );
  return count;
}

void
assembly_queue_init( void* udata )
{

  tf_enter("assembly_queue_init");

  sll_utraverse(assembly_queue, assembly_init, udata);

  tf_leave();
}


/*---------------------------------------------------------------------*/
/*
  PROCEDURE

  Take a linked list of assembly structures and turn it into an array
  of assembly structs.

  RETURN
  An array of Assembly structures.  Returns NULL if there are no assemblies.

 */
AssemblyArray*
assembly_queue_array_new( size_t* n_assemblies )
{
  AssemblyArray* array;
  AssemblyArray* tarray;

  tf_enter( "assembly_queue_array_new" );

  *n_assemblies = sll_count( assembly_queue );

  array = malloc( *n_assemblies * sizeof( AssemblyArray ) );
  if ( NULL == array )
    tf_exit( ExitERR_alloc, "Unable to allocate array of assemblies" );

  tarray = array;
  sll_utraverse( assembly_queue, assembly_cp, &tarray );

  tf_leave( );
  return array;
}

/* free an AssemblyArray  */
void
assembly_queue_array_free( size_t n_assemblies, AssemblyArray *array )
{
  size_t i;

  tf_enter( "assembly_queue_array_new" );

  for ( i = 0; i < n_assemblies; i++ )
    aperture_queue_array_free( array[i].apertures );

  free( array );

  tf_leave();
}


/*---------------------------------------------------------------------*/

/*
  PROCEDURE
  begin_assembly - create a new assembly structure

  DESCRIPTION This routine creates an assembly structure and attaches
  it to the assembly queue.  It creates the assembly's aperture queue and
  duplicates the current global coordinate transformation matrix (via
  check_in).

*/
static void
  begin_assembly( int block )
{
  Assembly *assembly;

  tf_enter("begin_assembly");
  
  /*
     To check for whether assembly-subassembly-aperture were started
     and ended in the correct order.
  */
  assembly_enter( ASSEMBLY );
  xstack_dup( );

  /* allocate new Assembly structure */
  if ( NULL == (assembly = malloc(sizeof(Assembly))) )
    tf_exit( ExitERR_alloc, "unable to allocate new assembly structure");

  /* establish the linked list of leaves */
  assembly->leaves = aperture_queue_new( );

  assembly->id = assembly_id;
  assembly->block = block;

  /* insert new assembly into queue */
  if ( sll_insert_tail(assembly_queue, assembly) )
    tf_exit( ExitERR_alloc, "unable to insert assembly in queue");

  tf_leave();
}

static void
begin_assembly_lua( void )
{
  char *block_pass = NULL;

  tf_enter( "begin_assembly_lua" );

  slua_args( errfunc, "begin_assembly",
	     "block/pass", SLua_STRING | SLua_OPTIONAL, &block_pass,
	     NULL );

  if ( block_pass != NULL )
  {
    if      (! strcmp( block_pass, "block" ) )
      begin_assembly( 1 );
    else if (! strcmp( block_pass, "pass"  ) )
      begin_assembly( 0 );
    else
      tf_exit( ExitERR_illegalArg,
	       "unrecognized argument to begin_assembly: %s", block_pass);

    free( block_pass );
  }
  else
  {
    /* default is to block */
    begin_assembly( 1 );
  }

  tf_leave();
}


/*---------------------------------------------------------------------*/
static void
end_assembly( void )
{

  tf_enter("end_assembly");

  assembly_leave( ASSEMBLY );
  xstack_pop( 1 );

  assembly_id++;

  tf_leave();
}

static void
  end_assembly_lua( void )
{
  slua_args( errfunc, "end_assembly", NULL );
  end_assembly( );
}


/*---------------------------------------------------------------------*/

static void
begin_subassembly( void )
{
  tf_enter("end_subassembly");

  assembly_enter( SUBASSEMBLY );
  xstack_dup();

  tf_leave();
}

static void
begin_subassembly_lua( void )
{
  slua_args( errfunc, "begin_subassembly", NULL );
  begin_subassembly( );
}

/*
  PROCEDURE
    end_subassembly

  DESCRIPTION
    Pop the top element of the transformation stack, via the function 
    pop_stack.  The function pop_stack shall check for consistency of the
    order which the subassemblies are built.

  AUTHOR
    Dan Nguyen
*/
static void
end_subassembly( void )
{
  tf_enter("end_subassembly");
  
  assembly_leave( SUBASSEMBLY );
  xstack_pop( 1 );

  tf_leave();
}


static void
end_subassembly_lua( void )
{
  slua_args( errfunc, "end_subassembly", NULL );
  end_subassembly( );
}



/*--------------------------------------------------------------------*/
static void
assembly_print_lua( void )
{
  slua_args( errfunc, "print_assembly", NULL );

  if ( is_in_assembly() )
  {
    Assembly* assembly = sll_tail( assembly_queue );
    tf_assert( NULL != assembly );

    assembly_print( assembly );
  }
  else
    tf_message("print_assembly: not in assembly\n");
}


/*---------------------------------------------------------------------*/

void
assembly_queue_register_lua( void )
{
  lua_register( "begin_assembly",    begin_assembly_lua );
  lua_register(   "end_assembly",      end_assembly_lua );
  lua_register( "begin_subassembly", begin_subassembly_lua );
  lua_register(   "end_subassembly",   end_subassembly_lua );
  lua_register( "print_assembly",    assembly_print_lua);
}

