/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include "config.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <lua.h>
#include <lualib.h>
#include <luasup/luasup.h>

#include "params.h"

#include "lua_register.h"

#include "aperture.h"
#include "assembly_check.h"
#include "xstack.h"

#include "common.h"


static void luaerrfunc( const char *fmt, ... );
SLua_ErrFunc errfunc=&luaerrfunc;

/*---------------------------------------------------------------------*/

#define LUA_ERROR_FUNC "tf_lua_error"
static void
tf_lua_error( void )
{
  tf_assert( lua_getparam( 1 ) != LUA_NOOBJECT );
  tf_die(  "%s", lua_getstring( lua_getparam( 1 ) ) );
}

static
void luaerrfunc( const char *fmt, ... ) {
  va_list ap;
  char buffer[1024];

  va_start( ap, fmt );

  vsprintf( buffer, fmt, ap );

  va_end( ap );

  /* this doesn't return */
  lua_error( buffer );
}


/*---------------------------------------------------------------------*/


void register_lua( Params *io )
{
  tf_enter("register_lua");

  /* register standard lua libraries */
  lua_iolibopen( );
  lua_strlibopen( );
  lua_mathlibopen( );

  /* register aperture definition functions */
  {
    ApertureModule **mod;
    for (mod = modules ; *mod ; mod++ )
      lua_register((*mod)->name, (*mod)->define);
  }

  /* register other systems */
  assembly_queue_register_lua();
  transform_register_lua();
  xstack_register_lua();

  /* register error function */
  lua_register( LUA_ERROR_FUNC, tf_lua_error );
  lua_pushobject(lua_getglobal( LUA_ERROR_FUNC ));
  lua_seterrormethod( );
}

/*---------------------------------------------------------------------*/

void exec_lua( Params *io )
{
  /* set up override function, if specified */
  if ( '\0' == io->override[0] )
    tf_assert( 0 == lua_dostring( "function override()\n end" ) );

  else if ( '@' == io->override[0] )
  {
    /* an '@' as the first letter signifies a file name follows */
    if ( lua_dofile( &io->override[1] ) )
      tf_die( "Error while parsing override function in file `%s'",
	      &io->override[1] );
  }

  else
  {
    char filename[ 9 ];
    int fd;
    FILE *file = NULL;

    strcpy(filename, "apXXXXXX" );

    fd = mkstemp( filename );
    if ( -1 == fd )
      tf_die( "unable to create source file for override() function\n" );
    

    file = fdopen( fd, "w" );
    if ( NULL == file )
    {
      close(fd);
      unlink( filename );
      tf_die( "error getting file handle for override() file\n" );
    }
	 
    fprintf( file, "function override()\n%s\nend",
	     io->override);
    fclose( file );

    if ( lua_dofile( filename ) )
    {
      unlink( filename );
      tf_die( "Error while creating override function" );
    }

    unlink( filename );
  }

  if ( lua_dofile( io->lua ) )
    tf_die( "Error while executing definition file %s", io->lua );

  /* make sure that things were setup correctly */
  check_assemblies( );

  xstack_check( );

  tf_leave();
}
