/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lua.h>
#include <luasup/luasup.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>

#include "module.h"
#include "ray_transforms.h"
#include "aperture.h"
#include "common.h"

#define MIN(x,y) ((x<y)?(x):(y))
#define MAX(x,y) ((x>y)?(x):(y))

/* forward declaration of function local to this file */
static void polygon(void);
static void init(Aperture *aperture, void* ami);
static ProcessResult process(Aperture *aperture);
static void cleanup(Aperture *aperture);


/*
   Define the ApertureModule structure for this module.
*/
ApertureModule polygon_module =
{
  "polygon",		/* the lua name of the instantiation function */
  polygon,		/* the C lua-callable instantiation function */
  init,			/* the aperture initialization function */
  process,		/* the aperture photon processing function */
  cleanup		/* the aperture cleanup function */
};


/*
   this structure contains all of the information necessary
   to describe an instantiation of this aperture.
*/

typedef struct {

  DVector2 *vertices;
  int    nvertices;
  int	 interior;
  ProcessResult result;

} Polygon;


/*
   this structure contains the offsets into the data packet of
   the data required by this aperture.  it is initialized by
   the init() function in this file.
*/
   
static struct
{
  DVector3 *position;
  DVector3 *direction;
} photon;

/*
   this function is called from lua to create an instantiation of
   this aperture type.
*/
static void
polygon( void ) {

  Polygon  *data;			/* do not free */
  double   transparent, nvertices, interior;
  lua_Object vertices = LUA_NOOBJECT, index = LUA_NOOBJECT, optargs = LUA_NOOBJECT;
  char     *label=NULL;    
  int idx;
  
  tf_enter("polygon");

  if ( NULL == (data = malloc( sizeof(Polygon) )) )
    tf_exit( ExitERR_alloc, "unable to allocate data for polygon" );

  slua_args( errfunc, "polygon",
	     "vertices",    SLua_TABLE,  &vertices,
	     "interior",    SLua_NUMBER, &interior,
	     "transparent", SLua_NUMBER, &transparent,
	     "optargs",	SLua_TABLE | SLua_OPTIONAL, &optargs,
	     NULL );

  vertices = 
    vertices == LUA_NOOBJECT ? lua_createtable() : slua_tbl_copy( vertices );

  data->nvertices = slua_tbl_nelem( vertices) ;

  if ( data->nvertices < 3 )
    tf_die( "must have greater than 2 vertices in polygon" );

  data->interior  = interior;
  data->result = transparent ? PASSED : BLOCKED;

  optargs = 
    optargs == LUA_NOOBJECT ? lua_createtable() : slua_tbl_copy( optargs );

  label = slua_tbl_el_si( optargs, APERTURELABEL );

  if ( NULL == (data->vertices = malloc( data->nvertices*sizeof(DVector2) )) )
    tf_exit( ExitERR_alloc, "unable to allocate data for polygon vertices" );

  for ( idx = 0; idx < data->nvertices; idx++ ) {

    lua_Object vertex, x, y, vertexindex = LUA_NOOBJECT;

    if ( LUA_NOOBJECT == (vertex = slua_tbl_next( vertices, &index )) )
      tf_die( "expected vertex not found" );
    
    if ( ! lua_istable( vertex ) )
      tf_die( "vertex %d is not a table", idx );

    if ( LUA_NOOBJECT == (x = slua_tbl_next( vertex, &vertexindex )) )
      tf_die( "expected X component of vertex not found" );

    if ( ! lua_isnumber( x ) )
      tf_die( "X component of vertex %d is not a number", idx );

    data->vertices[idx].x = lua_getnumber( x );

    if ( LUA_NOOBJECT == (y = slua_tbl_next( vertex, &vertexindex )) )
      tf_die( "expected Y component of vertex not found" );
   
    if ( ! lua_isnumber( y ) )
      tf_die( "Y component of vertex %d is not a number", idx );

    data->vertices[idx].y = lua_getnumber( y );
  }
  
  aperture_new( &polygon_module, data, label );

  tf_leave();
}

/*

   This routine is called once, before any photons arrive.  It should
   be used to determine offsets to whatever data this aperture needs
   from the photon data packet.  It should also initialize anything
   that needs initialization.
   
*/

static void 
init(Aperture *aperture, void* amip)
{
  ApertureModuleInfo* ami = amip;
  DpktField *dpktf_pos, *dpktf_dir;

  static int initialized = 0;

  /* do this only once to save some CPU */
  if (! initialized)
  {
    tf_enter("polygon init");

    if ( NULL == (dpktf_pos = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_pos )) )
      tf_die( "couldn't find field `" BP_AXAF_phot_pos "' in input stream" );

    if ( NULL == (dpktf_dir = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_dir )) )
      tf_die( "couldn't find field `" BP_AXAF_phot_dir "' in input stream" );

    photon.position  = (DVector3 *) bpipe_dpktf_data( dpktf_pos, ami->dpkt );
    photon.direction = (DVector3 *) bpipe_dpktf_data( dpktf_dir, ami->dpkt );
    
    initialized = 1;

    tf_leave();
  }
}

/*

   This routine is called for every photon.  It should be as efficient
   as possible.
   
*/

static ProcessResult
process( Aperture *aperture )
{

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */

  /* PTINPOLY variables */
  int crossings, j, yflag0, yflag1, xflag0;
  DVector2 vtx0, vtx1;

  Polygon *data = aperture->data_object;

  /* apply transformation to the photon */
  TRANSFORM_POSITION ( aperture->forward, photon.position, &p );
  TRANSFORM_DIRECTION( aperture->forward, photon.direction, &d );

  /* translate it to the z = 0 plane */
  PROJECT_RAY(&p, &d, 0.0);

  /* begin PTINPOLY routine */
  /* ptinpoly point inclusion in polygon test */
  /* ftp://ftp-graphics.stanford.edu/pub/Graphics/GraphicsGems/GemsIV/GGemsIV.tar.Z */

  vtx0 = data->vertices[data->nvertices - 1];
  /* get test bit for above/below X axis */
  yflag0 = ( vtx0.y > p.y );
  vtx1 = data->vertices[0];

  crossings = 0;

  for ( j = 0; j < data->nvertices; ) 
  {
    yflag1 = ( vtx1.y > p.y );
    /* check if endpoints straddle (are on opposite sides) of X axis
     * (i.e. the Y's differ); if so, +X ray could intersect this edge.
     * if ( y0 < y < y1 || y1 < y < y0 )
     */
    if ( yflag0 != yflag1 )
    {
      xflag0 = ( vtx0.x > p.x );
      /* check if endpoints are on same side of the Y axis (i.e. X's
       * are the same); if so, it's easy to test if edge hits or misses.
       * if ( ( x0 > x && x1 > x ) || ( x0 <= x && x1 <= x ) )
       */
      if ( xflag0 == ( vtx1.x > p.x ) ) 
      {
	/* if edge's X values both right of the point, must hit */
	if ( xflag0 ) 
	{
	  ( crossings ) ? ( crossings-- ) : ( crossings++ );
	}
      }
      else
      {
	/* compute intersection of pgon segment with +X ray, note
	 * if >= point's X; if so, the ray hits it.
	 */
	if ( (vtx1.x - (vtx1.y-p.y)*( vtx0.x-vtx1.x)/(vtx0.y-vtx1.y)) > p.x ) 
	{
	  ( crossings ) ? ( crossings-- ) : ( crossings++ );
	}
      }
    }
      
    /* move to next pair of vertices, retaining info as possible */
    yflag0 = yflag1;
    vtx0   = data->vertices[j];
    vtx1   = data->vertices[++j];
  }

  if ( data->interior )
  {
    return( ( crossings ) ? ( data->result ) : ( NOTAPPLICABLE ) );
  }
  else
  {
    return( ( ! crossings ) ? ( data->result ) : ( NOTAPPLICABLE ) );
  }
  /* end PTINPOLY routine*/
}

/*
   This routine is called once for every instantiatian of this aperture.
   It should free instantiation specific resources (like those allocated
   in the lua-called instantiation creation routine).
*/

static void
cleanup (Aperture *aperture)
{
  free(aperture->data_object);
}
