/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

#include <lua.h>
#include <luasup/luasup.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>

#include "module.h"
#include "ray_transforms.h"
#include "aperture.h"
#include "common.h"

/* forward declaration of function local to this file */
static void print_photon(void);
static void init(Aperture *aperture, void* amip);
static void cleanup(Aperture *aperture);

static ProcessResult process_print_xfrm(Aperture *aperture);
static ProcessResult process_print_raw(Aperture *aperture);
static ProcessResult process_print_raw_project(Aperture *aperture);
static ProcessResult process_print_project(Aperture *aperture);

/*
   Define the ApertureModule structure for this module.
*/
ApertureModule print_photon_module =
{
  "print_photon",	/* the lua name of the instantiation function */
  print_photon,		/* the C lua-callable instantiation function */
  init,			/* the aperture initialization function */
  NULL,			/* the aperture photon processing function */
  cleanup		/* the aperture cleanup function */
};


/*
   this structure contains the offsets into the data packet of
   the data required by this aperture.  it is initialized by
   the init() function in this file.
*/
   
static struct
{
  DVector3 *position;
  DVector3 *direction;
} photon;

static int print = 1;


/*
   this function is called from lua to create an instantiation of
   this aperture type.
*/
static void
print_photon( void ) {

  char *test;		/* the test to perform */

  ApertureModule *module;

  tf_enter("print_photon");

  if ( NULL == (module = malloc( sizeof(ApertureModule) )) )
    tf_exit( ExitERR_alloc, "unable to allocate module for photon_print" );
    
  slua_args( errfunc, "print_photon",
	     "test", SLua_STRING, &test,
	     NULL );

  if      ( ! strcmp(test, "xfrm") )
    print_photon_module.process = process_print_xfrm;
  else if ( ! strcmp(test, "raw") )
    print_photon_module.process = process_print_raw;
  else if ( ! strcmp(test, "raw_project") )
    print_photon_module.process = process_print_raw_project;
  else if ( ! strcmp(test, "project") )
    print_photon_module.process = process_print_project;
  else if ( ! strcmp(test, "on") )
  {
    print = 1;
    tf_leave();
    return;
  }
  else if ( ! strcmp(test, "off") )
  {
    print = 0;
    tf_leave();
    return;
  }
  else
    tf_exit( ExitERR_illegalArg, "bogus argument to print_photon" );

  *module = print_photon_module;
  
  aperture_new( module, module, NULL );

  free(test);

  tf_leave();
}

/*

   This routine is called once, before any photons arrive.  It should
   be used to determine offsets to whatever data this aperture needs
   from the photon data packet.  It should also initialize anything
   that needs initialization.
   
*/

static void 
init(Aperture *aperture, void* amip)
{
  ApertureModuleInfo* ami = amip;
  DpktField *dpktf_pos, *dpktf_dir;

  static int initialized = 0;

  /* do this only once to save some CPU */
  if (! initialized)
  {
    tf_enter("print_photon init");

    if ( NULL == (dpktf_pos = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_pos )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_pos
	                           "' in input stream" );

    if ( NULL == (dpktf_dir = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_dir )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_dir
	                           "' in input stream" );

    photon.position  = (DVector3 *) bpipe_dpktf_data( dpktf_pos, ami->dpkt );
    photon.direction = (DVector3 *) bpipe_dpktf_data( dpktf_dir, ami->dpkt );
    
    initialized = 1;

    tf_leave();
  }
}

/*

   This routine is called for every photon.  It should be as efficient
   as possible.
   
*/

static void
print_out_photon(
             unsigned long assembly,
	     unsigned long aperture,
	     DVector3 *p,
	     DVector3 *d
	     )
{
    tf_message("%03d.%03d: %.14g\t%.14g\t%.14g :\t%.14g\t%.14g\t%.14g\n",
	       assembly, aperture,
	       p->x, p->y, p->z, d->x, d->y, d->z);
}

static ProcessResult
process_print_xfrm( Aperture *aperture )
{

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */

  if (print)
  {
    /* apply transformation to the photon */
    TRANSFORM_POSITION ( aperture->forward, photon.position, &p );
    TRANSFORM_DIRECTION( aperture->forward, photon.direction, &d );

    print_out_photon(aperture->id.assembly, aperture->id.aperture, &p, &d);
  }

  return NOTAPPLICABLE;
}

static ProcessResult
process_print_project( Aperture *aperture )
{

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */

  if (print)
  {
    /* apply transformation to the photon */
    TRANSFORM_POSITION ( aperture->forward, photon.position, &p );
    TRANSFORM_DIRECTION( aperture->forward, photon.direction, &d );

    /* translate it to the z = 0 plane */
    PROJECT_RAY(&p, &d, 0.0);

    print_out_photon(aperture->id.assembly, aperture->id.aperture, &p, &d);
  }

  return NOTAPPLICABLE;
}

static ProcessResult
process_print_raw_project( Aperture *aperture )
{

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */

  DVector3 p_r;		/* position of vector in module ref frame */
  DVector3 d_r;		/* direction of vector in module ref frame */

  if (print)
  {
    /* apply transformation to the photon */
    TRANSFORM_POSITION ( aperture->forward, photon.position, &p );
    TRANSFORM_DIRECTION( aperture->forward, photon.direction, &d );

    /* translate it to the z = 0 plane */
    PROJECT_RAY(&p, &d, 0.0);

    /* now translate it back to the external frame */
    TRANSFORM_POSITION ( aperture->backward, &p, &p_r );
    TRANSFORM_DIRECTION( aperture->backward, &d, &d_r );

    print_out_photon(aperture->id.assembly, aperture->id.aperture, &p_r, &d_r);
  }

  return NOTAPPLICABLE;
}

static ProcessResult
process_print_raw( Aperture *aperture )
{

  if (print)
    print_out_photon(aperture->id.assembly, aperture->id.aperture,
		 photon.position, photon.direction);

  return NOTAPPLICABLE;
}


/*
   This routine is called once for every instantiatian of this aperture.
   It should free instantiation specific resources (like those allocated
   in the lua-called instantiation creation routine).
*/

static void
cleanup (Aperture *aperture)
{
  free(aperture->data_object);
}








