/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <linklist/linklist.h>

#include "module.h"
#include "assembly_queue.h"
#include "common.h"
#include "params.h"
#include "process.h"

typedef struct {
  void *dpkt;
  int first_valid;
} PhotonStack;

static double* weight;
static size_t dpkt_core_size;
static void *dpkt;
static void *dpkt_copy;
BPipe *bpipe;
BPipeOutput *bpo;

/* photons that are generated by apertures are kept on this stack */
static SLinkList photon_stack = NULL;

/* the assembly id that's valid for the current photon */
static int valid_assembly;

static void write_accumulated_statistics( char* filename, char* mode,
					  size_t num_assemblies,
					  AssemblyArray* assembly_array );
/*--------------------------------------------------------------------*/
static void
init_bpipe( Params* io )
{
  DpktField* dpkdtf_wt = NULL;

  tf_enter("init_bpipe");

  if ( NULL == ( bpipe = bpipe_new( ) ) )
    tf_exit( ExitERR_alloc, "unable to create bpipe" );

  if ( bpipe_input( bpipe, io->in ) )
    tf_exit( ExitERR_alloc, "unable to attach input stream `%s' to bpipe: %s",
	     io->in, bpipe_strerror( bpipe_errno ) );

  if ( NULL == ( bpo = bpipe_output( bpipe, io->out ) ) )
    tf_exit( ExitERR_alloc, "unable to attach output stream `%s' to bpipe: %s",
	     io->out, bpipe_strerror( bpipe_errno ) );

  dpkt_core_size = bpipe_map( bpipe );
  bpipe_write_hdr( bpipe );                            /* write out header */

  if ( (dpkt = malloc(dpkt_core_size)) == NULL )
    tf_exit( ExitERR_alloc, "Unable to allocate `dpkt'");

  if ( (dpkt_copy = malloc(dpkt_core_size)) == NULL )
    tf_exit( ExitERR_alloc, "Unable to allocate copy of dpkt");

  if ( NULL == (dpkdtf_wt = bpipe_dpktf( bpipe, BP_AXAF_phot_wt )) )
    tf_die( "couldn't find field `" BP_AXAF_phot_wt "' in input stream" );

  weight = (double*) bpipe_dpktf_data( dpkdtf_wt, dpkt );

  tf_leave();
}

static void
cleanup_bpipe( void )
{
  free( dpkt );
  free( dpkt_copy );
  bpipe_delete( bpipe );
}

static int
get_photon( void )
{
  PhotonStack *photon;
  SLLNode top_of_stack;

  /* If photon stack is empty then get photon from the binary pipe */
  if ( sll_count( photon_stack ) == 0 ) {
    valid_assembly = 0;
    return bpipe_read_dpkts( bpipe, dpkt, 1 ) ;
  } 
  
  /* get from the photon stack instead */
  top_of_stack = sll_head_node( photon_stack );

  tf_assert( top_of_stack != NULL );

  photon = sll_destroy_node( photon_stack, top_of_stack );
  tf_assert5( photon != NULL );

  memcpy( dpkt, photon->dpkt, dpkt_core_size );

  valid_assembly = photon->first_valid;

  free( photon->dpkt );
  free( photon );
  return 1;
}

enum { NEXT_LEAF, NEXT_PHOTON, NEXT_ASSEMBLY, SAME_ASSEMBLY };

static int process_aperture( Aperture* aperture ) {

  ProcessResult status;

  tf_assert5(aperture != NULL);

  /* ensure that dpkt_copy hasn't been corrupted */
  tf_assert5( ! memcmp(dpkt, dpkt_copy, dpkt_core_size) );

  status = aperture->module->process( aperture );
  switch( status ) {

  case NOTAPPLICABLE:       /* on to next aperture */
    tf_assert5( ! memcmp(dpkt, dpkt_copy, dpkt_core_size) );
    return NEXT_LEAF;			

  case PASSED:              /* on to next assembly */
    tf_assert5( ! memcmp(dpkt, dpkt_copy, dpkt_core_size) );
    return NEXT_ASSEMBLY;

  case MODIFIED:
    /* same as PASSED, but must copy photon to system buffer */
    memcpy( dpkt, dpkt_copy, dpkt_core_size );
    return NEXT_ASSEMBLY;

  case BLOCKED:             /* on to next photon */
    return NEXT_PHOTON;

  case GENERATE:            /* on to next photon, from the stack */
    return NEXT_PHOTON;
    
  case REDIRECTED:
    memcpy( dpkt, dpkt_copy, dpkt_core_size );
    /* if not in loop mode move on to next assembly.  if in loop mode
       restart this assembly. */
    return loop_it ? SAME_ASSEMBLY : NEXT_ASSEMBLY;

  default:
    tf_assert5( ! memcmp(dpkt, dpkt_copy, dpkt_core_size) );
    tf_enter("process_aperture");
    tf_exit( ExitERR_illegalArg,
	    "illegal status (%d)", status );
  }

  /* shouldn't get here! */
  tf_assert(0);

  /* NOT REACHED */

  /* for the complaining compilers */

  return NOTAPPLICABLE;
}

static int process_assembly( AssemblyArray* assembly ) {

  int val = NOTAPPLICABLE;
  int max_loop = 100 ;

  /*  if this isn't the appropriate assembly for this photon, signal it */
  if ( assembly->id < valid_assembly )
    return 0;

  do {
    int ii=0;
    for ( ii = 0; ii < assembly->num_apertures; ii++ ) {
      val = process_aperture( &(assembly->apertures[ ii ]) );
      if ( 0 != val ) {
	assembly->apertures[ ii ].status[ val ] += 1;
	assembly->apertures[ ii ].weight[ val ] += *weight;
	break;
      }
    }
  } while( SAME_ASSEMBLY == val && --max_loop );

  if (max_loop == 0) {
    tf_enter("process_assembly");
    tf_exit(ExitERR_misc,
	    "infinite redirect loop detected in Assembly %d", assembly->id);
  }

  tf_assert5( val == NEXT_ASSEMBLY || val == 0 || val == NEXT_PHOTON );

  /*
    val == 0 implies that it didn't get processed by any leaves.
    discard it the photon if the assembly is in block mode, else
    pass it on to the next assembly.
  */
  return  ( (val == 0 && assembly->block ) || val == NEXT_PHOTON ) ? 1 : 0;
}



void process_all_assemblies( Params* io )
{

  size_t num_assemblies=0;
  int i = 0;
  ApertureModuleInfo ami;
  AssemblyArray* assembly_array;

  tf_enter("process_all_assemblies");

  if ( 0 == assembly_queue_napertures() )
    tf_die( "No apertures were specified\n" );

  init_bpipe( io );

  ami.bpipe = bpipe;
  ami.dpkt  = dpkt_copy;

  assembly_queue_init( &ami );

  assembly_array = assembly_queue_array_new( &num_assemblies );

  while ( get_photon( ) ) {

    int val = 0;

    memcpy( dpkt_copy, dpkt, dpkt_core_size );

    /* loop through assemblies, and if it made it through,
       write the photon out */
    for ( i = 0; i < num_assemblies; i++ ) {
      val = process_assembly( &(assembly_array[ i ]) );
      if ( 0 != val )
	break;
    }

    if ( 0 == val ) 
      if ( bpipe_write_dpkt( bpipe, dpkt, bpo )  )
	tf_die( "error writing photon to output stream (%s): ",
		io->out, bpipe_strerror( bpipe_errno ));
  }

  if ( bpipe_errno )
    tf_die( "Error reading from input stream (%s): %s",
	    io->in, bpipe_strerror( bpipe_errno ) );

  write_accumulated_statistics( io->statistics, "w",
				num_assemblies, assembly_array );


  cleanup_bpipe();

  assembly_queue_array_free( num_assemblies, assembly_array );

  /* leaves are cleaned up in cleanup_assembly_queue called from main() */
  
  tf_leave();
}

void
push_photon( Aperture *aperture, void *new_photon ) 
{

  PhotonStack *photon;  /* do not free */

  tf_enter("push_photon");

  /* allocate a photon-stack if one does not exist */
  if ( photon_stack == NULL && NULL == (photon_stack = sll_new( NULL )) )
      tf_exit( ExitERR_alloc, "Unable to allocate photon_stack" );

  /* do not free photon */
  if ( NULL == (photon = malloc( sizeof( PhotonStack ) )) )
    tf_exit( ExitERR_alloc, "Unable to allocate `photon'" );

  /* do not free dpkt */
  if ( NULL == (photon->dpkt = malloc( dpkt_core_size )) )
    tf_exit( ExitERR_alloc, "Unable to allocate `dpkt'");

  memcpy( photon->dpkt, new_photon, dpkt_core_size );  

  /* 
     If in loop mode (loop_it == yes) then the current assembly is the
     first valid assembly for this photon, else use the next available
     assembly
  */

  photon->first_valid = loop_it ? aperture->id.assembly : aperture->id.assembly + 1;

  if ( sll_insert_tail( photon_stack, photon ) )
    tf_exit(ExitERR_alloc, "unable to insert photon in photon stack");

  tf_leave();
}


/* write out the accumulated statistics */
static void write_accumulated_statistics( char* filename, char* mode,
					  size_t num_assemblies,
					  AssemblyArray* assembly_array ) {

  int i;
  FILE* fp = NULL;

  tf_enter( "write_accumulated_statistics" );

  fp = fopen( filename, mode );
  if ( NULL == fp )
    tf_exit( ExitERR_misc, "Unable to open file `%s'", filename );

  fprintf( fp, 
	   "name\tnblocked\twblocked\tnpassed\twpassed"
	   "\tnredirected\twredirected\tngenerate\twgenerate"
	   "\tnmodified\twmodified\n" );
  fprintf( fp, "S\tN\tN\tN\tN\tN\tN\tN\tN\tN\tN\n" );
  for( i = 0; i < num_assemblies; i++ ) {
    int j;
    for ( j = 0; j < assembly_array[ i ].num_apertures; j++ ) {
      int k;
      fprintf( fp, "%s\t",
	       assembly_array[ i ].apertures[ j ].id.label );
      for ( k = 1; k < MAXPROCESSRESULT; k++ ) {
	fprintf( fp, "%lu\t%g",
		 assembly_array[ i ].apertures[ j ].status[ k ],
		 assembly_array[ i ].apertures[ j ].weight[ k ] );
	if ( k < MAXPROCESSRESULT - 1 )
	  fprintf( fp, "\t" );
	else
	  fprintf( fp, "\n" );
      }

    }
  }

  fclose( fp );

  tf_leave( );
}

