/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lua.h>
#include <luasup/luasup.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>

#include "module.h"
#include "ray_transforms.h"
#include "aperture.h"
#include "common.h"
#include "process.h"

/* forward declaration of function local to this file */
static void test1(void);
static void init(Aperture *aperture, void* amip);
static void cleanup(Aperture *aperture);

static ProcessResult process_pass_xfrm(Aperture *aperture);
static ProcessResult process_redirect( Aperture *aperture );
static ProcessResult process_shiftp( Aperture *aperture );
static ProcessResult process_shiftm( Aperture *aperture );
static ProcessResult process_generate( Aperture *aperture );

/*
   Define the ApertureModule structure for this module.
*/
ApertureModule test1_module =
{
  "test1",		/* the lua name of the instantiation function */
  test1,		/* the C lua-callable instantiation function */
  init,			/* the aperture initialization function */
  NULL,			/* the aperture photon processing function */
  cleanup		/* the aperture cleanup function */
};


/*
   the pointer to the data packet.  it is initialized by the
   init() function in this file
*/
static void *dpkt;

/*
   this structure contains the offsets into the data packet of
   the data required by this aperture.  it is initialized by
   the init() function in this file.
*/
   
static struct
{
  DVector3 *position;
  DVector3 *direction;
  double   *weight;
} photon;


/*
   this function is called from lua to create an instantiation of
   this aperture type.
*/
static void
test1( void ) {

  char *test;	  /* the test to perform */
  ApertureModule *module;

  tf_enter("test1");

  slua_args( errfunc, "test1", "test", SLua_STRING, &test, NULL );

  if ( NULL == (module = malloc( sizeof(ApertureModule) )) )
    tf_exit( ExitERR_alloc, "unable to allocate data for test1" );
    
  if      ( ! strcmp(test, "pass-xfrm") )
    test1_module.process = process_pass_xfrm;
  else if ( ! strcmp(test, "redirect") )
    test1_module.process = process_redirect;
  else if ( ! strcmp(test, "shiftp") )
    test1_module.process = process_shiftp;
  else if ( ! strcmp(test, "shiftm") )
    test1_module.process = process_shiftm;
  else if ( ! strcmp(test, "generate") )
    test1_module.process = process_generate;
  else
    tf_exit( ExitERR_illegalArg, "bogus argument to test1" );

  *module = test1_module;

  aperture_new( module, module, NULL );

  free(test);

  tf_leave();
}

/*

   This routine is called once, before any photons arrive.  It should
   be used to determine offsets to whatever data this aperture needs
   from the photon data packet.  It should also initialize anything
   that needs initialization.
   
*/

static void 
init(Aperture *aperture, void* amip)
{
  ApertureModuleInfo* ami = amip;
  DpktField *dpktf_pos, *dpktf_wt, *dpktf_dir;

  static int initialized = 0;

  /* do this only once to save some CPU */
  if (! initialized)
  {
    tf_enter("test1 init");

    if ( NULL == (dpktf_pos = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_pos )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_pos
	                           "' in input stream" );

    if ( NULL == (dpktf_dir = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_dir )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_dir
	                           "' in input stream" );

    if ( NULL == (dpktf_wt = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_wt )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_wt
	                           "' in input stream" );

    dpkt = ami->dpkt;

    photon.position  = (DVector3 *) bpipe_dpktf_data( dpktf_pos, dpkt );

    photon.direction = (DVector3 *) bpipe_dpktf_data( dpktf_dir, dpkt );

    photon.weight    = (double *)   bpipe_dpktf_data( dpktf_wt , dpkt );
    
    initialized = 1;

    tf_leave();
  }
}

/*

   This routine is called for every photon.  It should be as efficient
   as possible.
   
*/

static ProcessResult
process_pass_xfrm( Aperture *aperture )
{

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */

  /* apply transformation to the photon */
  TRANSFORM_POSITION ( aperture->forward, photon.position, &p );
  TRANSFORM_DIRECTION( aperture->forward, photon.direction, &d );

  *photon.position = p;
  *photon.direction = d;

  /* accept everything, but don't transform back */
  return MODIFIED;
}

static ProcessResult
process_generate( Aperture *aperture )
{
  static int first = 1;
  if (first)
  {
    photon.position->x += 1;
    photon.position->y += 2;
    push_photon( aperture, dpkt );

    photon.position->x += 1;
    photon.position->y += 2;
    push_photon( aperture, dpkt );

    photon.position->x += 1;
    photon.position->y += 2;
    push_photon( aperture, dpkt );

    photon.position->x += 1;
    photon.position->y += 2;
    push_photon( aperture, dpkt );

    first = 0;

    return GENERATE;
  }

  return NOTAPPLICABLE;
}

static ProcessResult
process_redirect( Aperture *aperture )
{
  return REDIRECTED;
}

static ProcessResult
process_shiftp( Aperture *aperture )
{
  photon.position->x += 4.0;
  photon.position->y += 4.0;

  return MODIFIED;
}

static ProcessResult
process_shiftm( Aperture *aperture )
{
  photon.position->x -= 4.0;
  photon.position->y -= 4.0;

  return MODIFIED;
}

/*
   This routine is called once for every instantiatian of this aperture.
   It should free instantiation specific resources (like those allocated
   in the lua-called instantiation creation routine).
*/

static void
cleanup (Aperture *aperture)
{
  free(aperture->data_object);
}








