/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>

#include <bpipe/bpipe.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <mst_nrutil/mst_nrutil.h>
#include <luasup/luasup.h>

#include <lua.h>
#include <lualib.h>

#include "lua_register.h"
#include "aperture.h"

#include "xmatrix.h"
#include "xstack.h"

#include "common.h"




/*---------------------------------------------------------------------*/
/*
  PROCEDURE
    Rotation of Co-ordinate Axes.

  DESCRIPTION
    Calculate the 3-D axes rotation, the return pointer of the
    procedure.  The axes are rotated anti-clockwise through an angle 
    theta radians about an  axis specified by the argument `axis'.
    If `axis' is XAXIS, YAXIS or ZAXIS then the rotation shall be about
    the axis x, y or z, respectively.

  AUTHOR
    Dan Nguyen
*/
static void
  rotate(
	 XYZaxis axis,		/* axis of rotation */ 
	 double  angle		/* angle to rotate, in radians */
	 )
{
  XMatrix mat;

  tf_enter("rotate");
  
  /*
    rotation angle for photon is *opposite* of that for the
    aperture.  since this is called when constructing aperture, negate it
  */

  mat = Xmatrix_rotation( axis, -angle );
  xstack_update( mat );
  XMatrix_free( mat );

  tf_leave();
  return;
}

/*---------------------------------------------------------------------*/
/*
  PROCEDURE
    Rotation about X-Co-ordinate axis.

  DESCRIPTION
    Calculate the 3-D axis rotation about the x-axis of `angle' radian.
    The orientation of the rotation is anti-clockwise.

  AUTHOR
    Dan Nguyen
*/
static void
rotate_x_lua( void )
{
  double angle;
  tf_enter("rotate_x");

  slua_args( errfunc, "rotate_x",
	     "theta", SLua_NUMBER, &angle, NULL );
  rotate( XAXIS , angle );

  tf_leave();
}

/*
  PROCEDURE
    Rotation about Y-Co-ordinate axis.

  DESCRIPTION
    Calculate the 3-D axis rotation about the y-axis of `angle' radian.
    The orientation of the rotation is anti-clockwise.

  AUTHOR
    Dan Nguyen
*/
static void
rotate_y_lua( void )
{
  double angle;
  tf_enter("rotate_y");

  slua_args( errfunc, "rotate_y",
	     "theta", SLua_NUMBER, &angle, NULL );
  rotate( YAXIS , angle );

  tf_leave();
}

/*
  PROCEDURE
    Rotation about Z-Co-ordinate axis.

  DESCRIPTION
    Calculate the 3-D axis rotation about the z-axis of `angle' radian.
    The orientation of the rotation is anti-clockwise.

  AUTHOR
    Dan Nguyen
*/
static void
rotate_z_lua( void )
{
  double angle;
  tf_enter("rotate_z");

  slua_args( errfunc, "rotate_z",
	     "theta", SLua_NUMBER, &angle, NULL );
  rotate( ZAXIS , angle );

  tf_leave();
}

/*
   a lua function to translate the matrix
*/
static void
translate_lua( void )
{
  XMatrix mat;
  double transl[DIM3];

  tf_enter("translate");

  slua_args( errfunc, "translate",
	     "x", SLua_NUMBER, &transl[0],
	     "y", SLua_NUMBER, &transl[1],
	     "z", SLua_NUMBER, &transl[2],
	     NULL );

  /*
    transformation matrix for photon is *opposite* of that for the
    aperture.  since this is called when constructing aperture, negate it
  */

  transl[0] = - transl[0];
  transl[1] = - transl[1];
  transl[2] = - transl[2];
  
  mat = Xmatrix_translation( transl );
  xstack_update( mat );
  XMatrix_free( mat );

  tf_leave();
}

/*---------------------------------------------------------------------*/

void
transform_register_lua( void )
{
  lua_register( "translate",  translate_lua );
  lua_register( "rotate_x",   rotate_x_lua  );
  lua_register( "rotate_y",   rotate_y_lua  );
  lua_register( "rotate_z",   rotate_z_lua  );
}

/*---------------------------------------------------------------------*/

#if 0
void
  transform_position( XMatrix tran, DVector3 *in, DVector3 *out )
{
  out->x = in->x * tran[0][0] + in->y * tran[1][0] + in->z * tran[2][0] + 
    tran[3][0];

  out->y = in->x * tran[0][1] + in->y * tran[1][1] + in->z * tran[2][1] + 
    tran[3][1];

  out->z = in->x * tran[0][2] + in->y * tran[1][2] + in->z * tran[2][2] + 
    tran[3][2];

  return;
}

void
  transform_direction( XMatrix tran, DVector3 *in, DVector3 *out )
{
  out->x = in->x * tran[0][0] + in->y * tran[1][0] + in->z * tran[2][0] ;

  out->y = in->x * tran[0][1] + in->y * tran[1][1] + in->z * tran[2][1] ;

  out->z = in->x * tran[0][2] + in->y * tran[1][2] + in->z * tran[2][2] ;

  return;
}
#endif


