/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

#include <lua.h>
#include <luasup/luasup.h>
#include <mathconst/mathconst.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>

#include "module.h"
#include "ray_transforms.h"
#include "aperture.h"
#include "common.h"

/* forward declaration of function local to this file */
static void wedge(void);
static void init(Aperture *aperture, void* amip);
static ProcessResult process(Aperture *aperture);
static void cleanup(Aperture *aperture);

/*
   Define the ApertureModule structure for this module.
*/
ApertureModule wedge_module =
{
  "wedge",		/* the lua name of the instantiation function */
  wedge,		/* the C lua-callable instantiation function */
  init,			/* the aperture initialization function */
  process,		/* the aperture photon processing function */
  cleanup		/* the aperture cleanup function */
};


/*
   this structure contains all of the information necessary
   to describe an instantiation of this aperture.
*/

typedef struct {

  int sign_min;
  int sign_max;
  int flipped;
  double xmin;
  double xmax;
  double ymin;
  double ymax;

  ProcessResult result;

} Wedge;


/*
   this structure contains the offsets into the data packet of
   the data required by this aperture.  it is initialized by
   the init() function in this file.
*/
   
static struct
{
  DVector3 *position;
  DVector3 *direction;
} photon;

/*
   this function is called from lua to create an instantiation of
   this aperture type.
*/
static void
wedge( void ) {

  Wedge *	data;			/* do not free */
  double	interior;
  double	xtest, ytest;
  double	theta_min, theta_max, mid_theta;
  double	transparent;
  lua_Object optargs = LUA_NOOBJECT;
  char     *label=NULL;    


  tf_enter("wedge");

  if ( NULL == (data = malloc( sizeof(Wedge) )) )
    tf_exit( ExitERR_alloc, "unable to allocate data for wedge" );

  slua_args( errfunc, "wedge",
	     "theta_min",   SLua_NUMBER, &theta_min,
	     "theta_max",   SLua_NUMBER, &theta_max,
	     "interior",    SLua_NUMBER, &interior,
	     "transparent", SLua_NUMBER, &transparent,
	     "optargs",	SLua_TABLE | SLua_OPTIONAL, &optargs,
	     NULL );

  /* this routine works by creating two lines which bound the wedge's
     edges.  the lines split the plane into four sections.  points in
     each section are uniquely tagged by a pair of numbers which
     denote on which side of the two lines that the points lie.  if the
     wedge is less than 180 degrees in width, it occupies one of the
     four sections.  if it is greater than 180 degrees, it occupies
     three of the four sections.  the algorithm determines if a point
     is in the wedge by constructing a test point which lies in the
     inclusive section (if the wedge is less than 180 degrees) or the
     exclusive section(if the wedge is greater than 180 degrees) and
     comparing the unique id pair of the test point with that of the
     ray.
   */

  /* generate comparison points for the angles */

  /* remove 2 PI degeneracy */
  theta_min = fmod(theta_min, M_2PI);
  theta_max = fmod(theta_max, M_2PI);

  /* if the exterior is requested, just swap the min and max angles */
  if ( ! interior )
  {
    double temp = theta_min;
    theta_min = theta_max;
    theta_max = temp;
  }

  /* if the wedge straddles the discontinuity, make it continuous */
  if (theta_min > theta_max) theta_max += M_2PI;

  /* if the wedge is greater than 180 degrees in width, make the
     exclusive section the wedge and flip the notion of rejection and
     acceptance
   */
  data->flipped = 0;

  if (theta_max - theta_min > M_PI)
  {
    double temp = theta_max;
    theta_max = theta_min;
    theta_min = temp;
    data->flipped++;

    /* ensure that theta_min < theta_max */
    theta_max += M_2PI;
  }

  /* determine the direction cosines of the two lines */
  data->xmin = cos(theta_min);
  data->ymin = sin(theta_min);

  data->xmax = cos(theta_max);
  data->ymax = sin(theta_max);

  /* construct test point exactly half way between edges of wedge */
  mid_theta = (theta_min + theta_max) / 2.0;
  xtest = cos(mid_theta);
  ytest = sin(mid_theta);

  /* determine which side of the lines the test point is */
  data->sign_min = (xtest * data->ymin - ytest * data->xmin) > 0.0;
  data->sign_max = (xtest * data->ymax - ytest * data->xmax) > 0.0;

  data->result = transparent ? PASSED : BLOCKED;

  optargs = 
    optargs == LUA_NOOBJECT ? lua_createtable() : slua_tbl_copy( optargs );

  label = slua_tbl_el_si( optargs, APERTURELABEL );

  aperture_new( &wedge_module, data, NULL );

  tf_leave();
}

/*

   This routine is called once, before any photons arrive.  It should
   be used to determine offsets to whatever data this aperture needs
   from the photon data packet.  It should also initialize anything
   that needs initialization.
   
*/

static void 
init(Aperture *aperture, void *amip )
{
  ApertureModuleInfo* ami = amip;
  DpktField *dpktf_pos, *dpktf_dir;

  static int initialized = 0;

  /* do this only once to save some CPU */
  if (! initialized)
  {
    tf_enter("wedge init");

    if ( NULL == (dpktf_pos = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_pos )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_pos
	                           "' in input stream" );

    if ( NULL == (dpktf_dir = bpipe_dpktf( ami->bpipe, BP_AXAF_phot_dir )) )
      tf_exit( ExitERR_misc, "couldn't find field `" BP_AXAF_phot_dir
	                           "' in input stream" );

    photon.position  = (DVector3 *) bpipe_dpktf_data( dpktf_pos, ami->dpkt );
    photon.direction = (DVector3 *) bpipe_dpktf_data( dpktf_dir, ami->dpkt );
    
    initialized = 1;

    tf_leave();
  }
}

/*

   This routine is called for every photon.  It should be as efficient
   as possible.
   
*/

static ProcessResult
process( Aperture *aperture )
{

  DVector3 p;		/* position of vector in module ref frame */
  DVector3 d;		/* direction of vector in module ref frame */
  Wedge *data = aperture->data_object;
  int	in_wedge;

  /* apply transformation to the photon */
  TRANSFORM_POSITION ( aperture->forward, photon.position, &p );
  TRANSFORM_DIRECTION( aperture->forward, photon.direction, &d );

  /* translate it to the z = 0 plane */
  PROJECT_RAY(&p, &d, 0.0);

  in_wedge =
    ((p.x * data->ymin) - (p.y * data->xmin) > 0.0) == data->sign_min &&
    ((p.x * data->ymax) - (p.y * data->xmax) > 0.0) == data->sign_max;

  if ( data->flipped )
    return in_wedge ? NOTAPPLICABLE :data->result;
  else
    return in_wedge ? data->result : NOTAPPLICABLE;
}

/*
   This routine is called once for every instantiatian of this aperture.
   It should free instantiation specific resources (like those allocated
   in the lua-called instantiation creation routine).
*/

static void
cleanup (Aperture *aperture)
{
  free(aperture->data_object);
}
