/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <stdio.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <mst_nrutil/mst_nrutil.h>

#include "xmatrix.h"

XMatrix
XMatrix_new( void )
{
  XMatrix matrix;

  if ( NULL == ( matrix = dmatrix_r ( 0L, DIM3, 0L, DIM3 ) ) )
  {
    tf_enter( "XMatrix_new" );
    tf_exit(ExitERR_alloc, "unable to allocate  matrix");
  }

  return matrix;
}

void
XMatrix_free( XMatrix matrix )
{
  free_dmatrix( matrix,  0L, DIM3, 0L, DIM3 );
}


XMatrix
XMatrix_dup( XMatrix from )
{
  double **to;

  tf_enter("XMatrix_dup");

  to = XMatrix_new();

  XMatrix_copy( from, to );

  tf_leave();
  return to;
}


/*
  PROCEDURE
    To print the transformation matrix.
*/
void
  XMatrix_print( double **a )
{
  int i, j;

  for ( i = 0; i < DIM4; i++ ) {
    tf_message( "" );
    for ( j = 0; j < DIM4; j++ )
      fprintf(stderr, " %-6.3f", a[i][j] );
    fputc('\n', stderr);
  }
  return;
}

/*
  PROCEDURE
    Multiplication of two double matrices of size DIM4 by DIM4.

  DESCRIPTION
    Calculates the matrix multiplication of DIM4 by DIM4 matrix `a'
    by a matrix DIM4 by DIM4 matrix `b'.  The results will be stored
    in the matrix matrix `ab', the matrix ab will be allocated within
    the routine.

  RETURNS
    The pointer to the matrix whose elements are the product of matrix
    a times b.  If it cannot allocate the result matrix, it prints
    an error to stderr via tf_exit() and exits

  AUTHOR
    Dan Nguyen
*/
XMatrix
  XMatrix_mult( XMatrix a  /* The first matrix in: a * b */, 
		XMatrix b  /* The second matrix in: a * b */ )
{
  int i, j, k;
  double tmp;
  XMatrix ab;

  tf_enter("Xmatrix_mult");

  ab = XMatrix_new();

  for ( i = 0; i < DIM4; i++ )
    for ( j = 0; j < DIM4; j++ ) {
      for ( tmp = 0.0, k = 0; k < DIM4; k++ ) 
	tmp += a[i][k] * b[k][j];
      ab[i][j] = tmp;
    }

  tf_leave();

  return ab;
}

/*---------------------------------------------------------------------*/
/*
   PROCEDURE
   Xmatrix_identity - return an identity transformation matrix

   RETURNS
   It returns a pointer to a dynamically allocated transformation matrix.
   The caller is responsible for freeing the matrix.
*/
XMatrix 
  Xmatrix_identity( void )
{
  int i;
  XMatrix trans;

  tf_enter( "Xmatrix_identity" );

  trans = XMatrix_new();

  XMatrix_set( trans, 0.0);

  for ( i = 0; i < DIM4; i++ )
    trans[i][i] = 1.0;

  tf_leave();
  return trans;
}
 

/*---------------------------------------------------------------------*/
/*
  PROCEDURE
    Find the inverse of a generalized 4x4 transformation matrix.

  DESCRIPTION
    Calculates the inverse

  RETURNS
    It returns a pointer to a transformation matrix containing the
    inverse of the passed matrix.  If it cannot allocate memory
    for the matrix, it prints a message to stderr and exits (via
    tf_exit).

*/
XMatrix 
Xmatrix_inverse( XMatrix a )
{
 
  double tmp;
  int i, j;
  XMatrix a_inverse = XMatrix_new();

  if (a_inverse == NULL)
  {
    tf_enter("Xmatrix_inverse");
    tf_exit(ExitERR_alloc, "unable to allocate matrix inverse result matrix");
  }

  XMatrix_set( a_inverse, 0.0);

  a_inverse[DIM3][DIM3] = 1.0;

  /*
    Rotation component of the inverse matrix = 
             Transpose of the rotation matrix component.
  */
  for ( i = 0; i < DIM3; i++ )
    for ( j = 0; j < DIM3; j++ )
      a_inverse[i][j] = a[j][i];

  /*
     Translation component of the inverse matrix = 
             - Translation component times the transpose of the rotation matrix
  */
  for ( i = 0; i < DIM3; i++ ) {
    for ( tmp = 0.0, j = 0; j < DIM3; j++ )
	tmp += a[DIM3][j] * a[i][j];
    a_inverse[DIM3][i] = - tmp;
  }

  return a_inverse;
}


/*---------------------------------------------------------------------*/
/*
  PROCEDURE
    Rotation of Co-ordinate Axes.

  DESCRIPTION
    Calculate the 3-D axes rotation, the return pointer of the
    procedure.  The axes are rotated anti-clockwise though an angle 
    theta radians about an axis specified by the argument `axis'. 

  RETURNS
    A pointer to the matrix whose elements are the rotation matrix.

  AUTHOR
    Dan Nguyen
*/
XMatrix 
  Xmatrix_rotation(
	   XYZaxis axis,     /* The axis of rotation. The variable
			       axis can be one of: 0,1,2 to indicate a
			       rotation about the x,y,z axis
			       respectively */
	   double theta  /* Rotation about of angle theta, in radians */
	   )
{

  int i, j;
  double cos_theta=cos(theta);
  double sin_theta=sin(theta);
  XMatrix rotation_matrix = XMatrix_new();


  if ( rotation_matrix == NULL)
  {
    tf_enter("Xmatrix_rotation");
    tf_exit(ExitERR_alloc, "unable to allocate rotated matrix");
  }

  tf_assert(axis == XAXIS || axis == YAXIS || axis == ZAXIS);

  XMatrix_set( rotation_matrix, 0 );

  rotation_matrix[axis][axis] = 1.0;
  rotation_matrix[DIM3][DIM3] = 1.0;

  i = ( axis + 1 ) % 3;
  j = ( i + 1 ) % 3;

  rotation_matrix[i][i] = cos_theta;
  rotation_matrix[j][j] = cos_theta;
  rotation_matrix[i][j] = sin_theta;
  rotation_matrix[j][i] = -sin_theta;

  return rotation_matrix;
}

/*---------------------------------------------------------------------*/
/*
  PROCEDURE
    Translation of Origin.


  DESCRIPTION
    Calculate the 3-D translation matrix `mat'. If the origin of

  RETURNS
    The translational matrix.

  AUTHOR
    Dan Nguyen
*/
XMatrix 
  Xmatrix_translation( double *tran /* amount to translate the origin by */ )
{
  int i;
  XMatrix mat;

  mat = XMatrix_new( );

  if ( mat == NULL )
  {
    tf_enter("Xmatrix_translation");
    tf_exit(ExitERR_alloc,
	    "unable to allocate matrix translation result matrix");
  }

  XMatrix_set( mat, 0.0 );

  for ( i = 0; i < DIM4; i++ )
    mat[i][i] = 1.0;

  for ( i = 0; i < DIM3; i++ )
    mat[DIM3][i] = tran[i];

  return mat;
}




/*

with(linalg);
Let r denote the 3x3 rotation matrix, t the 1x3 translation displacement.

a:=matrix(2,2,[r,0,t,1]);

                                      [ r  0 ]
                                 a := [      ]
                                      [ t  1 ]

assume that matrix b is of the same form as a

ainverse:=matrix(2,2,[rr,0,tt,1]);

                                         [ rr  0 ]
                             ainverse := [       ]
                                         [ tt  1 ]

multiply(a,ainverse);

                                [    r rr    0 ]
                                [              ]
                                [ t rr + tt  1 ]

The matrix above must be equal to the Identity matrix.
Hence, the following conditions must be satisfied:
                  1) r * rr = I 
		  and 
		  2) t * rr + tt = 0

The first condition is satisfied by requiring that:  rr = transpose(r), 
after all r is the rotation matrix.

The second condition is satisfied by solving t*rr + tt = 0, thus tt must
be -t * rr
  
solve(t*rr+tt,tt);

                                     - t rr

                                  [  transpose(r)       0 ]
                      ainverse := [                       ]
                                  [ - t * transpose(r)  1 ]

*/
