/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of aperture
 *
 * aperture is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * aperture is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <linklist/linklist.h>
#include <luasup/luasup.h>

#include "xmatrix.h"

#include "init.h"
#include "lua_register.h"
#include "xstack.h"

#include "transform.h"

#include "common.h"

/* the transformation stack.  it's empty inbetween processing assemblies */
static SLinkList xstack = NULL;       

/*---------------------------------------------------------------------*/
/*
   PROCEDURE
   xstack_init - initialize global structures before assemblies are
   built
*/
void
initialize_xstack( void )
{
  tf_enter("xstack_init");

  /* allocate transformation stack */
  tf_assert(xstack == NULL);

  if ( NULL == (xstack = sll_new( NULL )) )
    tf_die( "error creating transformation matrix stack" );

  if ( sll_insert_head( xstack, Xmatrix_identity( ) ) )
    tf_exit( ExitERR_alloc, "unable to insert global coord matrix in stack");

  tf_leave();
}

void
cleanup_xstack( void )
{
  sll_delete( xstack, (SLL_DELETE) XMatrix_free );
}

static XMatrix
xstack_top( void )
{
  XMatrix stack_top;

  stack_top = sll_head( xstack );
  tf_assert(stack_top != NULL);

  return stack_top;
}

/* return a copy of the top xfrm matrix */
XMatrix
xstack_top_copy( void )
{
  XMatrix stack_top;

  stack_top = sll_head( xstack );
  tf_assert(stack_top != NULL);

  return XMatrix_dup( stack_top );
}


/*

  PROCEDURE
  xstack_dup

  DESCRIPTION
  duplicates the top transformation matrix on the stack

*/
void
xstack_dup( void )
{
  XMatrix copy;

  tf_enter( "xstack_dup" );

  copy = xstack_top_copy( );

  if ( sll_insert_head( xstack, copy ) )
    tf_die( "couldn't insert duplicate node into stack" );

  tf_leave( );
}



/*

  PROCEDURE
  xstack_pop

  DESCRIPTION
  Pop off the top CTM from the transformation stack.  The value of the
  parameter empty_is_fatal determines how it reacts to an empty stack.
  If empty_is_fatal is true, an empty stack causes the program to exit.
  If it is false, an empty stack causes it to return non-zero.  If the
  stack isn't empty, it'll return zero.

  Note that an "empty" stack really has one element on it, the initial
  CTM pushed there by xstack_init.  This should, of course only
  be removed by cleanup_aperture.  Thus, an empty stack ain't empty.

*/
int
xstack_pop(
  int empty_is_fatal	/* if true, an empty stack is a fatal error,
			   else, return non-zero if empty stack */
)
{
  SLLNode head_node;
  XMatrix stack_top;

  tf_enter("xstack_pop");

  head_node = sll_head_node( xstack );
  tf_assert( head_node != NULL );

  if ( sll_count( xstack ) <= 1 )
  {
    if ( empty_is_fatal )
      tf_assert( sll_count(xstack) <= 1 );
    else 
      return 1;
  }

  stack_top = sll_destroy_node( xstack, head_node );
  tf_assert( stack_top != NULL );

  XMatrix_free( stack_top );

  tf_leave();

  return 0;
}


/*
  PROCEDURE
    Update the top element of the stack.

  DESCRIPTION
    Pop the top element of the stack.
    Multiply the element which was popped off with the incoming matrix.
    Store the resulting matrix as the top element of the stack.

  RETURNS
   void
*/
void xstack_update( XMatrix mat )
{
  SLLNode head_node;
  XMatrix stack_top;
  XMatrix top;

  tf_enter("xstack_update");

  /*
     get the matrix from the top of the stack.  leave the node on the
     xstack, we'll just update it later
  */
  head_node = sll_head_node( xstack );
  tf_assert(head_node != NULL);
  
  stack_top = sll_node_get_data( head_node );
  tf_assert(stack_top != NULL);

  /* multiply the old top with the matrix in the argument */
  top = XMatrix_mult( stack_top, mat );			/* do not free */

  /* now update the data in the node at the top of the stack */
  sll_node_put_data( head_node, top );

  /* free the old top of stack matrix */
  XMatrix_free( stack_top );

  tf_leave();
}



/*
  PROCEDURE
  xstack_check

  DESCRIPTION
  This routine is called after the lua script has been executed.
  If there are still entries on the transformation stack, there
  were too many push_CTM's, or not enough pop_CTM's performed
*/
void
xstack_check( void )
{
  /* there should be one CTM left on the stack when all is done */
  if ( sll_count( xstack ) > 1 )
    tf_die( "unbalanced push_CTM/pop_CTM calls: too few pop_CTM calls" );
}

void
  xstack_print( void )
{
  XMatrix mat;
  SLLNode node;
  int i;

  if ( xstack )
    for ( i = 0, node = sll_head_node( xstack );
	 node ;
	 i++, node = sll_next_node( xstack, node ) ) {
      mat = sll_node_get_data( node );
      tf_assert(mat != NULL);
      tf_message( "Transformation stack, i = %d\n", i );
      XMatrix_print( mat );
    }
  return;
}

static void
xstack_print_lua( void )
{
  slua_args( errfunc, "print_TM_stack", NULL );
  xstack_print( );
}

void
xstack_print_top ( void )
{
  if ( xstack )
  {
    tf_message("Transformation matrix:\n");
    XMatrix_print(xstack_top());
  }
  else
    tf_message("no current transformation\n");
}

static void
xstack_print_top_lua ( void )
{
  slua_args( errfunc, "print_CTM", NULL );
  xstack_print_top( );
}


/*---------------------------------------------------------------------*/

void
xstack_register_lua( void )
{
  lua_register( "print_CTM",      xstack_print_top_lua );
  lua_register( "print_TM_stack", xstack_print_lua );
}

#if 0
/*

  PROCEDURE
  pop_CTM_lua

  DESCRIPTION
  A lua callable version of pop_CTM.  It is illegal to call this
  inside of an assembly.

*/
void
pop_CTM_lua( void )
{
  slua_args( errfunc, "pop_CTM", NULL );
  if ( is_in_assembly( ) )
    tf_die( "illegal call of pop_CTM from within an assembly" );

  if ( pop_CTM( 0 ) )
    tf_die( "unbalanced push_CTM/pop_CTM calls: too many pop_CTM calls" );
}

/*

  PROCEDURE
  push_CTM_lua

  DESCRIPTION
  A lua callable version of xstack_dup.  It is illegal to call this
  within an assembly.

*/
void
push_CTM_lua( void )
{
  slua_args( errfunc, "push_CTM", NULL );
  if ( is_in_assembly( ) )
    tf_die( "illegal call of push_CTM from within an assembly" );

  xstack_dup( );
}



#endif

