// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of src
//
// src is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// src is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include "Bp2Fits.h"

#include "FunParamWrap.h"

Bp2Fits::Bp2Fits(
  const string& bpipe, 
  const string& fits, 
  const string& rdb,
  bool debug
  ) throw ( Bp2FitsErr, exception ) 
  : _bpipe(NULL), _dpkt(NULL), _last(NULL), _fun(NULL),
    _rdb(rdb), _data(NULL), _ndpktfs(0), _ncols(0), _nrows(0),
    _precision(20), _debug(debug),
    _doublepass(false), _longstrn(false), _reserved(false),
    _vectors(false) { 

  TraceFct tf( "Bp2Fits::Bp2Fits" );


  // Open BPipe...
  char* bpipefile = new char[bpipe.size( ) + 1];
  strcpy( bpipefile, bpipe.c_str( ) );
  
  _bpipe = bpipe_new( );
  bpcheck( );

  bpipe_input( _bpipe, bpipefile );
  bpcheck( );

  delete [] bpipefile;

  // Open FITS...
  char* fitsfile = new char[fits.size( ) + 1];
  strcpy( fitsfile, fits.c_str( ) );
  char mode[] = { 'w', '\0' };
  if ( ! (_fun = FunOpen( fitsfile, mode, NULL )) )
    throw( Bp2FitsErr( string( "Error opening FITS file `" ) + fits + "'" ) );

  delete [] fitsfile;

}

Bp2Fits::~Bp2Fits( 
  void 
  ) { 

  TraceFct tf( "Bp2Fits::~Bp2Fits" );


  // Clean up the BPipe...
  if ( _bpipe )
    bpipe_delete( _bpipe );

  // Clean up the FITS file...
  if ( _fun )
    FunClose( _fun );

  vector<ColumnInfo*>::iterator cip;
  for ( cip = _ci.begin(); cip != _ci.end(); cip++ )
  {
    ColumnInfo *ci = *cip;

    switch( ci->dtype )
    {
    case 'A' : 
      delete (char*) ci->min;
      delete (char*) ci->max;
      if ( ci->allocate )
	delete [] (char*) (ci->data); 
      break;

    case 'D' : 
      delete (double*) ci->min;
      delete (double*) ci->max;
      if ( ci->allocate )
	delete [] (double*) (ci->data);
      break;

    case 'J' : 
      delete (int*) ci->min;
      delete (int*) ci->max;
      if ( ci->allocate )
	delete [] (int*) (ci->data);
      break;

    case 'V' : 
      delete (unsigned int*) ci->min;
      delete (unsigned int*) ci->max;
      if ( ci->allocate )
	delete [] (unsigned int*) (ci->data);
      break;
    }

    // now the actual structure
    delete ci;
  }

  // cleanup up override info
  std::map<std::string,OverrideInfo*>::iterator orpos;
  for ( orpos = _orinfo.begin() ; orpos != _orinfo.end() ; orpos++ )
    delete orpos->second;


  // get rid of list of pointers to the data
  delete [] _data;

}

void Bp2Fits::bpcheck( 
  void
  ) throw ( Bp2FitsErr, exception ) {
#ifdef DEBUG
  TraceFct tf( "Bp2Fits::check" );
#endif

  if ( BPNOERROR != bpipe_errno ) 
    throw( Bp2FitsErr( bpipe_strerror( bpipe_errno ) ) );

}


void Bp2Fits::writeTable( 
  void 
  ) throw ( Bp2FitsErr, exception ) {

  TraceFct tf( "Bp2Fits::writeTable" );


  // Read the RDB translation file...
  readOverRide( );   

  // Get the column names from the data packets...
  createTable( );


  // Write the Primary HDU...
  writeHDU_comments( true );	// comments
  writeHDU_longstrn( true );	// longstrn

  // Write the Table HDU...
  writeHDU_comments( );		// comments
  writeHDU_longstrn(  );	// longstrn
  writeHDU_hdrf( );		// header fields
  writeHDU_colinfo( );		// extra column info

  // Add columns to the FITS object...
  addColumns( );

  // Read bpipe and write FITS...
  readwriteData( );

  // Rewrite the HDU...
  rewriteHDU( );
}



void Bp2Fits::createTable( 
  void
  ) throw ( Bp2FitsErr, exception ) {
#ifdef DEBUG
  TraceFct tf( "Bp2Fits::createTable" );
#endif

  _dpkt = bpipe_map_alloc( _bpipe, 1, NULL );
  bpcheck();

  bpipe_read_dpkts( _bpipe, _dpkt, 1 );

  /* HUH???
     if ( BPEIOERR == bpipe_errno )
     bpipe_errno = BPNOERROR;
     else 
  */

  bpcheck( );

  DpktField* dpktf;

  // Get all dpktf names...
  _last = NULL;
  while ( (dpktf = bpipe_dpktf_next( _bpipe, &_last )) ) {
    _ifields.insert( _ifields.end( ), bpipe_dpktf_name( dpktf ) );
    bpcheck( );
  }

  // Compare against the requested fields...
  suplib::colselect( _ifields, _add_exact_dpktfs, _add_regex_dpktfs, _del_exact_dpktfs, _del_regex_dpktfs, _ofields );

  if ( 0 == _ofields.size() )
    throw( Exception("no data packet fields selected?" ) );

  // Get all the data packet field names and convert them to FITS columns...
  _last = NULL;
  while ( (dpktf = bpipe_dpktf_next( _bpipe, &_last)) ) {
    const char* name  = bpipe_dpktf_name( dpktf );
    bpcheck( );

    if ( _ofields.end( ) != find( _ofields.begin( ), _ofields.end( ), name ) ) {

      convertDpktf2Col( name, dpktf );

      _ndpktfs++;
    }
  }


  // now, generate list of pointers to the data. this is passed to
  // funtools
  _data = new void* [_ncols];
  for ( vector<ColumnInfo*>::size_type idx = 0; idx < _ci.size() ; idx++ )
    _data[idx] = _ci[idx]->data;
}

void Bp2Fits::convertDpktf2Col(
  const char* bpname,
  DpktField* dpktf
  ) throw ( Bp2FitsErr, exception ) {

  TraceFct tf( "Bp2Fits::convertDpktf2Col" );


  string none[] = { "" };
  string xyz[]  = { "x", "y", "z" };
  string ri[]   = { "r", "i" };
  string qri[]  = { "q1_r", "q1_i", "q2_r", "q2_i", "q3_r", "q3_i" };

  BPDataType bptype  = bpipe_dpktf_type( dpktf );
  bpcheck( );

  void* bpdata = bpipe_dpktf_data( dpktf, _dpkt );
  bpcheck( );

  switch ( bptype ) {

  case BPDType_char:
    convertDpktf2ColTYPE( bpname, 'A', 'A', DT_char, dpktf, (char *)bpdata, 1, none );
    break;

  case BPDType_double:
    convertDpktf2ColTYPE( bpname, 'D', 'D', DT_double, dpktf, (double *)bpdata, 1, none );
    break;

  case BPDType_int:
    convertDpktf2ColTYPE( bpname, 'J', 'I', DT_int, dpktf, (int *)bpdata, 1, none );
    break;

  case BPDType_uint:
    ParamPut( _fun, "TSCAL", _ncols+1, 1, "", 1 );
    ParamPut( _fun, "TZERO",  _ncols+1, (double) 2147483648U, "", 1, 10 );
    convertDpktf2ColTYPE( bpname, 'J', 'I', DT_uint, dpktf, (unsigned int *)bpdata, 1, none );
    break;

  case BPDType_DVector2:
    convertDpktf2ColTYPE( bpname, 'D', 'D', DT_double, dpktf, (double *)bpdata, 2, xyz );
    break;

  case BPDType_DVector3:
    convertDpktf2ColTYPE( bpname, 'D', 'D', DT_double, dpktf, (double *)bpdata, 3, xyz );
    break;

  case BPDType_IVector2:
    convertDpktf2ColTYPE( bpname, 'J', 'I', DT_int, dpktf, (int *)bpdata, 2, xyz );
    break;

  case BPDType_IVector3:
    convertDpktf2ColTYPE( bpname, 'J', 'I', DT_int, dpktf, (int *)bpdata, 3, xyz );
    break;

  case BPDType_UIVector2:
    convertDpktf2ColTYPE( bpname, 'J', 'I', DT_int, dpktf, (unsigned int *)bpdata, 2, xyz );
    break;

  case BPDType_UIVector3:
    convertDpktf2ColTYPE( bpname, 'J', 'I', DT_uint, dpktf, (unsigned int *)bpdata, 3, xyz );
    break;

  case BPDType_DComplex:
    convertDpktf2ColTYPE( bpname, 'D', 'D', DT_double, dpktf, (double *)bpdata, 2, ri );
    break;

  case BPDType_DCVector2:
    convertDpktf2ColTYPE( bpname, 'D', 'D', DT_double, dpktf, (double *)bpdata, 4, qri );
    break;

  case BPDType_DCVector3:
    convertDpktf2ColTYPE( bpname, 'D', 'D', DT_double, dpktf, (double *)bpdata, 6, qri );
    break;

  default:
    throw( Bp2FitsErr( string( "Unhandled datatype `" ) + bpipe_datatype_name( bptype )+ "'" ) );
  }
}



template<typename Type>
void Bp2Fits::convertDpktf2ColTYPE( 
  const char* bpname,
  const char fitsform,
  const char fitsdisp,
  DataType cxxdt,
  DpktField* dpktf,
  Type* bpdata,
  const size_t vecsize,	// number of members in the dpktf structure
  string sfx[]
  ) {

  TraceFct tf( "Bp2Fits::convertDpktf2ColTYPE" );

  OverrideInfo* override = NULL;

  if ( _orinfo.end() != _orinfo.find( bpname ) )
    override = _orinfo.find( bpname )->second;

  // default name
  string name = bpname;

  // if this is a vector and we're splitting it out into
  // separate columns, add an underscore.
  if ( vecsize > 1 && ! _vectors )
    name += '_';

  // process column name override information (from fitsfield column
  // in options RDB table
  if ( override ) { 

    vector<string> tokens;
    suplib::tok( tokens, override->fitsfield, ",", false );

    // if there's anything there, use it
    if ( tokens.size( ) ) {

      // override the name if the field isn't blank
      if ( tokens[0].size() )
      {
	// if the name token is the character '!', the vector column
	// names have no prefix, just the member names.
	if ( "!" == tokens[0] )
	  name.erase( );

	else
	  name = tokens[0];
      }


      for ( string::size_type jdx = 1; jdx < tokens.size(); jdx++ )
      {

	if ( "!" == tokens[jdx] )
	  sfx[jdx-1].erase();

	else if ( tokens[jdx].size() )
	  sfx[jdx-1] = tokens[jdx];
      }

    } 
  }

  // the number of columns that this dpktf is split into
  size_t ncols = _vectors ? 1 : vecsize;

  // the shape of the data packet field
  BPMatrix* bpmatrix = bpipe_dpktf_matrix( dpktf, BPDSite_CORE, NULL );
  bpcheck( ); 


  // we only need to allocate memory if all of the following are true:
  //   1) it's a real vector 
  //   2) it's being split across multiple columns
  //   3) there is more than one member in the dpktf structure
  bool allocate = ncols > 1 && vecsize > 1 && bpmatrix->ne > 1;

  // the number of structure members stored in these columns
  size_t nelem = bpmatrix->ne * (_vectors ? vecsize : 1);

  // the byte offset between members. if we're splitting across
  // columns, then we're striding between vectors, not within.
  // in the latter case, we assume the gap between vectors is the same
  // as between elements in the vector.
  size_t stride = _vectors ? sizeof(Type) : bpipe_datatype_size( bpipe_dpktf_type( dpktf ) );

  for ( size_t jdx = 0; jdx < ncols ; jdx++) {

    stringstream ss;
    ColumnInfo *ci = new ColumnInfo;

    ci->cxxdt = cxxdt;
    ci->allocate = allocate;
    ci->nelem = nelem;
    ci->stride = stride;

    ci->scale = "1";
    ci->zero  = "0";
    ci->mode  = "w";

    if ( ci->allocate )
    {
      // where we have to copy the data
      ci->data = (void*) new Type[nelem];

      // the start of the data for this column. 
      ci->src  = (void*) ( bpdata + jdx );
    }

    // if we're not allocating data, we may be splitting a single
    // vector across multiple columns.  set the data pointer to the
    // correct location in the dpktf to grab the appropriate vector
    // element. note that we've made the assumption
    // that all elements of a vector are of the same type, which is
    // what has been done so far everywhere.
    // 
    else
    {
      // no copying; dest == src
      ci->src = ci->data = (void*) ( bpdata + jdx );
    }

    // these will hold the minimum and maximum values of the field
    // initialize them to the first value of the field 
    Type* min = new Type;
    Type* max = new Type;
    *min = *max = bpdata[jdx];
    
    ci->min = (void*) min;
    ci->max = (void*) max;

    ci->dtype = fitsform;
    ss.str("");
    ss << '@' << nelem << fitsform;
    ci->form = ss.str();

    ci->type = name;
    if ( ! _vectors ) ci->type += sfx[jdx];

    ss.str("");
    switch( fitsdisp )
    {
    case 'D' : ss << fitsdisp << _precision + 4 << '.' << _precision - 1; break;
    default:   ss << fitsdisp << _precision; break;
    }
    ci->disp = ss.str();


    // generate TDIM; looks like (a,b,c)
    ss.str("");
    ss << '(' ;

    if ( _vectors )
      ss << vecsize << ',';

    for ( size_t idx = 0; idx < bpmatrix->nd ; idx++ )
      ss << bpmatrix->extent[idx] << ',';

    ci->dim = ss.str();
    ci->dim.replace(ci->dim.size()-1, 1, ")" );

    if ( override ) {

      // these fields are never generated by us, so just use the
      // override, as it'll be empty if not specified.
      
      ci->unit = override->unit;
      ci->null = override->null;
      ci->lmin = override->lmin;
      ci->lmax = override->lmax;

      ci->dmin = override->dmin;
      ci->dmax = override->dmax;

      // these are generated above, so make sure to override only
      // if it was specified.
      if ( override->scale.size() ) ci->scale = override->scale;
      if ( override->zero.size()  ) ci->zero = override->zero;
      if ( override->disp.size()  ) ci->disp = override->disp;

    }

    // add the new column
    ci->offset = _ncols * sizeof(void*);
    _ci.push_back( ci );
    _ncols++;
  }


  bpipe_matrix_delete( bpmatrix );
  bpcheck( );
}

