#ifndef Bp2Fits_h
#define Bp2Fits_h

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of src
//
// src is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// src is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <cstring>
#include <map>
#include <pwd.h>
#include <set>
#include <string>
#include <vector>
#include <strings.h>
#include <Exception/Exception.h>
#include <bpipe/bpipe.h>
#include <bpipe/datatypes.h>
#include <funtools.h>
#include <suplibxx/colselect.h>
#include <suplibxx/str.h>

#include <rdbxx/RDB.h>
#include <tracefctxx/TraceFct.h>
#include "Bp2FitsErr.h"

using namespace std;

typedef enum
{
  DT_int,
  DT_uint,
  DT_double,
  DT_char
} DataType;


typedef struct
{
  string  mode;
  int    offset;

  // this is where the data originate in the bpipe core image
  void*  src;

  // this is where Funtools will find the data. it may be the
  // same as src, or may point at allocated space if the data are in a
  // matrix
  void*  data;

  // the minimum and maximum values in the column.  these point at
  // allocated space.
  void*  min;
  void*  max;

  // C++ datatype
  DataType cxxdt;

  bool   allocate;

  // FITS data type
  char   dtype;

  string scale;
  string  zero;

  string  type;
  string  form;
  string  unit;
  string  null;
  string  disp;
  string  dim;
  string  lmin;
  string  lmax;
  string  dmin;
  string  dmax;

  size_t  nelem;
  size_t  stride;
} ColumnInfo;

typedef struct
{
  string fitsfield;
  string unit;
  string null;
  string scale;
  string zero;
  string disp;
  string lmin;
  string lmax;
  string dmin;
  string dmax;
} OverrideInfo;

class Bp2Fits {
  
  template<typename Type>
    void convertDpktf2ColTYPE( 
      const char* bpname,
      const char fitsform,
      const char fitsdisp,
      DataType cxxdt,
      DpktField* dpktf,
      Type* bpdata,
      const size_t vecsize,
      string sfx[]
      );

  template<class Type>
    friend void transferDataTYPE( Bp2Fits& bp2fits, BPMatrix* bpmatrix, Type* bpdata, size_t vecsize, size_t elemsize );
  
 public:
  Bp2Fits( const string& fits, const string& bpipe, const string& rdb, bool debug=false ) throw ( Bp2FitsErr, exception );
  ~Bp2Fits( void );

  void writeTable( void ) throw ( Bp2FitsErr, exception );
  void rewriteHDU( void ) throw ( Bp2FitsErr, exception );
  
  void debug(      bool debug      ) { _debug      = debug;      };
  void doublepass( bool doublepass ) { _doublepass = doublepass; };
  void longstrn(   bool longstrn   ) { _longstrn   = longstrn;   };
  void reserved(   bool reserved   ) { _reserved   = reserved;   };
  void vectors(    bool vectors    ) { _vectors    = vectors;    };
  void verbose(    bool verbose    ) { _verbose    = verbose;    };

  void add_exact_dpktfs( const string& dpktfs  ) { suplib::tok( _add_exact_dpktfs, dpktfs, ", "); };
  void add_regex_dpktfs( const string& dpktfs  ) { suplib::tok( _add_regex_dpktfs, dpktfs, ", "); };
  void del_exact_dpktfs( const string& dpktfs  ) { suplib::tok( _del_exact_dpktfs, dpktfs, ", "); };
  void del_regex_dpktfs( const string& dpktfs  ) { suplib::tok( _del_regex_dpktfs, dpktfs, ", "); };
  
 private:
  void readOverRide( void ) throw ( Bp2FitsErr, exception );
  void addColumns( void ) throw( Bp2FitsErr, exception );
  void readwriteData( void ) throw( Bp2FitsErr, exception );
  void writeHDU_longstrn( bool primary=false );
  void writeHDU_colinfo( void ) throw ( Bp2FitsErr, exception );
  void writeHDU_hdrf( void ) throw ( Bp2FitsErr, exception );
  void writeHDU_comments( bool primary=false ) throw ( Bp2FitsErr, exception );

  char* makeKeyWord( const string& keyword );
  char* makeKeyValue( const string& keyword, const string& value, size_t& offset, size_t maxsize, bool contchar );
  void createTable( void ) throw ( Bp2FitsErr, exception );
  void convertDpktf2Col(const char* bpname, DpktField* dpktf) throw ( Bp2FitsErr, exception );

  void bpcheck( void ) throw ( Bp2FitsErr, exception );
  
  BPipe* _bpipe;
  void* _dpkt;
  void* _last;
  
  Fun _fun;
  vector<ColumnInfo*> _ci;
  map<string,OverrideInfo*> _orinfo;

  vector<string> _ifields;
  vector<string> _ofields;
  
  RDB _rdb;

  void** _data;
  size_t _ndpktfs;
  size_t _ncols;
  size_t _nrows;

  size_t _precision;
  
  bool _debug;
  bool _doublepass;
  bool _longstrn;
  bool _reserved;
  bool _vectors;
  bool _verbose;

  vector<string> _add_exact_dpktfs;
  vector<string> _add_regex_dpktfs;
  vector<string> _del_exact_dpktfs;
  vector<string> _del_regex_dpktfs;
  
};

#endif
