/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bp2rdb
 *
 * bp2rdb is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bp2rdb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <string.h>

#include <bpipe/bpipe.h>
#include <linklist/linklist.h>
#include <suplib/str.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"

static size_t  parse_matrix_element( char* name, BPMatrix *bpm, char* element );
static void    parse_field_spec( char *spec, char** name, char** matrix );

static int
key_node_compare(const void *key, const void *fi)
{
  return strcmp( key, bpipe_dpktf_name( ((FieldInfo*)fi)->dpktf) );
}

static int
move_fields( void* data, void* udata )
{
  FieldInfo* fi = data;
  FieldInfo*** fip = udata;

  /* copy the innards of the FieldInfo struct over */
  **fip = fi;

  (*fip)++;

  return 0;
}


/*
  PROCEDURE
  get_output_dpktfs - get a list of data packet fields to output
  
  DESCRIPTION

  This routine is passed a list of requested field names (which may be
  additive or subtractive) and returns a linked list of the resultant
  data packets.  If any of the passed fields is an additive field, the
  initial list of fields is empty.  Subtractive fields are indicated
  by having a @code{-} as their first character.  Subtractive fields
  which do not exist in the bpipe stream are silently ignored.

  RETURNS
  It returns a SLinkList containing pointers to the data packet fields
  to output.

*/
FieldInfo**
get_output_dpktfs(
  BPipe* bpipe,		/* bpipe to convert */
  int nfields,		/* number of requested fields */
  char* requested_field[]	/* the requested fields */
  )
{
  int ii;
  int all_fields = 0;
  int additive = 0;
  FieldInfo** field_arr = NULL;
  SLinkList  output_fields;

  tf_enter( "get_output_dpktfs" );


  output_fields = sll_new( SLL_NULL_CMP );
  tf_assert( NULL != output_fields );

  /* figure out if the output list starts with the full set of fields */
  for ( ii = 0; ii < nfields ; ii++ )
  {
    if ( 0 == strcmp( "%all", requested_field[ii] ) )
    {
      all_fields = 1;
      break;
    }

    if ( requested_field[ii][0] != '-' )
      additive++;
  }

  /* if there are no additive fields, or %all was specified,
     use all fields */
  if ( !additive )
    all_fields = 1;


  /* if all_fields, get current set of fields */
  if ( all_fields )
  {
    DpktField *dpktf;
    void *last = NULL;

    while( (dpktf = bpipe_dpktf_next( bpipe, &last )) )
    {
      FieldInfo *fi = new_FieldInfo( dpktf );
      tf_assert( NULL != fi );

      sll_insert_tail( output_fields, fi );
    }

  }

  for ( ii = 0; ii < nfields ; ii++ )
  {
    char *name;
    char *element;
    char *buf;

    if ( 0 == strcmp( "%all", requested_field[ii] ) )
      continue;

    buf = str_dup( requested_field[ii] );
    tf_assert( NULL != buf );
    parse_field_spec( buf, &name, &element );

    /* subtractive */
    if ( name[0] == '-' )
    {
      FieldInfo *fi = sll_destroy( output_fields, ++name, key_node_compare );

      /* silently ignore non-existent fields.  either they're not in
	 the output list, or they don't exist at all.
      */
      if ( fi )
	free_FieldInfo( fi );
    }

    /* additive */
    else
    {
      FieldInfo *fi;

      /* it is possible that this field is already in the output
	 list.  this can come about either because the user specified
	 it multiple times, or because %all was specified, and all of
	 the fields are in the list.  it might be considered an error
	 for the user to specify a field multiple times, but
	 distinguishing that from %all is too much of a pain.  so,
	 simply override any previous matrix element specifications --
	 the last one specified wins.
      */
      
      /* is this field already in the output list? */
      fi = sll_search( output_fields, name, key_node_compare );

      /* first time */
      if ( !fi )
      {
	DpktField *dpktf = bpipe_dpktf( bpipe, name );

	if ( NULL == dpktf )
	  tf_die( "%s: non-existent field", name );

	fi = new_FieldInfo( dpktf );
	tf_assert( NULL != fi );

	sll_insert_tail( output_fields, fi );
      }

      /* update its (byte) offset */
      if ( NULL != element )
	fi->offset = fi->size * parse_matrix_element( name, fi->matrix, element );
    }

    free( buf );
  }


  /* Finally, turn this into a NULL terminated array to optimize access */
  if ( sll_count( output_fields ) )
  {
    FieldInfo **fi = field_arr = malloc( sizeof(FieldInfo*) *
					 ( 1 + sll_count( output_fields ) ) );
    tf_assert( NULL != field_arr );

    /* move fieldinfo data to new array */
    sll_utraverse( output_fields, move_fields, &fi );
    *fi = NULL;
  }

  /* delete the list; the data have been moved to field_arr */
  sll_delete( output_fields, SLL_NULL_DELETE );

  tf_leave();
  return field_arr;
}


/*
  PROCEDURE
  parse_field_spec - split out a field name and matrix element spec
  in a string

  DESCRIPTION
  This routine is passed a buffer containing a string of the types
  @code{name} or @code{name[matrix]}.  It returns pointers to the
  beginning of @code{name} and @code{matrix}.  It modifies the passed
  buffer

*/

static
void parse_field_spec( char *spec, char** name, char** matrix )
{
  char* bra;

  *name = spec;
  *matrix = NULL;

  bra = strchr( spec, '[' );
  if ( bra )
  {
    char *ket;

    bra++;

    ket = bra + strlen(bra) - 1;
    if ( *ket != ']' )
      tf_die( "%s: malformed subscript", spec);
    *ket = '\0';

    /* seal off name. don't do it earlier so can use
       spec in the above error message */
    bra[-1] = '\0';
    *matrix = bra;
  }

}


static
char *matrix_range( BPMatrix *matrix, char *buf )
{
  int ii;
  char tbuf[256];

  strcpy( buf, "[" );

  for ( ii = 0 ; ii < matrix->nd ; ii++ )
    strcat( buf, "0," );
  buf[strlen(buf)-1] = ']';

  strcat( buf, " -> [");

  for ( ii = 0 ; ii < matrix->nd ; ii++ )
  {
    sprintf( tbuf, "%lu,", matrix->extent[ii] - 1 );
    strcat( buf, tbuf );
  }
  buf[strlen(buf)-1] = ']';

  return buf;
}


static size_t
parse_matrix_element( char* name, BPMatrix *bpm, char* element )
{
  size_t offset = 0;
  int ntoks, tok;
  char **strings;

  char tbuf[1024];

  char *err_buf = str_join( NULL, 4, name, "[", element, "]" );


  tf_enter( "parse_field" );


  ntoks = str_tokenize( element, ",", &strings, 0, 0 );

  if (0 == ntoks )
    tf_die( "%s: malformed matrix element specification", err_buf );

  if ( ntoks != bpm->nd )
    tf_die( "%s: incorrect number of dimensions;  expected %d",
	    err_buf, bpm->nd );

  /* dimensions are given in least significant to most significant
     the same as the internal representation of bpipe.
  */

  for ( offset = 0, tok = ntoks-1 ; tok >= 0; tok-- )
  {
    char *ptr;
    int l;

    /* empty strings are not appreciated */
    if ( strings[tok][0] == '\0' )
      tf_die( "%s: malformed subscript", err_buf );

    /* not an integer? */
    l = strtol( strings[tok], &ptr, 10 );
    if ( ptr[0] != '\0' )
      tf_die( "%s: malformed subscript", err_buf );

    /* bad subscript */
    if ( l < 0 || l >= bpm->extent[tok] )
      tf_die( "%s: subscript out of range. allowed range: %s\n",
	      err_buf, matrix_range( bpm, tbuf ) );
      
    if ( tok == 0 ) 
      offset += l;
    else
      offset = bpm->extent[tok-1] * ( offset + l );

  }


  free(err_buf);

  return offset;
}
