/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>

#include <rbtree/rbtree.h>

#include "bpipep.h"


/*

  PROCEDURE
  bpipe_new - Allocate and initialize a binary pipe structure.

  DESCRIPTION
  @code{bpipe_new} allocates and initializes a binary pipe structure.
  It sets up the linked-lists and binary trees for the header
  parameter and data packet field definitions.

  RETURNS
  It returns a pointer to a dynamically allocated binary pipe structure.
  It prints a message to stderr and exits upon error.

  Upon failure it returns @code{NULL}, and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

   @item BPENOMEM
     a memory allocation failed

   @end table

*/
BPipe *
bpipe_new(void)
{
  BPipe *bpipe;

  /* allocate binary pipe structure */
  if ( NULL == ( bpipe = (BPipe *) malloc( sizeof(BPipe) ) ) ||
       NULL == ( bpipe->output = sll_new( SLL_NULL_CMP ) )   ||
       bpipe_hdr_setup( bpipe ) || bpipe_dpkt_setup( bpipe )
     )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  bpipe->input = NULL;

  bpipe->output_arr = NULL;

  bpipe->xmap = NULL;
  bpipe->nmap = 0;

  bpipe->dpkt = NULL;

  return bpipe;
}


/*

  PROCEDURE
  outchannel_close_delete - Close and delete an output channel.

  DESCRIPTION
  This is an action routine passed to the @code{linklist} package for
  the purpose of closing and deleting a list of output channels

*/

static void
outchannel_close_delete( void *node )
{
  BPipeOutput *bpo = (BPipeOutput *) node;

  bpipe_iochannel_close( bpo->channel ) ;
  bpipe_output_delete( bpo );
}


/*

  PROCEDURE
  bpipe_delete - Close the channels attached to a binary pipe and destroy
  the pipe.

  DESCRIPTION
  This routine closes the input and output channels and deletes the
  associated data structures.  It deletes the pipe's header parameter
  and data packet field binary trees.  Finally, it frees the memory
  associated with the binary pipe structure itself.

*/
void
bpipe_delete(
  BPipe *bpipe		/* the binary pipe to close */
)
{

  /* close the input channel and delete it */
  bpipe_iochannel_close( bpipe->input );
  bpipe_iochannel_delete( bpipe->input );

  /* close off the output channels and delete them */
  sll_delete( bpipe->output, outchannel_close_delete );

  if ( bpipe->output_arr ) free( bpipe->output_arr );
  if ( bpipe->xmap ) free( bpipe->xmap );
  if ( bpipe->dpkt) free( bpipe->dpkt );

  bpipe_hdr_cleanup ( bpipe );
  bpipe_dpkt_cleanup( bpipe );

  free( bpipe );
}
