#ifndef BPIPEP_H
#define BPIPEP_H

/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdio.h>
#include <limits.h>

#include <rbtree/rbtree.h>
#include <linklist/linklist.h>

/* magic to allow bpipe.h and bpipep.h to suck in the same header file easily*/
typedef struct BPipe BPipe;
typedef struct BPMatrixMap BPMatrixMap;

#include <bpipe/common.h>

typedef struct DpktFieldPrivate DpktField;

struct BPMatrixMap
{
  size_t *	src_off;	/* offset into source matrix */
  size_t *	dst_off;	/* offset into dest matrix */
  size_t *	extent;		/* extent of source matrix */
};


/* an abstraction of data sources and sinks */
typedef struct IOchannel
{
  char *	path;		/* path to channel */
  FILE *	dsc;		/* channel descriptor */
} IOchannel;

/* all the info for a header data field */
typedef struct
{
  char *	name;		/* field name */
  size_t	index;		/* the index of this field */
  size_t	hi_index;	/* the highest index used */

  int		new;		/* true if this field was not in the
				   input stream */

  BPDataType	type;		/* data type */
  void *	data;		/* pointer to data */
  int		alloc_data;	/* true if data was allocated by bpipe */

  BPMatrix *	matrix;		/* storage description */

  SLinkList	same;
} HdrField;

typedef struct BPMatrixOverride
{
  DpktField *	dpktf;
  BPMatrix *	matrix;
  BPMatrixMap *	map;
} BPMatrixOverride;

/* a mapping table between two structure layouts */
typedef struct Xmap
{
  size_t	src;		/* offset of source object */
  size_t	dst;		/* offset of destination object */
  size_t	size;		/* size of field in bytes */
} Xmap;

typedef struct BPipeOutputDpktFInfo
{
  DpktField *	field;
  size_t	output_offset;
  size_t	n;
} BPipeOutputDpktFInfo;

/* output channel specific information */
typedef struct BPipeOutput
{
  IOchannel *	channel;	/* output channel */
  RBTree	dpktf_dels;	/* binary tree of pointers to deleted field
			   	   structures */
  SLinkList	dpktf;		/* included field structures */
  RBTree	matrix;		/* overriding matrix/map specs for
				   specific fields */

  Xmap *	xmap;		/* mapping from core image to output image */
  size_t	nmap;		/* number of entries in xmap */

  size_t	dpkt_size;	/* size of a data packet's output image */
  char *	buf;		/* the output transfer buffer */
} BPipeOutput;

/*

    the main binary pipe structure. contains information on the global
    header and data packet definitions.
*/
struct BPipe
{
  IOchannel *	input;	/* input channel */

  SLinkList	output;	/* linked list of output channel descriptors */

  BPipeOutput **output_arr;	/* array of output channel descriptors.
				   created by bpipe_write_hdr, deallocated
				   by bpipe_delete.  for efficiency
				   in bpipe_write_dpkt_n */

  BPipeOutput **output_max;	/* pointer to last channel descriptor
				   in output_arr.  for efficiency in
				   bpipe_write_dpkt_n */

  SLinkList	hdrf_ll;	/* linked list of header fields, in order of
				   creation */

  RBTree	hdrf_tree;	/* tree containing header fields, sorted
				   by name */

  RBTree	dpktf;		/* tree containing data packet fields, sorted
				   by name */

  size_t	dpktf_index;	/* highest index of data packet field
				   (0 means none).  incremented
				   whenever a data packet field is
				   created either via an input stream
				   or by the application. */

  size_t	dpkt_in_size;	/* size of a data packet's input image */
  size_t	dpkt_core_size;	/* size of a data packet's core image */

  char *	dpkt_inbuf;	/* Input buffer for data packet
				   fields.  allocated by bpipe_map,
				   deallocated by bpipe_dpkt_cleanup */
  size_t	dpkt_inbuf_len;	/* I/O buffer for data packet fields.
				   allocated by bpipe_map, deallocated
				   by bpipe_dpkt_cleanup */

  Xmap *	xmap;		/* transformation map from INPUT to CORE state.
				   created by bpipe_map */
  size_t	nmap;		/* number of entries in xmap */

  /* flags to prevent user stupidity */
  int		hdr_read;	/* true while in bpipe_read_hdr */
  int		hdr_written;	/* bpipe_write_hdr was called */
  int		hdr_mapped;	/* bpipe_map was called */
  int		hdr_dpkts;	/* the number of data packets defined in the
				   input stream */

  void *	dpkt;		/* a pointer to the (optionally) allocated
				   data packet, allocated by bpipe_map_alloc */

};

#include <bpipe/protos.h>

#endif /* BPIPEP_H */
