/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <rbtree/rbtree.h>

#include "bpipep.h"

/* ------------------------------------------------------------------ */

static void dpktf_delete(DpktField *field);
static DpktField *dpktf_new(const char *name, BPDataType type);
static int dpktf_output_destroy(void *data, void *udata);
static int dpktf_name_cmp(const void *name, const void *dpktf);

/* ------------------------------------------------------------------ */


/*
   PROCEDURE
   bpipe_dpkt_setup - initialize dpkt data structure

   DESCRIPTION
   This routine is called by @code{bpipe_new} to initialize the data packet
   structures in a new @code{BPipe}.

   RETURNS
   It returns zero upon success, non-zero upon failure.  It sets
   @code{bpipe_errno} upon failure.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following:
   @table @code

   @item BPENOMEM
     a memory allocation failed

   @end table

*/
int
bpipe_dpkt_setup(
  BPipe *bpipe		/* binary pipe */
)
{

  /* create header field binary tree and linked list */
  if ( NULL == ( bpipe->dpktf = rbtree_new(dpktf_node_name_cmp)) )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }

  bpipe->dpkt_in_size = 0;
  bpipe->dpkt_core_size = 0;

  bpipe->dpktf_index = 0;

  bpipe->dpkt_inbuf = NULL;

  bpipe->hdr_dpkts = 0;

  return 0;
}

/*
   PROCEDURE
   bpipe_dpkt_cleanup - Cleanup @code{BPipe} data packet structures

   DESCRIPTION
   This routine is called by @code{bpipe_delete} to cleanup data
   packet structures in the @code{BPipe}.

*/
void
bpipe_dpkt_cleanup(
  BPipe *bpipe		/* binary pipe */
)
{
  if ( bpipe->dpkt_inbuf )
    free( bpipe->dpkt_inbuf );

  rbtree_delete( bpipe->dpktf, (RBTREE_DELETE) dpktf_delete, PRE_ORDER,
		 LEFT_TO_RIGHT );
}

/*
   PROCEDURE
   bpipe_dpkt_n - return the number of data packet fields

   RETURNS
   The number of data packet fields.

*/
size_t
bpipe_dpkt_n(
  BPipe *bpipe		/* binary pipe */
)
{
  return rbtree_count( bpipe->dpktf );
}

/*

   PROCEDURE
   dpktf_delete - Delete a @code{DpktField} structure.

   DESCRIPTION
   Free the memory associated with a data packet field specification
   structure, including its associated data (name, type, input, core,
   output, input-to-core and core-to-output fields).

   If any pointers are @code{NULL} (including the passed pointer to the data
   packet field specification), they are skipped.

*/
static void
dpktf_delete(
  DpktField *field	/* the field structure to delete */
)
{
  if (field)
  {
    if (field->name)
      free( field->name );

    /* the following delete routines accept NULL arguments */
    bpipe_matrix_delete( field->input );
    bpipe_matrix_delete( field->core );
    bpipe_matrix_delete( field->output );

    bpipe_matrix_map_delete( field->input_to_core );
    bpipe_matrix_map_delete( field->core_to_output );

    free( field );
  }
}


/*

   PROCEDURE
   dpktf_new - Create and initialize a data packet field.

   DESCRIPTION
   This routine creates and initializes a data packet field structure.
   It duplicates the passed field name, and type. They are all assumed
   to be valid.

   RETURNS
   It returns a pointer to a dynamically allocated and initialized
   @code{DpktField} structure.  Upon error it returns @code{NULL} and
   sets @code{bpipe_errno}.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following:
   @table @code

   @item BPENOMEM
     a memory allocation failed

   @end table

*/
static DpktField *
dpktf_new(
  const char *name,		/* name of data packet field */
  BPDataType type	/* storage class of field */
)
{
  DpktField *field = NULL;

  assert( NULL != name );
  assert( BPDType_NOTYPE != type );
  assert( type <  _BPDType_num );

  if ( NULL == ( field = (DpktField *)  malloc(sizeof(DpktField)) ) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  if ( NULL == ( field->name = (char *) malloc(strlen(name) + 1) ) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  strcpy( field->name, name );

  field->type		= type;

  field->input		= NULL;
  field->core		= NULL;
  field->output		= NULL;
  field->input_to_core	= NULL;
  field->core_to_output	= NULL;

  field->input_offset	= 0;
  field->core_offset	= 0;

  field->global_delete	= 0;
  field->output_delete	= 0;

  return field;
}



/*

  PROCEDURE
  bpipe_dpktf - Return an access handle for a data packet field.

  DESCRIPTION
  This routine is used to obtain an access handle for a data packet
  field required by routines which extract data from the data packet
  field structure.

  RETURNS
  It returns a pointer to the data packet field's structure.  It
  returns @code{NULL} if the field doesn't exist.

*/
DpktField *
bpipe_dpktf(
  BPipe *bpipe,		/* the binary pipe with which the field is
			   associated */
  const char *name	/* the name of the field */
)
{
  return rbtree_search(bpipe->dpktf, name, dpktf_name_cmp);
}


/*
  PROCEDURE
  bpipe_dpktf_next - Return a handle to the next core data packet field.

  DESCRIPTION
  This routine is used when traversing the list of data packet fields.
  It requires an external pointer which it uses to keep track of where
  it is in the list.  The first call whould be done with that pointer
  set equal to @code{NULL}.  Note that the @emph{address} of the
  pointer is passed to @code{bpipe_dpktf_next}, not the pointer
  itself.  The value shouldn't be changed between calls to this
  routine.  Additions or deletions should not be made to the list in
  between calls to this routine.

  RETURNS
  It returns a pointer to the data packet field's structure.  This is
  used by the data packet field information extraction routines.  If
  the entire list of data packet fields has been traversed, it returns
  @code{NULL}.

*/
DpktField *
bpipe_dpktf_next(
  BPipe *bpipe,		/* the binary pipe with which the field is
			   associated */
  void **p_last		/* the address of a pointer used to traverse
			   the list */
)
{
  if ( NULL == *p_last )
    *p_last = rbtree_min_node(bpipe->dpktf);
  else
    *p_last = rbtree_next_node(bpipe->dpktf, *p_last, LEFT_TO_RIGHT);

  return ( NULL == *p_last ) ? NULL : rbtree_node_get_data(*p_last);
}


/*

  PROCEDURE
    bpipe_dpktf_type - return the datatype of a data packet field

  RETURNS
    Returns the data type of a data packet field, if it exists.  If it
    doesn't exist, it returns @code{BPDType_NOTYPE}.

*/
BPDataType
bpipe_dpktf_type(
  DpktField *dpktf	/* a pointer to the data packet field */
)
{
  return ( NULL == dpktf ) ? BPDType_NOTYPE : dpktf->type;
}

/*

  PROCEDURE
  bpipe_dpktf_datap - return a pointer to a data packet field's data.

  DESCRIPTION

  This routine returns a pointer to the location in the specified core
  image of the specified data packet field.  It is passed the name of
  the field (rather than a handle to it, as required by
  @code{bpipe_dpktf_data}) and the required data type.  If the passed
  type does not match the actual type in the data packet, an error is
  returned.

  Unlike @code{bpipe_dpktf_data}, @code{bpipe_dpktf_datap} is a
  function, not a C preprocessor macro.

  RETURNS
  It returns a pointer to the data in the data packet.  Upon error it
  returns @code{NULL} and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
    the requested data packet field doesn't exist

  @item BPEBADPIPE
    the data types didn't match

  @end table
*/
void *
bpipe_dpktf_datap(
  BPipe *bpipe,		/* the binary pipe that contains the data packet */
  void *core_image,	/* a pointer to the core image of the data packet */
  const char * name,	/* the name of the field */
  BPDataType type	/* the suspected data type of the field */
)
{
  DpktField *field = bpipe_dpktf( bpipe, name );

  if ( field == NULL )
  {
    bpipe_errno = BPEBADARG;
    return NULL;
  }

  if ( bpipe_dpktf_type( field ) != type )
  {
    bpipe_errno = BPEBADPIPE;
    return NULL;
  }

  return bpipe_dpktf_data( field, core_image );
}


/*

  PROCEDURE
  bpipe_dpktf_name - Retrieve the name of a data packet field.

  RETURNS
  This routine returns a pointer to the name of a data packet field,
  if it exists.  If it doesn't exist, it returns @code{NULL}.  Please note
  that the pointer returned points to the original copy of the field's
  name.  @strong{Do not change this data!}.

*/
const char *
bpipe_dpktf_name(
  DpktField *dpktf	/* a pointer to the data packet field */
)
{
  return ( NULL == dpktf ) ? BPDType_NOTYPE : dpktf->name;
}


/*

   PROCEDURE
   bpipe_dpktf_matrix - retrieve a copy of one of a data packet field's
   matrix descriptions

   DESCRIPTION
   This routine makes a copy of the matrix specification associated with
   the specified image of a data packet field and returns a pointer to it.

   RETURNS
   This routine returns a pointer to a @code{BPMatrix} structure upon
   success.  If an error occurs, it returns @code{NULL} and sets
   @code{bpipe_errno}.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following:
   @table @code

   @item BPEBADARG

   @itemize @bullet
   @item an input matrix was requested and none existed.
   @item the passed site was bogus
   @end itemize

   @item BPENOMEM
     a memory allocation failed

   @end table

*/
BPMatrix *
bpipe_dpktf_matrix(
  DpktField *dpktf,	/* a pointer to the data packet field structure from
			   which to extract the matrix */
  BPDataSite site,	/* the image site the matrix spec refers to */
  BPipeOutput *channel	/* the output channel handle if requesting matrix spec
			   for an output channel */
)
{
  BPMatrixOverride *over;

  switch(site)
  {
  case BPDSite_OUTPUT:
    over = (BPMatrixOverride *)
      sll_search( channel->matrix, dpktf, bpipe_dpktf_matrix_override_cmp );
    if ( over )
      return bpipe_matrix_dup( over->matrix );

    /* FALL THROUGH TO NEXT case */

  case BPDSite_CORE:
    if ( dpktf->core )
      return bpipe_matrix_dup( dpktf->core );

    /* FALL THROUGH TO NEXT case */

  case BPDSite_INPUT:
    if ( dpktf->input )
      return bpipe_matrix_dup( dpktf->input );

    /* FALL THROUGH TO NEXT case */

  default:
      bpipe_errno = BPEBADARG;
  }

  return NULL;
}

/*

  PROCEDURE
  bpipe_dpktf_newP - add a data packet field to a data packet.

  DESCRIPTION
  Adds a new field to either the input or core images of a data packet
  field.  It does not allow duplicate fields to be created.  It
  creates the necessary structures, replicates the user data (name,
  type) and places the parameter at the end of the list of
  fields as well as in the binary table of fields.  If the matrix
  specification is @code{NULL}, a matrix specification for a scalar is
  constructed.

  @code{bpipe_dpktf_newP} uses the passed matrix structure as is, and
  directs the @code{BPipe} cleanup code to take responsibility for
  freeing it.  The calling procedure should @emph{not} free it, nor
  should it pass it to any other @code{BPipe} routine which takes
  responsibility for freeing it.

  This is a private routine, and should be called only be bpipe
  internal routines.  The public routine, @code{bpipe_dpktf_add},
  calls this with the site set to @code{DataSite_CORE}.

  RETURNS
  It returns @samp{0} upon success.  If a field with the same name
  already exists, it returns @samp{1}.  It returns @samp{-1} upon
  error, and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
   @table @code

   @item BPEBADARG
     one of the arguments was bogus

   @item BPENOMEM
     a memory allocation failed

   @end table

*/
int
bpipe_dpktf_newP(
  BPipe *bpipe,		/* the binary pipe to which to add the data packet
			   field */
  const char *name,	/* the name of the new field */
  BPDataType type,	/* the new field's data type */
  BPMatrix *matrix,	/* a description of the storage requirements of the
			   field */
  size_t offset,	/* offset from start of input object */
  BPDataSite site	/* the data site where the field is to be located */
)
{
  DpktField *dpktf;

  if (
       NULL == name || BPDType_NOTYPE == type ||
      (site != BPDSite_INPUT && site != BPDSite_CORE)
      )
  {
    bpipe_errno = BPEBADARG;
    return -1;
  }

  /* check if a data packet with the same name already exists */
  if ( rbtree_search(bpipe->dpktf, name, dpktf_name_cmp) )
    return 1;

  /* check matrix for correctness. a NULL matrix is ok */
  if ( bpipe_matrix_verify( matrix ) )
  {
    bpipe_errno = BPEBADARG;
    return -1;
  }

  if ( NULL == ( dpktf = dpktf_new( name, type ) ) )
    return -1;

  matrix = ( NULL == matrix ) ? bpipe_matrix_new_va( (size_t) 1, (size_t) 1 ) : matrix;

  if ( NULL == matrix )
  {
    bpipe_errno = BPENOMEM;
    return -1;
  }

  if ( BPDSite_INPUT == site )
    dpktf->input = matrix;
  else
    dpktf->core  = matrix;

  dpktf->input_offset = offset;

  dpktf->index = bpipe->dpktf_index++;

  dpktf->bpipe = bpipe;

  /* insert it into the data packet trees */
  if ( rbtree_insert( bpipe->dpktf, dpktf ) )
  {
    bpipe_errno = BPENOMEM;
    return -1;
  }

  return 0;
}


/*

  PROCEDURE
  bpipe_dpktf_add - Add a data packet field to the core image of a
  data packet.

  DESCRIPTION
  This routine adds a new field to the definition of the core image of
  a data packet.  If no similarly named field exists, a new field with
  the given name is created and placed at the end of the list of
  current fields.  The name data is duplicated.  If the matrix
  specification is @code{NULL}, a matrix specification for a scalar is
  constructed.

  If the calling procedure constructs a matrix specification (a
  @code{BPMatrix} structure), @code{bpipe_dpktf_add} uses this
  structure as is and directs the @code{BPipe} cleanup code to take
  responsibility for freeing it.  The calling procedure should
  @emph{not} free it, nor should it pass it to any other @code{BPipe}
  routine which takes responsibility for freeing it.  It should
  not alter it in any way after passing it to @code{bpipe_dpktf_add}.

  RETURNS
  It returns @samp{0} upon success.  If a field with the same name
  already exists, it returns @samp{1}.  It returns @samp{-1} upon
  error, and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:

  @table @code

  @item BPEBADARG

  @itemize @bullet
  @item the input matrix was bogus
  @item the passed name was empty or @code{NULL} or had illegal
        characters in it.
  @end itemize

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
int
bpipe_dpktf_add(
  BPipe *bpipe,		/* the binary pipe to which to add the new
			   data packet field */
  const char *name,	/* the name of the new field */
  BPDataType type,	/* the data type of the new field */
  BPMatrix *matrix	/* the field's matrix description */
)
{
  if ( bpipe_check_field_name( name ) )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }
  return bpipe_dpktf_newP(bpipe, name, type, matrix, (size_t) 0, BPDSite_CORE);
}


/*

  PROCEDURE
  bpipe_dpktf_array_add - Add a one-dimensional array field to the
  core image of a data packet.

  DESCRIPTION
  This routine adds a one dimensional array field to the definition of
  the core image of a data packet.  If no similarly named field
  exists, a new field with the given name is created and placed at the
  end of the list of current fields.  The name data is duplicated.

  This routine simplifies the addition of one-dimensional arrays.
  Normally one must construct a matrix specification and call
  @code{bpipe_dpktf_add}.  This routine takes care of that.

  RETURNS
  It returns @samp{0} upon success.  If a field with the same name
  already exists, it returns @samp{1}.  It returns @samp{-1} upon
  error, and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:

  @table @code

  @item BPEBADARG

  @itemize @bullet
  @item the passed name was empty or @code{NULL} or had illegal
        characters in it.
  @end itemize

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
int
bpipe_dpktf_array_add(
  BPipe *bpipe,		/* the binary pipe to which to add the new
			   data packet field */
  const char *name,	/* the name of the new field */
  BPDataType type,	/* the data type of the new field */
  size_t extent		/* the extent of the array */

)
{
  BPMatrix *matrix;

  if ( bpipe_check_field_name( name ) )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }
  if ( NULL == (matrix = bpipe_matrix_new_va((size_t) 1, extent )) )
    return 1;
  return bpipe_dpktf_newP(bpipe, name, type, matrix, (size_t) 0, BPDSite_CORE);
}


/*

  PROCEDURE
  bpipe_dpktf_delete - Delete a data packet field.

  DESCRIPTION
  This routine deletes a field from the definition of either the core
  or output images of a data packet.  The @code{site} argument
  specifies the data site from which to delete the field, and should
  be either of the constants @code{BPDSite_CORE} or
  @code{BPDSite_OUTPUT} (denoting the core and output images,
  respectively).  In the latter case, it is deleted from the specified
  output channel.  If the output channel is the constant
  @code{BPOutputChannel_ALL} (or @code{NULL}), the field is deleted from
  all output channels.

  Note that once a field is deleted from the core image, it's gone!
  If a field of the same name is added, it's a completely new field,
  not the one that was in the input image.

  RETURNS
  It returns @samp{0} if the deletion was successful, @samp{1} if the
  field doesn't exist.  Upon error it returns @samp{-1} and
  sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
  @itemize @bullet
  @item the matrix specification was bogus
  @item the header was already been written
  @item the @code{BPipe} has been mapped
  @end itemize

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
int
bpipe_dpktf_delete(
  BPipe *bpipe,		/* the binary pipe with which the data packet is
			   associated */
  const char *name,	/* the name of the field to delete */
  BPDataSite site,	/* the data site from which to delete the field */
  BPipeOutput *channel	/* the output channel for which to delete this field.
			   (only relevant for deletion from output sites) */
)
{
  DpktField *dpktf;

  if ( NULL == bpipe || NULL == name ||
       bpipe->hdr_written || bpipe->hdr_mapped  ||
      ( site != BPDSite_CORE && site != BPDSite_OUTPUT) )
  {
    bpipe_errno = BPEBADARG;
    return -1;
  }

  switch (site)
  {
    case BPDSite_CORE:		/* delete from the core object */
      /* delete node in core tree.  if it doesn't exist, return that */
      if ( NULL ==
	   (dpktf = rbtree_destroy(bpipe->dpktf, name, dpktf_name_cmp) ) )
        return 1;

      /* for safety, delete reference to this field from all output
         deletion lists */
      if (dpktf->output_delete)
	sll_utraverse( bpipe->output, dpktf_output_destroy, dpktf );

      dpktf_delete( dpktf );

      break;

    case BPDSite_OUTPUT:	/* delete from output object */
      /* see if it's in the core tree.  if not, say so */
      if ( NULL ==
	   ( dpktf = rbtree_search( bpipe->dpktf, name, dpktf_name_cmp ) ) )
        return 1;

      dpktf->output_delete++;	/* set flag indicating deletion status */

      /*
         if it's a general delete, attach to the general list. it may
         already be in specific output lists, but that doesn't cause
         any harm
       */
      if ( BPOutputChannel_ALL == channel )
        dpktf->global_delete = 1;
      else	/* stick it into a particular output channel's list */
      {
        /* ensure it's not in list twice */
        if ( ! rbtree_search(channel->dpktf_dels, dpktf, RBTREE_NULL_CMP) &&
               rbtree_insert(channel->dpktf_dels, dpktf) )
	{
	  bpipe_errno = BPENOMEM;
	  return -1;
	}
      }
      break;

    default:
      assert( 0 );

  }

  return 0;
}

/*

  PROCEDURE
  bpipe_dpktf_init_core - Initialize a data packet field's core image.

  DESCRIPTION
  This routine will replicate a datatype initialization structure
  throughout the core image of a data packet field.  It is a simple
  utility routine which replaces calls to @code{bpipe_dpktf_matrix},
  and @code{bpipe_datatype_init} or @code{bpipe_memfill}.  If the user
  wants more control over how the data is initialized, these functions
  should be called instead.

  This routine will fill a data packet's field image with either default
  initialization values (zeroes of the appropriate type) or the data
  from a user provided data element (which will be replicated for each
  element in the data packet field's array).  It must
  be called @emph{after} @code{bpipe_map} has been called.

*/
void
bpipe_dpktf_init(
  DpktField *	field,	/* the field to initialize */
  void *	dpkt,	/* the core image, allocated by the user after
			   calling @code{bpipe_map} */
  void *	init	/* an (optional) initialization data element.
			   Set to @code{NULL} to use the default initialization
			   value */
)
{
  char *data = (char *)dpkt + field->core_offset;
  BPMatrix *matrix = field->core ? field->core : field->input;

  if ( init )
    bpipe_memfill( data, matrix->ne, init, (size_t) 1,
		   bpipe_datatype_size( field->type ) );
  else
    bpipe_datatype_init( data, field->type, matrix->ne );
}

/*

  PROCEDURE
  dpktf_output_destroy - Remove a data packet field from an output channel's
  deletion list.

  DESCRIPTION
  This routine is an action routine called when traversing the output
  channel list to remove a data packet field from all output channels'
  deletion lists.

*/
static int
dpktf_output_destroy( void *data, void *udata )
{
  rbtree_destroy( ((BPipeOutput *) data)->dpktf_dels, udata,
			  RBTREE_NULL_CMP );
  return 0;
}

/*

  PROCEDURE
  dpktf_name_cmp - Compare a name to a data packet field's name.

  DESCRIPTION
  This routine compares a name to a data packet field's name using
  string collating sequence.  The first parameter is the name, the
  second a pointer to the data packet field structure.

  RETURNS
  It returns @samp{-1}, @samp{0}, @samp{1} if the name is
  respectively less than, equal to, or greater than the field's name.

*/
static int
dpktf_name_cmp(
  const void *name,		/* name */
  const void *dpktf		/* data packet field */
)
{
  return strcmp( (const char *) name, ((DpktField *) dpktf)->name );
}


/*

   PROCEDURE
   dpktf_node_name_cmp - compare two data packet fields by their names

   DESCRIPTION
   This routine compares two data packet fields by their names,
   using string collating sequence

   RETURNS
   It returns @samp{-1}, @samp{0}, @samp{1} if the first field's name
   is respectifely less than, equal to, or greater than that of the second
   field.

*/
int
dpktf_node_name_cmp(
  const void *dpktf1,		/* first data packet field */
  const void *dpktf2		/* second data packet field */
)
{
  return strcmp( ((DpktField *) dpktf1)->name, ((DpktField *) dpktf2)->name );
}


/*

   PROCEDURE
   bpipe_dpktf_size_cmp - Compare two data packet fields based upon the size
   of their data types.

   DESCRIPTION
   This routine compares two data packet fields based upon the size of
   their data types.  In the case of ties, it uses the fields' index
   to break them (in ascending order).  The size comparison is in
   reverse order.  It assumes that the types are legitimate.

   RETURNS
   Returns @samp{-1}, @samp{0}, @samp{1} if the second is less
   than, equal to, or greater than the first, respectively.

*/
int
bpipe_dpktf_size_cmp(
  const void *dpktf1,	/* pointer to first data packet field */
  const void *dpktf2	/* pointer to second data packet field */
)
{
  size_t size1 = bpipe_datatype_size( ((DpktField *) dpktf1)->type );
  size_t size2 = bpipe_datatype_size( ((DpktField *) dpktf2)->type );

  if      (size1 < size2) return 1;
  else if (size1 > size2) return -1;

  return ((DpktField *) dpktf1)->index < ((DpktField *) dpktf2)->index;
}
