/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <string.h>

#include "bpipep.h"

/*

  PROCEDURE
  bpipe_read_dpkts - Read data packets from the input channel of a
  binary pipe.

  DESCRIPTION
  @code{bpipe_read_dpkt} reads a user specified number of data packets
  from a binary pipe input channel into a user-provided memory buffer.
  It converts the input data packet images into core data packet
  images.

  RETURNS
  It returns the number of data packets read.  Upon error it
  returns @samp{0} and sets @code{bpipe_errno} to @code{BPEIOERR}.
*/
size_t
bpipe_read_dpkts(
  BPipe *bpipe,		/* the binary pipe from which to read */
  void *buf,		/* pointer to a user-allocated memory buffer
			   into which the data packet core images will
			   be placed */
  size_t n_pkts		/* the number of data packets to read */
)
{
  BPerrno bpipe_errno_s = bpipe_errno;
  size_t n = 0;

  /* if there's no map, we shouldn't be reading from the input
     stream, which means that the user ignored the error return from
     bpipe_map */
  if ( NULL == bpipe->xmap )
    return 0;

  bpipe_errno = BPNOERROR;

  /* check if more than one entry in this map list */
  if ( bpipe->nmap > 1 )
  {
    /* non-contiguous src & dest chunks */

    char *dest = buf;
    char *dest_lim = dest + n_pkts * bpipe->dpkt_core_size;

    for ( ;
	 dest < dest_lim &&
	   bpipe_iochannel_read( bpipe->dpkt_inbuf, bpipe->dpkt_in_size,
				 (size_t) 1, bpipe->input ) == 1 ;
	 dest += bpipe->dpkt_core_size )
    {
      Xmap *xmap = bpipe->xmap;
      Xmap *xmap_lim = xmap + bpipe->nmap;

      for ( ; xmap < xmap_lim ; xmap++ )
	memcpy(dest + xmap->dst, bpipe->dpkt_inbuf + xmap->src, xmap->size);
    }
    n = (dest - (char *) buf) / bpipe->dpkt_core_size ;
  }
  else if (bpipe->xmap[0].size != bpipe->dpkt_in_size)
  {
    /* single chunk. length not equal to dpkt_in_size. */

    char *dest = (char *) buf + bpipe->xmap[0].dst;
    char *dest0 = dest;
    char *dest_lim = dest + n_pkts * bpipe->dpkt_core_size;
    char *src = bpipe->dpkt_inbuf + bpipe->xmap[0].src;

    for ( ;
	 dest < dest_lim &&
	   bpipe_iochannel_read( bpipe->dpkt_inbuf, bpipe->dpkt_in_size,
				 (size_t) 1, bpipe->input ) == 1 ;
	 dest += bpipe->dpkt_core_size
	)
      memcpy( dest, src, bpipe->xmap[0].size );
    n = (dest - dest0) / bpipe->dpkt_core_size ;
  }
  else if ( bpipe->xmap[0].dst != 0 )
  {
    /* single chunk, length equal to dpkt_in_size, offset destination chunks.
       xmap[0].src MUST be zero! */

    char *dest = (char *) buf + bpipe->xmap[0].dst;
    char *dest0 = dest;
    char *dest_lim = dest + n_pkts * bpipe->dpkt_core_size;

    for ( ;
	 dest < dest_lim &&
	   bpipe_iochannel_read( dest, bpipe->dpkt_in_size, (size_t) 1,
				 bpipe->input ) == 1 ;
	 dest += bpipe->dpkt_core_size
	)
      ;
    n = (dest - dest0) / bpipe->dpkt_core_size ;
  }
  else
  {
    /*
      single chunk, no destination offset, length equal to
      dpkt_in_size, xmap[0].src MUST be zero!
    */
    n = bpipe_iochannel_read(buf, bpipe->dpkt_in_size, n_pkts, bpipe->input);
  }

  if ( bpipe_errno != BPNOERROR )
    return 0;

  bpipe_errno = bpipe_errno_s;
  return n;
}


/*

  PROCEDURE
  bpipe_write_dpkt_n - Write data packets to a binary pipe's output channels.

  DESCRIPTION
  @code{bpipe_write_dpkt_n} writes a user specified number of data
  packets to one or all of a binary pipe's output channels.  It must
  be called after a call to @code{bpipe_write_hdr}.  It converts the
  packets' core images into output images.

  If @code{channel} is the constant @code{BPOutputChannel_ALL} (or
  @code{NULL}), the packets are written to all of the output channels.
  It is much more efficient to call @code{bpipe_write_dpkt} or
  @code{bpipe_write_dpkts} if the situation warrants it.

  RETURNS
  It returns zero upon success, non-zero upon failure.

*/
size_t
bpipe_write_dpkt_n(
  BPipe *bpipe,		/* the binary bpipe to which to write the packets */
  void *buf,		/* a pointer to a user-provided memory buffer
			   containing the data packets to be written */
  size_t n_pkts,	/* the number of data packets to write */
  BPipeOutput *bpo	/* the output channel to which to write the packets. */
)
{
  if ( bpo == BPOutputChannel_ALL )
  {
    BPipeOutput **bpop 	  = bpipe->output_arr;
    BPipeOutput **bpop_max = bpipe->output_max;

    if ( n_pkts == 1 )
    {
      for( ; bpop <= bpop_max ; bpo++ )
	if ( bpipe_write_dpkt( bpipe, buf, *bpop ) )
	  return 1;
    }
    else
    {
      for( ; bpop <= bpop_max ; bpo++ )
	if ( bpipe_write_dpkts( bpipe, buf, n_pkts, *bpop ) )
	  return 1;
    }
  }
  else
  {
    if ( n_pkts == 1 )
    {
      if ( bpipe_write_dpkt( bpipe, buf, bpo ) )
	return 1;
    }
    else
    {
      if (  bpipe_write_dpkts( bpipe, buf, n_pkts, bpo ) )
	return 1;
    }
  }

  return 0;
}



/*

  PROCEDURE
  bpipe_write_dpkt - Write a data packet to a binary pipe output channel.

  DESCRIPTION
  @code{bpipe_write_dpkt} writes a single data packet to a single
  binary pipe output channel.  Contrast it with
  @code{bpipe_write_dpkts} and @code{bpipe_write_dpkt_n}. It must be
  called after a call to @code{bpipe_write_hdr}.  It converts the
  packets' core images into output images.

  RETURNS
  It returns zero upon success, non-zero upon failure.  Upon failure
  @code{bpipe_errno} is set to @code{BPEIOERR}.

*/
int
bpipe_write_dpkt(
  BPipe *bpipe,		/* the binary bpipe to which to write the packets */
  void *buf,		/* a pointer to a user-provided memory buffer
			   containing the data packets to be written */
  BPipeOutput *bpo	/* the output channel to which to write the packets. */
)
{

  if ( bpo->nmap > 1 )
  {
    /* non-contiguous src & dest chunks */

    char *src = buf;
    Xmap *xmap = bpo->xmap;
    Xmap *xmap_lim = xmap + bpo->nmap;

    /* zero output buffer and copy to it, then write it out */
    memset( bpo->buf, 0, bpo->dpkt_size );

    for ( ; xmap < xmap_lim ; xmap++ )
      memcpy(bpo->buf + xmap->dst, src + xmap->src, xmap->size);

    return
      1 != bpipe_iochannel_write( bpo->buf, bpo->dpkt_size,
				  (size_t) 1, bpo->channel );
  }
  else if ( bpo->xmap->size == bpo->dpkt_size ) /* single copy is entire dpkt */
  {
     return 1 != bpipe_iochannel_write( (char *) buf + bpo->xmap->src,
					bpo->dpkt_size, (size_t) 1,
					bpo->channel );
  }
  else
  {
    /* only one map, but it isn't the same size as the output image,
       which means that there's padding around here somewhere */

    memcpy(bpo->buf + bpo->xmap->dst,
	   (char *) buf + bpo->xmap->src, bpo->xmap->size);

    return
      1 != bpipe_iochannel_write( bpo->buf, bpo->dpkt_size,
				  (size_t) 1, bpo->channel );
  }
}


/*

  PROCEDURE
  bpipe_write_dpkts - Write several data packets to a binary pipe
  output channel.

  DESCRIPTION
  @code{bpipe_write_dpkts} writes one or more data packets to a single
  binary pipe output channel.  Contrast it with
  @code{bpipe_write_dpkt} and @code{bpipe_write_dpkt_n}. It must be
  called after a call to @code{bpipe_write_hdr}.  It converts the
  packets' core images into output images.  It is more efficient to
  call @code{bpipe_write_dpkt} if there is only one data packet to be
  written.

  RETURNS
  It returns zero upon success, non-zero upon failure.  Upon failure
  @code{bpipe_errno} is set to @code{BPEIOERR}.

*/
int
bpipe_write_dpkts(
  BPipe *bpipe,		/* the binary bpipe to which to write the packets */
  void *buf,		/* a pointer to a user-provided memory buffer
			   containing the data packets to be written */
  size_t n_pkts,	/* the number of data packets to write */
  BPipeOutput *bpo	/* the output channel to which to write the packets. */
)
{
  char *pkt = (char *) buf;
  char *pkt_lim = pkt + n_pkts * bpipe->dpkt_core_size;

  if ( bpo->nmap > 1 )
  {
    /* non-contiguous src & dest chunks */

    /* zero output buffer and copy to it, then write it out */
    memset( bpo->buf, 0, bpo->dpkt_size );

    for ( ; pkt < pkt_lim ; pkt += bpipe->dpkt_core_size )
    {
      Xmap *xmap = bpo->xmap;
      Xmap *xmap_lim = xmap + bpo->nmap;

      for ( ; xmap < xmap_lim ; xmap++ )
	memcpy(bpo->buf + xmap->dst, pkt + xmap->src, xmap->size);

      if ( 1 != bpipe_iochannel_write( bpo->buf, bpo->dpkt_size,
				       (size_t) 1, bpo->channel ) )
	break;
    }
    return (pkt - (char *) buf) / bpipe->dpkt_core_size != n_pkts;
  }
  else if ( bpo->xmap->size == bpipe->dpkt_core_size )
            /* single copy is entire dpkt and is same size as core image */
  {
    return n_pkts !=
        bpipe_iochannel_write( buf, bpipe->dpkt_core_size, n_pkts, bpo->channel );
  }
  else if ( bpo->xmap->size == bpo->dpkt_size ) /* single copy is entire dpkt */
  {
    for ( ;
	  pkt < pkt_lim && 1 ==
	    bpipe_iochannel_write( pkt + bpo->xmap->src, bpo->dpkt_size,
				     (size_t) 1, bpo->channel ) ;
	  pkt += bpipe->dpkt_core_size )
      ;

    return (pkt - (char *) buf) / bpipe->dpkt_core_size != n_pkts;
  }
  else
  {
    /* only one map, but it isn't the same size as the output image,
       which means that there's padding around here somewhere */

    for ( ; pkt < pkt_lim ; pkt += bpipe->dpkt_core_size )
    {
      memcpy(bpo->buf + bpo->xmap->dst, pkt + bpo->xmap->src,
	     bpo->xmap->size);

      if ( 1 !=  bpipe_iochannel_write( bpo->buf, bpo->dpkt_size,
					(size_t) 1, bpo->channel ) )
	break;
    }
    return (pkt - (char *) buf) / bpipe->dpkt_core_size != n_pkts;
  }
}
