/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <rbtree/rbtree.h>

#include "bpipep.h"


/*

  PROCEDURE
  bpipe_dpktf_resize_core - Resize the core image of a data packet field.

  DESCRIPTION
  This routine allows one to specify the size of a data packet field's
  core image and the mapping from its input image.  The input state is
  mapped onto the core state by copying a submatrix of the input data.
  The default mapping is to simply duplicate the entirety of the input
  matrix, without changing its extents or dimensions.

  The field's core image size is determined by the passed matrix
  specification.  It may be created with @code{bpipe_matrix_new} or
  @code{bpipe_matrix_new_va}, or may be extracted from an existing
  data packet field. The matrix specification may also by @code{NULL},
  indicating that the core image should have the same size as the
  input image.

  The mapping is composed of specifications of the offset and extent
  of the source submatrix and the offset of the destination submatrix.
  These are passed as arrays.  If an offset array is @code{NULL}, the
  offsets in each dimension are set to @samp{0}.  If the submatrix
  extent is @code{NULL}, the largest possible submatrix, as determined
  by the sizes of the input and core matrices and the submatrix
  offsets, is copied.

  This routine may be called multiple times for the same data field;
  previous mappings are replaced.  To return to the default input to
  core mapping, the matrix specification and the submatrix offsets
  and extent should be passed as @code{NULL}.

  Note that @code{bpipe_dpktf_resize_core} uses the matrix
  specification structure and all of the offset and extent arrays
  directly, directing the @code{BPipe} cleanup code to take
  responsibility for freeing them.  The application should @emph{not}
  free them, nor should it pass them to any other @code{BPipe} routine
  which takes responsibility for freeing them.  After passing them to
  @code{bpipe_dpktf_resize_core}, it should not alter them in any
  fashion.

  Note that this routine does not do the actual mapping, it just
  stores the mapping specifications.  @code{bpipe_map} does the
  mapping.  Once a @code{BPipe} is mapped with @code{bpipe_map},
  resizing data packet fields has no effect.

  RETURNS
  It returns zero upon success, non-zero upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
  @itemize @bullet
  @item the passed data packet field doesn't have a input image
  @item the passed matrix is bogus
  @item the input and core matrix sizes and the submatrix offsets and
        extents are inconsistent
  @end itemize

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
int
bpipe_dpktf_resize_core(
  DpktField *dpktf,	/* a pointer to the data packet field to resize */

  BPMatrix *matrix,	/* specification for the core site's storage
			   description. */

  size_t    src_off[],	/* An array containing the offset coordinates
                           of the source submatrix to copy to the new
                           site.  The number of elements is equal to
                           the dimensions of the source matrix.  If
                           it's @code{NULL}, the origin is assumed. */

  size_t    dst_off[],	/* An array containing the offset coordinates
                           of the destination submatrix in the new
                           site.  The number of elements is equal to
                           the dimensions of the target state's
                           matrix.  If it's @code{NULL}, the origin is
                           assumed. */

  size_t    extent[]	/* An array containing the extents of the
                           submatrix to copy.  The number of elements
                           is equal to the dimensions of the source
                           matrix.  If it's @code{NULL}, the largest
                           submatrix which will fit is copied. */
)
{
  BPMatrix    *core = NULL;
  BPMatrixMap *map  = NULL;

  /* check if this data packet field actually has an input matrix */
  if ( NULL == dpktf->input )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  /* check if reestablishing default matrix size */
  if (matrix)
  {
    /* new setup */
    if ( bpipe_matrix_verify(matrix) )
    {
      bpipe_errno = BPEBADARG;
      return 1;
    }
    core = matrix;
  }

  bpipe_matrix_delete(dpktf->core);
  dpktf->core = core;

  /* delete previous map */
  bpipe_matrix_map_delete( dpktf->input_to_core );

  if ( src_off || dst_off || extent )	 /* create new map */
  {
    /* must resolve core matrix if default matrix is requested */
    map =
      bpipe_matrix_map_new( dpktf->input, src_off,
                            dpktf->core ? dpktf->core : dpktf->input,
			    dst_off, extent );
  }

  dpktf->input_to_core = map;

  /*
    if there's a map, ensure that it's compatible with the
    input and core matrices
  */
  if ( dpktf->input_to_core &&
       bpipe_check_matrix_copy( dpktf->input,
				dpktf->core ? dpktf->core : dpktf->input,
				dpktf->input_to_core ) )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  return 0;
}


/*

  PROCEDURE
  bpipe_dpktf_resize_output - Resize the output image of a data
  packet field.

  DESCRIPTION
  This routine allows one to specify the size of a data packet field's
  output image and the mapping from its core image.  The core state is
  mapped onto the output state by copying a submatrix of the core
  data.  The default mapping is to simply duplicate the entirety of
  the core matrix, without changing its extents or dimensions.

  The field's output image size is determined by the passed matrix
  specification, which is duplicated.  It may be created with
  @code{bpipe_matrix_new} or @code{bpipe_matrix_new_va}, or may be
  extracted from an existing data packet field.  The matrix
  specification may also by @code{NULL}, indicating that the output
  image should have the same size as the core image.

  The mapping is composed of specifications of the offset and extent
  of the source submatrix and the offset of the destination submatrix.
  These are passed as arrays.  If an offset array is @code{NULL}, the
  offsets in each dimension are set to @samp{0}.  If the submatrix
  extent is @code{NULL}, the largest possible submatrix, as determined
  by the sizes of the core and output matrices and the submatrix
  offsets, is copied.

  This routine may be called multiple times for the same data field;
  previous mappings are replaced.  To return to the default core to
  output mapping, the matrix specification and the submatrix offsets
  and extent should be passed as @code{NULL}.

  This routine should be called after @emph{all} calls to
  @code{bpipe_dpktf_resize_core} for this data packet field have been
  made.  The sizes of the core and output images must be known in
  in order to validate the mapping between them.

  Note that @code{bpipe_dpktf_resize_output} uses the matrix
  specification structure and all of the offset and extent arrays as
  is, directing the @code{BPipe} cleanup code to take responsibility
  for freeing them.  The application should @emph{not} free them, nor
  should it pass them to any other @code{BPipe} routine which takes
  responsibility for freeing them.  After passing them to
  @code{bpipe_dpktf_resize_output}, it should not alter them in any
  fashion.

  Note that this routine does not do the actual mapping, it just
  stores the mapping specifications.  @code{bpipe_map} does the
  mapping.  Once a @code{BPipe} is mapped with @code{bpipe_map},
  resizing data packet fields has no effect.

  RETURNS
  It returns zero upon success, non-zero upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
  @itemize @bullet
  @item the passed matrix is bogus
  @item the input and core matrix sizes and the submatrix offsets and
        extents are inconsistent

  @end itemize

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
int
bpipe_dpktf_resize_output(
  DpktField *dpktf,	/* a pointer to the data packet field to resize */

  BPMatrix *matrix,	/* specification for storage description at
			   destination site. */

  size_t    src_off[],	/* An array containing the offset coordinates
                           of the source submatrix to copy to the new
                           site.  The number of elements is equal to
                           the dimensions of the source matrix.  If
                           it's @code{NULL}, the origin is assumed. */

  size_t    dst_off[],	/* An array containing the offset coordinates
                           of the destination submatrix in the new
                           site.  The number of elements is equal to
                           the dimensions of the target state's
                           matrix.  If it's @code{NULL}, the origin is
                           assumed. */

  size_t    extent[],	/* An array containing the extents of the
                           submatrix to copy.  The number of elements
                           is equal to the dimensions of the source
                           matrix.  If it's @code{NULL}, the largest
                           submatrix which will fit is copied. */

  BPipeOutput *bpo	/* the output channel */
)
{
  BPMatrixOverride *over;

  /* check the validity of the passed matrix */
  if ( matrix && bpipe_matrix_verify(matrix) )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  /* delete the channel specific override if it already exists */
  bpipe_matrix_override_delete( (BPMatrixOverride *)
    sll_search( bpo->matrix, dpktf, bpipe_dpktf_matrix_override_cmp ) );


  /* if all of the passed specs are NULL, the user wants the default
     matrix and mapping, so don't bother creating a new override
     structure */
  if ( matrix || src_off || dst_off || extent )
  {
    /*
      create new map.  the map is only created if one of src_off,
      dst_off, or extent is non-NULL.  in that case, both src and dst
      matrices must be specified, so must resolve "default" matrices.

      if none of src_off, dst_off, or extent are non-NULL, then just
      the matrix is specified.

      in either case bpipe_matrix_override_new knows what to do

      */

    BPMatrix *src = dpktf->core ? dpktf->core : dpktf->input;
    BPMatrix *dst = matrix ? matrix : src;

    assert( src != NULL && dst != NULL);


    if ( NULL ==
	 ( over = bpipe_matrix_override_new( dpktf,
					     src, src_off,
					     dst, dst_off, extent ) ) )
      return 1;

    /*
      ensure that the map is compatible with the core and output
      matrices
    */
    if ( bpipe_check_matrix_copy( src, dst, over->map ) )
    {
      bpipe_errno = BPEBADARG;
      return 1;
    }

    if ( sll_insert( bpo->matrix, over ) )
    {
      bpipe_errno = BPENOMEM;
      return 1;
    }
  }

  return 0;
}
