/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>

#include <rbtree/rbtree.h>
#include <linklist/linklist.h>

#include "bpipep.h"

/* ------------------------------------------------------------------ */

static void hdrf_delete(HdrField *field);
static HdrField *hdrf_new(char *name, BPDataType type, BPMatrix *matrix, void *data, int copy);
static HdrField *hdrf_get(BPipe *bpipe, char *name, size_t index);
static void hdrf_ll_delete(void *node);
static void hdrf_all_delete(void *field, void *ll);
static int hdrf_name_cmp(const void *name, const void *field);
static int hdrf_index_cmp(const void *indexp, const void *field);
static int hdrf_node_name_cmp(const void *field1, const void *field2);
static int matrix_memcpy(void *udata, size_t dest, size_t src, size_t nbytes);

/* ------------------------------------------------------------------ */

/*

   PROCEDURE
   hdrf_delete - Delete a header data field structure and its
   associated data.

   DESCRIPTION
   This routine deletes the memory associated with a header data field
   structure.  It does @emph{not} remove it from the header lists;
   this must be done prior to calling @code{hdrf_delete}.  It
   is safe to call this routine with a @code{NULL} pointer.

*/
static void
hdrf_delete(
  HdrField *field	/* the header field structure to delete */
)
{
  if (field)
  {
    if ( field->name )
      free( field->name );

    if ( field->data && field->alloc_data )
      free( field->data );

    /* bpipe_matrix_delete accepts NULL arguments */
    bpipe_matrix_delete( field->matrix );

    free( field );
  }
}

/*

   PROCEDURE
   hdrf_new - Create and initialize a header data field.

   DESCRIPTION
   This routine creates and initializes a header data field structure.
   It duplicates the passed field name and type.  It assumes
   responsibility for the passed matrix specification.  All of the
   passed data are assumed to be valid.  The matrix specification may
   be @code{NULL}, in which case a matrix specification for a scalar
   is generated.  If the argument @code{copy} is true, then the data
   will be copied.  Otherwise, the memory pointed to by the
   @code{data} pointer will be used.  If @code{data} is @code{NULL}
   enough memory to hold the matrix is allocated, regardless of the
   state of the @code{copy} variable.

   RETURNS
   It returns a pointer to a dynamically allocated and initialized
   structure.  Upon error it returns @code{NULL} and sets
   @code{bpipe_errno}.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following:
   @table @code

   @item BPENOMEM
     a memory allocation failed

   @end table

*/
static HdrField *
hdrf_new (
  char *name,		/* the field's name */
  BPDataType type,	/* the field's data type */
  BPMatrix *matrix,	/* the field's storage spec */
  void *data,		/* a pointer to the data */
  int copy		/* if true, make a copy of the data, else
			   just stash the pointer */
)
{
  HdrField *field = NULL;

  assert( name != NULL );
  assert( ! bpipe_matrix_verify( matrix ) );
  assert( type != BPDType_NOTYPE );

  if ( NULL == ( field = malloc( sizeof(HdrField) ) ) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  /* initialize field structure */
  field->name	= NULL;
  field->data	= NULL;
  field->matrix	= NULL;
  field->same	= NULL;

  if ( NULL == ( field->name = malloc( strlen(name) + 1 ) ) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  /*
    if input matrix pointer is NULL, create a scalar matrix specification,
    else duplicate the passed matrix description
  */
  field->matrix = (matrix == NULL) ? bpipe_matrix_new_va((size_t) 1,
							 (size_t) 1) : matrix;

  if ( NULL == field->matrix )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  strcpy(field->name, name);

  field->type = type;

  if ( NULL == data || copy)
  {
    field->data = bpipe_data_dup(data, field->matrix->ne,
				 bpipe_datatype_size(type), 1);

    if (field->data == NULL)
    {
      bpipe_errno = BPENOMEM;
      return NULL;
    }

    field->alloc_data = 1;
  }
  else
  {
    field->data = data;
    field->alloc_data = 0;
  }

  return field;
}

/*

  PROCEDURE
  hdrf_get - Get a pointer to a header field.

  RETURNS
  It returns a pointer to the header field if it exists, @code{NULL}
  if it doesn't.  Upon error it returns @code{NULL} and sets
  @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
     the index specified was illegal

  @end table

*/
static HdrField *
hdrf_get(
  BPipe *bpipe,		/* binary pipe with which this field is associated */
  char *name,		/* the field's name */
  size_t index		/* the field's index.  set to @code{BPHdrfIdx_LAST} to
			   select the last one. */
)
{
  HdrField *field = NULL;

  if ( (field = rbtree_search(bpipe->hdrf_tree, name, hdrf_name_cmp)) )
  {
    field = (index == BPHdrfIdx_LAST) ?
      sll_tail( field->same ) :
      sll_search( field->same, &index , hdrf_index_cmp);

    /* if field is NULL, it was an illegal index */
    if ( NULL == field )
      bpipe_errno = BPEBADARG;
  }

  return field;
}


/*

  PROCEDURE
  bpipe_hdr_setup - Initialize hdr data structure.

  DESCRIPTION
  This routine is called by @code{bpipe_new} to initialize the header
  data structures of a new @code{BPipe}.

  RETURNS
  It returns zero upon success, non-zero upon failure.  It sets
  @code{bpipe_errno} upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
int
bpipe_hdr_setup(
  BPipe *bpipe		/* binary pipe */
)
{

  /* create header field binary tree and linked list */
  if ( NULL == ( bpipe->hdrf_tree = rbtree_new(hdrf_node_name_cmp) ) ||
       NULL == ( bpipe->hdrf_ll = sll_new( SLL_NULL_CMP ) ) )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }

  bpipe->hdr_written = 0;
  bpipe->hdr_mapped = 0;

  return 0;
}

/*
   PROCEDURE
   bpipe_hdr_cleanup - Cleanup header structure and deallocate header fields.

   DESCRIPTION
   This routine is called to cleanup a @code{BPipe} header.
   It frees all memory associated with the header (fields, etc.).

*/
void
bpipe_hdr_cleanup(
  BPipe *bpipe		/* binary pipe whose header is to be initialized */
)
{
  /* delete header field binary tree and linked list */
  sll_delete( bpipe->hdrf_ll, SLL_NULL_DELETE );
  rbtree_delete( bpipe->hdrf_tree, hdrf_ll_delete, POST_ORDER, LEFT_TO_RIGHT);

  bpipe->hdrf_ll = bpipe->hdrf_tree = NULL;
}


/*

  PROCEDURE
  bpipe_hdrf_n - Return the number of header fields with a given name.

  RETURNS
  It returns the number of header data fields which have the given name.
  If none match, returns @samp{0}.

*/
size_t
bpipe_hdrf_n(
  BPipe *bpipe,		/* binary pipe with which this field is associated */
  char *name		/* the field's name */
)
{
  HdrField *field;

  if ( (field = rbtree_search(bpipe->hdrf_tree, name, hdrf_name_cmp)) )
    return sll_count( field->same );

  return 0;
}

/*

  PROCEDURE
  bpipe_hdrf_type - Return the datatype of a header field.

  RETURNS
  It returns the data type of a header field, if it exists.  If it doesn't
  exist, it returns @code{BPDType_NOTYPE}.  Upon error it returns
  @code{BPDType} and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:

  @table @code

  @item BPEBADARG
     the index specified was illegal

  @end table

*/
BPDataType
bpipe_hdrf_type(
  BPipe *bpipe,		/* binary pipe with which this field is associated */
  char *name,		/* the field's name */
  size_t index		/* the field's index.  set to @code{BPHdrfIdx_LAST} to
			   select the last one. */
)
{
  HdrField *field;

  field = hdrf_get( bpipe, name, index );

  return (field == NULL) ? BPDType_NOTYPE : field->type;
}

/*

  PROCEDURE
  bpipe_hdrf_data - Get a pointer to a header field's data.

  DESCRIPTION
  This routine looks up a header data field and returns a pointer to
  the data in the field.  The calling routine may use this pointer to
  directly change the header field's data.

  RETURNS
  Returns a pointer to the data upon success.  If the field doesn't
  exist, it returns @code{NULL}.  If there's an error, it returns
  @code{NULL} and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:

  @table @code

  @item BPEBADARG
    the index specified was illegal

  @end table

*/
void *
bpipe_hdrf_data(
  BPipe *bpipe,		/* binary pipe with which this field is associated */
  char *name,		/* the field's name */
  size_t index		/* the field's index.  set to @code{BPHdrfIdx_LAST} to
			   select the last one. */
)
{
  HdrField *field;

  field = hdrf_get( bpipe, name, index );

  return (NULL == field) ? NULL : field->data;
}


/*

  PROCEDURE
  bpipe_hdrf_matrix - Retrieve a copy of a header data field's matrix
  specification.

  DESCRIPTION
  This routine makes a copy of the matrix specification associated with
  a header data field and returns a pointer to it.

  RETURNS
  This routine returns a pointer to a @code{BPMatrix} structure upon
  success.  If the field doesn't exist, it returns @code{NULL}.  If
  an error occurs, it returns @code{NULL} and sets
  @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:

  @table @code

  @item BPENOMEM
  a memory allocation failed

  @item BPEBADARG
  the index specified was illegal

  @end table

*/
BPMatrix *
bpipe_hdrf_matrix(
  BPipe *bpipe,		/* binary pipe with which this field is associated */
  char *name,		/* the field's name */
  size_t index		/* the field's index.  set to @code{BPHdrfIdx_LAST} to
			   select the last one. */
)
{
  HdrField *field;
  BPMatrix *matrix;
#ifndef NDEBUG
  BPerrno bpipe_errno_s = bpipe_errno;
  bpipe_errno = BPNOERROR;
#endif

  if ( NULL == ( field = hdrf_get(bpipe, name, index) ) )
    return NULL;

  matrix = bpipe_matrix_dup( field->matrix );

  assert( BPEBADARG != bpipe_errno );

#ifndef NDEBUG
  if ( BPNOERROR == bpipe_errno )
    bpipe_errno = bpipe_errno_s;
#endif

  return matrix;
}


/*

  PROCEDURE
  bpipe_hdrf_resize - Resize the storage space associated with a
  header field.

  DESCRIPTION
  This routine resizes the storage space associated with a header data
  field.  The user must have created a @code{BPMatrix} structure
  (using one of the @code{bpipe_matrix_new} routines or
  @code{bpipe_matrix_dup}) for the new representation of the data.
  This routine will allocate space for the new matrix, copy over the
  specified submatrix, and delete the old space.

  The submatrix to be copied is specified via the @code{src_off},
  @code{dst_off}, and @code{extent} arguments.  They specify,
  respectively, the offset of the origin of the submatrix in the
  original matrix, its offset from the origin of the new matrix, and
  the extents of the submatrix.  The submatrix offsets are arrays of
  integers, one element per dimension.  Either may be @code{NULL},
  indicating that the submatrix's origin coincides with that of the
  matrix.  The extent argument is an array with as many elements as
  there are dimensions in the original matrix.  It may be @code{NULL},
  in which case the maximum extents possible will be automatically
  calculated.

  Before any data is copied from the old field into the resized field,
  the new field is first initialized, so that any holes in the data
  will have deterministic values.  By default the field is filled
  with zeroes of the appropriate type.  The calling routine may also
  specify an array of data to be used to initialize the field.  If the
  number of data elements in this array is less than that needed to
  initialize the field, it will be repeated.

  Note that @code{bpipe_hdrf_resize} uses the matrix specification
  structure and all of the offset and extent arrays as is, directing
  the @code{BPipe} cleanup code to take responsibility for freeing
  them.  The application should @emph{not} free them, nor should it
  pass them to any other @code{BPipe} routine which takes
  responsibility for freeing them.  After passing them to
  @code{bpipe_hdrf_resize}, it should not alter them in any fashion.

  RETURNS
  It returns zero upon success and non-zero if the data field doesn't exist.
  If there was an error it returns non-zero and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
  @itemize @bullet
  @item the new matrix specification was bogus
  @item the matrix copy wasn't consistent with the sizes of the
        old and new matrix specifications
  @item the passed index was out of bounds
  @end itemize

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
int
bpipe_hdrf_resize(
  BPipe *	bpipe,		/* binary pipe with which this field
				   is associated */
  char *	name,		/* the field's name */
  size_t	index,		/* the field's index.  set to
				   @code{BPHdrfIdx_LAST} to select the
				   last one. */
  BPMatrix *	matrix,		/* the new matrix specification */
  size_t 	src_off[],	/* offset of source submatrix to copy
				   to destination.  extent is equal to
				   the dimensions of the source
				   matrix. */
  size_t	dst_off[],	/* offset of destination submatrix.
				   extent is equal to the dimensions
				   of the destination matrix. */
  size_t	extent[],	/* the extents of the submatrix to
				   copy.  the extent of this array is
				   equal to the dimensions of the
				   source matrix. */
  void *	init,		/* an array of data which
				   will be used to initialize the new
				   field array elements before copying
				   data from the old field array.  The
				   data must have the same type as
				   the header field.  Set to @code{NULL}
				   to use the default initialization. */
  size_t	n_init		/* the number of initializing elements */

)
{
  HdrField *field;
  BPMatrixMap *map;
  void *data;
  char *start[2];

  /* if the parameter doesn't exist, return */
  if ( NULL == (field = hdrf_get( bpipe, name, index )) )
    return 1;

  if ( bpipe_matrix_verify( matrix ) )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  if ( NULL ==
        ( map = bpipe_matrix_map_new(field->matrix, src_off, matrix,
				    dst_off, extent) ) )
    return 1;

  if ( NULL ==
        ( map = bpipe_matrix_map_expand( field->matrix, matrix, map ) ) )
    return 1;


  /* allocate storage for new matrix */
  if ( NULL ==
       (data = malloc(matrix->ne * bpipe_datatype_size(field->type))) )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }

  /* initialize storage */
  if ( init )
    bpipe_memfill( data, matrix->ne, init, n_init,
		   bpipe_datatype_size( field->type ) );
  else
    bpipe_datatype_init( data, field->type, matrix->ne );


  /* initialize source and destination for matrix_memcpy */
  start[0] = data;
  start[1] = field->data;

  if ( bpipe_matrix_copy(field->type, BPXMap_Core_to_Core,
			 field->matrix, matrix, map, start,
			 matrix_memcpy) )
  {
    free(data);
    bpipe_matrix_map_delete( map );
    return 1;

  }

  if ( field->alloc_data )
    free( field->data );
  field->data = data;
  field->alloc_data = 1;

  /* delete old matrix */
  bpipe_matrix_delete( field->matrix );
  field->matrix = matrix;

  bpipe_matrix_map_delete( map );
  return 0;
}


/*

   PROCEDURE
   bpipe_hdrf_next - Get the name of the next header field in the header.

   DESCRIPTION
   It may be necessary to process the header data fields in the order
   that they appear in the header.  This function provides the name
   and index of the next field.  It requires that the user allocate
   three variables: a @code{char *} which will point to the name of
   the field, an @code{int} which will contain the field's index, and
   a @code{void *}, which is used by @code{bpipe_hdrf_next} to keep
   track of where it is.  The last variable should be set to
   @code{NULL} the first time @code{bpipe_hdrf_next} is called.  The
   invoking routine passes @emph{addresses} of these three variables
   to this routine.

   Do not add or delete header data fields in between calls to
   @code{bpipe_hdrf_next}.

   RETURNS
   It returns non-zero if a field is available, zero if the entire
   list has been traversed.  The name and index are returned via the
   @code{p_name} and @code{p_index} arguments.  Please note that the
   pointer returned via @code{p_name} points to the original copy of
   the field's name.  @strong{Do not change this data!}.

*/
int
bpipe_hdrf_next(
  BPipe *bpipe,		/* binary pipe with which this data is associated */
  char **p_name,	/* a place to store a pointer to the field name */
  size_t  *p_index,	/* the index of the field */
  void **p_last		/* the address of a pointer used to traverse
			   the list */
)
{
  HdrField *field = NULL;

  *p_last = ( NULL == *p_last ) ?
            sll_head_node( bpipe->hdrf_ll ) :
            sll_next_node( bpipe->hdrf_ll, *p_last );

  if ( NULL == *p_last )
    return 0;

  field = (HdrField *) sll_node_get_data( *p_last );

  assert( field != NULL );

  if ( p_name )
    *p_name = field->name;

  if ( p_index )
    *p_index = field->index;

  return 1;
}


/*

   PROCEDURE
   bpipe_hdrf_delete - Remove a header field.

   DESCRIPTION
   This routine will remove a field from the header and free the
   memory used by the header data.  It does not renumber the indices
   of the fields which have the same name as the field.  If requested,
   it can delete all of the fields of a given name.

   RETURNS
   It returns zero upon success, non-zero if the field didn't exist.
   If an error occurs (the index was bad), it returns non-zero and sets
   @code{bpipe_errno}.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following:
   @table @code

   @item BPEBADARG
     the index specified was illegal

   @end table

*/
int
bpipe_hdrf_delete(
  BPipe *bpipe,		/* binary pipe with which this field is associated */
  char *name,		/* the field's name */
  size_t index		/* the field's index.  set to @code{BPHdrfIdx_LAST} to
			   delete the last one, @code{BPHdrfIdx_ALL} to delete
			   all of them */
)
{
  RBNode rbnode;
  HdrField *field, *field0;
  SLinkList same;


  if ( NULL ==
       ( rbnode = rbtree_search_node( bpipe->hdrf_tree, name, hdrf_name_cmp )))

    return 1;

  field = rbtree_node_get_data( rbnode );

  assert( field != NULL );

  /* keep this around as may have field deleted out from underneath us */
  same = field->same;

  if ( BPHdrfIdx_ALL == index )
  {
    /* delete all of the headers with this name */
    rbtree_destroy_node( bpipe->hdrf_tree, rbnode );
    sll_udelete( field->same, hdrf_all_delete, bpipe->hdrf_ll );
  }
  else
  {
    if ( BPHdrfIdx_LAST == index )
    {
      field0 = sll_destroy_tail( same );
      assert( field0 != NULL );
    }
    else
    {
      /* check for bogus index */
      if ( NULL == ( field0 = sll_destroy( same, &index, hdrf_index_cmp)) )
      {
	bpipe_errno = BPEBADARG;
	return 1;
      }
    }

    if ( sll_count( same ) > 0)	/* it wasn't the only field in the list */
    {
      /* check if the first field in the list is being destroyed */
      if ( field0 == field )
      {
	/* get the new first node */
	field = sll_head( same );
	assert( field != NULL );

	/*
	  stuff the new first field pointer into the rbtree node.  since
	  the rbtree is sorted by name, and this field has the same name
	  as the previous head field, s'ok
	 */
	rbtree_node_put_data( rbnode, field );

	/* stuff the same list in the new first link */
	field->same = same;

	/* update the hi_index */
	field->hi_index = field0->hi_index;
      }
    }
    else		/* it's the only field in the list */
    {
      /* kill the node */
      rbtree_destroy_node(bpipe->hdrf_tree, rbnode);

      /* kill the linked list, but not the field.  that's done below */
      sll_delete( same, SLL_NULL_DELETE );
    }

    /* remove the field from the bpipe's linked list of fields */
    sll_destroy( bpipe->hdrf_ll, field0, rbtree_node_cmp_v );

    /* remove it */
    hdrf_delete( field0 );
  }

  return 0;
}



/*

  PROCEDURE
  bpipe_hdrf_add - Add a new data field to the header.

  DESCRIPTION
  This routine adds a data field to the header associated with the
  passed @code{BPipe}.  Duplicate fields with the same name are
  allowed; they are given different index numbers.  The new field
  is added to the end of the list of existing fields.

  The passed name is duplicated.  If the matrix specification is
  @code{NULL}, a matrix specification for a scalar is constructed.  If
  the @code{copy} argument is true, a copy is made of the data pointed
  to by the @code{data} argument.  Otherwise, the header field will
  refer directly to the memory it points to.  This saves time and
  space for large data fields.  if the passed data pointer is
  @code{NULL}, memory is allocated for it, regardless of the state of
  the @code{copy} flag.

  If the calling procedure constructs a matrix specification (a
  @code{BPMatrix} structure), @code{bpipe_hdrf_add} uses
  this structure as is and directs the @code{BPipe} cleanup code to take
  responsibility for freeing it.  The calling procedure should
  @emph{not} free it, nor should it pass it to any other @code{BPipe}
  routine which takes responsibility for freeing it.  It should not
  alter it in any way after passing it to @code{bpipe_hdrf_add}.

  RETURNS
  It returns zero upon success.  Upon failure it returns non-zero,
  and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPEBADARG
  @itemize @bullet
  @item the matrix specification was bogus
  @item the name was illegal
  @item the type specification was illegal
  @end itemize

  @item BPEBADPIPE
  @itemize @bullet
  @item the header was already been written
  @item the @code{BPipe} has been mapped
  @end itemize

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
int
bpipe_hdrf_add(
  BPipe *bpipe,		/* binary pipe to which to add the field */
  char *name,		/* the name of the field */
  BPDataType type,	/* the field's datatype */
  BPMatrix *matrix,	/* the matrix specification */
  void *data,		/* a pointer to the field's data */
  int copy		/* if true, make a copy of the data */
)
{
  HdrField *field;
  HdrField *sfield;

  if ( bpipe->hdr_written || bpipe->hdr_mapped )
  {
    bpipe_errno = BPEBADPIPE;
    return 1;
  }

  /* check arguments for correctness. a NULL matrix is ok */
  if (
       bpipe_check_field_name( name ) ||
       BPDType_NOTYPE == type || type >= BPDType_num ||
       bpipe_matrix_verify(matrix)
       )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  /* hdrf_new deals with NULL matrices correctly */
  if ( NULL == ( field = hdrf_new(name, type, matrix, data, copy) ) )
    return 1;

  field->new = bpipe->hdr_read ? 0 : 1;

  /*
     insert the field into the binary tree. if it's
     already in there, attach to the node linked list and update
     the counters
  */
  if ( NULL == ( sfield = (HdrField *)
	      rbtree_search( bpipe->hdrf_tree, name, hdrf_name_cmp ) ) )
  {
    /* it's not in there yet, initialize some stuff */
    field->index = 0;
    field->hi_index = 0;
    if ( NULL == (field->same = sll_new( SLL_NULL_CMP )) )
    {
      bpipe_errno = BPENOMEM;
      return 1;
    }

    sfield = field;

    /* add it to the bpipe's binary tree */
    if ( rbtree_insert( bpipe->hdrf_tree, field ) )
    {
      bpipe_errno = BPENOMEM;
      return 1;
    }
  }
  else
  {
    /* it's in there, set the index and increment the lead field's
       hi_index counter */
    sfield->hi_index++;
    field->index = sfield->hi_index;
  }

  /*
    attach to end of the field's list of similarly named fields and
    to end of bpipe's list
   */
  if ( sll_insert_tail( sfield->same, field ) ||
       sll_insert_tail( bpipe->hdrf_ll, field )      )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }

  return 0;
}

/*

   PROCEDURE
   bpipe_hdrf_string_add - Add a new string data field to the header.

   DESCRIPTION

   This routine adds a character array data field to the header
   associated with the passed @code{BPipe}.  Duplicate fields with the
   same name are allowed; they are given different index numbers.  The
   new field is added to the end of the list of existing fields.
   The passed name specification is duplicated.

   This routine differs from @code{bpipe_hdrf_add} in that the passed
   data is expected to be a @code{C} style null-terminated string.
   It creates a one dimensional character array with an extent equal
   to the length of the passed string (including the terminating null).
   The calling procedure can specify an overriding string length, which
   must include space for the terminating null.

   If the @code{copy} argument is true, a copy is made of the string
   pointed to by the @code{data} argument.  In the case that this
   string is longer than the overriding length (if provided), it is
   truncated so that it will fit in the requested length and
   terminated with a @samp{\0}.  if the @code{copy} argument is false,
   the header field will refer directly to the memory it points to.
   This saves time and space for large data fields.  However, if the
   requested length is shorter than the actual length of the string,
   it will write a @samp{\0} in the passed string at the appropriate
   position.

   If the passed
   data pointer is @code{NULL}, memory is allocated for it, regardless
   of the state of the @code{copy} flag.

   RETURNS
   It returns zero upon success.  Upon failure it returns non-zero,
   and sets @code{bpipe_errno}.

   ERRORS
   Upon error @code{bpipe_errno} is set to one of the following:
   @table @code

   @item BPEBADARG
   @itemize @bullet
   @item
   @item the header was already been written
   @item the @code{BPipe} has been mapped
   @item @code{data} was @code{NULL} and the requested length
         was less than @samp{2}.
   @end itemize

   @item BPENOMEM
     a memory allocation failed

   @end table

   WARNINGS
   If the @code{copy} parameter is false, the @code{data} points to
   a string and the optional length is
   specified and is shorter than the actual length of the string,
   @code{bpipe_hdrf_string_add} will write a @samp{\0} into the
   passed string.  This may cause address violations if the
   string is a const array.

*/
int
bpipe_hdrf_string_add(
  BPipe *bpipe,		/* binary pipe to which to add the field */
  char *name,		/* the name of the field */
  char *data,		/* a pointer to the string */
  size_t len,		/* an optional length.  Set to @code{0} to use
			   the length of the passed string */
  int copy		/* if true, make a copy of the data */
)
{
  BPMatrix *matrix;
  char *string;
  size_t length = 0;

  if ( NULL == data && len < 2 )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  if ( data )
    length = strlen( data ) + 1;

  len = len ? len : length;

  /* create matrix for character array */
  if ( NULL == (matrix = bpipe_matrix_new_va( (size_t) 1, len )) ||
       bpipe_hdrf_add( bpipe, name, BPDType_char, matrix, data, copy ) )
    return 1;

  /* seal off end of string */
  if ( len < length )
  {
    string = bpipe_hdrf_data( bpipe, name, BPHdrfIdx_LAST );
    assert( string != NULL );
    string[len - 1] = '\0';
  }

  return 0;
}

/*

  PROCEDURE
  bpipe_hdrf_array_add - Add a one-dimensional array field to the header.

  DESCRIPTION
  This routine adds a one dimesional array field to the header
  associated with the passed @code{BPipe}.  Duplicate fields with the
  same name are allowed; they are given different index numbers.  The
  new field is added to the end of the list of existing fields.  The
  passed name specification is duplicated.

  This routine simplifies the addition of one-dimensional arrays.
  Normally one must construct a matrix specification and call
  @code{bpipe_hdrf_add}.  This routine takes care of that.

  If the @code{copy} argument is true, a copy is made of the data
  pointed to by the @code{data} argument.  Otherwise, the header field
  will refer directly to the memory it points to.  This saves time and
  space for large data fields.  if the passed data pointer is
  @code{NULL}, memory is allocated for it, regardless of the state of
  the @code{copy} flag.

  RETURNS
  It returns zero upon success.  Upon failure it returns non-zero, and
  sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:

  @table @code

  @item BPEBADARG
  @itemize @bullet
  @item
  @item the header was already been written
  @item the @code{BPipe} has been mapped
  @end itemize

  @item BPENOMEM
     a memory allocation failed

  @end table

*/
int
bpipe_hdrf_array_add(
  BPipe *bpipe,		/* binary pipe to which to add the field */
  char *name,		/* the name of the field */
  BPDataType type,	/* the field's datatype */
  void *data,		/* a pointer to the field's data */
  size_t extent,	/* the extent of the array */
  int copy		/* if true, make a copy of the data */
)
{
  BPMatrix *matrix;

  if ( NULL == (matrix = bpipe_matrix_new_va( (size_t) 1, extent )) ||
       bpipe_hdrf_add( bpipe, name, type, matrix, data, copy ) )
    return 1;

  return 0;
}

/*

  PROCEDURE
  hdrf_ll_delete - @code{SLinkList} node delete routine for header data fields.

  DESCRIPTION
  This routine will delete a header field structure and all of the
  sibling fields with the same name.  It does @emph{not} delete the
  link in the parent @code{bpipe} structure's @code{hdrf_ll} list.

*/
static void
hdrf_ll_delete(
  void *node		/* the first field to delete */
)
{
  sll_delete( ((HdrField *) node)->same, (SLL_DELETE) hdrf_delete );
}

/*

  PROCEDURE
  hdrf_all_delete - @code{SLinkList} node delete routine for header
  data fields.

  DESCRIPTION
  This routine will delete a header field structure and the associated
  link in the parent @code{bpipe} structure's @code{hdrf_ll} list.

*/
static void
hdrf_all_delete(
  void *field,		/* the field to delete */
  void *ll		/* the parent @code{bpipe->hdrf_ll} list */
)
{
  hdrf_delete( (HdrField *) field );
  sll_destroy( (SLinkList) ll, field, rbtree_node_cmp_v );
}


/*

   PROCEDURE
   hdrf_name_cmp - Compare a name to a header parameter's name using
   string collating sequence.

   RETURNS
   It returns @samp{-1}, @samp{0}, or @samp{1} if the name is less than,
   equal to, or greater than, the header parameters name, respectively

*/
static int
hdrf_name_cmp(
  const void *name,		/* a pointer to the name */
  const void *field		/* a pointer to the field */
)
{
  return strcmp( (char *)name, ((HdrField *)field)->name );
}


/*

   PROCEDURE
   hdrf_index_cmp - Compare an index to that in a header field.

   RETURNS
   It returns @samp{-1}, @samp{0}, or @samp{1} if the index is less than,
   equal to, or greater than, the header field's index, respectively

*/
static int
hdrf_index_cmp(
  const void *indexp,		/* a pointer to the name */
  const void *field		/* a pointer to the field */
)
{
  return *( (size_t *) indexp) > ((HdrField *)field)->index ? 1 :
         *( (size_t *) indexp) < ((HdrField *)field)->index ? -1 :
         0;
}


/*

  PROCEDURE
  hdrf_node_name_cmp - Compare two header fields using string collating
  sequence.

  RETURNS
  It returns @samp{-1}, @samp{0}, or @samp{1} if the first field's
  name is less than, equal to, or greater than, the second's,
  respectively.

*/
static int
hdrf_node_name_cmp(
  const void *field1,		/* a pointer to the first field */
  const void *field2		/* a pointer to the second field */
)
{
  return strcmp( ((HdrField *) field1)->name, ((HdrField *) field2)->name );
}


/*
   PROCEDURE
   matrix_memcpy - @code{memcpy} type process function for
   @code{bpipe_matrix_copy}

   DESCRIPTION
   This routine is a process routine for @code{bpipe_datatype_copy}.  It
   does an immediate copy of the data chunk.  The @code{udata} passed
   to it must be a character array with two elements.  The first
   should point to the beginning of the destination data, the
   second should point to the beginning of the source data.

*/
static int
matrix_memcpy(void *udata, size_t dest, size_t src, size_t nbytes)
{
  memcpy(((char **) udata)[0] + dest, ((char **) udata)[1] + src, nbytes);
  return 0;
}
