/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include "config.h"

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

#include "bpipep.h"


/*

  PROCEDURE
  bpipe_iochannel_new - create and initialize an IOchannel structure

  DESCRIPTION
  This routine allocates an @code{IOchannel} structure, making a copy of the
  passed path to the file or device to use as the I/O channel.  It
  does not open the I/O channel.

  RETURNS
  It returns a pointer to a dynamically allocated @code{IOchannel}
  structure.  It returns @code{NULL} if @code{path} is empty or
  @code{NULL}, or if it couldn't allocate memory for the structure..
  Upon error @code{bpipe_errno} is set and @code{NULL} is returned.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:

  @table @code

  @item BPENOMEM
     a memory allocation failed
  @end table

*/
IOchannel *
bpipe_iochannel_new(
  char *path		/* a path to the file or device to use as the
                           I/O channel */
)
{
  IOchannel *channel = NULL;

  if ( NULL == (channel = (IOchannel *) malloc(sizeof(IOchannel))) ||
       NULL == (channel->path = (char *) malloc(strlen(path) + 1)) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  strcpy(channel->path, path);

  channel->dsc = NULL;
  return channel;
}


/*

  PROCEDURE
  bpipe_iochannel_open - open an I/O channel

  DESCRIPTION
  This routine opens a previous allocated I/O channel.

  It recognizes the paths @samp{stdin} and @samp{stdout} as special,
  and connects the channel to @code{stdin} or @code{stdout}
  respectively.  In these cases, mode is ignored.

  RETURNS
  It returns zero upon succes, non-zero if it cannot open the path.
  @code{errno} should be checked upon error (not @code{bpipe_errno}), as
  it is set by @code{fopen}.

*/
int
bpipe_iochannel_open(
  IOchannel *ioc,	/* a pointer to the @code{IOchannel} to open */
  char *mode		/* the mode with which to open the channel, see
                           @code{fopen} for details */
)
{
  /* special treatment if input name is stdin */
  if      (! strcmp(ioc->path, "stdin"))
    ioc->dsc = stdin;
  else if (! strcmp(ioc->path, "stdout"))
    ioc->dsc = stdout;
  else if ( NULL == (ioc->dsc = fopen(ioc->path, mode)) )
      return -1;

  return 0;
}

/*

  PROCEDURE
  bpipe_iochannel_close - close an I/O channel

  DESCRIPTION
  Closes an I/O channel.

*/
void
bpipe_iochannel_close(
  IOchannel *ioc	/* the I/O channel to close */
)
{
  if ( ioc && ioc->dsc )
  {
    fclose( ioc->dsc );
    ioc->dsc = NULL;
  }
}


/*

  PROCEDURE
  bpipe_iochannel_delete - delete an I/O channel structure

  DESCRIPTION
  This routine deallocates the memory associated with an I/O channel.
  if the passed pointer is @code{NULL}, it returns quietly.

*/
void
bpipe_iochannel_delete(
  IOchannel *ioc		/* the channel to deallocate */
)
{
  if ( ioc )
  {
    if ( ioc->path )
      free( ioc->path );
    free( ioc );
  }
}

/*

  PROCEDURE
  bpipe_iochannel_read - read data from an I/O channel.

  DESCRIPTION
  This routine reads data from an I/O channel.  Because of the
  idiosyncracies of the various OS's implementation of @code{fread},
  it isn't as efficient as it can be, nor does it return as
  much information as might be available.

  RETURNS
  It returns the number of objects read.  Upon error it
  returns @samp{0} and sets @code{bpipe_errno} to @code{BPEIOERR}.

*/
size_t
bpipe_iochannel_read(
  void *buf,		/* the destination buffer to fill */
  size_t size,		/* the size of an object to be read */
  size_t n,		/* the number of objects to read */
  IOchannel *ioc	/* the I/O channel to read from */
)
{
  size_t nr = fread( buf, size, n, ioc->dsc );

  if ( nr != n && ferror( ioc->dsc) )
  {
    nr = 0;
    bpipe_errno = BPEIOERR;
  }

  return nr;
}


/*

  PROCEDURE
  bpipe_iochannel_write - Write data to an I/O channel.

  DESCRIPTION
  This routine writes data to an I/O channel.  Because of the
  idiosyncracies of the various OS's implementation of @code{fread},
  it isn't as efficient as it can be, nor does it return as
  much information as might be available.

  RETURNS
  It returns the number of objects written.  Upon error it
  returns @samp{0} and sets @code{bpipe_errno} to @code{BPEIOERR}.

*/
size_t
bpipe_iochannel_write(
  void *buf,		/* the source buffer to write */
  size_t size,		/* the size of an object to write */
  size_t n,		/* the number of objects to write */
  IOchannel *ioc	/* the I/O channel to write to */
)
{
  size_t nr = fwrite( buf, size, n, ioc->dsc );

  if ( nr != n )
  {
    nr = 0;
    bpipe_errno = BPEIOERR;
  }

  return nr;
}


/*

  PROCEDURE
  bpipe_iochannel_fgetrn - Read a record from an I/O channel.

  DESCRIPTION
  This routine reads data from an I/O channel until a newline
  character is read or the channel returns @code{EOF}.  It truncates
  the trailing newline character.  It dynamically resizes its buffers
  to encompass the data, and can thus read any length record (up to
  the memory limits on the machine).

  The calling procedure must provide a pointer to the buffer space and
  an integer of type @code{size_t} which will record the length of the
  buffer.  Upon first calling @code{bpipe_iochannel_fgetrn}, the
  buffer pointer should be set to @code{NULL}.  Note that the
  @emph{addresses} of these variables are passed to
  @code{bpipe_iochannel_fgetrn}.

  RETURNS
  It returns @code{NULL} upon end of file, else a pointer to the input
  buffer.  Upon error it returns @code{NULL} and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:
  @table @code
  @item BPENOMEM
      a memory allocation failed
  @end table

*/
char *
bpipe_iochannel_fgetrn(
  char **buf,		/* a two element array of pointers which will
			   be set to point to the the buffers used to
			   store the input string.  the buffers are
			   dyamically re-allocated, hence the need for
			   the address of the user-supplied
			   pointer. buf[0] should be NULL upon the
			   first call with it. */
  size_t *buf_len,	/* the length of the buffers */
  IOchannel *ioc	/* the channel from which to read the data */
)
{
  char *buffer;
  char *eos;		/* pointer to *<EOS>* in input buffer */
  char *s;		/* pointer to current position in input buffer */

/* size of block by which to extend buffer when full */
#define BUFEXTEND 256

  /* if first time through, allocate input buffer */
  if ( NULL == *buf )
  {
    *buf_len = BUFEXTEND + 1;
    if ( NULL == (*buf = (char *) malloc(*buf_len)) )
    {
      bpipe_errno = BPENOMEM;
      return NULL;
    }
  }

  buffer = *buf;

  eos = buffer + *buf_len - 1;
  s = buffer;

  for (;;)
  {
    /* check for end of file */
    for (; s < eos ; s++)
    {
      /* check for end of file */
      if  ( bpipe_iochannel_read( s, (size_t) 1, (size_t) 1, ioc ) != 1)
      {
        /* reached end of file */
        *s = '\0';		/* seal off end of string */
        /* if no characters were read, return end of file */

        return (s - buffer) ? buffer : NULL;
      }

      /* reached end of string */
      if ( *s == '\n' )
      {
	*s = '\0';
	return buffer;
      }
    }

    /* ran out of room, possibly more to come, reallocate buf */
    if ( NULL == (buffer = (char *) realloc( buffer, *buf_len + BUFEXTEND ) ) )
    {
      bpipe_errno = BPENOMEM;
      return NULL;
    }

    *buf = buffer;
    s = buffer + *buf_len - 1;
    *buf_len += BUFEXTEND;
    eos = s + BUFEXTEND;
  }

  /* NOT REACHED */
}
