/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdarg.h>

#include "bpipep.h"


/*

  PROCEDURE
  bpipe_matrix_override_new - Create a matrix override structure.

  DESCRIPTION
  This routine allocates memory for an output channel matrix override
  structure.  It uses the passed matrix description and offset and
  extent arrays and directs the @code{BPipe} cleanup code to take
  responsibility for freeing them.  The calling procedure should
  @emph{not} free them, nor should it pass them to any other
  @code{BPipe} routine which takes responsibility for freeing them.
  It assumes that the matrix description is legit.

  RETURNS
  Upon success it returns a pointer to the newly allocated override
  structure, else it returns @code{NULL} and sets @code{bpipe_errno}.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following:
  @table @code

  @item BPENOMEM
    a memory allocation failed

  @end table

*/
BPMatrixOverride *
bpipe_matrix_override_new(
  DpktField *dpktf,	/* data packet field to which matrix descriptor
  			   belongs */
  BPMatrix *src,	/* src matrix */
  size_t src_off[],	/* offset of source submatrix to copy to destination.
                           extent is equal to the dimensions of the source
                           matrix. */
  BPMatrix *dst,	/* destination matrix spec */
  size_t dst_off[],	/* offset of destination submatrix.  extent is equal
                           to the dimensions of the destination matrix. */
  size_t extent[]	/* the extents of the submatrix to copy.  the
  			   extent of this array is equal to the
  			   dimensions of the source matrix. */
)
{
  BPMatrixOverride *over;

  if ( NULL ==
       ( over = (BPMatrixOverride *) malloc(sizeof(BPMatrixOverride) ) ) )
  {
    bpipe_errno = BPENOMEM;
    return NULL;
  }

  over->dpktf = dpktf;
  over->matrix = dst;

  if ( src_off || dst_off || extent )
    over->map = bpipe_matrix_map_new( src, src_off, dst, dst_off, extent );
  else
    over->map = NULL;

  return over;
}


/*

  PROCEDURE
  bpipe_matrix_override_delete - Delete a matrix override structure.

  DESCRIPTION
  This routine frees the memory allocated with a matrix override
  structure and the associated matrix description structure.  if the
  passed pointer is @code{NULL}, it returns quietly.

*/
void
bpipe_matrix_override_delete(
  BPMatrixOverride *over	/* the matrix override structure to delete */
)
{
  if ( over )
  {
    bpipe_matrix_delete( over->matrix );
    bpipe_matrix_map_delete( over->map );
    
    free( over );
  }
}


/*

  PROCEDURE
  dpktf_matrix_override_cmp - Compare a data packet field to a
  matrix override structure.

  DESCRIPTION
  This routine compares a data packet field pointer to that in
  a matrix override structure.

  RETURNS
  It returns @samp{-1}, @samp{0}, @samp{1} if the pointer is
  respectively less than, equal to, or greater than that in the
  matrix override structure.

*/
int
bpipe_dpktf_matrix_override_cmp(
  const void *d,		/* name */
  const void *m	        	/* data packet field */
)
{
  DpktField *dpktf = (DpktField *) d;
  BPMatrixOverride *over = (BPMatrixOverride *) m;
  
  if      ( dpktf < over->dpktf ) return -1;
  else if ( dpktf > over->dpktf ) return 1;
  return 0;
}
