/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include <rbtree/rbtree.h>
#include <linklist/linklist.h>

#include <suplib/io.h>
#include <suplib/str.h>

/* ------------------------------------------------------------------- */

typedef struct
{
  SLinkList members;
  char *type;
  char *name;
  size_t size;
} Struct;

typedef struct
{
  char *name;
  char *type;
  size_t size;
} Member;

typedef struct
{
  char *type;
  char *name;
} TypeToName;

/* ------------------------------------------------------------------- */

char *next_line(FILE *fin, void *recbuf);
Struct *new_struct(void);
Member *new_member(size_t n, char *tokens[]);
int struct_cmp(const void *st1, const void *st2);
int struct_type_cmp(const void *type, const void *str);
void insert_default_struct( char *type, size_t size);
void insert_default_structs( void );
int main(int argc, char *argv[]);
int dump_struct(void *data);
int dump_member_size(void *data);
int dump_struct_members(void *data);
int dump_member(void *data, void *udata);
int struct_size(void *data, void *udata);
int member_size(void *data);
int add_members_sizes(void *data, void *udata);
char *struct_name( char *type );
int dump_struct_members_init(void *data);
int dump_member_init(void *data);

/* ------------------------------------------------------------------- */

size_t line;

RBTree structs;
SLinkList structs_in_order;

TypeToName type_to_name[] =
{
  { "unsigned int",	"uint" },
  { NULL,		NULL }
};


/* ------------------------------------------------------------------- */

char *
next_line( FILE *fin, void* recbuf )
{
  char * buf = fget_rec( fin, recbuf );
  if ( NULL == buf )
  {
    fprintf( stderr, "unexpected eof at line %lu\n", (unsigned long) line );
    exit( 1 );
  }
  str_prune( buf );
  ++line;

  return buf;
}

Struct *
new_struct( )
{
  Struct *str = (Struct *) malloc( sizeof(Struct) );
  assert( (long) (str != NULL) );

  str->members = sll_new( SLL_NULL_CMP );

  assert( str->members != NULL );

  str->size = 0;

  return str;
}

Member *
new_member( size_t n, char *tokens[] )
{
  Member *member = (Member *) malloc( sizeof(Member) );

  assert( member != NULL );

  if ( n == 2 )
    member->type = str_dup( tokens[0] );
  else if (n == 3)
  {
    member->type =
      (char *) malloc( strlen(tokens[0]) + strlen(tokens[1]) + 2 );
    assert( member->type != NULL );
    strcpy( member->type, tokens[0] );
    strcat( member->type, " " );
    strcat( member->type, tokens[1] );
  }

  tokens[n-1][strlen(tokens[n-1]) - 1] = '\0';

  member->name = str_dup( tokens[n-1] );

  member->size = 0;

  return member;
}

int
struct_cmp( const void *st1, const void *st2 )
{
  return strcmp( ((Struct *) st1)->type, ((Struct *) st2)->type );
}

int
struct_type_cmp( const void *type, const void *str )
{
  return strcmp( (char *) type, ((Struct *) str)->type );
}

char *
struct_name( char *type )
{
  char *name;
  TypeToName *ttn = type_to_name;

  for ( ttn = type_to_name ; NULL != ttn->type ; ttn++ )
  {
    if (! strcmp( type, ttn->type ) )
    {
      name = str_dup( ttn->name );
      assert( name != NULL );
      return name;
    }
  }

  name = str_dup( type );
  assert( name != NULL );
  return name;
}

void
insert_default_struct( char *type, size_t size )
{
  Struct *str;

  str = new_struct( );
  str->type = str_dup( type );
  str->name = struct_name( type );
  str->size = size;

  assert( ! rbtree_insert( structs, str ) );
  assert( ! sll_insert_tail( structs_in_order, str ) );
}

#define INSERT(type) insert_default_struct( #type, sizeof(type) )
void
insert_default_structs( )
{
  INSERT(char);
  INSERT(double);
  INSERT(int);
  INSERT(unsigned int);
}

int
main( int argc, char *argv[] )
{
  size_t n, n_old;
  FILE *fin = stdin;
  Struct *str;
  void *recbuf;
  char* buf;

  recbuf = fget_rec_new( (size_t) 1024, (size_t) 1024 );

  structs = rbtree_new( struct_cmp );
  structs_in_order = sll_new( SLL_NULL_CMP );

  assert( structs != NULL );
  assert( structs_in_order != NULL );

  insert_default_structs( );

  line = 0;
  
  

  while(( buf = fget_rec( fin, recbuf ) ))
  {
    ++line;
    /* start scan for typedef line */
    str_prune( buf );
    if ( strcmp( buf, "typedef struct" ) )
      continue;

    /* create new structure */
    str = new_struct( );

    /* look for '{' on a single line */
    while ( (buf = next_line( fin, recbuf )) )
    {
      if ( 0 == buf[0] )
	continue;

      if ( strcmp( buf, "{" ) )
      {
	fprintf(stderr, "expected { at line %lu\n", (unsigned long) line );
	exit( 1 );
      }
      break;
    }

    /* grab succeeding lines until one which begins with a '}' */
    while ( (buf = next_line( fin, recbuf )) )
    {
      char *tokens[3];

      if ( '}' == buf[0] )
	break;

      n = toksplit( buf, tokens, " \t", 3 );

      if ( n > 3 || n < 2 )
      {
	fprintf(stderr, "wrong number of tokens in decl at line %lu\n",
		(unsigned long) line );
	exit( 1 );
      }

      assert( ! sll_insert_tail( str->members, new_member( n, tokens ) ) );
      
    }

    /* grab type of structure */
    /* ignore ';' at end of line */
    buf[strlen(buf) - 1] = '\0';

    str->type = str_dup( buf + 1 );
    str->name = struct_name( str->type );

    assert( ! rbtree_insert( structs, str ) );
    assert( ! sll_insert_tail( structs_in_order, str ) );
  }

  /* traverse tree, determining size of structures */
  n = 0;
  n_old = 0;

  for(;;)
  {
    sll_utraverse( structs_in_order, struct_size, &n );

    if ( n == 0 || n == n_old )
      break;

    n_old = n;
    n = 0;
  }

  if ( n != 0 )
  {
    fprintf( stderr, "couldn't resolve all sizes\n" );
    exit( 1 );
  }

  sll_traverse( structs_in_order, dump_struct_members );

  sll_traverse( structs_in_order, dump_struct_members_init );

  puts( "STRUCT STRUCT_ARRAY[]=" );
  puts( "{" );

  /* first entry is a dummy entry */
  puts( "  { (char *) NULL, 0, 0, (MEMBER *) NULL, 0, NULL}," );
  
  sll_traverse( structs_in_order, dump_struct );

  /* last entry is a dummy entry */
  putchar( '\n' );
  puts( "  { (char *) NULL, 0, 0, (MEMBER *) NULL, 0, NULL }" );

  puts  ( "};\n" );
  return 0;
}

int
dump_struct( void *data )
{
  Struct *str = (Struct *) data;

  putchar( '\n' );
  printf( "  { \"%s\", sizeof(%s), \n    ", str->name, str->type );

  if ( sll_count( str->members ) )
  {
    sll_traverse( str->members , dump_member_size );
    puts  ( "0," );
  }
  else
    printf( "sizeof(%s),\n", str->type );

  printf( "    %s_members, sizeof(%s_members) / sizeof(MEMBER), &%s_init }, \n",
	 str->name, str->name, str->name);

  return 0;
}

int
dump_member_size( void *data )
{
  Member *member = (Member *) data;
  Struct *str = rbtree_search( structs, member->type, struct_type_cmp );

  if ( sll_count( str->members ) )
    sll_traverse( str->members, dump_member_size );
  else
    printf( "sizeof(%s) + ", member->type );

  return 0;
}


int
dump_struct_members( void *data )
{
  Struct *str = (Struct *) data;
  char *info[2];

  info[0] = str->type;
  info[1] = "";

  printf( "static MEMBER %s_members[] =\n" , str->name );
  puts  ( "{" );

  if ( sll_count( str->members ) )
    sll_utraverse( str->members, dump_member, info );
  else
    printf( "  { sizeof(%s), 0 }, \n", str->type );
  
  puts  ( "};\n" );

  return 0;
}

int
dump_member( void *data, void *udata )
{
  Member *member = (Member *) data;
  char **info = (char **) udata;
  Struct *str = rbtree_search( structs, member->type, struct_type_cmp );

  if ( sll_count( str->members ) )
  {
    char *new_info[2];

    if ( info[1][0] != '\0' )
    {
      new_info[1] =
	(char *) malloc( strlen( info[1] ) + strlen( member->name ) + 2);
      assert( new_info[1] != NULL );
      sprintf(new_info[1], "%s.%s", info[1], member->name);
    }
    else
      new_info[1] = str_dup( member->name );

    new_info[0] = info[0];
    sll_utraverse( str->members, dump_member, new_info);

    free( new_info[1] );
  }
  else
  {
    if ( info[1][0] != '\0' )
      printf( "  { sizeof(%s), offsetof(%s, %s.%s) },\n",
	      member->type, info[0], info[1], member->name );
    else
      printf( "  { sizeof(%s), offsetof(%s, %s) },\n",
	      member->type, info[0], member->name );
  }

  return 0;
}

int
struct_size( void *data, void *udata )
{
  Struct *str = (Struct *) data;
  size_t *n   = (size_t *) udata;
  size_t size;

  if ( str->size == 0 )
  {
    /* determine sizes of members */
    if ( sll_traverse( str->members, member_size ) )
    {
      (*n)++;
      return 0;
    }

    /* determine size of structure */
    size = 0;
    sll_utraverse( str->members, add_members_sizes, &size );
    str->size = size;
  }

  return 0;
}

int
member_size( void *data )
{
  Member *member = (Member *) data;
  Struct *str = rbtree_search( structs, member->type, struct_type_cmp );

  if ( NULL == str )
  {
    fprintf( stderr, "unable to find struct %s\n", member->type );
    exit( 1 );
  }

  if ( str->size == 0 )
    return 1;

  member->size = str->size;

  return 0;
}

int
add_members_sizes( void *data, void *udata )
{
  Member *member = (Member *) data;
  size_t *size = (size_t *) udata;

  *size += member->size;

  return 0;
}





int
dump_struct_members_init( void *data )
{
  Struct *str = (Struct *) data;

  printf( "static %s %s_init =\n", str->type, str->name );
  puts  ( "{" );

  if ( sll_count( str->members ) )
    sll_traverse( str->members, dump_member_init );
  else
    if ( strcmp( "double", str->type ) )
      fputs( "  0,", stdout );
    else
      fputs( "  0.0,", stdout );
  
  puts  ( "\n};\n" );

  return 0;
}

int
dump_member_init( void *data)
{
  Member *member = (Member *) data;

  Struct *str = rbtree_search( structs, member->type, struct_type_cmp );

  if ( sll_count( str->members ) )
  {
    fputs( "  {", stdout );
    sll_traverse( str->members, dump_member_init);
    fputs( "},", stdout );
  }
  else
      fputs( "  0.0,", stdout );

  return 0;
}

