/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of bpipe
 *
 * bpipe is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * bpipe is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "bpipep.h"

/*----------------------------------------------------------------*/

static int write_hdrf_def(void *data, void *udata);
static int write_dpktf_def(void *data, void *udata);
static int write_hdrf_data(void *data);
static int hdrf_channel_write(void *udata, size_t src, size_t dst,
			      size_t nbytes);
static int write_defn(IOchannel *channel, char *name, BPDataType type, BPMatrix *matrix);

/*----------------------------------------------------------------*/

static IOchannel *	hdrf_channel_write_ioc;

/*----------------------------------------------------------------*/

/*

  PROCEDURE
  bpipe_write_hdr - Write header information to a binary pipe's output
  channels.

  DESCRIPTION
  @code{bpipe_write_hdr} writes header information to a binary pipe's
  output channels.  It must be called after a call to @code{bpipe_map}
  and before any data packets are written to the output channel(s).
  no manipulation of header parameter definitions and data or data
  packet field definitions should be done after invoking this
  function.

  RETURNS
  It returns zero upon success, non-zero upon failure.
  It sets @code{bpipe_errno} upon failure.

  ERRORS
  Upon error @code{bpipe_errno} is set to one of the following errors:
  @table @code
  @item BPEBADARG
  @itemize @bullet
  @item the header had already been written
  @item the @code{BPipe} hasn't been mapped (by @code{bpipe_map})
  @item there are no output channels
  @end itemize

  @item BPEIOERR
      an I/O error occurred

  @item BPENOMEM
      a memory allocation failed

  @end table

*/
int
bpipe_write_hdr(
  BPipe *bpipe		/* the binary pipe for which to output the header */
)
{
  size_t n_bpo;
  size_t n;
  char newline = '\n';
  SLLNode node;

  if ( bpipe->hdr_written ||
       ! bpipe->hdr_mapped ||
       sll_count( bpipe->output ) == 0
     )
  {
    bpipe_errno = BPEBADARG;
    return 1;
  }

  bpipe->hdr_written++;

  n_bpo = sll_count( bpipe->output );

  /* create array of output channels */
  if ( NULL == (bpipe->output_arr = (BPipeOutput **)
		malloc( n_bpo * sizeof( BPipeOutput *))))
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }
  bpipe->output_max = bpipe->output_arr + n_bpo;

  node = sll_head_node( bpipe->output );
  
  /* loop over output channels */
  for ( n = 0 ; n < n_bpo ; n++, node = sll_next_node( bpipe->output, node ) )
  {
    BPipeOutput *bpo = sll_node_get_data( node );

    bpipe->output_arr[n] = bpo;

    /* first write definitions of global header parameters */
    if ( sll_utraverse( bpipe->hdrf_ll, write_hdrf_def, bpo ) )
    {
      bpipe_errno = BPEIOERR;
      return 1;
    }

    /* write end of definition marker */
    if ( bpipe_iochannel_write( &newline, (size_t) 1, (size_t) 1,
				bpo->channel ) != 1 )
    {
      bpipe_errno = BPEIOERR;
      return 1;
    }

    /* do the same for the data packet fields */
    if ( sll_utraverse( bpo->dpktf, write_dpktf_def, bpo ) )
    {
      bpipe_errno = BPEIOERR;
      return 1;
    }

    /* write end of definition marker */
    if ( bpipe_iochannel_write( &newline, (size_t) 1, (size_t) 1,
				bpo->channel ) != 1 )
    {
      bpipe_errno = BPEIOERR;
      return 1;
    }

    /* dump out header field data */
    hdrf_channel_write_ioc = bpo->channel;
    
    if ( sll_traverse( bpipe->hdrf_ll, write_hdrf_data) )
    {
      bpipe_errno = BPEIOERR;
      return 1;
    }

  }


  return 0;

}

/*

  PROCEDURE
  write_hdrf_def - Write the definition of a header field to an output
  channel

  DESCRIPTION
  This is an action routine used when traversing the list of header
  fields which writes out the field definitions.

*/
static int
write_hdrf_def( void *data, void *udata )
{
  HdrField *field = (HdrField *) data;
  BPipeOutput *bpo = (BPipeOutput *) udata;

  return write_defn( bpo->channel, field->name, field->type,
			   field->matrix );
}
    
/*

  PROCEDURE
  write_dpktf_def - Write the definition of a data packet field to an output
  channel

  DESCRIPTION
  This is an action routine used when traversing the list of data packet
  fields which writes out the field definitions.

*/
static int
write_dpktf_def( void *data, void *udata )
{
  BPipeOutputDpktFInfo *info = (BPipeOutputDpktFInfo *) data;
  DpktField *field = info->field;
  BPipeOutput *bpo = (BPipeOutput *) udata;
  BPMatrixOverride *over = sll_search( bpo->matrix, field,
					  bpipe_dpktf_matrix_override_cmp );

  return write_defn( bpo->channel, field->name, field->type,
		     over ? over->matrix :
		       ( field->output ? field->output :
		             (field->core ? field->core : field->input ) ) );
}
    
/*

  PROCEDURE
  write_hdrf_data - Write a header field's data to an output channel.

  DESCRIPTION
  This is an action routine used when traversing the list of header
  fields which writes out the header field data.

  RETURNS
  It returns zero upon success, non-zero upon failure.  It sets
  @code{bpipe_errno} to @code{BPEIOERROR} upon failure.

*/
static int
write_hdrf_data( void *data )
{
  HdrField *field = (HdrField *) data;

  if ( bpipe_datatype_copy( (size_t) 0, (size_t) 0, field->matrix->ne,
			    field->type, BPXMap_Core_to_Output,
			    field->data, hdrf_channel_write ) )
  {
    bpipe_errno = BPEIOERR;
    return 1;
  }

  return 0;
}

/*

  PROCEDURE
  hdrf_channel_write - Write data from memory to an output channel.

  DESCRIPTION
  This is a callback routine for @code{bpipe_datatype_copy} which
  writes the requested number of bytes from the memory pointed to by
  the passed @code{udata} parameter (with the appropriate offset as
  given by @code{bpipe_datatype_copy} to the output channel in the
  file global variable @code{hdrf_channel_output_ioc}.  It ignores the
  @code{dst} argument.

  RETURNS
  It returns zero upon success, non-zero upon failure

*/

static int
hdrf_channel_write( void *udata, size_t src, size_t dst, size_t nbytes )
{
  return nbytes !=
    bpipe_iochannel_write( (char *) udata + src, (size_t) 1, nbytes,
			   hdrf_channel_write_ioc );
}

/*

  PROCEDURE
  write_defn - Write a field definition to an I/O channel.

  DESCRIPTION
  This routine writes a definition (either header field or data
  packet field) to an I/O channel.  It's kludgy to avoid overrunning strings.
  one big sprintf would look nicer. maybe a bpipe_iochannel_printf?

  DIAGNOSTICS
  It returns zero upon success, non-zero upon error.

*/
static int
write_defn(
  IOchannel *	channel,	/* the iochannel */
  char *	name,		/* the name of the data */
  BPDataType	type,		/* the type of the data */
  BPMatrix *	matrix		/* the matrix */
)
{
  size_t i;
  char delim = ' ';
  char eos = '\n';
  BPMatrix *sq;

  if ( bpipe_datatype_write(channel, type) ||
       bpipe_iochannel_write(&delim, (size_t) 1, (size_t) 1, channel) != 1 ||
       bpipe_iochannel_write(name, (size_t) 1, strlen(name), channel) != strlen(name)
       )
    return 1;

  sq = bpipe_matrix_crunch( matrix );
  if ( NULL == sq )
  {
    bpipe_errno = BPENOMEM;
    return 1;
  }

  for ( i = 0 ; i < sq->nd ; i++ )
  {
    char nc[256];
    sprintf( nc, "[%lu]", (unsigned long) sq->extent[i] );
    if ( bpipe_iochannel_write(&nc, (size_t) 1, strlen(nc), channel)
	 != strlen(nc) )
      return 1;
  }
      
  if ( bpipe_iochannel_write(&eos, (size_t) 1, (size_t) 1, channel) != 1 )
    return 1;

  bpipe_matrix_delete( sq );

  return 0;
}





