// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of fits2bp
//
// fits2bp is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// fits2bp is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <string.h>
#include <ctype.h>
#include <limits>

#include <fitsio.h>

#include <tracefctxx/TraceFct.h>

#include "bpmap.h"
#include "fits2bp.h"
#include "input.h"
#include "fits_error.h"

static BPDataType map_types( int fitstype );

static BPipeMap_mapList_Entry *
create_entry( SLinkList bp_list, 
	      SLinkList map_list, 
	      const string& name,
	      long repeat, int fitstype, int col );

static int	repeat_calc( int fitstype, BPDataType type, int repeat );

Input
open_fits( const char* file, SLinkList bp_list, SLinkList map_list, 
	   char* extname, int auto_create )
{
  char err_pfx[1024];
  Input input;
  int status = 0;

  TraceFct tf("open_fits");

  fits_open_file( &input.fptr, file, READONLY, &status );
  fits_error( "", status );

  fits_movnam_hdu( input.fptr, BINARY_TBL, extname, 0, &status );
  if ( BAD_HDU_NUM == status )
    tf.exit( 1, "%s: no binary table with EXTNAME=`%s'",
	     file, extname );
  
  int ncols;
  fits_get_num_cols( input.fptr, &ncols, &status );
  fits_error( "", status );

  for( int col=1 ; col <= ncols ; col++ )
  {
    int fitstype;
    long repeat;
    long width;
    int dummy;
    char tpl[256];
    char name_c[256];

    sprintf( err_pfx, "col %d", col );

    sprintf( tpl, "%d", col );
    fits_get_colname( input.fptr, 0, tpl, name_c, &dummy, &status);
    fits_error( err_pfx, status );

    // lower case name
    for ( size_t i = 0 ; i < strlen( name_c ) ; i++ )
      name_c[i] = tolower(name_c[i]);

    fits_get_eqcoltype( input.fptr, col, &fitstype, &repeat, &width,
		      &status );

    fits_error( err_pfx, status );

    // see if column is in map_list
    string name = name_c;
    BPipeMap_mapList_Entry* entry = (BPipeMap_mapList_Entry*)
      sll_search( map_list, (void*) &name, mapList_key_cmp );

    // At present, cannot correctly handle repeats > 1
    if (   
            (entry       && 1 < repeat_calc( fitstype, entry->type, repeat ) )
	 || (auto_create && 1 < repeat )
       )
      tf.exit( 1, 
	       "column %s has more than one element."
	       "fits2bp cannot (yet) handle this\n", name.c_str() );

    if ( entry )
    {
      entry->col = col;

      int extent = entry->bp->mx ? entry->bp->mx->extent[0] : 1;

      // make sure bits get counted correctly
      repeat = repeat_calc( fitstype, entry->type, repeat );

      if ( repeat > extent )
	tf.exit( 1, "required input column `%s' extent (%d) larger than previously specified (%d)",
		   name.c_str(), repeat, extent );

      entry->extent = repeat;
    }
    else if ( auto_create )
    {
      entry = create_entry( bp_list, map_list, name, repeat, fitstype, col );
    }

  }
  return input;
}

void 
close_fits( Input input )
{
  int status = 0;
  fits_close_file( input.fptr, &status );
  fits_error( "", status );
}

static BPipeMap_mapList_Entry *
create_entry( SLinkList bp_list, 
	      SLinkList map_list, 
	      const string& name,
	      long repeat, int fitstype, int col )
{
  TraceFct tf( "create_entry" );
  BPDataType type = map_types( fitstype );

  // if can't handle it, just return.
  if ( _BPDType_NOTYPE == type )
    return NULL;

  if ( bpipe_check_field_name( name.c_str() ) )
    tf.exit( 1, "can't convert fits name to bpipe name: `%s'", name.c_str() );

  // check if the bpipe field already exists
  if ( sll_search( bp_list, (void *) &name, bpList_key_cmp ) )
    tf.exit( 1, "can't auto create bpipe field `%s': already exists",
	     name.c_str() );

  // make sure bits get counted correctly
  repeat = repeat_calc( fitstype, type, repeat );

  BPipeMap_bpList_Entry* bps = new_bpList_Entry( name, name, type, repeat, 0 );

  if ( sll_insert( bp_list, bps ) )
    tf.exit(1, "error inserting new bpipe field for `%s'", name.c_str() );

  // now create map_list entry
  BPipeMap_mapList_Entry* mps = new_mapList_Entry( name, name, bps, 0,
						   type, col);

  mps->alien_type = fitstype;
  mps->extent = repeat;

  if ( sll_insert( map_list, mps ) )
    tf.exit(1, "error inserting new column mapping for `%s'", name.c_str() );

  return mps;
}

// handle bits gently; they're stored packed so lenth of array to
// request is smaller than number of bits
static int
repeat_calc( int fitstype, BPDataType type, int repeat )
{
  if ( TBIT == fitstype )
  {
    int nbits = bpipe_datatype_size( type ) * CHAR_BIT;
    repeat = ( repeat + (nbits - 1 )) / nbits;
  }

  return repeat;
}

// determine bpipe type which corresponds to fitstype
static BPDataType
map_types( int fitstype )
{
  BPDataType type = _BPDType_NOTYPE;

  
  /*
//    TBIT            X
//    TBYTE           B
//    TLOGICAL        L
    TSTRING         A
//    TSHORT          I
//    TINT32BIT|TLONG J
//    TFLOAT          E
//    TDOUBLE         D
//    TCOMPLEX        C
//    TDBLCOMPLEX     M
//    TINT            J
    TSBYTE          S
//    TUINT           V
//    TUSHORT         U
//    TULONG          V
    TLONGLONG       K
  */

  switch( fitstype )
  {
  case TSHORT:
    type = BPDType_int;
    break;

  case TLONG:
    type = BPDType_int;
    break;

  case TFLOAT:
    type = BPDType_double;
    break;

  case TDOUBLE:
    type = BPDType_double;
    break;

  case TLOGICAL:
    type = BPDType_int;
    break;

    // see also the code in create_entry which knows about the
    // mapping to char
  case TBIT:
    type = BPDType_char;
    break;

  case TBYTE:
    type = BPDType_char;
    break;

  case TSTRING:
    type = BPDType_char;
    break;

  case TCOMPLEX:
    // Maybe you want to warn about this...
    type = BPDType_DComplex;
    break;

  case TDBLCOMPLEX:
    type = BPDType_DComplex;
    break;

  case TINT:
    type = BPDType_int;
    break;

  case TUINT:
    type = BPDType_uint;
    break;

  case TUSHORT:
    type = BPDType_uint;
    break;

  case TULONG:
    type = BPDType_uint;
    break;

  }

  // fall through's get _BPDType_NOTYPE


  return type;
}


