// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of fits2bp
//
// fits2bp is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// fits2bp is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <fstream>
#include <string>
#include <cstring>
#include <vector>

using namespace std;

#include <tracefctxx/TraceFct.h>
#include <linklist/linklist.h>
#include <suplib/str.h>
#include <bpipe/bpipe.h>
#include <bpipe/datatypes.h>

#include <suplibxx/io.h>
#include <suplibxx/str.h>

using namespace suplib;

#include "bpmap.h"
#include "parse_spec.h"

static void create_bp_entry( const string& err_pfx,
			     vector<string>& tokens,
			     SLinkList bp_list );

static BPDataType is_legal_dtype( string& dtype );

static size_t
dtype_offset( const string& err_pfx, BPDataType type, int index, 
	      const string& member, BPDataType& atomic_type );

static int dtype_has_member( BPDataType type );

static void create_map_entry( const string& err_pfx, 
			      vector<string>& tokens,
			      SLinkList map_list, 
			      SLinkList bp_list );

static BPipeMap_bpList_Entry* create_bp_spec( const string& err_pfx, 
					      const string& bp_name, 
					      BPDataType type,
					      const string &extent_s,
					      SLinkList bp_list );

void
parse_spec( const char *file, SLinkList& bp_list, SLinkList& map_list ) 
{
  TraceFct tf("parse_spec");

  // construct lists of bpipe and map specs
  bp_list  = sll_new( bpList_node_cmp );
  map_list = sll_new( mapList_node_cmp );

  if ( NULL == bp_list || NULL == map_list )
    tf.exit( 1, "unable to create bpipe or map spec lists" );

  // if no file is specified, that's ok
  if ( 0 == strcmp( file, "none" ) || 0 == strlen(file) )
    return;

  ifstream ifs( file );
  if ( !ifs )
    tf.exit( 1, "unable to open spec file `%s'", file );

  int line = 0;
  string record;
  vector<string> tokens(20);
  while ( ! getrecord(ifs, record, READ_PHYS | STRIP ).eof() )
  {
    line++;
    char dummy[512];
    sprintf( dummy, "%s: %d: ", file, line );
    string err_pfx = dummy;

    prune(err_pfx);

    // ignore comments
    if ( iscomment( record ) || 0 == record.size() )
      continue;

    tokens.clear();
    tok( tokens, record );

    // create a bpipe entry
    if ( "bpipe" == tokens[0] )
      create_bp_entry( err_pfx, tokens, bp_list );

    // map a field
    else if ( "map" == tokens[0] )
      create_map_entry( err_pfx, tokens, map_list, bp_list );

    // nothing that we know about!
    else
      tf.exit( 1, err_pfx + 
	       "syntax error: unknown directive `" + tokens[0] + "'" );
  }

}


static void
create_bp_entry( const string& err_pfx, 
  vector<string> &tokens, 
		  SLinkList bp_list )
{
  TraceFct tf("create_bp_entry");

  int ntok = tokens.size();
  if ( ntok < 3 || ntok > 4 )
    tf.exit( 1, "%s: syntax error: incorrect number of fields: %d",
	     err_pfx.c_str(), ntok );

  int tokn = 1;
  string bp_name = tokens[tokn];
  // check if name is already in bpipe list. if so it's an error
  if ( sll_search( bp_list, (void*) &bp_name, bpList_key_cmp ) )
    tf.exit( 1, "%s: bpipe field `%s' already specified", 
	     err_pfx.c_str(), bp_name.c_str() );

  // check if type is one of the recognized/supported ones
  tokn++;
  BPDataType type = is_legal_dtype(tokens[tokn]);
  if ( -1 == type )
    tf.exit( 1, "%s: bpipe type '%s' not supported",
	     err_pfx.c_str(), tokens[tokn].c_str() );

  // see if there's an extent
  tokn++;
  string extent_s;
  if ( ntok > tokn) 
    extent_s = tokens[tokn];

  create_bp_spec( err_pfx, bp_name, type, extent_s, bp_list );
}

static void
create_map_entry( const string& err_pfx, 
	          vector<string>& tokens, 
		  SLinkList map_list, 
		  SLinkList bp_list )
{
  TraceFct tf("create_map_entry");

  int ntok = tokens.size();
  if ( ntok < 3 || ntok > 4 )
    tf.exit( 1, "%s: syntax error: incorrect number of fields: %d", 
	     err_pfx.c_str(), ntok );

  int tokn = 1;
  string name = tokens[tokn];
  // check if name is already in map list. if so it's an error
  if ( sll_search( map_list, (void*) &name, mapList_key_cmp ) )
    tf.exit( 1, "%s: map for input field `%s' already specified", 
	     err_pfx.c_str(), name.c_str() );

  // resolve bp_name into the possible components
  // name, index, member, i.e. foo[30].x
  string bp_name;
  string member;
  string::size_type index = 0;

  tokn++;
  {
    string::size_type end_name = string::npos;
    string::size_type end = 0;
    string::size_type bra = tokens[tokn].find('[');

    if ( bra == 0 )
	tf.exit( 1, "%s: syntax error: missing name: `%s'",
		 err_pfx.c_str(), tokens[tokn].c_str() );

    if ( string::npos != bra )
    {
      end_name = bra;
      string::size_type ket = tokens[tokn].find(']', bra + 1 );

      if ( string::npos == ket || bra + 1 == ket)
	tf.exit( 1, "%s: syntax error: incomplete array index: `%s'",
		 err_pfx.c_str(), tokens[tokn].c_str() );

      try {
	index = str2l( tokens[tokn].substr(bra+1, ket - bra - 1).c_str() );
      }
      catch ( Exception& e )
      {
	e.set_message( err_pfx + 
		       "syntax error: illegal index specification: " +
		       tokens[tokn] );
	tf.exit( 1, e );
      }      
      end = ket + 1;
    }

    // check for structure member
    string::size_type dot = tokens[tokn].find( '.', end );

    if ( dot == 0 )
	tf.exit( 1, "%s: syntax error: missing name: `%s'",
		 err_pfx.c_str(), tokens[tokn].c_str() );

    if ( string::npos != dot )
    {
      if ( string::npos == end_name )
	end_name = dot;
      if ( dot == tokens[tokn].size() - 1 )
	tf.exit(1, err_pfx + 
		"syntax error: illegal member specification: "
		+ tokens[tokn] );

      member = tokens[tokn].substr( dot + 1 );
    }

    bp_name = tokens[tokn].substr(0,end_name);
  }

  // see if this is an existing field
  BPipeMap_bpList_Entry *bps = 
    (BPipeMap_bpList_Entry*) sll_search( bp_list, (void*) &bp_name,
					 bpList_key_cmp );

  // see if a datatype was specified
  tokn++;
  BPDataType type;

  if ( bps )
    type = bps->type;
  else
    type = _BPDType_NOTYPE;

  if ( ntok > tokn )
  {
    // check if type is one of the recognized/supported ones
    if ( -1 == ( type = is_legal_dtype(tokens[tokn]) ) )
      tf.exit( 1, "%s: bpipe type '%s' not supported",
	       err_pfx.c_str(), tokens[tokn].c_str() );
    
    if ( bps && type != bps->type )
      tf.exit( 1, 
	       "%s: type `%s' does not match that of pre-existing bpipe field",
	       err_pfx.c_str(), tokens[tokn].c_str() );
  }

  // if no type was specified, must be an existing field.
  if ( _BPDType_NOTYPE == type && !bps )
    tf.exit(1, "%sattempt to map to non-existant bpipe field (`%s')",
	    err_pfx.c_str(), bp_name.c_str() );

  // if there's no entry in the bp_list, create one.
  if ( !bps )
  {
    string extent_s;
    bps = create_bp_spec( err_pfx, bp_name, type, extent_s, bp_list );
  }

  // if there's an index specified, and it's larger than what's
  // possible, complain
  if ( index && ( ! bps->mx || index > bps->mx->extent[0]-1 ) )
    tf.exit(1, "%sarray index (%d) larger than allowed by array extent (%d)\n", 
	    err_pfx.c_str(), index, bps->mx ? bps->mx->extent[0] : 0);

  // have a type by now.  if it's a type that has members (i.e. a
  // structure) make sure that a member was specified
  if ( dtype_has_member(type) && 0 == member.size() )
    tf.exit(1, "%smissing member specification for `%s'",
	    err_pfx.c_str(), name.c_str() );

  // now, assemble map structure
  {
    BPDataType atomic_type;
    size_t offset = dtype_offset( err_pfx, type, index, member,
				  atomic_type);
    BPipeMap_mapList_Entry * mps = 
      new_mapList_Entry( err_pfx, name, bps, offset, atomic_type, 0 );

    if ( sll_insert( map_list, mps ) )
      tf.exit(1, err_pfx + "error inserting new column mapping" );
  }
}

static BPipeMap_bpList_Entry*
create_bp_spec( const string& err_pfx, 
		const string& bp_name, 
		BPDataType type, 
		const string& extent_s, 
		SLinkList bp_list )
{
  TraceFct tf( "create_bp_spec" );

  // make sure that it's a legal bpipe name
  if ( bpipe_check_field_name( bp_name.c_str() ) )
    tf.exit( 1, err_pfx + "illegal field name: " + bp_name );

  int ext = 0;
  
  if ( extent_s.size() )
  {
    try {
      ext = str2i( extent_s.c_str() );
    }
    catch ( Exception& e )
    {
      e.set_message( err_pfx + "syntax error: illegal extent: " +
		     extent_s );
      tf.exit(1, e );
    }
  }

  BPipeMap_bpList_Entry* bps =
    new_bpList_Entry( err_pfx, bp_name, type, ext, 0 );

  if ( sll_insert( bp_list, bps ) )
    tf.exit(1, err_pfx + "error inserting new BPipeMap_bpList_Entry" );

  return bps;
}


// a very limited list of supported datatypes.  eventually this
// should go away and all of the bpipe data types should automatically
// be supported

static TokListToken dtype[] =
{
  { "DVector2",	BPDType_DVector2 },
  { "DVector3",	BPDType_DVector3 },
  { "IVector2",	BPDType_IVector2 },
  { "IVector3",	BPDType_IVector3 },
  { "char",	BPDType_char	 },
  { "double",	BPDType_double	 },
  { "int",	BPDType_int	 },
  { "uint",	BPDType_uint	 }
};
TokList dtypes = GenTokList(dtype);

// check if type is one of the recognized/supported ones
static BPDataType
is_legal_dtype( string &dtype_s )
{
  return (BPDataType) tokmatch( dtype_s.c_str(), &dtypes );
}

// this should be done by bpipe, not here!
static size_t
dtype_offset( const string& err_pfx, BPDataType type, int index, 
	      const string& member, BPDataType& atomic_type )
{
  TraceFct tf( "dtype_offset" );
  size_t size = bpipe_datatype_size( type );
  size_t offset = size * index;

  switch( type )
  {
  case BPDType_DVector2:
    if      ( "x" == member )
      offset += offsetof( DVector2, x );
    else if ( "y" == member )
      offset += offsetof( DVector2, y );
    else
      tf.exit(1, "%s: DVector2 doesn't have a `%s' member", 
	      err_pfx.c_str(), member.c_str() );
    atomic_type = BPDType_double;
    break;
    
  case BPDType_DVector3:
    if      ( "x" == member )
      offset += offsetof( DVector3, x );
    else if ( "y" == member )
      offset += offsetof( DVector3, y );
    else if ( "z" == member )
      offset += offsetof( DVector3, z );
    else
      tf.exit(1, "%s: DVector3 doesn't have a `%s' member",
	      err_pfx.c_str(), member.c_str() );
    atomic_type = BPDType_double;
    break;

  case BPDType_IVector2:
    if      ( "x" == member )
      offset += offsetof( IVector2, x );
    else if ( "y" == member )
      offset += offsetof( IVector2, y );
    else
      tf.exit(1, "%s: IVector2 doesn't have a `%s' member", 
	      err_pfx.c_str(), member.c_str() );
    atomic_type = BPDType_int;
    break;
    
  case BPDType_IVector3:
    if      ( "x" == member )
      offset += offsetof( IVector3, x );
    else if ( "y" == member )
      offset += offsetof( IVector3, y );
    else if ( "z" == member )
      offset += offsetof( IVector3, z );
    else
      tf.exit(1, "%s: IVector3 doesn't have a `%s' member", 
	      err_pfx.c_str(), member.c_str() );
    atomic_type = BPDType_int;
    break;

  default:
    if ( member.size() )
      tf.exit(1, err_pfx + "bpipe field is not a structure" );
    atomic_type = type;
  }

  return offset;
}


static BPDataType DT_has_member[] =
{ BPDType_DVector2,
  BPDType_DVector3,
  BPDType_IVector2,
  BPDType_IVector3 
};

#define N_DT_HM (sizeof(DT_has_member) / sizeof(BPDataType))

static int
dtype_has_member( BPDataType type )
{
  for( size_t i = 0 ; i < N_DT_HM ; i++ )
    if ( type == DT_has_member[i] )
      return 1;

  return 0;
}
