/****************************************************************
 * Source     : matrix_ops.c
 */

/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of flat
 *
 * flat is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * flat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <stdio.h>

#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>
#include <bpipe/datatypes.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <mst_nrutil/mst_nrutil.h>

#include "flatbpipe.h"


/*
  PROCEDURE
    Find the inverse of a generalized 4x4 transformation matrix.

  DESCRIPTION
    Calculates the inverse

  RETURNS
    It returns a pointer to a transformation matrix containing the
    inverse of the passed matrix.  If it cannot allocate memory
    for the matrix, it prints a message to stderr and exits (via
    tf_exit).

*/
double **
  inverse_transformation( double **a )
{
 
  double tmp;
  int i, j;
  double **a_inverse = dmatrix_r( 0, DIM3, 0, DIM3 );

  if (a_inverse == NULL)
  {
    tf_enter("inverse_transformation");
    tf_exit(ExitERR_alloc, "unable to allocate matrix inverse result matrix");
  }

  for ( i = 0; i < DIM4; i++ )
    for ( j = 0; j < DIM4; j++ )
      a_inverse[i][j] = 0.0;

  a_inverse[DIM3][DIM3] = 1.0;

  /*
    Rotation component of the inverse matrix = 
             Transpose of the rotation matrix component.
  */
  for ( i = 0; i < DIM3; i++ )
    for ( j = 0; j < DIM3; j++ )
      a_inverse[i][j] = a[j][i];

  /*
     Translation component of the inverse matrix = 
             - Translation component times the transpose of the rotation matrix
  */
  for ( i = 0; i < DIM3; i++ ) {
    for ( tmp = 0.0, j = 0; j < DIM3; j++ )
	tmp += a[DIM3][j] * a[i][j];
    a_inverse[DIM3][i] = - tmp;
  }

  return a_inverse;
}

/*
  PROCEDURE
    Multiplication of two double matrices of size DIM4 by DIM4.

  DESCRIPTION
    Calculates the matrix multiplication of DIM4 by DIM4 matrix `a'
    by a matrix DIM4 by DIM4 matrix `b'.  The results will be stored
    in the matrix matrix `ab', the matrix ab will be allocated within
    the routine.

  RETURNS
    The pointer to the matrix whose elements are the product of matrix
    a times b.  If it cannot allocate the result matrix, it prints
    an error to stderr via tf_exit() and exits

  AUTHOR
    Dan Nguyen
*/
double **
  matrix_mult( double **a  /* The first matrix in: a * b */, 
	      double **b   /* The second matrix in: a * b */ )
{
  int i, j, k;
  double tmp;

  double **ab = dmatrix_r( 0, DIM3, 0, DIM3 );

  
  if (ab == NULL)
  {
    tf_enter("matrix_mult");
    tf_exit(ExitERR_alloc, "unable to allocate matrix mult result matrix");
  }

  for ( i = 0; i < DIM4; i++ )
    for ( j = 0; j < DIM4; j++ ) {
      for ( tmp = 0.0, k = 0; k < DIM4; k++ ) 
	tmp += a[i][k] * b[k][j];
      ab[i][j] = tmp;
    }

  return ab;
}

void print_4x4_matrix( double** mat, char* str ) {

  int iindex, jindex;
  
  if ( str )
    fprintf( stderr, "%s :\n", str );

  for ( iindex = 0; iindex < DIM4; iindex++ )
    for( jindex = 0; jindex < DIM4; jindex++ ) {
      fprintf( stderr, "%-10.5e", mat[iindex][jindex] );
      if ( jindex < DIM4-1 )
	fprintf( stderr, " " );
      else
	fprintf( stderr, "\n" );
    }
  fprintf( stderr, "\n" );

  return;
}

/*
  PROCEDURE
    Rotation of Co-ordinate Axes.

  DESCRIPTION
    Calculate the 3-D axes rotation, the return pointer of the
    procedure.  The axes are rotated anti-clockwise though an angle 
    theta radians about an axis specified by the argument `axis'. 

  RETURNS
    A double pointer to the matrix whose elements are the
    The rotation matrix.

  AUTHOR
    Dan Nguyen
*/
double **
  rotation(
	   XYZaxis axis,     /* The axis of rotation. The variable
			       axis can be one of: 0,1,2 to indicate a
			       rotation about the x,y,z axis
			       respectively */
	   double theta  /* Rotation about of angle theta, in radians */
	   )
{

  int i, j;
  double cos_theta=cos(theta);
  double sin_theta=sin(theta);
  double **rotation_matrix = dmatrix_r( 0, DIM3, 0, DIM3 );

  if ( rotation_matrix == NULL)
  {
    tf_enter("rotation");
    tf_exit(ExitERR_alloc, "unable to allocate rotated matrix");
  }

  tf_assert(axis == XAXIS || axis == YAXIS || axis == ZAXIS);

  for ( i = 0; i < DIM4; i++ )
    for( j = 0; j < DIM4; j++ )
      rotation_matrix[i][j] = 0.0;

  rotation_matrix[axis][axis] = 1.0;
  rotation_matrix[DIM3][DIM3] = 1.0;

  i = ( axis + 1 ) % 3;
  j = ( i + 1 ) % 3;

  rotation_matrix[i][i] = cos_theta;
  rotation_matrix[j][j] = cos_theta;
  rotation_matrix[i][j] = sin_theta;
  rotation_matrix[j][i] = -sin_theta;

  return rotation_matrix;
}

/*
  PROCEDURE
    Rotation of Co-ordinate Axes.

  DESCRIPTION
    Calculate the 3-D axes rotation, the return pointer of the
    procedure.  The axes are rotated anti-clockwise through an angle 
    theta radians about an  axis specified by the argument `axis'.
    If `axis' is XAXIS, YAXIS or ZAXIS then the rotation shall be about
    the axis x, y or z, respectively.

  AUTHOR
    Dan Nguyen
*/
double **
  rotate(
	 XYZaxis axis,		/* axis of rotation */ 
	 double  angle		/* angle to rotate, in radians */
	 )
{
  double **mat;

  tf_enter("rotate");
  
  /*
    rotation angle for photon is *opposite* of that for the
    aperture.  since this is called when constructing aperture, negate it
  */

  mat = rotation( axis, -angle );
   
  tf_leave();
  return mat ;
}

/*
  PROCEDURE
    Translation of Origin.
  
  
  DESCRIPTION
    Calculate the 3-D translation matrix `mat'. If the origin of
  
  RETURNS
    The translational matrix.

  AUTHOR
    Dan Nguyen
*/
double **
  translation( double tran[ DIM3 ]  )
{
  int i, j ;
  double **mat;
 
  mat = dmatrix_r( 0, DIM3, 0, DIM3 );

  if ( mat == NULL )
  {
    tf_enter("translation");
    tf_exit(ExitERR_alloc,
	    "unable to allocate matrix translation result matrix");
  }

  for ( i = 0; i < DIM4; i++ ) {
    for( j = 0; j < DIM4; j++ ) {
      mat[i][j] = 0.0;
    }
  }

  for ( i = 0; i < DIM4; i++ )
    mat[i][i] = 1.0;

  for ( i = 0; i < DIM3; i++ )
    mat[DIM3][i] = - tran[i];  /* NOTE MINUS SIGN */

  return mat;
}

void
transform_position( double **tran, DVector3 *in, DVector3 *out )
{ /* ROW vectors */
  out->x = tran[0][0] * in->x + tran[1][0] * in->y  + tran[2][0] * in->z +
    tran[3][0];
  out->y = tran[0][1] * in->x + tran[1][1] * in->y  + tran[2][1] * in->z +
    tran[3][1];
  out->z = tran[0][2] * in->x + tran[1][2] * in->y  + tran[2][2] * in->z +
    tran[3][2];

return;
}

void
transform_direction( double **tran, DVector3 *in, DVector3 * out)
{  /* ROW vectors */
  out->x = tran[0][0] * in->x + tran[1][0] * in->y + tran[2][0] * in->z ;
  out->y = tran[0][1] * in->x + tran[1][1] * in->y + tran[2][1] * in->z ;
  out->z = tran[0][2] * in->x + tran[1][2] * in->y + tran[2][2] * in->z ;
  return;
}

/*

with(linalg);
Let r denote the 3x3 rotation matrix, t the 1x3 translation displacement.

a:=matrix(2,2,[r,0,t,1]);

                                      [ r  0 ]
                                 a := [      ]
                                      [ t  1 ]

assume that matrix b is of the same form as a

ainverse:=matrix(2,2,[rr,0,tt,1]);

                                         [ rr  0 ]
                             ainverse := [       ]
                                         [ tt  1 ]

multiply(a,ainverse);

                                [    r rr    0 ]
                                [              ]
                                [ t rr + tt  1 ]

The matrix above must be equal to the Identity matrix.
Hence, the following conditions must be satisfied:
                  1) r * rr = I 
		  and 
		  2) t * rr + tt = 0

The first condition is satisfied by requiring that:  rr = transpose(r), 
after all r is the rotation matrix.

The second condition is satisfied by solving t*rr + tt = 0, thus tt must
be -t * rr
  
solve(t*rr+tt,tt);

                                     - t rr

                                  [  transpose(r)       0 ]
                      ainverse := [                       ]
                                  [ - t * transpose(r)  1 ]

*/
