/****************************************************************************
 * Source     : process.c
 * Author     : David Grumm
 * Description: main driver for bpipe flat
 */

/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of flat
 *
 * flat is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * flat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <mst_nrutil/mst_nrutil.h>
#include <bpipe/datatypes.h>    
#include <fullray/fullray.h> 
#include <mathconst/mathconst.h>    

#include "flatbpipe.h"
#include "main.h"
#include "common.h"

void
process(void)
{
  double transl[ DIM3 ];
  double rot_angle1, rot_angle2, rot_angle3;
  DVector3 inmirror;      /* point in mirror in global system */
  DVector3 inter_transf;   /* intersection in tranformed system */
  DVector3 normal;
  Ray      ray;
  Bpipe    pipe;
  
  int num_incident=0, num_transmitted=0;
  double rayscalar;
  double **translate_mat=NULL;
  double **rot_mat1=NULL, **rot_mat2=NULL, **rot_mat3=NULL;
  double **total_mat=NULL;
  double **inv_total_mat=NULL;
  DVector3 odir_flat;  /* original ray direction in flat system */
  DVector3 rdir_global;  /* reflected ray direction in global system */
  
  tf_enter("process");
  
  normal.x = src.normx;  /* components of normal */
  normal.y = src.normy;
  normal.z = src.normz;
  
  transl[ 0 ] = src.x0;   /* coordinates of center of flat */
  transl[ 1 ]=  src.y0;
  transl[ 2 ] = src.z0;
  
  /*  make sure the normal vector is normalized */
  normalize( &normal );
  
  init_bpipe( src.input, src.output, &pipe, &ray );
  
  /*  translate coordinate system to center of flat*/ 
  translate_mat = translation( transl );
  
  /* rotate about Z axis (1st rotation to align with normal) */    
  rot_angle1 = atan2( normal.y, normal.x ) - PI/2.;
  rot_mat1 = rotate( ZAXIS, rot_angle1 );
    
  /* rotate about X axis (2nd rotation to align with normal) */    
  rot_angle2 =  - atan2( sqrt( normal.x * normal.x +
			       normal.y * normal.y ) , normal.z);
  rot_mat2 = rotate( XAXIS, rot_angle2 );
    
  /* rotate about Z axis (rotation of flat about its center) */
  /* (convert from user-input degrees to radians) */    
  rot_angle3 =  src.rotation * PI/180;
  rot_mat3 = rotate( ZAXIS, rot_angle3);
    
  /* multiply transformations; total transformation matrix is total_mat */
  {
    double** temp_mat1 = matrix_mult( rot_mat2, rot_mat3 );
    double** temp_mat2 = matrix_mult( rot_mat1, temp_mat1 );
    total_mat = matrix_mult( translate_mat, temp_mat2 );
    
    free_dmatrix( temp_mat1, 0, DIM3, 0, DIM3 );
    free_dmatrix( temp_mat2, 0, DIM3, 0, DIM3 );
    free_dmatrix( translate_mat, 0, DIM3, 0, DIM3 );
    free_dmatrix( rot_mat1, 0, DIM3, 0, DIM3 );
    free_dmatrix( rot_mat2, 0, DIM3, 0, DIM3 );
    free_dmatrix( rot_mat3, 0, DIM3, 0, DIM3 );
  }

  /* calculate inverse matrix for later */
  inv_total_mat = inverse_transformation( total_mat );

  /*  obtain data from bpipe */
  while( bpipe_read_dpkts( pipe.bpipe, pipe.data, 1 ) ) {
    
    num_incident++; /*   increment # of incident rays */    
    
    /*  calculate scalar rayscalar */  
    rayscalar = calc_rayscalar( &ray, &inmirror, &normal );
    
    /*  find intersection of ray and plane of flat in global system */
    find_intersection( &ray, rayscalar, &inmirror );
    
    /* transform position of ray intersection into flat system */
    transform_position( total_mat, &inmirror, &inter_transf );  

    /* assign normal vector for ray */
    *(ray.norm) = normal; 
    
    /* check if ray hit optic */
    if ( check_optic_hit( &ray, &inter_transf, &normal ) ) {
      /* photon has  hit optic, reset ray position to intersection point and
	 transform direction of ray. */
      *(ray.pos) = inmirror;
      
      /************* reflection code commented out for now ********/      
      /*
	transform_direction( total_mat, ray.dir, &odir_flat );    
	reflect_ray( &odir_flat ) ;
	*(ray.dir) = odir_flat; 
	transform_direction( inv_total_mat, ray.dir, &rdir_global );
        *(ray.dir) = rdir_global;
	*/
      /*****************  end of reflection code *******************/      
      
    } else {
      num_transmitted++;  /* increment # of missed rays */    
    }
    /* write ray to bpipe */
    if ( bpipe_write_dpkt( pipe.bpipe, pipe.data, pipe.outstream ) ) {
      tf_enter("bpipe_write_dpkt()");  
      tf_exit(ExitERR_fread, "error while writing ray to `%s'",
	      src.output,"bpipe errno is %s", bpipe_strerror(bpipe_errno));  
    }
  }
  
  if ( src.printdebug )
    fprintf( stderr, "%d\n", num_transmitted );  
  
  free_dmatrix( total_mat, 0, DIM3, 0, DIM3 );
  free_dmatrix( inv_total_mat, 0, DIM3, 0, DIM3 );   
  
  /*  free pipe resources  */
  bpipe_delete(pipe.bpipe);
  free(pipe.data);
  
  tf_leave();
}

