/****************************************************************
 * Source     : readbp.c
 */

/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of flat
 *
 * flat is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * flat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>

#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>
#include <bpipe/datatypes.h>
#include <tracefct/exiterrvals.h>
#include <tracefct/tracefct.h>

#include "readbp.h"
#include "flatbpiperay.h"


/* simple routine to get the data pointer for a named data packet field */
static char *dpktf_data( Bpipe *bpipe, char *name )
{
  DpktField *field = bpipe_dpktf( bpipe->bpipe, name );

  if ( field == NULL )
    tf_exit( ExitERR_misc, name );

  return bpipe_dpktf_data( field, bpipe->data );
}

/* ensure existence of fields  */
void create_field( Bpipe *bpipe, char *name, BPDataType type) {  

  switch( bpipe_dpktf_add( bpipe->bpipe, name, type, NULL ) )
    {
    case -1: /* error creating field */	
      tf_exit( ExitERR_misc, "error creating data packet field: bpipe errno is  %s", bpipe_strerror( bpipe_errno) ); 
      break;
    case 0: /* successful creation */
      break;
    case 1: /* field already exists */
      break;
    default: /* impossible to get here */
      tf_assert ((bpipe_dpktf_add( bpipe->bpipe, name, type, NULL )  == -1) ||
		 (bpipe_dpktf_add( bpipe->bpipe, name, type, NULL)  == 0) ||
		 (bpipe_dpktf_add( bpipe->bpipe, name, type, NULL) == 1) );
      break;
    }
}

void init_bpipe( char* input, char* output,
		 Bpipe* pipe, Ray* photon_offset ) {
  
  unsigned int surface_num;
  size_t npos;
  DpktField *dpktf_pos, *dpktf_dir, *dpktf_wt ;
  size_t dpkt_core_size;


  tf_enter( "init_bpipe" );

  if( NULL == (pipe->bpipe=bpipe_new( )) )
     tf_exit( ExitERR_misc, "error on bpipe new: bpipe errno is %s",  
	     bpipe_strerror( bpipe_errno) ); 
  
  if ( bpipe_input( pipe->bpipe, input ) )
    tf_exit( ExitERR_misc, "error on bpipe input: bpipe errno is %s",  
	     bpipe_strerror( bpipe_errno) );
  if ( NULL == (pipe->outstream=bpipe_output( pipe->bpipe, output ) ) )
    tf_exit( ExitERR_misc, "error on bpipe output: bpipe errno is %s",  
	     bpipe_strerror( bpipe_errno) );
  create_field( pipe, BP_AXAF_surf_missed, BPDType_uint);  
  create_field( pipe, BP_AXAF_surf_norm, BPDType_DVector3);  

    /*  add surface number field and initialize */
  if ( 0 == bpipe_hdrf_n( pipe->bpipe, "srfno"))
    {
      surface_num = 1; 
      if ( bpipe_hdrf_add( pipe->bpipe,"srfno", BPDType_uint, NULL,
			   &surface_num, 1))
	{
	  tf_exit(ExitERR_misc, "bpipe_hdrf_add failure; bpipe errno is %s:",
		  bpipe_strerror(bpipe_errno));  
	}
    }
  
  /* map data packet fields */
  if ( NULL == ( dpkt_core_size = bpipe_map( pipe->bpipe ) ) &&
       bpipe_errno != BPNOERROR )
    tf_exit( ExitERR_misc, "error on bpipe map: bpipe errno is %s",  
	     bpipe_strerror( bpipe_errno) );  
  
  /* output header. must do this after calling bpipe_map */
  if ( bpipe_write_hdr( pipe->bpipe ) )
    tf_exit( ExitERR_misc, "error writing bpipe hdr: bpipe errno is %s",  
	     bpipe_strerror( bpipe_errno) );     

  /* space for instance of data packet */
  if ( NULL == (pipe->data = malloc(dpkt_core_size)))
    tf_exit( ExitERR_misc, "error on core malloc: bpipe errno is %s",  
	     bpipe_strerror( bpipe_errno) );  
  photon_offset->wt      = (double*) dpktf_data( pipe, BP_AXAF_phot_wt );
  photon_offset->energy  = (double *) dpktf_data( pipe, BP_AXAF_phot_energy );
  photon_offset->pos     = (DVector3*) dpktf_data( pipe, BP_AXAF_phot_pos );
  photon_offset->dir     = (DVector3*) dpktf_data( pipe, BP_AXAF_phot_dir ); 
  photon_offset->missed  = (bitmask_t*) dpktf_data( pipe, BP_AXAF_surf_missed);
  photon_offset->norm    = (DVector3*) dpktf_data( pipe, BP_AXAF_surf_norm );
  
  
  tf_leave();
  return;
}


