#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <bpipe/bpipe.h>
#include <bpipe/datatypes.h>
#include <bpipe/bpipe_axaf.h>

#include <fullray/fullray.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "dpktf_data.h"
#include "process.h"

void
process(Global *g)
{
  FILE *fin;
  BPipe *bpipe;
  BPipeOutput *bpo;
  fullray fr;

  size_t s_obj_out, nphot;

  void *obj_out;

  DVector3 *	position;
  DVector3 *	direction;
  DVector3 *	surface_norm;
  double *	weight;
  double *	graze_angle;
  double *	energy;
  double *	path_length;
  DCVector2 *	pol_cos;
  DCVector2 *	pol_sin;
  unsigned int	* id;
  int *		handy;
  
  unsigned int surface_num;
  int delga;
  int delnorm;
  int ga_offset;
  
  tf_enter("process");
  
  if (! strcmp(g->input, "stdin") )
    fin = stdin;
  else
    {
      fin = fopen(g->input, "r");
      if (NULL == fin)
	tf_exit(ExitERR_fopen, "couldn't open %s for input", g->input);
    }
  
  if ( NULL == ( bpipe = bpipe_new( ) ) )
    tf_exit(ExitERR_alloc, "unable to create bpipe" );
  
  if ( NULL == ( bpo = bpipe_output( bpipe, g->output ) ) )
    tf_exit(ExitERR_alloc, "unable to attach output channel `%s': %s",
	    g->output, bpipe_strerror(bpipe_errno));
  
  if ( 
      bpipe_dpktf_add( bpipe, BP_AXAF_phot_pos,  BPDType_DVector3, NULL )
      ||
      bpipe_dpktf_add(bpipe, BP_AXAF_phot_dir, BPDType_DVector3, NULL )
      ||
      bpipe_dpktf_add(bpipe, BP_AXAF_phot_wt, BPDType_double,NULL )
      ||
      bpipe_dpktf_add(bpipe, BP_AXAF_phot_energy, BPDType_double, NULL )
      ||
      bpipe_dpktf_add(bpipe, BP_AXAF_path_length, BPDType_double, NULL )
      ||
      bpipe_dpktf_add(bpipe, BP_AXAF_phot_id, BPDType_uint, NULL )
      ||
      bpipe_dpktf_add(bpipe, BP_AXAF_pol_cos, BPDType_DCVector2, NULL)
      ||
      bpipe_dpktf_add(bpipe, BP_AXAF_pol_sin, BPDType_DCVector2, NULL)
      ||
      bpipe_dpktf_array_add(bpipe, "handy", BPDType_int, 3)
      ||
      bpipe_dpktf_add( bpipe, BP_AXAF_surf_norm, BPDType_DVector3, NULL )
      ||
      bpipe_dpktf_array_add( bpipe, BP_AXAF_phot_graze_angle,BPDType_double, 1)
      )
    tf_exit( ExitERR_misc, "error creating data packet fields: %s",
   	     bpipe_strerror( bpipe_errno) );
  
  /* if *.par surface number >0, ensure existence (must be created)
     of bpipe hdr field for surface number  and set srfno = 1
     (set surface # in bpipe hdr field to 1)  */
  
  if ( g->surf_num > 0  )
    { /* surf number > 0 */
      surface_num = 1;
      if ( bpipe_hdrf_add( bpipe,"srfno", BPDType_uint, NULL,
			   &surface_num, 1))
	tf_exit(ExitERR_misc, " bpipe_hdrf_add failure: %s",
		bpipe_strerror(bpipe_errno));
      ga_offset = g->surf_num > 1 ? 1 : 0;
    }    
  else
    {
      /* since surface no. is 0, delete g.a. and surface_normal fields from 
	 output: */
      delga = bpipe_dpktf_delete( bpipe, BP_AXAF_phot_graze_angle,
				  BPDSite_OUTPUT, bpo );
      switch ( delga )
	{
	case -1:
	  tf_exit( ExitERR_misc, " bpipe_dpktf_delete failure: %s",
		   bpipe_strerror(bpipe_errno));
	  break;
	case 1:  /* shouldn't get here */ 
	  tf_assert( ( delga == -1 ) || ( delga == 0 ) ) ;
	  break;
	default:  /* = 0, successful deletion */
	  break;
	}
      
      delnorm = bpipe_dpktf_delete( bpipe, BP_AXAF_surf_norm,
				    BPDSite_OUTPUT, bpo );
      switch ( delnorm )
	{
	case -1:
	  tf_exit( ExitERR_misc, " bpipe_dpktf_delete failure: %s",
		   bpipe_strerror(bpipe_errno));
	  break;
	case 1:   /* shouldn't get here */ 
	  tf_assert( ( delnorm == -1 ) || ( delnorm == 0 ) ) ;
	  break;
	default:   /* = 0, successful deletion */
	  break;
	}

      ga_offset = 0;
    }
  
  if ( 0 == (s_obj_out = bpipe_map(bpipe) ) )
    tf_exit( ExitERR_misc, "error mapping bpipe: %s",
	     bpipe_strerror( bpipe_errno) );
  
  if ( NULL == (obj_out = malloc(s_obj_out)) )
    tf_exit(ExitERR_alloc, "unable to allocate object(s)");
  
  position	= (DVector3 *) dpktf_data( bpipe, BP_AXAF_phot_pos, obj_out );
  direction	= (DVector3 *) dpktf_data( bpipe, BP_AXAF_phot_dir, obj_out );
  weight	= (double *)   dpktf_data( bpipe, BP_AXAF_phot_wt, obj_out );
  energy	= (double *)   dpktf_data( bpipe, BP_AXAF_phot_energy, 
					   obj_out );
  path_length	= (double *)   dpktf_data( bpipe, BP_AXAF_path_length,
					   obj_out);
  id		= (unsigned int *) dpktf_data( bpipe,BP_AXAF_phot_id, obj_out);
  handy		= (int *)	dpktf_data(bpipe, "handy", obj_out );
  pol_cos	= (DCVector2 *)	dpktf_data( bpipe, BP_AXAF_pol_cos, obj_out );
  pol_sin	= (DCVector2 *)	dpktf_data( bpipe, BP_AXAF_pol_sin, obj_out );
  surface_norm  = (DVector3 *) dpktf_data( bpipe, BP_AXAF_surf_norm, obj_out );
  graze_angle   = (double *) dpktf_data( bpipe, BP_AXAF_phot_graze_angle,
					 obj_out );
  
  bpipe_write_hdr(bpipe);
  
  for ( nphot = 1; fread( &fr, sizeof(fullray), 1, fin ) == 1 ; nphot++ )
    {
      position->x = fr.pos.r[0];
      position->y = fr.pos.r[1];
      position->z = fr.pos.r[2];
      
      direction->x = fr.pos.v[0];
      direction->y = fr.pos.v[1];
      direction->z = fr.pos.v[2];
      
      weight[0] = fr.wt;
      
      energy[0] = fr.energy;
      path_length[0] = fr.path;
      
      pol_cos->q1.r = fr.c2[0].r;
      pol_cos->q1.i = fr.c2[0].i;
      pol_cos->q2.r = fr.c2[1].r;
      pol_cos->q2.i = fr.c2[1].i;
      
      pol_sin->q1.r = fr.s2[0].r;
      pol_sin->q1.i = fr.s2[0].i;
      pol_sin->q2.r = fr.s2[1].r;
      pol_sin->q2.i = fr.s2[1].i;
      
      id[0] = fr.id;
      
      handy[0] = fr.handy[0];
      handy[1] = fr.handy[1];
      handy[2] = fr.handy[2];
      
      surface_norm->x = fr.norm[0];
      surface_norm->y = fr.norm[1];
      surface_norm->z = fr.norm[2];
      
      graze_angle[0] = fr.g_ang[ga_offset];
      /*    graze_angle[1] = fr.g_ang[0]; do not uncomment because the extent 
	    of the file g.a. field created is only 1 */
      
      if ( bpipe_write_dpkt( bpipe, obj_out, bpo ) )
	tf_exit( ExitERR_IOerr, "error writing photon %d to output channel",
		 nphot );
    }
  
  if ( ferror( fin ) )
    tf_exit( ExitERR_fread, "error reading photon %d from input stream",
	     nphot );
  
  fclose( fin );
  bpipe_delete( bpipe );
  
  free( obj_out );
  
  tf_leave( );
}

