/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of genphot
 *
 * genphot is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * genphot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

/* gparam.c - get parameters */

#include <string.h>

#include <stdlib.h>
#include <stdio.h>

#include <mathconst/mathconst.h>
#include <suplib/debug.h>
#include <suplib/str.h>
#include <tparam/tparam.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

static TokListToken
energy_dist[] =
{
  { "ave", EnDist_ave },
  { "max", EnDist_max },
  { "min", EnDist_min },
  { "picket", EnDist_picket },
  { "sample", EnDist_sample }
};
TokList energy_dists = GenTokList(energy_dist);

void
gparam( int argc, char *argv[])
{
  char dummy[ 1024 ];
  paramfile pf;
  size_t i;

  tf_enter( "gparam" );

  pf = tparamopen( NULL, argv, argc, "r" );

  /* get debug flags */
  tpgetstr( pf, "debug", dummy, NAME_LENGTH, 0 );
  debug_init( dummy );
  
  c.version = tpgetb( pf, "version" );

  if ( paccess( pf, "help" ) )
    c.help = tpgetb( pf, "help" );
  else
  {
    tf_message( "you must update your parameter file\n" );
    c.help = 0;
  }

  if ( !c.version && !c.help ) {

    if ( paccess( pf, "spectrum" ) ) 
      tpgetstr( pf, "spectrum", c.spectrum, NAME_LENGTH, 1 );
    else
    {
      tf_message( "you must update your parameter file\n" );
      strcpy( c.spectrum, "none" );
    }
    
    /* output file */
    tpgetstr( pf, "output", c.output, NAME_LENGTH, 1 );
    
    /* output geometry */
    tpgetstr( pf, "geo", dummy, NAME_LENGTH, 1 );
    for ( i = 0; i < nOutputGeometries &&
	    strcmp( OutputGeometries[i]->name, dummy );
	  i++ );
    
    if ( i == nOutputGeometries )
      tf_exit( ExitERR_illegalArg, "unknown output geometry `%s'",
	       dummy );
    c.geometry = OutputGeometries[i];
    c.geometry->gparam( pf );
    
    /* output photon density distribution */
    tpgetstr( pf, "dist", dummy, NAME_LENGTH, 1 );
    for ( i = 0; i < nOutputDistributions &&
	    strcmp( OutputDistributions[i]->name, dummy );
	  i++ );
    
    if ( i == nOutputDistributions )
      tf_exit( ExitERR_illegalArg, "unknown output distribution `%s'",
	       dummy );
    c.distribution = OutputDistributions[i];
    c.distribution->gparam( pf );
    
    if ( paccess( pf, "energy_dist" ) )
    {
      tpgetstr( pf, "energy_dist", dummy, NAME_LENGTH, 1 );
      c.en_dist = tokmatch( dummy, &energy_dists );
      if ( -1 == c.en_dist )
	tf_exit( ExitERR_illegalArg, "unknown energy distributin `%s'",
		 dummy );
    }
    else
    {
      tf_message( "you must update your parameter file\n" );
      c.en_dist = EnDist_sample;
    }

    /* random seed */
    c.seed1 = tpgeti( pf, "seed1" );
    c.seed2 = tpgeti( pf, "seed2" );
    c.block = tpgeti( pf, "block" );
    
    /* center of distribution */
    c.x = tpgetd( pf, "x" );
    c.y = tpgetd( pf, "y" );

    if ( paccess( pf, "z" ) )
      c.z = tpgetd( pf, "z" );
    else
    {
      tf_message( "you must update your parameter file; defaulting to z=0\n" );
      c.z = 0;
    }

    c.phi = tpgetd( pf, "phi" ) * M_PI / 180.0;
    
    /* photon density */
    c.phot_dens = tpgetd( pf, "photdens" );
    
    /* photon energy */
    c.energy = tpgetd( pf, "energy" );	
    
    /*
     * photon spherical coordinates angles: theta is the 'elevation' angle
     * from the Z+ axis phi is the 'azimuth' angle after the rotation with
     * the above phi and translation with the above x, y were already
     * performed
     */
    
    c.ray_theta = tpgetd( pf, "ray_theta" ) * M_PI / 180.0;
    c.ray_phi = tpgetd( pf, "ray_phi" ) * M_PI / 180.0;
    
  }
  paramclose( pf );
  tf_leave(  );
}

