/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of genphot
 *
 * genphot is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * genphot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>

#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include <bpipe/datatypes.h>

/****************************************************************************
 *
 * Source     : raywt.c
 * Author     : Terry Gaetz
 *
 * Description: evaluate ray weight based on OSAC-style ``cosine'' 
 *              and ``sine'' polarization amplitudes.
 *
 */


static double
eval_ray_weight( const DCVector2 *c2, const DCVector2 *s2,
                 int is_discrete_polarization )
{
  
  gsl_complex ctm0, ctmp;
  gsl_complex c2q1, c2q2, s2q1, s2q2;

  GSL_SET_COMPLEX( &c2q1, c2->q1.r, c2->q1.i );
  GSL_SET_COMPLEX( &c2q2, c2->q2.r, c2->q2.i );

  GSL_SET_COMPLEX( &s2q1, s2->q1.r, s2->q1.i );
  GSL_SET_COMPLEX( &s2q2, s2->q2.r, s2->q2.i );
  
  ctm0 = gsl_complex_conjugate( c2q1 );
  ctmp = gsl_complex_mul( ctm0, c2q1 );


  ctm0 = gsl_complex_conjugate( c2q2 );
  ctm0 = gsl_complex_mul( ctm0, c2q2 );
  ctmp = gsl_complex_add( ctmp, ctm0 );

  ctm0 = gsl_complex_conjugate( s2q1 );
  ctm0 = gsl_complex_mul( ctm0, s2q1 );
  ctmp = gsl_complex_add( ctmp, ctm0 );

  ctm0 = gsl_complex_conjugate( s2q2 );
  ctm0 = gsl_complex_mul( ctm0, s2q2 );
  ctmp = gsl_complex_add( ctmp, ctm0 );

  if ( is_discrete_polarization )
  {
    gsl_complex ctm1;

    ctm0 = gsl_complex_conjugate( s2q1 );
    ctm0 = gsl_complex_mul( ctm0, c2q1 );

    ctm1 = gsl_complex_conjugate( s2q2 );
    ctm1 = gsl_complex_mul( ctm1, c2q2 );

    ctm0 = gsl_complex_add( ctm0, ctm1 );
    ctm0 = gsl_complex_mul_real( ctm0, -2.0 );
    ctmp = gsl_complex_add( ctmp, ctm0 );
  }

  return GSL_REAL(ctmp);
}



void
setup_polarization( DCVector2 *pol_cos,  DCVector2 *pol_sin, double *weight)
{
  double ecc, cpsi, spsi, sqre, xtmp;
  gsl_complex ctmp0, ctmp1;
    
  /* hardwire unpolarized state */
  ecc  = 1.0;
  cpsi = 1.0;
  spsi = 0.0;
  sqre = sqrt(2.0);

  pol_cos->q1.r =  cpsi / sqre;
  pol_cos->q1.i =   0.0;

  pol_cos->q2.r =  spsi / sqre;
  pol_cos->q2.i =   0.0;

  xtmp        = -spsi / sqre;
  GSL_SET_COMPLEX( &ctmp0, xtmp, 0.0 );
  ctmp1 = gsl_complex_mul_real( ctmp0, ecc );
  pol_sin->q1.r = GSL_REAL(ctmp1);
  pol_sin->q1.i = GSL_IMAG(ctmp1);
  
  xtmp        =  cpsi / sqre;
  GSL_SET_COMPLEX( &ctmp0, xtmp, 0.0 );
  ctmp1 = gsl_complex_mul_real( ctmp0, ecc );
  pol_sin->q2.r = GSL_REAL(ctmp1);
  pol_sin->q2.i = GSL_IMAG(ctmp1);

  *weight = eval_ray_weight( pol_cos, pol_sin, 0 );
}
