/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of genphot
 *
 * genphot is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * genphot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

/* process.c */

#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <mathconst/mathconst.h>

#include <fcntl.h>    
#include <sys/types.h> 

#include <suplib/debug.h>   
#include <randlib_c/randlib.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <bpipe/bpipe_axaf.h>
#include <bpipe/bpipe.h>
#include <bpipe/datatypes.h>

#include "main.h"
#include "common.h"
#include "flux.h"

BPipe *
setup_bpipe( size_t* nphot )
{
  BPipe *bpipe;

  if ( NULL == ( bpipe = bpipe_new(  ) ) )
    tf_exit( ExitERR_misc, "couldn't open bpipe: %s",
	    bpipe_strerror( bpipe_errno ) );

  if ( bpipe_dpktf_add(  bpipe, BP_AXAF_phot_wt, BPDType_double, NULL ) )
    tf_exit( ExitERR_misc, "couldn't add %s: %s", BP_AXAF_phot_wt,
	    bpipe_strerror( bpipe_errno ) );

  if ( bpipe_dpktf_add( bpipe, BP_AXAF_phot_pos, BPDType_DVector3, NULL ) )
    tf_exit( ExitERR_misc, "couldn't add %s: %s", BP_AXAF_phot_pos,
	    bpipe_strerror( bpipe_errno ) );

  if ( bpipe_dpktf_add( bpipe, BP_AXAF_phot_dir, BPDType_DVector3, NULL ) )
    tf_exit( ExitERR_misc, "couldn't add %s: %s", BP_AXAF_phot_dir,
	    bpipe_strerror( bpipe_errno ) );

  if ( bpipe_dpktf_add( bpipe, BP_AXAF_phot_energy, BPDType_double, NULL ) )
    tf_exit( ExitERR_misc, "couldn't add %s: %s", BP_AXAF_phot_energy,
	    bpipe_strerror( bpipe_errno ) );

  if ( bpipe_dpktf_add( bpipe, BP_AXAF_pol_sin,  BPDType_DCVector2, NULL )  )
    tf_exit( ExitERR_misc, "couldn't add %s: %s", BP_AXAF_pol_sin,
	    bpipe_strerror( bpipe_errno )  );

  if ( bpipe_dpktf_add( bpipe, BP_AXAF_pol_cos,  BPDType_DCVector2, NULL )  )
    tf_exit( ExitERR_misc, "couldn't add %s: %s", BP_AXAF_pol_cos,
	    bpipe_strerror( bpipe_errno )  );

  if ( bpipe_dpktf_add( bpipe, BP_AXAF_phot_id,  BPDType_uint, NULL )  )
    tf_exit( ExitERR_misc, "couldn't add %s: %s", BP_AXAF_phot_id,
	    bpipe_strerror(bpipe_errno )  );

  return bpipe;
}

void
process( void )
{
  size_t n;
  size_t nphot;
  int specflag;        /*  = 1 if requested input spectrum is found */
  unsigned int *rayid;
  long nlines;
  double cos_phi, sin_phi;
  double *weight, *energy;

  DVector3 *position, *direction;
  DCVector2 *     pol_sin;
  DCVector2 *     pol_cos;
  
  BPipe *bpipe;
  BPipeOutput *bpo;
  void *obj_out;

  Flux *data = NULL;

  tf_enter( "process" );

  nphot = c.distribution->init( c.geometry->init() );

  /* get spectral data from rdb table */
  if ( strcmp( "none", c.spectrum  )  ) {
    specflag = 1;         /* input spectrum is NOT 'none' */
    data = spectrum( c.spectrum, &nlines, c.en_dist );

    /* fudge number of photons if one per spectrum bin was requested */
    if ( EnDist_sample != c.en_dist && EnDist_picket != c.en_dist )
      nphot = nlines;
        
  } else {
    specflag = 0;         /* input spectrum is 'none' */
  }

  /********************************************************************/
  /* set up bpipe */
  bpipe = setup_bpipe( &nphot );

  if ( NULL == ( bpo = bpipe_output( bpipe, c.output ) ) )
    tf_exit( ExitERR_misc, "couldn't open output channel: %s",
	    bpipe_strerror( bpipe_errno ) );

  if ( NULL == ( obj_out = bpipe_map_alloc( bpipe, 1, NULL ) ) )
    tf_exit(  ExitERR_alloc, "unable to allocate memory for data packet: %s",
	     bpipe_strerror( bpipe_errno ) );

  /********************************************************************/
  /* retrieve data pointers in bpipe data packet */
  weight = ( double * )
    bpipe_dpktf_data( bpipe_dpktf( bpipe, BP_AXAF_phot_wt ), obj_out );

  energy = ( double * )
    bpipe_dpktf_data( bpipe_dpktf( bpipe, BP_AXAF_phot_energy ), obj_out );

  position = ( DVector3 * )
    bpipe_dpktf_data( bpipe_dpktf( bpipe, BP_AXAF_phot_pos ), obj_out );

  direction = ( DVector3 * )
    bpipe_dpktf_data( bpipe_dpktf( bpipe, BP_AXAF_phot_dir ), obj_out );

  pol_cos = ( DCVector2 * )
    bpipe_dpktf_data( bpipe_dpktf( bpipe, BP_AXAF_pol_cos  ), obj_out  );
  
  pol_sin = ( DCVector2 * )
    bpipe_dpktf_data( bpipe_dpktf( bpipe, BP_AXAF_pol_sin  ), obj_out  );

  rayid = ( unsigned int* )
    bpipe_dpktf_data( bpipe_dpktf( bpipe, BP_AXAF_phot_id  ), obj_out  );

  if ( bpipe_write_hdr( bpipe )  )
    tf_exit( ExitERR_misc,"bpipe_write_hdr: %s",bpipe_strerror( bpipe_errno));

  /* initialize random number generator and get to correct block */
  setall( c.seed1, c.seed2 );
  for ( n = 0; n < c.block; n++ )
    initgn( 1 );

  /********************************************************************/
  /* these quantities are the same for each ray, so do them outside
     of the loop */

  /* default ray weight and polarization */
  setup_polarization( pol_cos, pol_sin, weight  );

  /* optical axis position */
  position->z = c.z;

  /* direction cosines. rays are parallel. */
  direction->x = sin( c.ray_theta ) * cos( c.ray_phi );
  direction->y = sin( c.ray_theta ) * sin( c.ray_phi );
  direction->z = cos( c.ray_theta );

  /********************************************************************/
  /* do this once, for speed */
  cos_phi = cos( c.phi );
  sin_phi = sin( c.phi );

  /* starting ray id */
  *rayid = 0;

  /* if a spectrum is used, this is overriden in the loop below */
  *energy = c.energy;
  
  /* loop over all photons, generating position and energy */
  for ( n = 0 ; n < nphot ; n++ )
  {
    double x0, y0;
    
    ++( *rayid );   /* increment ray id */

    /*****************************************/
    /* ray position */
    if ( c.geometry->genphot( &x0, &y0 ) )
      break;

    /* rotate and shift them */
    /* z is set above */
    position->x = x0 * cos_phi - y0 * sin_phi + c.x;
    position->y = y0 * cos_phi + x0 * sin_phi + c.y;

    /*****************************************/
    /* ray energy preset to constant above, or get random energy from
       rdb table input */
    if ( specflag )
      *energy = g_energy( data, nlines, n, c.en_dist);


    /*****************************************/
    if ( bpipe_write_dpkt( bpipe, obj_out, bpo ) )
      tf_exit( ExitERR_misc, "error writing photon %d: %s", nphot,
	      bpipe_strerror( bpipe_errno ) );
  }

  bpipe_delete( bpipe );

  free( data ) ;
  tf_leave(  );
}
