/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of genphot
 *
 * genphot is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * genphot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

/* rdbread.c; read rdb table */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <mst_rdb/mst_rdb.h>

#include "flux.h"

int 
qsort_cmp( const void *vp, const void *vq )
{
  double p = ( ( Flux * ) vp)->emin;
  double q = ( ( Flux * ) vq)->emin;
  double diff = p - q;

  if ( diff < 0.0 )
    return -1;
  else if ( diff > 0.0 )
    return 1;
  else
    return 0;
}


/* number of rows to allocate if size of rdb file is uknown.
   also used as an increment when extending the array */

#define BUF_NROW	1000

#define SETFIELD(var, vartype)				\
do{							\
    fields[nfields].name = #var;			\
    fields[nfields].type = vartype;			\
    fields[nfields].offset = offsetof( Flux, var );	\
    nfields++;						\
}while(0)

Flux *
readrdb(
  const char *spectrum_file,
  long *nlines,
  EnergyDist energy_dist )
{

  size_t index;
  size_t nrow;

  FILE *infile = NULL;		/* input rdb table */
  Flux *data = NULL;		/* data structure array */

  rdbHeader *head;
  DataColumnMap_st *map;

  RDBFieldStInfo fields[3];
  int nfields = 0;

  tf_enter( "rdbread" );

  switch( energy_dist )
  {
    /* need emin and emax, but not counts */
  case EnDist_ave:
    SETFIELD( emin, RDB_Num );
    SETFIELD( emax, RDB_Num );
    break;
    
    /* just need emax */
  case EnDist_max:
    SETFIELD( emax, RDB_Num );
    break;

    /* just need emin */
  case EnDist_min:
    SETFIELD( emin, RDB_Num );
    break;
    
  case EnDist_sample:
    SETFIELD( emin, RDB_Num );
    SETFIELD( emax, RDB_Num );
    SETFIELD( counts, RDB_Num );
    break;

    /* don't need emax */
  case EnDist_picket:
    SETFIELD( emin, RDB_Num );
    SETFIELD( counts, RDB_Num );
    break;

  default:
    tf_assert(0);
  }


  if ( 0 == strcmp( spectrum_file, "stdin" ) )
  {
    infile = stdin;
  }
  else
  {
    infile = fopen( spectrum_file, "r" );
    if ( NULL == infile )
      tf_exit( ExitERR_fopen, "Unable to open rdb table `%s'",
	       spectrum_file );
  }

  head = rdb_rd_hdr( infile );

  map = rdb_map_cols_stst( head, nfields, fields );

  nrow = rdb_count( infile, head );

  if ( -1 == nrow )
    nrow = BUF_NROW;

  data = (Flux *) malloc( nrow * sizeof( Flux ) );
  if ( NULL == data)
    tf_exit( ExitERR_alloc, "unable to allocate space for spectrum data");

  /* read data from rdb table */
  index = 0;
  while ( rdb_col_read_st( infile, head, map, &data[ index ] ) )
  {
    index++;
    if ( index == nrow && ! feof(infile) )
    {
      Flux *tdata = realloc( data, (nrow + BUF_NROW) * sizeof(Flux));
      if ( NULL == tdata )
      {
	free( data );
	tf_exit( ExitERR_alloc,
		 "unable to reallocate space for spectrum data" );
      }
      nrow += BUF_NROW;
      data = tdata;
    }
  }

  *nlines = index;

  /* sort data */
  qsort( ( void * ) data, *nlines, sizeof( Flux ), qsort_cmp );

  fclose( infile );
  
  rdb_free_hdr( head );
  rdb_free_map( map );
    
  tf_leave();

  return data;

}
