/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of genphot
 *
 * genphot is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * genphot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mst_rdb/mst_rdb.h>
#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>
#include <suplib/debug.h>

typedef struct
{
  double emin, emax, counts;
} Bin;

RDBFieldStInfo fields[] = {
  RDBentry( emin,   RDB_Num, Bin ),
  RDBentry( emax,   RDB_Num, Bin ),
  RDBentry( counts, RDB_Num, Bin )
};
#define NFIELDS ( sizeof(fields) / sizeof(RDBFieldStInfo ) )

/* qsort bin comparison routine */
int qcmp( const void *vb1, const void *vb2 )
{
  Bin* b1 = (Bin*) vb1;
  Bin* b2 = (Bin*) vb2;

  return
      b1->emin < b2->emin ? -1
    : b1->emin > b2->emin ?  1
    :                        0;
}

/* bsearch bin comparison routine */
int bincmp( const void *key, const void *vbin )
{
  double energy = *( (double*)key );
  Bin *bin = (Bin*) vbin;
  return
      energy <  bin->emin ? -1
    : energy >= bin->emax ?  1
    :                        0;
}

/* read in the spectrum */
int read_spectrum( const char* file, Bin**bin , size_t *nbins )
{
  rdbHeader* hdr;
  DataColumnMap_st *map;
  Bin* bptr;
  
  FILE *fin = fopen( file, "r" );
  

  if ( NULL == fin )
    die( "unable to open %s\n", file );

  hdr = rdb_rd_hdr( fin );
  map = rdb_map_cols_stst( hdr, NFIELDS, fields );
  *nbins  = rdb_count( fin, hdr );

  bptr = *bin = (Bin*) malloc( *nbins * sizeof(Bin) );
  if ( NULL == bptr )
    die( "unable to allocate %ul bins\n", *nbins );

  while( rdb_col_read_st( fin, hdr, map, bptr++ ) );

  rdb_free_hdr( hdr );
  rdb_free_map( map );
  /* sort bins  */
  qsort( *bin, *nbins, sizeof(Bin), qcmp );

  return EXIT_SUCCESS;
}


static char *
dpktf_data( BPipe *bpipe, char *name, void *data )
{
  DpktField *field = bpipe_dpktf( bpipe, name );

  if ( field == NULL )
    die( "%s isn't in input data stream\n", name);

  return bpipe_dpktf_data( field, data );

}

int main( int argc, char *argv[] )
{
  void *data;
  double *energy;
  double *weight;

  Bin* spectrum_in;
  Bin* spectrum_out;
  size_t nbins;
  BPipe* bpipe;
  
  const char* spec_file;
  const char* bp_file;

  if ( argc != 3 )
    die( "usage: %s <spectrum rdb file> <bpipe file>\n" );
  
  spec_file = argv[1];
  bp_file   = argv[2];


  /* ------------------------------------------------------------ */
  /* read in spectrum */
  if ( read_spectrum( spec_file, &spectrum_in, &nbins ) )
    die( "error reading spectrum: %s", spec_file);

#if 0
  {
    size_t ibin;
    for ( ibin = 0 ; ibin < nbins ; ibin++ )
      printf( "emin: %f  emax: %f counts: %f\n",
	      spectrum_in[ibin].emin,
	      spectrum_in[ibin].emax,
	      spectrum_in[ibin].counts
	      );
  }
#endif

  /* duplicate bins */
  spectrum_out = (Bin*) malloc( nbins * sizeof(Bin) );
  if ( NULL == spectrum_out )
    die( "unable to allocate %ul bins\n", nbins );

  memcpy( spectrum_out, spectrum_in, nbins * sizeof(Bin) );
  {
    size_t ibin;
    for ( ibin = 0 ; ibin < nbins ; ibin++ )
      spectrum_out[ibin].counts = 0;
  }

  /* ------------------------------------------------------------ */


  if ( NULL == ( bpipe = bpipe_new() ))
    die( "couldn't create bpipe: %s", bpipe_strerror(bpipe_errno));

  if ( bpipe_input(bpipe, bp_file) )
    die( "couldn't open input file `%s': %s", bp_file,
	 bpipe_strerror(bpipe_errno));

  if (NULL == (data = bpipe_map_alloc(bpipe,1, NULL)))
    die( "unable to allocate memory for data packet: %s",
	 bpipe_strerror(bpipe_errno));

  energy = (double*) dpktf_data( bpipe, BP_AXAF_phot_energy, data );
  weight = (double*) dpktf_data( bpipe, BP_AXAF_phot_wt,     data );

  while( bpipe_read_dpkts( bpipe, data, 1 ) )
  {
    Bin *bin = bsearch( energy, spectrum_out, nbins, sizeof(Bin), bincmp );

    if ( bin )
      bin->counts += *weight;

    else
      fprintf( stderr, "outside of bins: %f\n", *energy );
  }

  {
    size_t ibin;
    double chisqr = 0;
    double max_in  = 0;
    double max_out = 0;
    
    for ( ibin = 0 ; ibin < nbins ; ibin++ )
    {
      if ( max_in < spectrum_in[ibin].counts )
	max_in = spectrum_in[ibin].counts;
      
      if ( max_out < spectrum_out[ibin].counts )
	max_out = spectrum_out[ibin].counts;
    }

    for ( ibin = 0 ; ibin < nbins ; ibin++ )
    {
      double dcounts;
      if ( 0 == spectrum_out[ibin].counts )
	continue;
      
      spectrum_out[ibin].counts /= max_out;
      spectrum_in[ibin].counts  /= max_in;

      dcounts = spectrum_out[ibin].counts - spectrum_in[ibin].counts;
      chisqr += dcounts * dcounts / spectrum_out[ibin].counts;
    }

    printf( "%lu %f %f\n", nbins, chisqr, chisqr / (nbins - 2) );
  }

  return EXIT_SUCCESS;
}

