/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of linklist
 *
 * linklist is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * linklist is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>

#include "linklist.h"

/* macros to easily access node and list members */
#define NEXT(node)	((node)->next)
#define PREV(node)	((node)->prev)
#define DATA(node)	((node)->data)
#define COUNT(ll)	((ll)->count)
#define NIL(ll)		(&(ll)->sentinel)
#define FIRST(ll)	((ll)->sentinel.next)
#define LAST(ll)	((ll)->sentinel.prev)
#define CMP(ll)		((ll)->comp)
#define EMPTY(ll)	(COUNT(ll) == 0)


#define LINKLIST	DLinkList
#define NODE		DLLNode

typedef struct Node
{
  struct Node *next;
  struct Node *prev;
  void *data;
} Node;

typedef struct
{
  Node sentinel;
  size_t count;
  int (*comp) (const void *d1, const void *d2);
} LL;

static Node *Search(LL *ll, void *data,
		     int (*comp) (const void *, const void *));
static void UnLink(LL *ll, Node *node);
static void deleteNode(Node *node);
static Node *newNode(void *data);
static void Insert(LL *ll, Node *node);

/*

  PROCEDURE
  dll_new - Create a new doubly linked list.

  DESCRIPTION
  This routine creates a new list structure.  If the list is to have
  some intrinsic order, a function defining that order should be
  passed.  The comparison routine is called with two node data
  pointers as the arguments.  It must return @samp{-1}, @samp{0}, or
  @samp{1} if, respectively, the first argument is less than, equal
  to, or greater than the second.

  RETURNS
  It returns a pointer to a new linked list, or @code{NULL} if it can't
  allocate it.

*/
LINKLIST
dll_new(
  int (*cmp) (const void *, const void *)
  	/* a comparison function for use in searches of the
	   linked list, may be @code{DLL_NULL_CMP} for unordered lists */
)
{
  LL *ll = (LL *) malloc(sizeof(LL));

  if (ll)
  {
    FIRST(ll) = NIL(ll);
    LAST(ll) = NIL(ll);
    COUNT(ll) = 0;
    CMP(ll) = cmp;
  }

  return (LINKLIST) ll;
}

/*

  PROCEDURE
  dll_delete - Delete a doubly list.

  DESCRIPTION
  @code{dll_delete} traverses a linked list, calling a user-supplied function
  at each node in the list, then deleting the node.  Finally it
  deletes the list header.

*/
void
dll_delete(
  LINKLIST ull,			/* the linked list to be deleted */
  void (*ufree) (void *)	/* a function called at each
				   link to delete user data.
				   it may be @code{DLL_NULL_DELETE} */
)
{
  LL *ll = (LL *) ull;

  if (ll)
  {
    Node *next = FIRST(ll);

    while (next != NIL(ll))
    {
      Node *node = next;

      next = NEXT(node);

      if (ufree)
	ufree(DATA(node));
      free(node);
    }
    free(ll);
  }
}


/*

  PROCEDURE
  dll_udelete - Delete a doubly list.

  DESCRIPTION
  @code{dll_udelete} traverses a linked list, calling a user-supplied function
  at each node in the list, then deleting the node.  Finally it
  deletes the list header.  This routine differs from
  @code{dll_delete} in that it can pass along a pointer provided by
  the calling routine to the action routine, allowing arbitrary data
  to be available to the action routine.

*/
void
dll_udelete(
  LINKLIST ull,			/* the linked list to be deleted */
  void (*ufree) (void *, void *),	/* a function called at each
				   link to delete user data.
				   it may be @code{DLL_NULL_DELETE} */

  void *udata			/* a pointer to data to be
				   passed to the action routine */
)
{
  LL *ll = (LL *) ull;

  if (ll)
  {
    Node *next = FIRST(ll);

    while (next != NIL(ll))
    {
      Node *node = next;

      next = NEXT(node);

      if ( ufree )
	ufree( DATA(node), udata );
      free( node );
    }
    free(ll);
  }
}


/*

  PROCEDURE
  dll_destroy - Destroy a node in a doubly linked list.

  DESCRIPTION
  @code{dll_destroy} searches a linked list for a node which compares
  equivalently to the passed data.  It removes the node from the list
  and deallocates the memory associated with it.  The user must
  destroy the data referenced by the node's data pointer.

  It uses the passed comparison routine, if available.  If not, it
  uses that with which the list was initialized.  If @emph{that} is
  not available, a segmentation violation is unavoidable.  In the
  former case the passed data need not have the same form as the data
  stored in the node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  RETURNS
  It returns the node's data pointer if the node was found, @code{NULL}
  otherwise.

*/
void *
dll_destroy(
  LINKLIST ull, 	/* the list to search */
  void *data,   	/* the data to search for */
  int (*cmp) (const void *, const void *)
	      /* The address of a comparison function.
		 Set to @code{DLL_NULL_CMP} to use the list's initial
		 comparison function. */
)
{
  LL *ll = (LL *) ull;
  Node *node = Search(ll, data, cmp);

  data = NULL;
  if (node)
  {
    data = DATA(node);
    UnLink(ll, node);
    free(node);
  }

  return data;
}

/*

  PROCEDURE
  dll_destroy_head - Remove the head node from a doubly linked list
  and destroy it.

  DESCRIPTION
  This routine removes the head node from the specified list and
  deallocates the memory associated with it.  The user must destroy
  the data referenced by the node's data pointer.

  RETURNS
  It returns the node's data pointer.  If the list is empty, it
  returns @code{NULL}.

*/
void *
dll_destroy_head(
  LINKLIST ull			/* the list from which to remove the node */
)
{
  LL *ll = (LL *) ull;
  Node *node = FIRST(ll);
  void *data = NULL;

  if (node != NIL(ll))
  {
    data = DATA(node);
    UnLink(ll, node);
    deleteNode(node);
  }

  return data;
}

/*

  PROCEDURE
  dll_destroy_tail - Remove the tail node from a doubly linked list
  and destroy it.

  DESCRIPTION
  This routine removes the tail node from the specified list and
  deallocates the memory associated with it.  The user must destroy
  the data referenced by the node's data pointer.

  RETURNS
  It returns the node's data pointer.  If the list is empty, it
  returns @code{NULL}.

*/
void *
dll_destroy_tail(
  LINKLIST ull			/* the list from which to remove the node */
)
{
  LL *ll = (LL *) ull;
  Node *node = LAST(ll);
  void *data = NULL;

  if (node != NIL(ll))
  {
    data = DATA(node);
    UnLink(ll, node);
    deleteNode(node);
  }

  return data;
}

/*

  PROCEDURE
  dll_count - Determine the number of nodes in a doubly linked list.

  DESCRIPTION
  This function returns the number of nodes in a list.  It is an
  inexpensive routine to call.

  RETURNS
  It returns the number of nodes in the list, @samp{0} if the passed
  pointer is @code{NULL}.

*/
size_t
dll_count(
  LINKLIST ull		/* the list to count */
)
{
  return ull ? COUNT(((LL *) ull)) : 0;
}

/*

  PROCEDURE
  dll_destroy_dnode - Deallocate the memory associated with a detached node.

  DESCRIPTION
  This routine deallocates the memory associated with a detached node.
  The user must have already destroyed the data associated with the
  node.  The node handle must have been returned by
  @code{dll_detach_node}.

*/
void
dll_destroy_dnode(
  NODE unode	        /* the node to destroy */
)
{
  deleteNode((Node *) unode);
}

/*

  PROCEDURE
  dll_destroy_node - Remove a node from a doubly linked list and destroy it.

  DESCRIPTION
  This routine removes the passed node from the specified list and
  deallocates the memory associated with it.  The user must destroy
  the data referenced by the node's data pointer.

  RETURNS
  It returns the node's data pointer if the node is not @code{NULL},
  @code{NULL} otherwise.

*/
void *
dll_destroy_node(
  LINKLIST ull, 	/* the list from which to remove the node */
  NODE unode	        /* the node to remove */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;
  void *data = NULL;

  if (node)
  {
    data = DATA(node);
    UnLink(ll, node);
    deleteNode(node);
  }

  return data;
}

/*

  PROCEDURE
  dll_detach_node - Remove a node from a doubly linked list.

  DESCRIPTION
  This routine removes a node from a list without destroying it.  The
  user is responsible for its subsequent care.

  RETURNS
  It returns a handle for the detached node.

*/
void
dll_detach_node(
  LINKLIST ull,	        /* the list from which to detach the node */
  NODE unode	        /* the node to detach */
)
{
  if (unode)
    UnLink((LL *) ull, (Node *) unode);
}

/*

  PROCEDURE
  dll_insert - Create and insert a node into a doubly linked list.

  DESCRIPTION
  @code{dll_insert} creates a node, stores the passed data pointer in
  it, and inserts the node in the list in the collating order
  determined by the comparison function with which the list was
  initialized by @code{dll_new}.

  RETURNS
  It returns zero if the insert was successful, non-zero if it was unable
  to create the new node.

*/
int
dll_insert(
  LINKLIST ull,	        /* a handle to the list into which to insert
                           the node */
  void *data		/* a pointer to the new node's data */
)
{
  LL *ll = (LL *) ull;
  Node *node = newNode(data);

  if (!node)
    return 1;

  Insert(ll, node);
  return 0;
}

/*

  PROCEDURE
  dll_insert_node - Insert a detached node into a list.

  DESCRIPTION
  This routine inserts a detached node into a list in the collating
  sequence determined by the comparison function with which the list
  was initialized by @code{dll_new}.  It assumes that the list's
  insert/delete comparison function can be applied to the data in the
  passed node.  The node handle must have been obtained from
  @code{dll_detach_node}.

  RETURNS
  It returns non-zero if the node is @code{NULL}, zero otherwise.

*/
int
dll_insert_node(
  LINKLIST ull, 	/* a handle to the list into which to insert
                           the node */
  NODE unode	        /* the node to insert */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  if (!node)
    return 1;

  Insert(ll, node);
  return 0;
}

/*

  PROCEDURE
  dll_insert_head - Create and insert a node at the head of a doubly
  linked list.

  DESCRIPTION
  This routine creates a node, inserts the passed data pointer in it,
  and attaches the node to the head of the list.

  RETURNS
  It returns zero if the insert was successful, non-zero if it was unable
  to create the new node.

*/
int
dll_insert_head(
  LINKLIST ull, 	/* a handle to the list into which to insert
                           the node */
  void *data	        /* a pointer to the new node's data */
)
{
  Node *node = newNode(data);

  return node ? dll_insert_head_node(ull, (NODE) node) : 1;
}

/*

  PROCEDURE
  dll_insert_head_node - Insert a detached node at the head of a
  doubly linked list.

  DESCRIPTION
  This routine inserts a node at the head of a list.  The node handle
  must have been obtained from @code{dll_detach_node}.

  RETURNS
  It returns non-zero if the node is @code{NULL}, zero otherwise.

*/
int
dll_insert_head_node(
  LINKLIST ull, 	/* a handle to the list into which to insert
                           the node */
  NODE unode	        /* the node to insert */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  if (!node)
    return 1;

  NEXT(node) = FIRST(ll);
  PREV(node) = NIL(ll);
  FIRST(ll) = PREV(NEXT(node)) = node;
  COUNT(ll)++;

  return 0;
}

/*

  PROCEDURE
  dll_insert_tail - Create and insert a node at the tail of a doubly
  linked list.

  DESCRIPTION
  This routine creates a node, inserts the passed data pointer in it,
  and attaches the node to the tail of the list.

  RETURNS
  It returns zero if the insert was successful, non-zero if it was unable
  to create the new node.

*/
int
dll_insert_tail(
  LINKLIST ull,	        /* a handle to the list into which to insert
                           the node */
  void *data    	/* a pointer to the new node's data */
)
{
  Node *node = newNode(data);

  return node ? dll_insert_tail_node(ull, (NODE) node) : 1;
}

/*

  PROCEDURE
  dll_insert_tail_node - Insert a detached node at the tail of a  list.

  DESCRIPTION
  This routine inserts a node at the tail of a list.  The node handle
  must have been obtained from @code{dll_detach_node}.

  RETURNS
  It returns non-zero if the node is @code{NULL}, zero otherwise.

*/
int
dll_insert_tail_node(
  LINKLIST ull, 	/* a handle to the list into which to insert
                           the node */
  NODE unode    	/* the node to insert */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  if (!node)
    return 1;

  NEXT(node) = NIL(ll);
  PREV(node) = LAST(ll);
  LAST(ll) = NEXT(PREV(node)) = node;
  COUNT(ll)++;

  return 0;
}

/*

  PROCEDURE
  dll_head - Get the data of the head node of a list.

  RETURNS
  It returns the data of the head node of a list, @code{NULL} if the list
  is empty.

*/
void *
dll_head(
  LINKLIST ull		/* the list in question */
)
{
  LL *ll = (LL *) ull;

  return EMPTY(ll) ? NULL : DATA(FIRST(ll));
}

/*

  PROCEDURE
  dll_head_node - Get a node handle to the head node of a list.

  RETURNS
  It returns the node handle of the head node of a list, @code{NULL} if
  the list is empty.

 */
NODE
dll_head_node(
  LINKLIST ull	        /* the list in question */
)
{
  LL *ll = (LL *) ull;

  return EMPTY(ll) ? NULL : FIRST(ll);
}

/*
  PROCEDURE
  dll_detach_head_node - Detach the head node of a list.

  DESCRIPTION
  This routine detaches the head node of a list from the list and returns
  a handle to the detached node.

  RETURNS
  It returns a handle to the node upon success, @code{NULL} on failure.

*/
NODE
dll_detach_head_node(
  LINKLIST ull		/* the list in question */
)
{
  LL *ll = (LL *) ull;


  if ( EMPTY(ll) )
    return NULL;
  else
  {
    Node *node = FIRST(ll);
    UnLink( ll, node );
    return node;
  }
}

/*

  PROCEDURE
  dll_tail - Get the data of the tail node of a list.

  RETURNS
  It returns the data of the tail node of a list, @code{NULL} if the list
  is empty.

*/
void *
dll_tail(
  LINKLIST ull		/* the list in question */
)
{
  LL *ll = (LL *) ull;

  return EMPTY(ll) ? NULL : DATA(LAST(ll));
}

/*

  PROCEDURE
  dll_tail_node - Get a node handle to the tail node of a list.

  RETURNS
  It returns the node handle of the tail node of a list, @code{NULL} if
  the list is empty.

*/
NODE
dll_tail_node(
  LINKLIST ull  	/* the list in question */
)
{
  LL *ll = (LL *) ull;

  return EMPTY(ll) ? NULL : LAST(ll);
}


/*

  PROCEDURE
  dll_sizeof_node - Get the size of a node.

  DESCRIPTION
  This routine returns the size of the internal structure used for
  each node.  This does not include user supplied data.  Please note
  that this is a function call, not a macro!.

*/
size_t
dll_sizeof_node(void)
{
  return sizeof(Node);
}


/*

  PROCEDURE
  dll_next - Retrieve the data in the node following a given node in a list.

  RETURNS
  It returns @code{NULL} if there are no more nodes, else a pointer to
  the data.

*/
void *
dll_next(
  LINKLIST ull,		/* the list which contains the node */
  NODE unode		/* the preceding node */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  return (NEXT(node) == NIL(ll)) ? NULL : DATA(NEXT(node));
}

/*

  PROCEDURE
  dll_next_node - Get a handle to the node following a given node in a list.

  RETURNS
  It returns @code{NULL} if there are no more nodes, else a handle to the
  node.

*/
NODE
dll_next_node(
  LINKLIST ull, 	/* the list which contains the node */
  NODE unode	        /* the preceding node */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  return (NEXT(node) == NIL(ll)) ? NULL : NEXT(node);
}


/*

  PROCEDURE
  dll_prev - Retrieve the data in the node preceding a given node in a list.

  RETURNS
  It returns @code{NULL} if there are no more nodes, else a pointer to
  the data.

*/
void *
dll_prev(
  LINKLIST ull,		/* the list which contains the node */
  NODE unode		/* the following node */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  return (PREV(node) == NIL(ll)) ? NULL : DATA(PREV(node));
}

/*

  PROCEDURE
  dll_prev_node - Get a handle to the node preceding a given node in a list.

  RETURNS
  It returns @code{NULL} if there are no more nodes, else a handle to the node.

*/
NODE
dll_prev_node(
  LINKLIST ull, 	/* the list which contains the node */
  NODE unode	        /* the following node */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  return (PREV(node) == NIL(ll)) ? NULL : PREV(node);
}


/*

  PROCEDURE
  dll_node_get_data - Retrieve the data from a specified linked list node.

  RETURNS
  It returns the node's data pointer.

*/
void *
dll_node_get_data(
  NODE unode	        /* the node in question */
)
{
  return unode ? DATA((Node *) unode) : NULL;
}

/*

  PROCEDURE
  dll_node_put_data - Replace the data pointer in a given node with another.

  DESCRIPTION
  This function allows the calling routine to change the data that a
  node points to.  No check is made to insure that any existing
  ordering of the data is maintained.

*/
void
dll_node_put_data(
  NODE unode,	        /* the node in question */
  void *data	        /* the new data */
)
{
  if ( unode )
    DATA((Node *) unode) = data;
}

/*

  PROCEDURE
  dll_search - Search a linked list for a node with equivalent data.

  DESCRIPTION
  This routine searchs the list for the node for which the passed data
  compares equivalently with the node's data.  It uses the passed
  comparison routine, if available.  If not, it uses that with which
  the list was initialized.  If @emph{that} is not available, a
  segmentation violation is unavoidable.  In the former case the
  passed data need not have the same form as the data stored in the
  node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  RETURNS
  It returns the node's data pointer if the node was found, @code{NULL}
  otherwise.

*/
void *
dll_search(
  LINKLIST ull, 	/* the list to search */
  void *data,		/* the data to search for */
  int (*cmp) (const void *, const void *)
	      /* The address of a comparison function.
		 Set to @code{DLL_NULL_CMP} to use the list's initial
		 comparison function. */
)
{
  LL *ll = (LL *) ull;
  Node *node = Search(ll, data, cmp);

  return node ? DATA(node) : NULL;
}

/*

  PROCEDURE
  dll_search_node - Search a linked list for a node with equivalent data.

  DESCRIPTION
  This searchs the list for the node for which the passed data
  compares equivalently with the node's data.  It uses the passed
  comparison routine, if available.  If not, it uses that with which
  the list was initialized.  If @emph{that} is not available, a
  segmentation violation is unavoidable.  In the former case the
  passed data need nothave the same form as the data stored in the
  node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  RETURNS
  It returns a handle to the matching mode, or @code{NULL} if not found.

*/
NODE
dll_search_node(
  LINKLIST ull, 	/* the list to search */
  void *data,	        /* the data to search for */
  int (*cmp) (const void *, const void *)
	      /* The address of a comparison function.
		 Set to @code{DLL_NULL_CMP} to use the list's initial
		 comparison function. */
)
{
  LL *ll = (LL *) ull;

  return (NODE) Search(ll, data, cmp);
}

/*
  PROCEDURE
  dll_traverse - Traverse a doubly linked list either forwards or
  backwards, processing each node.

  DESCRIPTION
  This routine walks along a list, calling a user supplied action
  function at each node.The action routine is passed the node's
  data pointer.  If an invocation of the action routine returns non-zero,
  the traversal is aborted and the action routine's return value is
  returned.

  Compare this to @ref{dll_traverse_d}, @ref{dll_utraverse}, and
  @ref{dll_utraverse_d}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
dll_traverse(
  LINKLIST ull,         	/* A handle to the list to traverse */
  int (*action) (void *),	/* The user supplied action routine
                                   applied to each node */
  LinkTraverseDirection direction	/* the direction in which to
                                           traverse the list */
)
{
  LL *ll = (LL *) ull;
  Node *node = NIL(ll);
  int val = 0;

  switch (direction)
  {
    case HEAD_TO_TAIL:
      for (node = FIRST(ll);
	   node != NIL(ll) && ! (val = action(DATA(node))) ;
	   node = NEXT(node)
	   );
      break;

    case TAIL_TO_HEAD:
      for (node = LAST(ll);
	   node != NIL(ll) && ! (val = action(DATA(node))) ;
	   node = PREV(node));
      break;
  }

  return val;
}

/*
  PROCEDURE
  dll_traverse_d - Traverse a doubly linked list either forwards or
  backwards, processing each node.

  DESCRIPTION
  This routine walks along a list, calling a user supplied action
  function at each node.The action routine is passed the node's data
  pointer. If an invocation of the action routine returns non-zero, the
  traversal is aborted, the current node's data pointer is stored in
  the location specified by the @code{ndata} argument, and the action
  routine's return value is returned.

  Compare this to @ref{dll_traverse}, @ref{dll_utraverse}, and
  @ref{dll_utraverse_d}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
dll_traverse_d(
  LINKLIST ull,         	/* A handle to the list to traverse */
  int (*action) (void *),	/* The user supplied action routine
                                   applied to each node */
  LinkTraverseDirection direction,	/* the direction in which to
                                           traverse the list */
  void *	ndata		/* a pointer to a @code{void *} variable
  				   which will recieve the data pointer
  				   of the node which caused the traversal
  				   to be aborted.  The type really is 
				   @code{void **}. */
)
{
  LL *ll = (LL *) ull;
  Node *node = NIL(ll);
  int val = 0;

  switch (direction)
  {
    case HEAD_TO_TAIL:
      for (node = FIRST(ll);
	   node != NIL(ll) && ! (val = action(DATA(node))) ;
	   node = NEXT(node)
	   );
      break;

    case TAIL_TO_HEAD:
      for (node = LAST(ll);
	   node != NIL(ll) && ! (val = action(DATA(node))) ;
	   node = PREV(node));
      break;
  }

  if ( val )
    *( void **) ndata = DATA(node);

  return val;
}

/*
  PROCEDURE
  dll_utraverse - Traverse a doubly linked list either forwards or
  backwards, processing each node.

  DESCRIPTION
  This routine walks along a list, calling a user supplied action
  function at each node.  The action routine is passed the node's data
  pointer as well as a pointer provided by the calling routine, thus
  allowing arbitrary data to be available to the action routine.  If an
  invocation of the action routine returns non-zero the traversal is
  aborted and the action routine's return value is returned.

  Compare this to @ref{dll_traverse}, @ref{dll_traverse_d}, and
  @ref{dll_utraverse_d}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
dll_utraverse(
  LINKLIST ull,         	/* A handle to the list to traverse */
  int (*action) (void *node, void *udata),	/* The user supplied
						   action routine
						   applied to each
						   node */
  void *udata,				/* a pointer to data to be
					   passed to the action routine */
  LinkTraverseDirection direction	/* the direction in which to
                                           traverse the list */
)
{
  LL *ll = (LL *) ull;
  Node *node = NIL(ll);
  int val = 0;

  switch (direction)
  {
    case HEAD_TO_TAIL:
      for (node = FIRST(ll);
	   node != NIL(ll) && ! (val = action(DATA(node), udata)) ;
	   node = NEXT(node)
	   );
      break;

    case TAIL_TO_HEAD:
      for (node = LAST(ll);
	   node != NIL(ll) && ! (val = action(DATA(node), udata)) ;
	   node = PREV(node));
      break;
  }

  return val;
}

/*
  PROCEDURE
  dll_utraverse_d - Traverse a doubly linked list either forwards or
  backwards, processing each node.

  DESCRIPTION
  This routine walks along a list, calling a user supplied action
  function at each node.  The action routine is passed the node's data
  pointer as well as a pointer provided by the calling routine, thus
  allowing arbitrary data to be available to the action routine.
  If an invocation of the action routine returns non-zero, the traversal is
  aborted, the current node's data pointer is stored in the location
  specified by the @code{ndata} argument, and the action routine's
  return value is returned.

  Compare this to @ref{dll_traverse}, @ref{dll_traverse_d}, and
  @ref{dll_utraverse}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
dll_utraverse_d(
  LINKLIST ull,         	/* A handle to the list to traverse */
  int (*action) (void *node, void *udata),	/* The user supplied
						   action routine
						   applied to each
						   node */
  void *udata,				/* a pointer to data to be
					   passed to the action routine */
  LinkTraverseDirection direction,	/* the direction in which to
                                           traverse the list */
  void *	ndata		/* a pointer to a @code{void *} variable
  				   which will recieve the data pointer
  				   of the node which caused the traversal
  				   to be aborted.  The type really is 
				   @code{void **}. */
)
{
  LL *ll = (LL *) ull;
  Node *node = NIL(ll);
  int val = 0;

  switch (direction)
  {
    case HEAD_TO_TAIL:
      for (node = FIRST(ll);
	   node != NIL(ll) && ! (val = action(DATA(node), udata)) ;
	   node = NEXT(node)
	   );
      break;

    case TAIL_TO_HEAD:
      for (node = LAST(ll);
	   node != NIL(ll) && ! (val = action(DATA(node), udata)) ;
	   node = PREV(node));
      break;
  }

  if ( val )
    *( void **) ndata = DATA(node);

  return val;
}

/*

  PROCEDURE
  dll_join - Join two lists.

  DESCRIPTION
  This routine moves the nodes in a list to another.  If the destination
  list has a preferred order (if a comparison function was specified
  when the list was created), the new nodes are inserted in order.
  (This implies that the data in the source list have the same format as
  those in the destination list.)  If it has no preferred order, the source
  list is simply appended to the destination list.  The source list is
  @emph{not} destroyed, it is simply emptied.

*/
void
dll_join
(
  LINKLIST dst_ull,		/* the destination list */
  LINKLIST src_ull		/* the source list */
)
{
  LL *src = (LL *) src_ull;
  LL *dst = (LL *) dst_ull;
  Node *node = FIRST(src);

  if ( NULL != CMP(dst) )
    while( node != NIL(src) )
    {
      Node *next = NEXT(node);
      Insert( dst, node );
      node = next;
    }
  else  if ( ! EMPTY(src) )
  {
    NEXT(LAST(dst)) = FIRST(src);
    PREV(FIRST(src)) = LAST(dst);

    LAST(dst) = LAST(src);

    NEXT(LAST(dst)) = NIL(dst);
    COUNT(dst) += COUNT(src);
  }

  /* now reset destination list */
   FIRST(src) = NIL(src);
   LAST(src) = NIL(src);
   COUNT(src) = 0;
}

/*

  PROCEDURE
  Insert - Insert a node into a list using the user defined collating function.

*/
static void
Insert(
  LL *ll,
  Node *node
)
{
  Node *next = FIRST(ll);

  while (next != NIL(ll) && ll->comp(DATA(next), DATA(node)) < 0)
    next = NEXT(next);

  NEXT(node) = next;
  PREV(node) = PREV(next);
  NEXT(PREV(node)) = PREV(next) = node;

  COUNT(ll)++;
}

/*

  PROCEDURE
  Search - Search for a node using either a passed compare function or
  that used to initialize the linked list.

  RETURNS
  It returns the node pointer or @code{NULL} if no match.

*/
static Node *
Search(
  LL *ll,
  void *data,
  int (*comp) (const void *, const void *)
)
{

  Node *node = FIRST(ll);

  if (!comp)
    comp = CMP(ll);

  while (node != NIL(ll) && comp(data, DATA(node)))
    node = NEXT(node);

  return (node == NIL(ll)) ? NULL : node;
}

/*

  PROCEDURE
  UnLink - Remove a node from a linked list, repairing the list.

*/

static void
UnLink(
  LL *ll,		/* the link list in question */
  Node *node		/* the node to remove */
)
{
  NEXT(PREV(node)) = NEXT(node);
  PREV(NEXT(node)) = PREV(node);
  COUNT(ll)--;
}

/*

  PROCEDURE
  deleteNode - delete a node.

  DESCRIPTION
  This routine deallocates memory associated with a node.  If asked to
  delete a @code{NULL} node, it just returns.

*/
static void
deleteNode(
  Node *node
)
{
  if (node)
    free(node);
}

/*

  PROCEDURE
  newNode - Create a new node.

  RETURNS
  It returns a pointer to the new node, @code{NULL} if it was unable to
  allocate it.

*/
static Node *
newNode(
  void *data		/* data pointer for new node */
)
{
  Node *node = (Node *) malloc(sizeof(Node));

  if (node)
    DATA(node) = data;

  return node;
}
