/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of linklist
 *
 * linklist is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * linklist is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>

#include "linklist.h"

/* macros to easily access node and list members */
#define NEXT(node)	((node)->next)
#define DATA(node)	((node)->data)
#define COUNT(ll)	((ll)->count)
#define NIL(ll)		(&(ll)->hdr)
#define FIRST(ll)	((ll)->hdr.next)
#define LAST(ll)	((ll)->last)
#define CMP(ll)		((ll)->comp)
#define EMPTY(ll)	(COUNT(ll) == 0)


#define LINKLIST	SLinkList
#define NODE		SLLNode

typedef struct Node
{
  struct Node *next;
  void *data;
} Node;

typedef struct
{
  Node hdr;
  Node *last;
  size_t count;
  int (*comp) (const void *d1, const void *d2);
} LL;

static Node *Search(LL *ll, void *data,
		     int (*comp) (const void *, const void *));
static void UnLink(LL *ll, Node *node);
static void deleteNode(Node *node);
static Node *newNode(void *data);
static void Insert(LL *ll, Node *node);


/*

  PROCEDURE
  sll_new - Create a new singly linked list.

  DESCRIPTION
  This routine creates a new list structure.  If the list is to have
  some intrinsic order, a function defining that order should be
  passed.  The comparison routine is called with two node data
  pointers as the arguments.  It must return @samp{-1}, @samp{0}, or
  @samp{1} if, respectively, the first argument is less than, equal
  to, or greater than the second.

  RETURNS
  It returns a pointer to a new linked list, or @code{NULL} if it can't
  allocate it

*/
LINKLIST
sll_new(
  int (*cmp) (const void *, const void *)
  /* a comparison function for use in searches of the
     linked list, may be @code{SLL_NULL_CMP} for unordered lists */
)
{
  LL *ll = (LL *) malloc(sizeof(LL));

  if (ll)
  {
    FIRST(ll) = NIL(ll);
    LAST(ll)  = NIL(ll);
    COUNT(ll) = 0;
    CMP(ll)   = cmp;
  }

  return (LINKLIST) ll;
}

/*

  PROCEDURE
  sll_delete - Delete a singly linked list.

  DESCRIPTION
  @code{sll_delete} traverses a linked list, calling a user-supplied
  function at each node in the list, then deleting the node.  Finally
  it deletes the list header.

 */

void
sll_delete(
  LINKLIST ull,			/* the linked list to be deleted */
  void (*ufree) (void *)	/* a function called at each
				   link to delete user data.
				   it may be @code{SLL_NULL_DELETE} */
)
{
  LL *ll = (LL *) ull;

  if (ll)
  {
    Node *next = FIRST(ll);

    while (next != NIL(ll))
    {
      Node *node = next;

      next = NEXT(node);

      if (ufree)
	ufree(DATA(node));
      free(node);
    }
    free(ll);
  }
}


/*

  PROCEDURE
  sll_udelete - Delete a singly linked list.

  DESCRIPTION
  @code{sll_udelete} traverses a linked list, calling a user-supplied
  function at each node in the list, then deleting the node.  Finally
  it deletes the list header.  This routine differs from
  @code{sll_delete} in that it can pass along a pointer provided by
  the calling routine to the action routine, allowing arbitrary data
  to be available to the action routine.

 */
void
sll_udelete(
  LINKLIST ull,			/* the linked list to be deleted */

  void (*ufree) (void *, void *),	/* a function called at each
				   link to delete user data.
				   it may be @code{SLL_NULL_DELETE} */

  void *udata			/* a pointer to data to be
				   passed to the action routine */
)
{
  LL *ll = (LL *) ull;

  if (ll)
  {
    Node *next = FIRST(ll);

    while (next != NIL(ll))
    {
      Node *node = next;

      next = NEXT(node);

      if (ufree)
	ufree( DATA(node), udata );
      free( node );
    }
    free( ll );
  }
}


/*

  PROCEDURE
  sll_destroy - Destroy a node in a singly linked list.

  DESCRIPTION
  @code{sll_destroy} searches a linked list for a node which compares
  equivalently to the passed data.  It removes the node from the list
  and deallocates the memory associated with it.  The user must
  destroy the data referenced by the node's data pointer.

  It uses the passed comparison routine, if available.  If not, it
  uses that with which the list was initialized.  If @emph{that} is
  not available, a segmentation violation is unavoidable.  In the
  former case the passed data need not have the same form as the data
  stored in the node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  RETURNS
  It returns the node's data pointer if the node was found, @code{NULL}
  otherwise.

*/
void *
sll_destroy(
  LINKLIST ull,			/* the list to search */
  void *data,			/* the data to search for */
  int (*cmp) (const void *, const void *)
	      /* The address of a comparison function.
		 Set to @code{SLL_NULL_CMP} to use the list's initial
		 comparison function. */
)
{
  LL *ll = (LL *) ull;
  Node *node = Search(ll, data, cmp);

  data = NULL;
  if (node)
  {
    data = DATA(node);
    UnLink(ll, node);
    free(node);
  }

  return data;
}

/*

  PROCEDURE
  sll_destroy_head - Remove the head node from a singly linked list
  and destroy it.

  DESCRIPTION
  This routine removes the head node from the specified list and
  deallocates the memory associated with it.  The user must destroy
  the data referenced by the node's data pointer.

  RETURNS
  It returns the node's data pointer.  If the list is empty, it
  returns @code{NULL}.

*/
void *
sll_destroy_head(
  LINKLIST ull			/* the list from which to remove the node */
)
{
  LL *ll = (LL *) ull;
  Node *node = FIRST(ll);
  void *data = NULL;

  if (node != NIL(ll))
  {
    data = DATA(node);
    UnLink(ll, node);
    deleteNode(node);
  }

  return data;
}

/*

  PROCEDURE
  sll_destroy_tail - Remove the tail node from a singly linked list
  and destroy it.

  DESCRIPTION
  This routine removes the tail node from the specified list and
  deallocates the memory associated with it.  The user must destroy
  the data referenced by the node's data pointer.

  RETURNS
  It returns the node's data pointer.  If the list is empty, it
  returns @code{NULL}.

*/
void *
sll_destroy_tail(
  LINKLIST ull			/* the list from which to remove the node */
)
{
  LL *ll = (LL *) ull;
  Node *node = LAST(ll);
  void *data = NULL;

  if (node != NIL(ll))
  {
    data = DATA(node);
    UnLink(ll, node);
    deleteNode(node);
  }

  return data;
}

/*

  PROCEDURE
  sll_count - Determine the number of nodes in a singly linked list.

  DESCRIPTION
  This function returns the number of nodes in a list.  It is an
  inexpensive routine to call.

  RETURNS
  It returns the number of nodes in the list, @samp{0} if the passed
  pointer is @code{NULL}.

*/
size_t
sll_count(
  LINKLIST ull			/* the list to count */
)
{
  return ull ? COUNT(((LL *) ull)) : 0;
}

/*

  PROCEDURE
  sll_destroy_dnode - Deallocate the memory associated with a detached node.

  DESCRIPTION
  This routine deallocates the memory associated with a detached node.
  The user must have already destroyed the data associated with the
  node.  The node handle must have been returned by
  @w{@code{sll_detach_node}}.

*/
void
sll_destroy_dnode(
  NODE unode			/* the node to destroy */
)
{
  deleteNode((Node *) unode);
}

/*

  PROCEDURE
  sll_destroy_node - Remove a node from a singly linked list and destroy it.

  DESCRIPTION
  This routine removes the passed node from the specified list and
  deallocates the memory associated with it.  The user must destroy
  the data referenced by the node's data pointer.

  RETURNS
  It returns the node's data pointer if the node is not @code{NULL},
  @code{NULL} otherwise.

*/
void *
sll_destroy_node(
  LINKLIST ull,			/* the list from which to remove the node */
  NODE unode			/* the node to remove */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;
  void *data = NULL;

  if (node)
  {
    data = DATA(node);
    UnLink(ll, node);
    deleteNode(node);
  }

  return data;
}

/*
  PROCEDURE
  sll_detach_node - Remove a node from a singly linked list.

  DESCRIPTION
  This routine removes a node from a list without destroying it.  The
  user is responsible for its subsequent care.

  RETURNS
  It returns a handle for the detached node.

*/
void
sll_detach_node(
  LINKLIST ull,			/* the list from which to detach the node */
  NODE unode			/* the node to detach */
)
{
  if (unode)
    UnLink((LL *) ull, (Node *) unode);
}

/*

  PROCEDURE
  sll_insert - Create and insert a node into a singly linked list.

  DESCRIPTION
  @code{sll_insert} creates a node, stores the passed data pointer in
  it, and inserts the node in the list in the collating order
  determined by the comparison function with which the list was
  initialized by @code{sll_new}.

  RETURNS
  It returns zero if the insert was successful, non-zero if it was unable
  to create the new node

 */
int
sll_insert(
  LINKLIST ull,		/* a handle to the list into which to insert
			   the node */
  void *data		/* a pointer to the new node's data */
)
{
  LL *ll = (LL *) ull;
  Node *node = newNode(data);

  if (!node)
    return 1;

  Insert(ll, node);
  return 0;
}

/*

  PROCEDURE
  sll_insert_node - Insert a detached node into a list.

  DESCRIPTION
  This routine inserts a detached node into a list in the collating
  sequence determined by the comparison function with which the list
  was initialized by @code{sll_new}.  It assumes that the list's
  insert/delete comparison function can be applied to the data in the
  passed node.  The node handle must have been obtained from
  @code{sll_detach_node}.

  RETURNS
  It returns non-zero if the node is @code{NULL}, zero otherwise.

*/
int
sll_insert_node(
  LINKLIST ull,		/* a handle to the list into which to insert
			   the node */
  NODE unode		/* the node to insert */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  if (!node)
    return 1;

  Insert(ll, node);
  return 0;
}

/*
  PROCEDURE
  sll_insert_head - Create and insert a node at the head of a singly.
  linked list

  DESCRIPTION
  This routine creates a node, inserts the passed data pointer in it,
  and attaches the node to the head of the list.

  RETURNS
  It returns zero if the insert was successful, non-zero if it was unable
  to create the new node.

*/
int
sll_insert_head(
  LINKLIST ull,		/* a handle to the list into which to insert
			   the node */
  void *data		/* a pointer to the new node's data */
)
{
  Node *node = newNode(data);

  return node ? sll_insert_head_node(ull, (NODE) node) : 1;
}

/*

  PROCEDURE
  sll_insert_head_node - Insert a detached node at the head of a
  singly linked list.

  DESCRIPTION
  This routine inserts a node at the head of a list.  The node handle
  must have been obtained from @code{sll_detach_node}.

  RETURNS
  It returns non-zero if the node is @code{NULL}, zero otherwise.

*/
int
sll_insert_head_node(
  LINKLIST ull,		/* a handle to the list into which to insert
			   the node */
  NODE unode		/* the node to insert */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  if (!node)
    return 1;

  NEXT(node) = FIRST(ll);
  FIRST(ll) = node;
  if (NEXT(node) == NIL(ll))
    LAST(ll) = node;

  COUNT(ll)++;

  return 0;
}

/*

  PROCEDURE
  sll_insert_tail - Create and insert a node at the tail of a singly
  linked list.


  DESCRIPTION
  This routine creates a node, inserts the passed data pointer in it,
  and attaches the node to the tail of the list.

  RETURNS
  It returns zero if the insert was successful, non-zero if it was unable
  to create the new node.

*/
int
sll_insert_tail(
  LINKLIST ull,		/* a handle to the list into which to insert
			   the node */
  void *data		/* a pointer to the new node's data */
)
{
  Node *node = newNode(data);

  return node ? sll_insert_tail_node(ull, (NODE) node) : 1;
}

/*

  PROCEDURE
  sll_insert_tail_node - Insert a detached node at the tail of a  list.

  DESCRIPTION
  This routine inserts a node at the tail of a list.  The node handle
  must have been obtained from @code{sll_detach_node}.

  RETURNS
  It returns non-zero if the node is @code{NULL}, zero otherwise.

*/
int
sll_insert_tail_node(
  LINKLIST ull,		/* a handle to the list into which to insert
			   the node */
  NODE unode		/* the node to insert */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  if (!node)
    return 1;

  NEXT(node) = NIL(ll);
  NEXT(LAST(ll)) = node;
  LAST(ll) = node;

  COUNT(ll)++;

  return 0;
}

/*

  PROCEDURE
  sll_head - Get the data of the head node of a list.

  RETURNS
  It returns the data of the head node of a list, @code{NULL} if the list
  is empty.

*/
void *
sll_head(
  LINKLIST ull		/* the list in question */
)
{
  LL *ll = (LL *) ull;

  return EMPTY(ll) ? NULL : DATA(FIRST(ll));
}

/*
  PROCEDURE
  sll_head_node - Get a node handle to the head node of a list.

  RETURNS
  It returns the node handle of the head node of a list, @code{NULL} if
  the list is empty.

*/
NODE
sll_head_node(
  LINKLIST ull		/* the list in question */
)
{
  LL *ll = (LL *) ull;

  return EMPTY(ll) ? NULL : FIRST(ll);
}

/*
  PROCEDURE
  sll_detach_head_node - Detach the head node of a list.

  DESCRIPTION
  This routine detaches the head node of a list from the list and returns
  a handle to the detached node.

  RETURNS
  It returns a handle to the node upon success, @code{NULL} on failure.

*/
NODE
sll_detach_head_node(
  LINKLIST ull		/* the list in question */
)
{
  LL *ll = (LL *) ull;


  if ( EMPTY(ll) )
    return NULL;
  else
  {
    Node *node = FIRST(ll);
    UnLink( ll, node );
    return node;
  }
}

/*
  PROCEDURE
  sll_tail - Get the data of the tail node of a list.

  RETURNS
  It returns the data of the tail node of a list, @code{NULL} if the list
  is empty.

*/
void *
sll_tail(
  LINKLIST ull		/* the list in question */
)
{
  LL *ll = (LL *) ull;

  return EMPTY(ll) ? NULL : DATA(LAST(ll));
}

/*
  PROCEDURE
  sll_tail_node - Get a node handle to the tail node of a list.

  RETURNS
  It returns the node handle of the tail node of a list, @code{NULL} if
  the list is empty.

 */
NODE
sll_tail_node(
  LINKLIST ull		/* the list in question */
)
{
  LL *ll = (LL *) ull;

  return EMPTY(ll) ? NULL : LAST(ll);
}


/*
  PROCEDURE
  sll_sizeof_node - Get the size of a node.

  DESCRIPTION
  This routine returns the size of the internal structure used for
  each node.  This does not include user supplied data.  Please note
  that this is a function call, not a macro!

 */
size_t
sll_sizeof_node(void)
{
  return sizeof(Node);
}


/*
  PROCEDURE
  sll_next - Retrieve the data in the node following a given node in a
  list.

  RETURNS
  It returns @code{NULL} if there are no more nodes, else a pointer to
  the data.

*/
void *
sll_next(
  LINKLIST ull,		/* the list which contains the node */
  NODE unode		/* the preceding node */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  return (NEXT(node) == NIL(ll)) ? NULL : DATA(NEXT(node));
}


/*
  PROCEDURE
  sll_next_node - Get a handle to the node following a given node in a
  list.

  RETURNS
  It returns @code{NULL} if there are no more nodes, else a
  handle to the node.

*/
NODE
sll_next_node(
  LINKLIST ull,		/* the list which contains the node */
  NODE unode		/* the preceding node */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;

  return (NEXT(node) == NIL(ll)) ? NULL : NEXT(node);
}


/*
  PROCEDURE
  sll_prev - Retrieve the data in the node preceding a given node in a list.

  DESCRIPTION
  Search a singly linked list starting at the list head and determine
  the node previous to the passed node.  This is a very expensive
  operation!  Before using this, think about using a doubly linked
  list!

  RETURNS
  It returns @code{NULL} if there is none, else a pointer to the previous
  node's data.

*/
void *
sll_prev(
  LINKLIST ull,		/* the list which contains the node */
  NODE unode		/* the following node */
)
{
  Node *node = (Node *) sll_prev_node(ull, unode);

  return node ? DATA(node) : NULL;
}

/*
  PROCEDURE
  sll_prev_node - Get a handle to the node previous to a given node in
  a list.

  DESCRIPTION
  Search a singly linked list starting at the list head and determine
  the node previous to the passed node.  This is a very expensive
  operation!  Before using this, think about using a doubly linked
  list!

  RETURNS
  It returns @code{NULL} if there is none, else a handle to the node.

*/
NODE
sll_prev_node(
  LINKLIST ull,		/* the list which contains the node */
  NODE unode		/* the following node */
)
{
  LL *ll = (LL *) ull;
  Node *node = (Node *) unode;
  Node *prev = NIL(ll);
  Node *next = FIRST(ll);

  while (next != NIL(ll) && next != node)
  {
    prev = next;
    next = NEXT(next);
  }

  return prev == NIL(ll) ? NULL : prev;
}


/*
  PROCEDURE
  sll_node_get_data - Retrieve the data from a specified linked list node.


  RETURNS
  It returns the node's data pointer.

*/
void *
sll_node_get_data(
  NODE unode		/* the node in question */
)
{
  return unode ? DATA((Node *) unode) : NULL;
}

/*
  PROCEDURE
  sll_node_put_data - Replace the data pointer in a given node with another.

  DESCRIPTION
  This function allows the calling routine to change the data that a
  node points to.  No check is made to insure that any existing
  ordering of the data is maintained.

*/
void
sll_node_put_data(
  NODE unode,		/* the node in question */
  void *data		/* the new data */
)
{
  if ( unode )
    DATA((Node *) unode) = data;
}

/*
  PROCEDURE
  sll_search - Search a linked list for a node with equivalent data.

  DESCRIPTION
  This searchs the list for the node for which the passed data
  compares equivalently with the node's data.  It uses the passed
  comparison routine, if available.  If not, it uses that with which
  the list was initialized.  If @emph{that} is not available, a
  segmentation violation is unavoidable.  In the former case the
  passed data need nothave the same form as the data stored in the
  node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  RETURNS
  It returns the node's data pointer if the node was found,
  @code{NULL} otherwise.

*/
void *
sll_search(
  LINKLIST ull,		/* the list to search */
  void *data,		/* the data to search for */
  int (*cmp) (const void *, const void *)
	      /* The address of a comparison function.
		 Set to @code{SLL_NULL_CMP} to use the list's initial
		 comparison function. */
)
{
  LL *ll = (LL *) ull;
  Node *node = Search(ll, data, cmp);

  return node ? DATA(node) : NULL;
}

/*
  PROCEDURE
  sll_search_node - Search a linked list for a node with equivalent data.

  DESCRIPTION
  This searchs the list for the node for which the passed data
  compares equivalently with the node's data.  It uses the passed
  comparison routine, if available.  If not, it uses that with which
  the list was initialized.  If @emph{that} is not available, a
  segmentation violation is unavoidable.  In the former case the
  passed data need nothave the same form as the data stored in the
  node.

  The comparison routine is called with the passed data as the first
  argument and the node's data as the second argument.  It must return
  @samp{-1}, @samp{0}, or @samp{1} if, respectively, the first
  argument is less than, equal to, or greater than the second.

  RETURNS
  It returns a handle to the matching mode, or @code{NULL} if not found.

*/
NODE
sll_search_node(
  LINKLIST ull,		/* the list to search */
  void *data,		/* the data to search for */
  int (*cmp) (const void *, const void *)
	      /* The address of a comparison function.
		 Set to @code{SLL_NULL_CMP} to use the list's initial
		 comparison function. */
)
{
  LL *ll = (LL *) ull;

  return (NODE) Search(ll, data, cmp);
}

/*
  PROCEDURE
  sll_traverse - Traverse a singly linked list, processing each node.

  DESCRIPTION
  This routine walks along a list, calling a user supplied action
  function at each node.  The action routine is passed the node's data
  pointer.  If an invocation of the action routine returns non-zero,
  the traversal is aborted and the action routine's return value is
  returned.

  Compare this to @ref{sll_traverse_d}, @ref{sll_utraverse}, and
  @ref{sll_utraverse_d}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
sll_traverse(
  LINKLIST ull,			/* A handle to the list to traverse */
  int (*action) (void *)	/* The user supplied action routine
				   to be applied to each node */
)
{
  LL *ll = (LL *) ull;
  Node *node;
  int val = 0;

  for (node = FIRST(ll);
       node != NIL(ll) && ! (val = action(DATA(node))) ;
       node = NEXT(node)
       );

  return val;
}

/*
  PROCEDURE
  sll_traverse_d - Traverse a singly linked list, processing each node.

  DESCRIPTION
  This routine walks along a list, calling a user supplied action
  function at each node. The action routine is passed the node's data
  pointer.  If the action routine returns non-zero, the traversal is
  aborted, the current node's data pointer is stored in the location
  specified by the @code{ndata} argument, and the action routine's
  return value is returned.

  Compare this to @ref{sll_traverse}, @ref{sll_utraverse}, and
  @ref{sll_utraverse_d}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
sll_traverse_d(
  LINKLIST ull,			/* A handle to the list to traverse */
  int (*action) (void *),	/* The user supplied action routine
				   to be applied to each node */
  void *	ndata		/* a pointer to a @code{void *} variable
  				   which will recieve the data pointer
  				   of the node which caused the traversal
  				   to be aborted.  The type really is 
				   @code{void **}. */
)
{
  LL *ll = (LL *) ull;
  Node *node;
  int val = 0;

  for (node = FIRST(ll);
       node != NIL(ll) && ! (val = action(DATA(node))) ;
       node = NEXT(node)
       );

  if ( val )
    *( void **) ndata = DATA(node);

  return val;
}

/*

  PROCEDURE
  sll_utraverse - Traverse a singly linked list, processing each node.

  DESCRIPTION
  This routine walks along a list, calling a user supplied action
  function at each node.  The action routine is passed the node's data
  pointer as well as a pointer provided by the calling routine, thus
  allowing arbitrary data to be available to the action routine.  If an
  invocation of the action routine returns non-zero the traversal is
  aborted and the action routine's return value is returned.

  Compare this to @ref{sll_traverse}, @ref{sll_traverse_d}, and
  @ref{sll_utraverse_d}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
sll_utraverse(
  LINKLIST ull,				/* A handle to the list to traverse */
  int (*action) (void *node, void *udata),	/* The user supplied action
					   routine applied to each node */
  void *udata				/* a pointer to data to be
					   passed to the action routine */
)
{
  LL *ll = (LL *) ull;
  Node *node;
  int val = 0;

  for (node = FIRST(ll);
       node != NIL(ll) && ! (val = action(DATA(node), udata)) ;
       node = NEXT(node)
       );

  return val;
}

/*

  PROCEDURE
  sll_utraverse_d - Traverse a singly linked list, processing each node.

  DESCRIPTION
  This routine walks along a list, calling a user supplied action
  function at each node.  The action routine is passed the node's data
  pointer as well as a pointer provided by the calling routine, thus
  allowing arbitrary data to be available to the action routine.
  If an invocation of the action routine returns non-zero, the traversal is
  aborted, the current node's data pointer is stored in the location
  specified by the @code{ndata} argument, and the action routine's
  return value is returned.

  Compare this to @ref{sll_traverse}, @ref{sll_traverse_d}, and
  @ref{sll_utraverse}.

  RETURNS
  If the action routines all return @samp{0}, it returns @samp{0},
  else it returns the value returned by the last action routine
  called.

*/
int
sll_utraverse_d(
  LINKLIST ull,				/* A handle to the list to traverse */
  int (*action) (void *node, void *udata),	/* The user supplied action
					   routine applied to each node */
  void *udata,				/* a pointer to data to be
					   passed to the action routine */
  void *	ndata		/* a pointer to a @code{void *} variable
  				   which will recieve the data pointer
  				   of the node which caused the traversal
  				   to be aborted.  The type really is 
				   @code{void **}. */
)
{
  LL *ll = (LL *) ull;
  Node *node;
  int val = 0;

  for (node = FIRST(ll);
       node != NIL(ll) && ! (val = action(DATA(node), udata)) ;
       node = NEXT(node)
       );

  if ( val )
    *( void **) ndata = DATA(node);

  return val;
}

/*

  PROCEDURE
  sll_join - Join two lists.

  DESCRIPTION
  This routine moves the nodes in a list to another.  If the destination
  list has a preferred order (if a comparison function was specified
  when the list was created), the new nodes are inserted in order.
  (This implies that the data in the source list have the same format as
  those in the destination list.)  If it has no preferred order, the source
  list is simply appended to the destination list.  The source list is
  @emph{not} destroyed, it is simply emptied.

*/
void
sll_join
(
  LINKLIST dst_ull,		/* the destination list */
  LINKLIST src_ull		/* the source list */
)
{
  LL *src = (LL *) src_ull;
  LL *dst = (LL *) dst_ull;
  Node *node = FIRST(src);

  if ( NULL != CMP(dst) )
    while( node != NIL(src) )
    {
      Node *next = NEXT(node);
      Insert( dst, node );
      node = next;
    }
  else  if ( ! EMPTY(src) )
  {
    NEXT(LAST(dst)) = FIRST(src);
    LAST(dst) = LAST(src);
    NEXT(LAST(dst)) = NIL(dst);
    COUNT(dst) += COUNT(src);
  }

  /* now reset source list */
   FIRST(src) = NIL(src);
   LAST(src)  = NIL(src);
   COUNT(src) = 0;
}

/*

  PROCEDURE
  Insert - Insert a node into a list using the user defined collating function.

*/
static void
Insert(
  LL *ll,
  Node *node
)
{
  Node *prev = NIL(ll);
  Node *next = FIRST(ll);

  while (next != NIL(ll) && ll->comp(DATA(next), DATA(node)) < 0)
  {
    prev = next;
    next = NEXT(next);
  }

  NEXT(node) = next;
  NEXT(prev) = node;


  if (NEXT(node) == NIL(ll))
    LAST(ll) = node;

  COUNT(ll)++;
}

/*

  PROCEDURE
  Search - Search for a node using either a passed compare function or
  that used to initialize the linked list.

  RETURNS
  It returns the node pointer or @code{NULL} if no match.

*/
static Node *
Search(
  LL *ll,
  void *data,
  int (*comp) (const void *, const void *)
)
{

  Node *node = FIRST(ll);

  if (!comp)
    comp = CMP(ll);

  while (node != NIL(ll) && comp(data, DATA(node)))
    node = NEXT(node);

  return (node == NIL(ll)) ? NULL : node;
}

/*

  PROCEDURE
  UnLink - Remove a node from a linked list, repairing the list.

*/

static void
UnLink(
  LL *ll,		/* the link list in question */
  Node *node		/* the node to remove */
)
{
  Node *prev;

  /* search for previous node */
  for (prev = NIL(ll);
       NEXT(prev) != node && NEXT(prev) != NIL(ll);
       prev = NEXT(prev));

  if (NEXT(prev) != NIL(ll))
  {
    /* ensure that the last pointer in the header is kept up-to-date */
    if ( LAST(ll) == node )
      LAST(ll) = prev;

    NEXT(prev) = NEXT(node);
    COUNT(ll)--;
  }
}

/*

  PROCEDURE
  deleteNode - Delete a node.

  DESCRIPTION
  This routine deallocates memory associated with a node.  If asked to
  delete a @code{NULL} node, it just returns.

*/
static void
deleteNode(
  Node *node
)
{
  if (node)
    free(node);
}

/*

  PROCEDURE
  newNode - Create a new node.

  RETURNS
  It returns a pointer to the new node, @code{NULL} if it was unable to
  allocate it.

*/
static Node *
newNode(
  void *data		/* data pointer for new node */
)
{
  Node *node = (Node *) malloc(sizeof(Node));

  if (node)
    DATA(node) = data;

  return node;
}
