/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of linklist
 *
 * linklist is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * linklist is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#include <linklist.h>

#ifdef USE_SLL

#define LINKLIST	SLinkList
#define NODE		SLLNode
#define NEW		sll_new
#define HEAD		sll_head
#define TAIL		sll_tail
#define COUNT		sll_count
#define INSERT_TAIL	sll_insert_tail
#define INSERT_HEAD	sll_insert_head
#define HEAD_NODE	sll_head_node
#define NEXT_NODE	sll_next_node
#define DELETE		sll_delete
#define DESTROY		sll_destroy
#define DESTROY_HEAD	sll_destroy_head
#define DESTROY_TAIL	sll_destroy_tail

#else

#define LINKLIST	DLinkList
#define NODE		DLLNode
#define NEW		dll_new
#define HEAD		dll_head
#define TAIL		dll_tail
#define COUNT		dll_count
#define INSERT_TAIL	dll_insert_tail
#define INSERT_HEAD	dll_insert_head
#define HEAD_NODE	dll_head_node
#define NEXT_NODE	dll_next_node
#define DELETE		dll_delete
#define DESTROY		dll_destroy
#define DESTROY_HEAD	dll_destroy_head
#define DESTROY_TAIL	dll_destroy_tail

#endif 


LINKLIST nodes;

int list[] = {1, 3, 5, 7, 9, 0, 2, 4, 6, 8};
#define LEN(list) (sizeof(list) / sizeof(int))

int cmp( const void *i1, const void *i2)
{
  return *((int *) i1) - *((int *) i2);
}

int
print_node( void *data, void *udata )
{
  printf( "%s: %d\n", (char *) udata, *((int *) data) ); 
  return 0;
}

/* brute force counting */
int count_elements( LINKLIST ll )
{
  NODE node;
  int nelem = 0;

  for ( node = HEAD_NODE( ll );
	node ;
	node = NEXT_NODE( ll, node ) )
    nelem++;

  return nelem;
}

#define ALLOC_NODE(ptr, value)			\
  do{						\
    ptr = malloc( sizeof(int) );		\
    assert( NULL != ptr );			\
    *ptr = value;				\
    INSERT_TAIL( nodes, ptr );		\
  } while(0)

void
populate_list( LINKLIST ll )
{
  int i;
  int* ip;

  for ( i = 0 ; i < LEN(list) ; i++ )
  {
    ALLOC_NODE( ip, list[i] );
    INSERT_TAIL( ll, ip );
  }

  assert( COUNT( ll ) == LEN(list) );
  assert( count_elements( ll ) == LEN(list) );
}


int
main( int argc, char *argv[])
{
  int i;

  LINKLIST ll;

  ll = NEW( cmp );

  nodes = NEW( NULL );

  populate_list( ll );

  /***************************************************************/
  /* destroy list from front to back */
  for ( i = 0 ; i < LEN(list) ; i++ )
  {
    int *ptr = DESTROY( ll, &list[i], NULL );
    int len = LEN(list) - i - 1;

    assert( *ptr == list[i] );
    assert(      COUNT( ll ) == len);
    assert( count_elements( ll ) == len);

    /* make sure head and tail of list is ok */
    if ( i < LEN(list) - 1 )
    {
      assert( *((int*) HEAD(ll)) == list[i+1] );
      assert( *((int*) TAIL(ll)) == list[LEN(list)-1] );
    }

    /* insert at the head */
    ALLOC_NODE( ptr, list[i] );
    INSERT_HEAD( ll, ptr );

    assert(      COUNT( ll ) == len + 1 );
    assert( count_elements( ll ) == len + 1 );

    assert( *((int*) HEAD(ll))         == list[i] );
    assert( *((int*) TAIL(ll))         == list[LEN(list)-1] );
    assert( *((int*) DESTROY_HEAD(ll)) == list[i] );

    /* insert at the tail */
    ALLOC_NODE( ptr, list[i] );
    INSERT_TAIL( ll, ptr );
    assert(      COUNT( ll ) == len + 1 );
    assert( count_elements( ll ) == len + 1 );

    if ( i < LEN(list) - 1 )
      assert( *((int*) HEAD(ll)) == list[i+1] );

    assert( *((int*) TAIL(ll)) == list[i] );

    assert( *((int*) DESTROY_TAIL(ll)) == list[i] );
  }

  assert( COUNT( ll ) == 0 );
  assert( count_elements( ll ) == 0 );

  populate_list( ll );

  /***************************************************************/
  /* destroy list from back to front */
  for ( i = LEN(list)-1 ; i >=0 ; i-- )
  {
    int *ptr = DESTROY( ll, &list[i], NULL );
    int len = i;

    assert( *ptr == list[i] );
    assert(      COUNT( ll ) == len );
    assert( count_elements( ll ) == len );

    /* make sure head and tail of list is ok */
    if ( i > 0 )
    {
      assert( *((int*) HEAD(ll)) == list[0] );
      assert( *((int*) TAIL(ll)) == list[i-1] );
    }

    /* insert at the head */
    ALLOC_NODE( ptr, list[i] );
    INSERT_HEAD( ll, ptr );
    assert(      COUNT( ll ) == len + 1 );
    assert( count_elements( ll ) == len + 1 );

    assert( *((int*) HEAD(ll)) == list[i] );
    if ( i > 0 )
      assert( *((int*) TAIL(ll)) == list[i-1] );
    assert( *((int*) DESTROY_HEAD(ll)) == list[i] );

    /* insert at the tail */
    ALLOC_NODE( ptr, list[i] );
    INSERT_TAIL( ll, ptr );
    assert(      COUNT( ll ) == len + 1 );
    assert( count_elements( ll ) == len + 1 );

    assert( *((int*) HEAD(ll)) == list[0] );
    assert( *((int*) TAIL(ll)) == list[i] );

    assert( *((int*) DESTROY_TAIL(ll)) == list[i] );
  }

  assert( COUNT( ll ) == 0 );
  assert( count_elements( ll ) == 0 );

  DELETE( ll, NULL );
  DELETE( nodes, free );

  return 0;
}

