/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of luasup
 *
 * luasup is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * luasup is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdarg.h>

#include <lua.h>
#include <lualib.h>

#include <suplib/str.h>

#include "luasup.h"


/*

  PROCEDURE
  slua_args - read in the arguments passed to a C function from Lua

  DESCRIPTION
  This routine will read the arguments passed by Lua to a C function
  when the function is called from Lua.  It expects a list of triples,
  each of which contains the argument's name (for error reporting
  purposes), the argument's type (one of @code{SLua_STRING},
  @code{SLua_NUMBER} or @code{SLua_TABLE}), and a pointer to storage
  for the argument.  The list should be terminated by a @code{NULL}
  value.

  Numeric values are stored as @code{double}.  String values are
  returned as a pointer to a duplicate string.  The calling routine is
  responsible for deallocating this memory.

  As @code{slua_args} uses @code{slua_args_va}, see its write-up
  for more information.

  RETURNS
  It returns the number of arguments found, or @samp{-1} if an error
  ocurred.  If the number of arguments passed by Lua is different than
  in the passed list, an error message is generated.

  ERRORS
  Errors are returned via the passed @code{SLua_ErrFunc} function.

  AUTHOR
  Diab Jerius

  EXAMPLE
  @example
  @include lib/slua_args_c.texi
  @end example

*/
int
slua_args(
  SLua_ErrFunc errfunc, 		/* the error function */
  const char *func, 			/* the name of the Lua called
					   function.  Used for error
					   messages */
  ...					/* the list of arguments */
  )
{
  va_list	ap;
  int		n;

  va_start( ap, func );

  n = slua_args_va( errfunc, func, 1, ap );
  
  va_end( ap );

  if ( n > 0 && lua_getparam(n+1) != LUA_NOOBJECT )
    errfunc( "too many arguments in call to lua function `%s'", func);

  return n;
}

/*

  PROCEDURE
  slua_args_n - read in the arguments passed to a C function from Lua

  DESCRIPTION
  This routine will read the arguments passed by Lua to a C function
  when the function is called from Lua.  It expects a list of triples,
  each of which contains the argument's name (for error reporting
  purposes), the argument's type (one of @code{SLua_STRING},
  @code{SLua_NUMBER}, or @code{SLua_TABLE}), and a pointer to storage
  for the argument.  The list should be terminated by a @code{NULL}
  value.

  Numeric values are stored as @code{double}.  String values are
  returned as a pointer to a duplicate string.  The calling routine is
  responsible for deallocating this memory.

  It differs from @code{slua_args} in that it will start with the Lua
  argument at the position specified by @code{n}, and will not
  complain about extra passed arguments.

  As @code{slua_args_n} uses @code{slua_args_va}, see its write-up
  for more information.

  RETURNS
  It returns the number of arguments found, or @samp{-1} if an error
  ocurred.

  ERRORS
  Errors are returned via the passed @code{SLua_ErrFunc} function.

  AUTHOR
  Diab Jerius

  EXAMPLE
  @example
  @include lib/slua_args_n_c.texi
  @end example

*/

int
slua_args_n(
  SLua_ErrFunc errfunc, 		/* the error function */
  const char *func, 			/* the name of the Lua called
					   function.  Used for error
					   messages */
  int n,				/* the Lua argument position
					   at which to start. Unary
					   based. */
  ...					/* the list of arguments */
)
{
  int nargs;
  va_list	ap;

  va_start( ap, n );

  nargs = slua_args_va( errfunc, func, n, ap );

  va_end( ap );

  return nargs;
}


/*

  PROCEDURE
  slua_args_va - read in the arguments passed to a C function from Lua

  DESCRIPTION
  This routine will read the arguments passed by Lua to a C function
  when the function is called from Lua.  It expects a @code{va_list}
  of triples, each of which contains the argument's name (for error
  reporting purposes), the argument's type (one of @code{SLua_STRING},
  @code{SLua_NUMBER}, or @code{SLua_TABLE}) and a pointer to storage
  for the argument.  The list should be terminated by a @code{NULL}
  value.  Support is provided for optional arguements.  These must be
  at the end of the arguement list, and should have their type or'd
  with @code{SLua_OPTIONAL} (e.g. @code{SLua_TABLE | SLua_OPTIONAL}).
  To determine if an optional argument was present, use
  @code{slua_args_va}'s return value.

  Numeric values are stored as @code{double}.  String values are
  returned as a pointer to a duplicate string.  The calling routine is
  responsible for deallocating this memory.

  Arguments are returned beginning with the Lua argument at
  the position specified by @code{n}.

  RETURNS
  It returns the number of arguments found, or @samp{-1} if an error
  ocurred.

  ERRORS
  Errors are returned via the passed @code{SLua_ErrFunc} function.

  AUTHOR
  Diab Jerius

  EXAMPLE
  @example
  @include lib/slua_args_va_c.texi
  @end example

*/

int
slua_args_va(
  SLua_ErrFunc errfunc, 		/* the error function */
  const char *func, 			/* the name of the Lua called
					   function.  Used for error
					   messages */
  int n,				/* the Lua argument position
					   at which to start. Unary
					   based. */
  va_list ap				/* the list of arguments */
)
{
  lua_Object	obj;
  char *	name;
  int		nargs = 0;

  for ( ; (name = (char *) va_arg( ap, char *)) && name[0] != '\0' ; n++)
  {
    SLuaArgType	type = (SLuaArgType) va_arg( ap, SLuaArgType );
    void *	dest = (void *)     va_arg( ap, void * );
    int optional = type & SLua_OPTIONAL;
    type &= SLua_TYPES;

    /* if there are no more arguments in the list, and the requested
       argument wasn't optional, signal an error */
    if ( LUA_NOOBJECT == ( obj = lua_getparam( n ) )  )
    {

      /* this was optional, continue */
      if ( optional )
	continue;

      /* clean out va_list */
      while ( NULL != (char *) va_arg( ap, char *) )
      {
        type = (SLuaArgType) va_arg( ap, SLuaArgType );
        dest = (void *)     va_arg( ap, void * );
      }
      va_end( ap );
      errfunc("too few arguments in call to lua function `%s' " 
		   "at argument `%s'", func, name );
    }

    if ( SLua_NUMBER != type && SLua_TABLE != type && SLua_STRING != type )
      errfunc( "unknown lua object type: %d\n", type );

    /* number */
    if      ( SLua_NUMBER == type && lua_isnumber( obj ) )
      *((double *) dest) = lua_getnumber( obj );

    /* string */
    else if ( SLua_STRING == type && lua_isstring( obj ) )
      *(( char **) dest) = str_dup( lua_getstring( obj ) );

    /* table */
    else if ( SLua_TABLE  == type && lua_istable( obj  ) )
      *((lua_Object *) dest) = obj;

    else
    {
      while ( NULL != (char *) va_arg( ap, char *) )
      {
        SLuaArgType type = (SLuaArgType) va_arg( ap, SLuaArgType );
        dest = (void *)     va_arg( ap, void * );
      }
      va_end( ap );
      errfunc( "incorrect argument type in call to lua function `%s' "
		    "for argument `%s': expected a %s", func, name,
	       SLua_NUMBER == type ? "number" :
	       (SLua_STRING == type ? "string" : "table") );
      return -1;
    }

    nargs++;
  }

  return nargs;
}


/*

  PROCEDURE
  slua_check_n_arg - check the number of arguments passed to C from Lua.

  DESCRIPTION
  This routine compares the number of arguments expected (passed via
  the @code{n} parameter) to the actual number of arguments passed
  to the C function by Lua.  It's a simple wrapper around
  @code{lua_getparam}.  
  
  RETURNS
  If the number of arguments matches is not less than that passed via
  @code{n}, it returns zero.  If there are fewer arguments than
  expected, it calls the supplied error function and returns -1.
  
  AUTHOR
  Diab Jerius

  EXAMPLE
  @example
  @include lib/slua_args_n_c.texi
  @end example

 */

int
slua_check_n_arg(
  SLua_ErrFunc	errfunc,
  const char *	func,		/* the name of the lua function */
  int		n		/* the number of arguments expected */
)
{
  if ( lua_getparam( n+1 ) != LUA_NOOBJECT )
  {
    errfunc( "too many arguments in call to lua function `%s'", func);
    return -1;
  }

  return 0;
}
