/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of luasup
 *
 * luasup is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * luasup is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <suplib/str.h>

#include <lua.h>
#include <lualib.h>

#include <testlib/testlib.h>

#include <luasup/luasup.h>


int cmp_str( const void *s1, const void *s2 )
{
  return strcmp( *((const char**) s1), *((const char**) s2) ); 
}

void
dump_table( const char *name, lua_Object table, int level )
{
  lua_Object index = LUA_NOOBJECT;
  lua_Object value;
  int nelem;
  int i, j;

  char **key;
  nelem = slua_tbl_nelem( table );

  key = (char**) malloc( nelem * sizeof( char* ));

  index = LUA_NOOBJECT;
  for( i = 0 ; i < nelem ; i++ )
  {
    value = slua_tbl_next( table, &index );
    key[i] = str_dup( lua_getstring( index ) );
  }

  qsort( key, nelem, sizeof(char*), cmp_str );

  for ( i = 1 ; i < level ; i++)
    putchar( ' ' );
  printf( "table %s:\n", name );

  for( i = 0; i < nelem ; i++ )
  {

    /* get a handle to the element */
    lua_pushobject( table );
    lua_pushstring( key[i] );
    value = lua_gettable();

    if ( lua_istable( value ) )
    {
      dump_table( key[i], value, ++level );
      continue;
    }

    for ( j = 0 ; j < level ; j++)
      putchar( ' ' );

    if ( lua_isnumber( value ) )
      printf( "%s -> %f\n", key[i], lua_getnumber(value ) );
    else
      printf( "%s -> `%s'\n", key[i], lua_getstring(value ) );

    free( key[i] );
  }

  free( key );
}

void
errfunc( const char *fmt, ... )
{
  va_list ap;
  char buffer[1024];

  va_start( ap, fmt );

  vsprintf( buffer, fmt, ap );

  va_end( ap );

  /* this doesn't return */
  lua_error( buffer );
}

void
dump_table_lua( void )
{
  lua_Object table;
  char *name;

  slua_args( errfunc, "luacfunc",
	        "name", SLua_STRING, &name,
		"table", SLua_TABLE, &table,
	     NULL );

  dump_table( name, table, 1 );
}

int
main ( int argc, char *argv[] )
{
  lua_Object tbl1;
  lua_Object tbl2;
  lua_Object tbl3;
  char *cptr;
  double dbl;

  /* attach the standard lua libraries */
  lua_iolibopen( );
  lua_strlibopen( );
  lua_mathlibopen( );


  START( "luasup" );

  lua_register( "dump_table", dump_table_lua );

  /*---------------------------------------------------------------------------*/
  /* set up some tables */
  lua_dostring( "tbl1 = { a = 1, b = 2, c='nono' }" );
  lua_dostring( "tbl2 = { a = 1, b = 2, c='nono' }" );
  lua_dostring( "tbl3 = { a = 1, b = 2, d='frrr' }" );


  tbl1 = lua_getglobal( "tbl1" );
  tbl2 = lua_getglobal( "tbl2" );
  tbl3 = lua_getglobal( "tbl3" );

  /*---------------------------------------------------------------------------*/
  /* test table comparison routine; needed later */

  TEST( "slua_tbl_cmp: eq", slua_tbl_cmp( tbl1, tbl2 ), 0 );
  TEST( "slua_tbl_cmp: ne", slua_tbl_cmp( tbl1, tbl3 ), 1 );

  /*---------------------------------------------------------------------------*/
  /* test table string extraction */

        TEST( "slua_tbl_el_s: a exists", slua_tbl_el_s( tbl1, "a", &cptr ), 0 );
  TESTSTRING( "slua_tbl_el_s: a val", cptr, "1" );
  free( cptr );

        TEST( "slua_tbl_el_s: c exists", slua_tbl_el_s( tbl1, "c", &cptr ), 0 );
  TESTSTRING( "slua_tbl_el_s: c val", cptr, "nono" );
  free( cptr );

  TEST( "slua_tbl_el_s: nonexist", slua_tbl_el_s( tbl1, "C", &cptr ), 1 );

  /*---------------------------------------------------------------------------*/
  /* test table number extraction */

  TEST( "slua_tbl_el_n: a exists", slua_tbl_el_n( tbl1, "a", &dbl ), 0 );
  TEST( "slua_tbl_el_n: a val", dbl, 1 );


  TEST( "slua_tbl_el_n: c exists", slua_tbl_el_n( tbl1, "c", &dbl ), 0 );
  TEST( "slua_tbl_el_n: c val", dbl, 0 );

  TEST( "slua_tbl_el_n: nonexist", slua_tbl_el_n( tbl1, "C", &dbl  ), 1 );


  /*---------------------------------------------------------------------------*/
  /* test table table extraction */

  lua_dostring( "tbl1 = { tbl2 = { x = 1 } }" );
  lua_dostring( "tbl3 = { x = 1 }" );
  tbl1 = lua_getglobal( "tbl1" );
  tbl3 = lua_getglobal( "tbl3" );

  TEST_RUN( "slua_tbl_el_tbl: run", (tbl2 = slua_tbl_el_tbl( tbl1, "tbl2" )), (tbl2 == (lua_Object) NULL), 0 );
      TEST( "slua_tbl_el_tbl: val", slua_tbl_cmp( tbl2, tbl3 ), 0 );
  

  /*---------------------------------------------------------------------------*/
  /* table element existence checks */
  lua_dostring( "tbl1 = { a = 1 }" );
  tbl1 = lua_getglobal( "tbl1" );
  TEST("slua_tbl_el_exists: no",  slua_tbl_el_exists( tbl1, "C" ), 0 );
  TEST("slua_tbl_el_exists: yes", slua_tbl_el_exists( tbl1, "a" ), 1 );


  /*---------------------------------------------------------------------------*/
  /* delete table elements */
  slua_tbl_el_delete( tbl1, "a" );

  TEST( "slua_tbl_el_delete", slua_tbl_el_exists( tbl1, "a" ), 0  );


  /*---------------------------------------------------------------------------*/
  /* copy tables */
  tbl2 = slua_tbl_copy( tbl1 );
  TEST( "slua_tbl_copy", slua_tbl_cmp( tbl1, tbl2 ), 0 );
  
  lua_dostring( "tbl1 = { a = 1, c='33.2', tbl2 = { snake = 3 } }" );
  lua_dostring( "tbl3 = { snake = 3 }" );
  tbl1 = lua_getglobal( "tbl1" );
  tbl3 = lua_getglobal( "tbl3" );

  {
    int rv;
    double d = 0.0;
    char *s;

    TEST( "slua_tbl_el: nonexist, required",
	  rv = slua_tbl_el( NULL, tbl1, NULL, "d", SLua_NUMBER, &d ), SLua_ENOENT );

    TEST( "slua_tbl_el: nonexist, optional",
	  slua_tbl_el( NULL, tbl1, NULL, "d", SLua_NUMBER | SLua_OPTIONAL, &d ), SLua_ENOEL );

    TEST( "slua_tbl_el: number, run",
	  slua_tbl_el( NULL, tbl1, NULL, "a", SLua_NUMBER, &d ), SLua_ENOERR );
    TEST( "slua_tbl_el: number, val", d, 1 );

    TEST( "slua_tbl_el: number as string, run",
	  slua_tbl_el( NULL, tbl1, NULL, "a", SLua_STRING, &s ), SLua_ENOERR );
    TESTSTRING( "slua_tbl_el: number as string, val", s, "1" );

    TEST( "slua_tbl_el: string, run",
	  slua_tbl_el( NULL, tbl1, NULL, "c", SLua_STRING, &s ), SLua_ENOERR );
    TESTSTRING( "slua_tbl_el: string, val", s, "33.2" );

    TEST( "slua_tbl_el: string as number, run",
	  slua_tbl_el( NULL, tbl1, NULL, "c", SLua_NUMBER, &d ), SLua_ENOERR );
    TEST( "slua_tbl_el: string as number, val", d, 33.2 );

    TEST( "slua_tbl_el: string, copied, run",
	  slua_tbl_el( NULL, tbl1, NULL, "c", SLua_STRING|SLua_COPY, &s ), SLua_ENOERR );
    TESTSTRING( "slua_tbl_el: string, copied, val", s, "33.2" );
    if ( s ) free( s );

    TEST( "slua_tbl_el: table, run",
	  slua_tbl_el( NULL, tbl1, NULL, "tbl2", SLua_TABLE, &tbl2 ), SLua_ENOERR );
    TEST( "slua_tbl_el: table, val", slua_tbl_cmp( tbl2, tbl3 ), 0 );

    TEST( "slua_tbl_el: table as string",
	  slua_tbl_el( NULL, tbl1, NULL, "tbl2", SLua_STRING, &s ), SLua_ENOMATCH );

  }
  
  SUMMARY( );

  return FAILED( ) ? EXIT_FAILURE : EXIT_SUCCESS;

}

