#!@PERL@

# --8<--8<--8<--8<--
#
# Copyright (C) 2010 Smithsonian Astrophysical Observatory
#
# This file is part of make-world
#
# make-world is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# make-world is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#       Free Software Foundation, Inc.
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package MakeWorld::Package;

use strict;
use warnings;

use File::Path;
use File::Spec::Functions qw[ catdir catfile path file_name_is_absolute ];

use Cwd qw/ realpath /;

use MakeWorld::Global;
use MakeWorld::Interpolate;
use MakeWorld::Log;
use MakeWorld::Run;
use MakeWorld::Tarball;

use MakeWorld::Package::AutoConf;
use MakeWorld::Package::PerlModule;
use MakeWorld::Package::Unknown;
use MakeWorld::Package::Custom;


# environmental variables which are paths.  variables which end in PATH
# are automatically recognized
my %PathVar = map { $_ => 1 }
  qw[ TEXINPUTS PERLLIB PERL5LIB ];

sub new
{
    my ( $class, $name, %attr ) = @_;

    my $self = bless
               { name   => $name,
		 pkgspec => defined $attr{pkgspec} ? $attr{pkgspec} : undef,
		 tracked_files => [],
		 hook => { map { $_ => { pre => [], post => [] } } @Stages },
		 env => {},
	       }, $class;

    $self->set_option( $attr{option} ) if exists $attr{option};

    return $self;
}

sub apply_spec {

    my $self = shift;

    if ( defined $self->{pkgspec} )
    {
	my ( $name, $options ) = $self->{pkgspec}->parse_node( $self, 1 );

	die( "internal error: expected <package>, got <$name>\n" )
	  unless $name eq 'package';

	$self->set_option( $options );
    }

    return;
}


sub name    { $_[0]->{name}    }
sub option  { $_[0]->{option}  }
sub tarball { $_[0]->{tarball} }
sub base    { $_[0]->{tarball}->base }
sub hook    { $_[0]->{hook} }

sub set_option {
    my ( $self, $options ) = @_;

    $self->{option} = { %DefaultOptions, %$options, %OverrideOptions  };

};

sub set_env{

    my $self = shift;

    my %env = @_;

    $self->{env}{$_} = $env{$_} for keys %env;
    return;

}

sub add_to_hook {

    my ( $self, $stage, $location, $cmd, $action ) = @_;

    die( "uknown stage: $stage\n" )
      if ! defined $self->{hook}{$stage};

    die( "unkown stage location: $location\n" )
      if $location !~ /^(pre|post)/;

    push @{ $self->{hook}{$stage}{$location} }, [ $cmd, $action ];

    return;
}

sub find_tarball
{
    my $pkg = shift;
    $pkg->{tarball} =
      MakeWorld::Tarball->new( { name => $pkg->name,
		      matchre   => $pkg->{option}{matchre},
		      versionre => $pkg->{option}{versionre},
		    },  @_ );
}

sub process
{
    my $pkg = shift;

    my $name    = $pkg->name;
    my $option  = $pkg->option;

    local %ENV = %ENV;

    # set package specific environmental variables. variables with an empty
    # value should be deleted.  this is done twice; once here, and also
    # after merging with $opt{ENV} and $opt{oENV}.  FIXME: WHY?
    while( my ( $var, $value ) = each %{$pkg->{env}} )
    {
	if ( $value ne '' )
	{
	    $ENV{$var} = interpolate( $value, $pkg->tarball );
	}
	else
	{
	    delete $ENV{$var};
	}
    }

    # merged environmental variables
    while( my ( $var, $value ) = each %{$opt{ENV}} )
    {
	$value = interpolate( $value, $pkg->tarball );

	if ( exists $option->{env} && exists $option->{env}{$var} )
	{
	    $value = join( ($var =~ /PATH$/ || $PathVar{$var}) ? ':' : ' ',
			   $value, $option->{env}{$var} );
	}
	$ENV{$var} = $value;
    }

    # general overrides
    while( my ( $var, $value ) = each %{$opt{oENV}} )
    {
	$ENV{$var} = interpolate( $value, $pkg->tarball );
    }


    # package specific environmental variable deletions
    delete @ENV{
	map { $_ } grep { $pkg->{env}{$_} eq '' } keys %{$pkg->{env}}
    };

    # personalize options by interpolating against the tarball and
    # command line option values
    while( my ( $key, $value ) = each %{$option} )
    {
	next if $key eq 'env';
	$option->{$key} = interpolate( $value, $pkg->tarball );
    }

    # switch to the builddir
    chdir( $opt{builddir} )
      or die( "unable to cd to $opt{builddir}\n" );

    # extract the tarball into the builddir
    my $srcdir = $pkg->tarball->extract( $pkg->logfile( 'extract' ) );

    # switch to the source dir if possible
    if ( -d $srcdir )
    {
	chdir( $srcdir )
	  or die( "unable to cd to $srcdir\n" );

	# explicit specification of package type?  normally
	# this is only done for custom types, but we handle 'em all
	# anyway, 'cause we're that sort of folks.
	if ( $option->{pkg_type} )
	{

	    my %pkg_types =
	      map { lc($_) => __PACKAGE__ . "::$_" }
		qw [ AutoConf PerlModule Custom];

	    my $pkgtype =
	      exists $pkg_types{lc $option->{pkg_type}}
	          ?  $pkg_types{lc $option->{pkg_type}}
		  :  undef;

	    die( "unknown package type: $option->{pkg_type}\n" )
	      unless defined $pkgtype;

	    die( "package does not meet requirements for package type ",
		 "$option->{pkg_type}\n" )
	      unless $pkgtype->identify($pkg);
	}

	# determine what type of package we're dealing with
	# $pkg will get reblessed into that package and everything
	# will just magically work
	else
	{

	    for my $pkgclass ( map { __PACKAGE__ . "::$_" }
			       qw[ AutoConf PerlModule ] )
	    {
		last if $pkgclass->identify( $pkg );
	    }

	    # if $pkg is still in __PACKAGE__, we have no idea what it is.
	    die( "cannot determine package type;",
		 " don't know how to build it\n" )
	      if __PACKAGE__ eq ref $pkg;
	}
    }

    # no dir and not a dryrun? this shouldn't happen
    elsif ( ! $opt{dryrun} )
    {
	die( "extraction to $srcdir failed?\n" );
    }

    # no srcdir && dryrun; don't know which package it is, so use a
    # do-nothing noisy class
    else
    {
	MakeWorld::Package::Unknown->new( $pkg );
    }


    for my $stage ( @Stages )
    {
	next if
	    (exists $option->{"run_$stage"} && ! $option->{"run_$stage"} )
	    || ! ok_to_run($stage)
	    || ! $pkg->can( $stage );

	$pkg->prepost_stage( $option, $stage, 'pre' );

	msg( MakeWorld::Log::Notice, "stage: $stage" );
	no strict 'refs';
	$pkg->$stage;

	$pkg->prepost_stage( $option, $stage, 'post' );
    }

    # switch back to the builddir
    chdir( $opt{builddir} )
      or die( "unable to cd to $opt{builddir}\n" );

    # delete the builddir
    rmtree( $srcdir )
      if $opt{clean} && ! $opt{dryrun} && ok_to_run( $RunComponent[-1] );

    # remove any other files that are tracked
    if ( $opt{clean} && ! $opt{dryrun} )
    {
	unlink( $_ ) foreach @{$pkg->{tracked_files}};
    }
}

#----------------------------------------------------------------#

# run something before or after a stage

sub prepost_stage {

    my ( $pkg, $option, $stage, $location ) = @_;

    my @path = ( $stage, $location );

    return unless ok_to_run( @path );

    for my $action ( @{ $pkg->hook->{$stage}{$location} } )
    {
	my ( $cmd, $opt )  = @$action;
	next unless ok_to_run( @path, $cmd );


	my $stage = join(':', @path, $cmd);

	msg( MakeWorld::Log::Notice, "stage: $stage" );

	die( "error: can't perform $stage\n" )
	  unless $pkg->can($cmd);

      {
	  no strict 'refs';

	  # these options haven't yet been interpolated into
	  $opt = interpolate( $opt, $pkg->tarball );
	  $pkg->$cmd( $opt );
      }

    }

    return;
}

#----------------------------------------------------------------#

sub make { main::run( $opt{make_cmd}, @_ ); }

#----------------------------------------------------------------#

sub run_stage {

    my ( $self, $stage ) = @_;

    my $option = $self->{option}{$stage};
 
    return 0 unless $option->{exe};

    run( find_command($option->{exe}), $option->{arg} )
	or die( "error in $stage stage\n" );

    return 1;
}

#----------------------------------------------------------------#

# it's up to the Package subclasses to perform actions based upon package type
# these are called by those subclass methods to do package specific
# actions.  they return 0 if they haven't handled the action.

sub configure
{
    return $_[0]->run_stage( 'configure' );
}

sub build
{
    return $_[0]->run_stage( 'build' );
}

sub test
{
    return $_[0]->run_stage( 'test' );
}

sub install
{
    return $_[0]->run_stage( 'install' );
}

#----------------------------------------------------------------#

sub is_excluded
{
    my ( $self ) = @_;

    if ( exists $self->{option}{exclude_if_inpath}{exe} )
    {
	return 1 if $self->find_in_path( $self->{option}{exclude_if_inpath}{exe} );
    }

    return 0;
}

#----------------------------------------------------------------#

sub find_in_path
{
    my ( $self, $exe ) = @_;

    my $bindir = realpath( catdir( $opt{exec_prefix}, 'bin' ) );

    foreach my $path ( path() )
    {
	next if realpath($path) eq $bindir;

	return 1 if -x catfile( $path, $exe );
    }

    return;
}

#----------------------------------------------------------------#

sub exec
{
    my ( $self, $opt ) = @_;

    my $arg = $opt->{arg} || [];

    my $exe = $opt->{exe};
    my $file;

    # check if absolute path to file; test directly
    if ( file_name_is_absolute( $exe ) )
    {
	$file = $exe;
    }

    # not absolute; find it
    else
    {
	( $file ) = grep { -f $_ }
	  map { glob(catfile($_,$exe)) }
	    @{$opt{srcdir}}, split(':', $ENV{PATH} );
    }

    unless ( $file && -f $file )
    {
	return if $opt->{optional};
	die "exec: $exe: non-existent executable file\n";
    }

    die( "exec: $exe is not executable\n" )
      unless -x $file;

    run( $file, $arg )
	or die( "error running $exe\n" );
}

#----------------------------------------------------------------#

sub patch
{
    my ( $self, $opt ) = @_;

    my $arg = $opt->{arg} || '-p1';

    # find patch file
    ( my $file ) = grep { -f $_ }
                      map { glob(catfile($_,$opt->{patch})) } @{$opt{srcdir}};

    unless ( $file && -f $file )
    {
	return if $opt->{optional};
	die "patch: non-existent patch file: $opt->{patch}\n";
    }

    run( $opt{patch_cmd}, $arg, '-i', $file )
      or die( "error patching\n" );
}


#----------------------------------------------------------------#

sub find_command
{
    my ( $cmd ) = @_;


    ( my $file ) = grep { -x $_ } map { "$_/$cmd" } @{$opt{srcdir}}, '.';

    die( "unable to find command: $cmd \n" )
      unless $file;

    return $file;
}



#----------------------------------------------------------------#

sub logfile
{
    my ( $self, $ext ) = @_;

    my $logfile;

    if ( $opt{logredir} )
    {
	$logfile = catfile( $opt{logdir}, $_[0]->name . "_$ext.log" );
	push @{$self->{tracked_files}}, $logfile;
    }


    return $logfile;
}

sub logredir
{
    my ( $self, $ext ) = @_;

    return MakeWorld::IORedir->new( file => $self->logfile( $ext ) );
}

1;
