# --8<--8<--8<--8<--
#
# Copyright (C) 2010 Smithsonian Astrophysical Observatory
#
# This file is part of make-world
#
# make-world is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# make-world is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#       Free Software Foundation, Inc.
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--


package MakeWorld::PackageSpec;

use base qw[ Exporter ];

our @EXPORT_OK = qw[ load_pkgfile ];

use XML::Tiny qw[ parsefile ];

use strict;
use warnings;

use MakeWorld::Log;
use MakeWorld::Global;
use MakeWorld::Package;
use MakeWorld::Interpolate;
use MakeWorld::Run;

my %ConditionalVar;
my @VarStack;

# load package descriptions from package files
sub load_pkgfile
{
    my ( $pkgfiles ) = @_;

    my @pkgs;

    for my $pkgfile ( @$pkgfiles )
    {
	msg( MakeWorld::Log::Notice, "Loading package file $pkgfile" );

	my $specs = parsefile( $pkgfile );

	my $element = $specs->[0]{name};
	die( "expected top level <packages> element; got $element\n" )
	  unless $element eq 'packages';

	die( "illegal element type: $element\n" )
	  unless $specs->[0]{type} eq 'e';

	for my $pkgspec ( map { MakeWorld::PackageSpec->new( $_ ) }  @{ $specs->[0]{content} } )
	{
	    die( "expected <package> element, got <", $pkgspec->{name}, ">\n" )
	      unless $pkgspec->{name} eq 'package';

	    die( "unexpected content in <package> element: $pkgspec->{content}\n" )
	      unless $pkgspec->{type} eq 'e';

	    # need the top level package attributes (such as matchre, versionre)
	    # before the actual parsing gets done.
	    my %attr = $pkgspec->parse_attr;

	    my $pkg_name = $attr{name}
	      or die( qq{<package> must have a "name" attribute\n} );

	    push @pkgs, MakeWorld::Package->new( $pkg_name, pkgspec => $pkgspec, option => \%attr );

	}
    }

    return @pkgs;
}

#------------------------------------------------------------------------


sub new {

    my ( $class, $node ) = @_;

    return bless $node, $class;
}

#------------------------------------------------------------------------


sub parse_attr {

    my ( $self ) = @_;

    my %attr;
    while ( my ( $key, $value ) = each %{ $self->{attrib} } )
    {
	$value = 1 if lc($value) =~ /\A(?:1|true|on|yes)\z/;
	$value = 0 if lc($value) =~ /\A(?:0|false|off|no)\z/;
	$attr{$key} = $key eq 'arg' ? [ $value ] : $value;
    }

    return %attr;
}

#------------------------------------------------------------------------

=pod

=item B<parse_node>

  ($name, $content) = $node->parse_node( $node, $pkg, $level );

Turn an XML::Tiny node into a nested hash.

=cut

sub parse_node {

    my ( $node, $pkg, $level ) = @_;

    size_stack( $level );

    my $name = $node->{name};

    my %node = parse_attr( $node );

    # handle <arg> elements differently.  they can occur more than
    # once and thus get bundled into arrays.

    if ( $name eq 'arg' )
    {

	die( "<arg> element does not take attributes\n" )
	  if keys %node;

	die( "<arg> must have text\n" )
	  if ! @{$node->{content}}
	    || $node->{content}[0]{type} ne 't';

	my $res = $node->{content}[0]{content};
	$res =~ s/^\s+//;
	$res =~ s/\s+$//;
	return ( $name, $res );
    }


    # boolean check on specified variable or last %eval
    if ( $name =~ /^if(true|false)$/ )
    {
	my $bval = $1;

	my $varname = defined $node->{attrib}{var}
                    ? $node->{attrib}{var}
		    : '%eval';

	die( "<$name>: no variable with the name $varname\n" )
	  unless exists $VarStack[-1]{$varname};

	my $ok = $bval eq 'true'
	         ?   $VarStack[-1]{$varname}
		 : ! $VarStack[-1]{$varname};

	return undef if ! $ok;

	undef $name;
    }

    # generic if using Perl eval
    elsif ( $name =~ /^if$/ )
    {
	return undef unless _if( $name, $pkg, $node );

	undef $name;
    }

    # just eval, storing result
    elsif ( $name eq 'eval' )
    {
	_eval( $name, $pkg, $node );
	return undef;
    }

    # log a string at debug level
    elsif ( $name eq 'debug' )
    {
	_debug( $name, $pkg, $node );
	return undef;
    }

    # catch all rule; if it's <el>text</el>, treat it as an
    # attribute in the parent node, e.g. $parent->{attrib}{el}=text
    elsif ( @{$node->{content}} == 1 && $node->{content}[0]{type} eq 't' )
    {
	die( "unexpected text in element: $node->{content}\n" )
	  if $node->{type} eq 't';

	return ( $name, $node->{content}[0]{content} );
    }

    for my $sn ( @{ $node->{content} } )
    {
	my ($name, $res ) = parse_node( $sn, $pkg, $level + 1 );

	# nodes may be removed from the tree by parse_node; it returns
	# $res = undef in that instance.
	next unless defined $res;

	my $import;
	# node contents may be promoted up a level (e.g. a conditional
	# node's contents); parse_node returns $name = undef in this
	# case.

	$import = defined $name ? { $name => $res }
	                        : $res ;
	while ( my ( $key, $val ) = each %$import )
	{
	    if ( exists $node{$key} && 'ARRAY' eq ref $node{$key} )
	    {
		push @{$node{$key}}, 'ARRAY' eq ref $val ? @$val : $val;
	    }

	    else
	    {
		$node{$key} = $key eq 'arg' && 'ARRAY' ne ref $val 
		                  ? [ $val ] : $val;
	    }
	}
    }

    # short circuit ; if $name is undefined, the node is being promoted up a level
    return ( $name, \%node )
      if ! defined $name;

    # patch and exec commands get inserted into the package hooks list,
    # and get removed from the node tree
    if ( $name =~ /^(?:patch|exec)$/ )
    {
	die( "<$name> must have a hook attribute\n" )
	  if !defined $node{hook};

	# hook should look like "stage:location"
	2 == ( ($node{stage}, $node{location}) = $node{hook} =~ /(.*):(.*)/ )
	  or die( "invalid hook attribute value: $node{hook}\n" );

	grep { $node{stage} eq $_ } @Stages
	  or die( "unknown stage in hook attribute: $node{hook}\n" );

	$node{location} =~ /^(pre|post)$/
	  or die( "unknown location in hook attribute: $node{hook}\n" );

	die( "<$name> must have a exe attribute\n" )
	  if ( $name eq 'exec' && ! defined $node{exe} );

	die( "<$name> must have a patch attribute\n" )
	  if ( $name eq 'patch' && ! defined $node{patch} );

	$pkg->add_to_hook( $node{stage}, $node{location}, $name, \%node );
	return undef;
    }

    if ( $name eq 'env' )
    {
	die( "<env> may not have sub-elements\n" )
	  if grep { ref $_ } values %node;

	msg( MakeWorld::Log::Debug, "env: $_ = $node{$_}" )
	  for keys %node;

	$pkg->set_env( %node );
    }

    if ( $name eq 'run' )
    {
	_run( $name, $pkg, \%node );
	return ( undef, undef );
    }


    return $name, \%node;
}


#------------------------------------------------------------------------

sub size_stack {

    my ( $size ) = @_;

    my $osize = @VarStack;

    return if $size == $osize;

    $#VarStack = $size - 1;

    if ( $size > $osize )
    {
	$VarStack[-1] = { %{ $VarStack[-2] || {} } };
    }
    else
    {
	$VarStack[-1] ||= {};
    }

    return;
}


sub set_stack {

    my %vars = @_;

    my ( $key, $val );
    $VarStack[-1]{$key} = $val while ($key, $val ) = each %vars;

    return;
}

sub dump_stack {

    my ( @vars ) = @_;

    @vars = keys %{$VarStack[-1]} unless @vars;

    msg( MakeWorld::Log::Debug, "$_: $VarStack[-1]{$_}\n" )
      foreach @vars;

    return;
}

#------------------------------------------------------------------------

# execute a command and stash the results

sub _run {

    my ( $name, $pkg, $node ) = @_;

    die( "<$name> must have an exe attribute\n" )
      unless defined $node->{exe};

    my @keys = qw[ exit stdout stderr ];

    do { $keys[$_] = $node->{$keys[$_]} || $keys[$_] }
      for 0..@keys-1;

    $node = interpolate( $node, $pkg->tarball );

    my %res;
    @res{@keys} = capture( $node->{exe}, $node->{arg} );

    set_stack( %res );

    dump_stack( @keys );
}

sub _safe_eval {

    my ( $label, $expr ) = @_;

    msg( MakeWorld::Log::Debug, "$label: $expr\n" );

    require Safe;
    my $cpt = Safe->new;

    for my $vars ( $VarStack[-1], \%MakeWorld::Interpolate::InterpolateVars )
    {
	${$cpt->varglob($_)} = $vars->{$_}
	  foreach keys %$vars;
    }

    %{$cpt->varglob('ENV')} = %ENV;

    my $res = $cpt->reval( $expr ) || 0;
    die( "<eval>: error in expression: $expr\n$@\n" )
      if $@;

    return $res;
}

sub _if {

    my ( $name, $pkg, $node ) = @_;

    die( "<$name> requires an 'expr' attribute\n" )
      if ! defined $node->{attrib}{expr};

    return _safe_eval( $name, $node->{attrib}{expr} );
}

sub _eval {

    my ( $name, $pkg, $node ) = @_;

    my $expr = @{$node->{content}} == 1 && $node->{content}[0]{type} eq 't' 
             ? $node->{content}[0]{content}
             : $node->{attrib}{expr};

    die( "<$name> must specify an expression, either as an attribute or as its content\n" )
      unless defined $expr;

    my $varname = defined $node->{attrib}{var}
                ?  $node->{attrib}{var}
	        : '%eval';

    my $res = _safe_eval( $name, $expr );

    set_stack( $varname => $res );
    dump_stack( $varname );

    return;
}

sub _debug {

    my ( $name, $pkg, $node ) = @_;

    my $str = @{$node->{content}} == 1 && $node->{content}[0]{type} eq 't' 
             ? $node->{content}[0]{content}
             : $node->{attrib}{str};

    die( "<$name> must specify a string\n" )
      unless defined $str;

    msg( MakeWorld::Log::Debug, $str );

    return;
}
1;


__END__

=head1 NAME

MakeWorld::PackageSpec -- create MakeWorld::Package objects from package specification files

=head1 DESCRIPTION


=head2 Syntax


  <package name="name"/> 

  <package name="name"> 

    <command arg="argument" />

    <command>
      <arg>argument</arg>
      <arg>argument</arg>
    </command>

  </package>
