package MakeWorld::Run;

use File::Basename;

use MakeWorld::Global;
use MakeWorld::Log;

use Capture::Tiny qw[];

use base qw[ Exporter ];

our @EXPORT = qw[
		    ok_to_run
		    run
		    capture
	       ];


#----------------------------------------------------------------#

# figure out if the stage operation has been allowed

# @path is a stage heirarchy (such as qw( configure pre patch ) )

# Logic; order is Allow, Deny
# if --run is specified only those stages specified are allowed, else
# all are allowed.
# if a component in the stage heirarchy is {*} all things at that level
# are allowed.  if a component is {**}, all things at that level and below
# are allowed.

# Stages specified by --dont are not run.
# if a component in the stage heirarchy is {*} all things at that level
# (and below) are not run.

sub ok_to_run {

    my ( @path ) = @_;

    my $allowed = 1;

    # if $opt{run} exists, then --run was specified.
    if ( exists $opt{run} )
    {
	my $run = $opt{run};

	my @tpath = @path;
	my $end = pop @tpath;

	$allowed =    exists $run->{join($;, @tpath, $end)}
	           || exists $run->{join($;, @tpath, '*')}
	           || exists $run->{join($;, @tpath, '**')};

	# if not explicitly allowed at the bottom-most level, look for
	# '**' in parent level
	while (! $allowed && @tpath)
	{
	    pop @tpath;

	    $allowed = exists $run->{join($;, @tpath, '**' )};
	}
    }


    # if $opt{dont} exists, then --dont was specified.
    if ( $allowed && exists $opt{dont} )
    {

	my $dont = $opt{dont};

	my @tpath = @path;
	my $end = pop @tpath;

	$allowed = !(   exists $dont->{join($;, @tpath, $end)}
	             || exists $dont->{join($;, @tpath, '*')}
	             || exists $dont->{join($;, @tpath, '**')}
		    );

	# if not explicitly disallowed at the bottom-most level, look in parent levels
	while ($allowed && @tpath)
	{
	    pop @tpath;

	    $allowed = !(    exists $dont->{join($;, @tpath, '**' )}
			  || exists $dont->{join($;, @tpath, '*' )}
			);
	}
    }

    return $allowed;
}

#----------------------------------------------------------------#

# switches for programs which can handle "dry run" flags
our  %dryrun = (
                make       => '-n',
                gmake      => '-n',
                configure  => '-n',
               );

# wrapper for system() command which knows about dryruns
sub run
{
    my ( $exe, @args ) = @_;

    # @args may have an arrayref hidden in there, so take care of it
    @args = map { 'ARRAY' eq ref $_ ? @$_ : $_ } grep { defined $_ } @args;

    my $lexe = basename( $exe );

    unshift @args, $dryrun{$lexe}
      if $opt{dryrun} > 1 && exists $dryrun{$lexe};

    my $cmd = join(' ', $exe, @args );

    msg( $opt{dryrun} ? MakeWorld::Log::Info : MakeWorld::Log::Debug, 'executing: ' . $cmd );

    # this level of dryrun means just print what we'd do
    return 1 if 1 == $opt{dryrun};

    # if a command can do a dryrun, have it do so, otherwise
    # return
    return 1 if $opt{dryrun} && ! exists $dryrun{$lexe};

    return system( $exe, @args ) == 0;
}


# wrapper for system() command which knows about dryruns
# FIXME: this is pretty much a duplicate of run()
sub capture
{
    my ( $exe, @args ) = @_;

    # @args may have an arrayref hidden in there, so take care of it
    @args = map { 'ARRAY' eq ref $_ ? @$_ : $_ } grep { defined $_ } @args;

    my $lexe = basename( $exe );

    unshift @args, $dryrun{$lexe}
      if $opt{dryrun} > 1 && exists $dryrun{$lexe};

    my $cmd = join(' ', $exe, @args );

    msg( $opt{dryrun} ? MakeWorld::Log::Info : MakeWorld::Log::Debug, 'executing: ' . $cmd );

    # this level of dryrun means just print what we'd do
    return 1 if 1 == $opt{dryrun};

    # if a command can do a dryrun, have it do so, otherwise
    # return
    return 1 if $opt{dryrun} && ! exists $dryrun{$lexe};

    my $exit;
    my ( $stdout, $stderr ) = Capture::Tiny::capture { $exit = system( $exe, @args ) };

    return( $exit, $stdout, $stderr );
}


1;
