#!@PERL@

# --8<--8<--8<--8<--
#
# Copyright (C) 2010 Smithsonian Astrophysical Observatory
#
# This file is part of make-world
#
# make-world is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# make-world is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#       Free Software Foundation, Inc.
#       51 Franklin Street, Fifth Floor
#       Boston, MA  02110-1301, USA
#
# -->8-->8-->8-->8--

package MakeWorld::Tarball;

use strict;
use warnings;

use File::Basename;
use File::Temp qw( tempdir );
use File::Copy;
use File::Path;
use Cwd qw( realpath getcwd );
use File::Spec::Functions qw/ catdir /;

use MakeWorld::Global;
use MakeWorld::Log;
use MakeWorld::Run;

sub new
{
    my $class = shift;

    my $opts = shift;

    my $self = bless { %$opts }, $class;

    return $self->_find( @_ );
}

sub base { $_[0]->{base} }

sub _find
{
    my $self = shift;

    my ( $dirs ) = @_;

    my $name = $self->{name};

    my @files;
    my @candidates;
    my $maxcomps = 0;

    my $cwd = getcwd;

    msg( MakeWorld::Log::Debug, "searching for tarball" );

    # get a list of candidate files, either from a supplied RE
    # or by using our default glob
    if ( defined $self->{matchre} )
    {
	for my $dir ( @$dirs )
	{
	    chdir( $dir )
	      or die( "unable to chdir to $dir\n" );
	    push @candidates, 
	      map { "$dir/$_" }
	      grep { /$self->{matchre}/ } glob("*");
	    chdir( $cwd )
	      or die( "unable to chdir to $cwd\n" );
	}
    }
    else
    {
	@candidates = map { glob("$_/$name-*") } @$dirs;
    }

  FILE:
    for my $file ( @candidates )
    {

	my ( $base, $path, $suffix ) = fileparse( $file,
						  '.tar',
						  '.tar.gz',
						  '.tgz',
						  '.tar.bz2'
						);

	# only care about things we understand
	next FILE unless defined $suffix;

	# what's in this thing?
	# if it's a symbolic link, file looks at that, so get the realpath
	my $rfile = realpath($file);

	my ( $ztype ) = `file $rfile` =~ /[^:]+:.*\b(gzip|bzip2|zip)\b/i ;
	next FILE unless defined $ztype;

	# grab version info from end. assume everything after last - is
	# it
	my $version_re = $self->{versionre} || '-([^-]*)\z';
	my ( $version ) = $base =~ /$version_re/;
	if ( ! defined $version )
	{
	    warn( "unable to determine version for $file\n" );
	    next FILE;
	}

	# now split version on periods and underscores.
	# don't be too fancy with the error checking, and
	# keep the delimiters
	my @subvers = split( /([._])/, $version );

	$maxcomps = @subvers if $maxcomps < @subvers;

	# and make nice zero prefixed numbers out of those
	# bits that are numeric
	for my $subvers ( @subvers )
	{
	    if ( $subvers =~ /\A\d+\z/ )
	    {
		$subvers = sprintf( "%03d", $subvers )
	    }

	    # this magic is so that . is later in collating sequence
	    # than is _.
	    elsif ( $subvers eq '.' )
	    {
		$subvers = 1;
	    }
	    elsif ( $subvers eq '_' )
	    {
		$subvers = 0;
	    }

	    # if it ain't a delimiter we can parse, leave
	    # it like it is.
	    else
	    {
	    }
	}

	push @files, { file    => $file,
		       rfile   => $rfile,
		       ztype   => $ztype,
		       base    => $base,
		       path    => $path,
		       suffix  => $suffix,
		       version => $version,
		       subversion => \@subvers,
		     }
    }

    die( "unable to find tarball for $name\n" )
      unless @files;

    my %is_dup_file;

    # recreate versions, filling in empty components if necessary
    for my $file ( @files )
    {
	push @{$file->{subversion}}, ('1', '000' )
	  while @{$file->{subversion}} < $maxcomps;

	$file->{vstring} = join('', @{$file->{subversion}} );

	delete $file->{subversion};

	$is_dup_file{$file->{vstring}}++;
    }

    # now, sort by version and pluck the greatest one
    my $latest = ( sort { $a->{vstring} cmp $b->{vstring} } @files )[-1];
    die( "unable to find tarball for $name\n" )
      unless defined $latest;


    if ( $is_dup_file{$latest->{vstring}} > 1 )
    {
	my @dups = grep { $_->{vstring} eq $latest->{vstring} } @files;
	die( "duplicate versions of $name: \n\t", 
	     join(qq{\n\t}, map { $_->{file} } @dups ), "\n" );
    }

    %$self = %$latest;

    msg( MakeWorld::Log::Debug, "found tarball: " . $self->{file} );
    return $self;
}


#----------------------------------------------------------------#

sub extract
{
    my ( $tarball, $logfile )  = @_;

    my $base = $tarball->{base};
    my $file = $tarball->{file};

    my $cwd = getcwd;


    # temporary directory where the tarball is extracted
    my $tmpdir;

    # this is the name of the directory that'll get renamed as $base.
    # either $tmpdir or $tmpdir/subdir
    my $srcdir;

  EXTRACT:
    if ( ok_to_run( 'extract' ) )
    {
	# extract tarball into temporary directory.  this allows
	# processing of tarballs which don't unpack into
	# a properly named and versioned directory

	unless ( $opt{dryrun} )
	{
	    $tmpdir = tempdir( DIR => $cwd );
	    chdir( $tmpdir )
	      or die( "unable to chdir to $tmpdir\n" );
	}

	my $topts = $opt{verbose} > 1 ? "xvf" : "xf";
	my $ztype = lc $tarball->{ztype};
	my $rfile = $tarball->{rfile};

	# horrid, but make it work with unschooled tar commands
	my $cmd = 
	  'gzip'  eq $ztype ? "gunzip  -c $rfile | tar $topts -"
	: 'bzip2' eq $ztype ? "bunzip2 -c $rfile | tar $topts -"
	: 'zip'   eq $ztype ? "unzip $rfile"
        :                     "tar xvf $rfile";

	msg( MakeWorld::Log::Notice, "extracting $file" );

        {
	    my $rd = MakeWorld::IORedir->new( file => $logfile );
	    run( $cmd )
		or die( "error extracting package into $tmpdir\n" );
	}

	if ( ! $opt{dryrun} )
	{
	    # now, what's in the current directory?
	    my @dircontents = glob('*');


	    # only one thing and its a directory. cool.
	    if ( 1 == @dircontents && -d $dircontents[0] )
	    {
		$srcdir = catdir( $tmpdir, $dircontents[0] );
	    }

	    # more than one entry. rename $tmpdir to $base
	    else
	    {
		$srcdir = $tmpdir;
	    }
	}

	chdir( $cwd );
	# remove build directory if it exists
	if ( -d $base )
	{
	    msg( MakeWorld::Log::Info, "removing existing build in $cwd/$base" );
	    rmtree( $base )
	      unless $opt{dryrun};
	}

	unless ( $opt{dryrun} )
	{
	    move( $srcdir, $base ) 
	      or die( "error renaming $srcdir to $base\n" );

	    # remove tmpdir if it hasn't been renamed to $base
	    rmtree( $tmpdir ) if -d $tmpdir;
	}
    }

    return catdir( $cwd, $base );
}

1;
