// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of mergerays
//
// mergerays is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// mergerays is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <suplibxx/str.h>

#include "BpipeInput.h"
#include "DataField.h"

BpipeInput::~BpipeInput( ) {

  if ( bpipe ) {
    for ( size_t i = 0; i < num; i++ )
      bpipe_delete( bpipe[ i ] );
    delete [] bpipe;
    bpipe = NULL;
  }

  if ( data ) {
    /*
    for ( int i = 0; i < num; i++ )
      free( data[ i ] );
      */
    delete [] data;
    data = NULL;
  }

  if ( n_pkts ) {
    delete [] n_pkts;
    n_pkts = NULL;
  }

}

BpipeInput::BpipeInput( string& fnames, string& xfer, int& status,
			ostream& os ) : num( 0 ), bpipe( NULL ),
  data( NULL ), n_pkts_valid( 0 ), n_pkts( NULL ) {

    status = EXIT_FAILURE;

    vector< string > filenames;
    suplib::tok( filenames, fnames, "," );
    num = filenames.size( );

    if ( EXIT_FAILURE == allocate_bpipe( filenames, os ) )
      return;
    // allocate memory for the members bpipe and data.

    if ( EXIT_FAILURE == get_num_dpkts( os ) )
      return;
    // Make sure that the headers: BPHN_n_dpkts_valid and BPHN_n_dpkts
    // are present just in case the user wants to merge the files with
    // method=sample.

    if ( EXIT_FAILURE == header_to_datafield( xfer, os ) )
      return;
    // Transfer headers to the data field which are requested 
    // with the xfer parameter.

    status = EXIT_SUCCESS;

}

//
// allocate meory for the bpipe and data members.
//
int BpipeInput::allocate_bpipe( vector< string >& filenames, ostream& os ) { 

  bpipe = new BPipe*[ num ];
  if ( NULL == bpipe ) {
    os << "BpipeInput::allocate_bpipe( ostream& ) : Unable to allocate input";
    os << endl;
    return EXIT_FAILURE;
  }

  data = new void*[ num ];
  if ( NULL == data ) {
    os << "BpipeInput::allocate_bpipe( ostream& ) : Unable to allocate data";
    os << endl;
    return EXIT_FAILURE;
  }

  for ( size_t i = 0; i < num; i++ ) {
    
    // create a BPipe object
    bpipe[ i ] = bpipe_new( );
    if ( NULL == bpipe[ i ] ) {
      os << "BpipeInput::allocate_bpipe( ostream& ) : ";
      os << bpipe_strerror( bpipe_errno ) << endl;
      return EXIT_FAILURE;
    }

    // attach an input file to the BPipe
    if ( bpipe_input( bpipe[ i ],
		      const_cast< char* >( filenames[ i ].c_str( ) ) ) ) {
      os << "BpipeInput::allocate_bpipe( ostream& ) : ";
      os <<  bpipe_strerror( bpipe_errno ) << ' ' << filenames[ i ];
      os << endl;
      return EXIT_FAILURE;
    }

  }

  return EXIT_SUCCESS;

}

int BpipeInput::get_num_dpkts( ostream& os ) {

  n_pkts = new size_t[ num ];

  for ( size_t ii = 0; ii < num; ii++ ) {

    // make sure that the header field BPHN_n_dpkts_valid is present.
    int* ptr = (int*) bpipe_hdrf_data( bpipe[ ii ], BPHN_n_dpkts_valid, 0 );
    if ( ptr ) {

      if ( 0 != *ptr )
	++n_pkts_valid;

      size_t* npkts =
	(size_t*) bpipe_hdrf_data( bpipe[ ii ], BPHN_n_dpkts, 0 );
      if ( npkts )
	n_pkts[ ii ] = *npkts;
      else
	n_pkts[ ii ] = 0;

    }

  }

  return EXIT_SUCCESS;

}

int BpipeInput::header_to_datafield( string& xfer, ostream& os ) {

  vector< string > xfer_token;
  suplib::tok( xfer_token, xfer, "," );

  //
  // Loop over the input stream.
  //
  for ( size_t ii = 0; ii < num; ii++ ) {

    void* last = NULL;	             // initialize it for first time through
    char* name = NULL;
    size_t index;

    // loop until there are no more header fields available
    while ( bpipe_hdrf_next( bpipe[ ii ], &name, &index, &last ) ) {

      //
      // extract the data type and a pointer to the data.  these should
      // be valid since we've been handed name and index by bpipe_hdrf_next
      //
      BPDataType type = bpipe_hdrf_type( bpipe[ ii ], name, index );

      if ( BPNOERROR != bpipe_errno ) {
	os << "BpipeInput::header_to_datafield( " << xfer;
	os << ", ostream& ) : " << bpipe_strerror( bpipe_errno ) << endl;
	return EXIT_FAILURE;
      }

      //
      // For each input stream and header.  Loop over the tokens of the
      // xfer parameter to see if the current header needs to be transfer
      // to the data field or not.
      //
      for ( string::size_type i = 0; i < xfer_token.size( ); i++ ) {

	//
	// The header field is to be transferred to the data
	// field if the field name matches with one of the
	// xfer tokens.
	//
	if ( 0 == xfer_token[ i ].compare( name ) ) {

	  BPMatrix* matrix =
	    bpipe_hdrf_matrix ( bpipe[ ii ], name, index );

	  if ( NULL == matrix ) {
	    os << "BpipeInput::header_to_datafield( " << xfer;
	    os << ", ostream& ) : " << bpipe_strerror( bpipe_errno ) << endl;
	    return EXIT_FAILURE;
	  }

	  if ( bpipe_dpktf_add( bpipe[ ii ], name, type, matrix ) ) {
	    os << "BpipeInput::header_to_datafield( " << xfer;
	    os << ", ostream& ) : " << bpipe_strerror( bpipe_errno );
	    os << endl;
	    return EXIT_FAILURE;
	  }

	}
	  
      }

    }

  }

  return EXIT_SUCCESS;

}

int BpipeInput::resize_core( SLinkList data_fields, string& xfer, 
			     ostream& os ) {

  vector< string > xfer_tokens;
  suplib::tok( xfer_tokens, xfer, "," );

  // Loop over the input streams.
  for ( size_t ii = 0; ii < num; ii++ ) {

    // resize the core then map the core.
    for ( SLLNode node = sll_head_node( data_fields );
	  NULL != node; node = sll_next_node( data_fields, node ) ) {

      DataField* datafield = (DataField*) sll_node_get_data( node );
      BPMatrix*   matrix = datafield->get_matrix( );
      char*       name   = datafield->get_name( );

      DpktField*  dpktf = bpipe_dpktf( bpipe[ ii ], name );

      for ( string::size_type iii = 0; iii < xfer_tokens.size( ); iii++ ) {

	if ( xfer_tokens[ iii ].compare( name ) ) {
	  BPMatrix*   matrix_dup = bpipe_matrix_dup( matrix );
	  if ( bpipe_dpktf_resize_core( dpktf, matrix_dup,
					NULL, NULL, NULL ) ) {
	    os << "BpipeInput::resize_core( SLinkList, ostream& ) : ";
	    os << bpipe_strerror( bpipe_errno ) << "\n";
	    os << datafield << endl;
	    return EXIT_FAILURE;
	  }
	}

      }

    }

    data[ ii ] = bpipe_map_alloc( bpipe[ ii ], 1, NULL );

  }

  for ( size_t i = 0; i < num; i++ )

    for ( SLLNode node = sll_head_node( data_fields );
	  NULL != node; node = sll_next_node( data_fields, node ) ) {

      DataField* datafield = (DataField*) sll_node_get_data( node );
      char*       name   = datafield->get_name( );
      BPDataType type = datafield->get_type( );

      void* dst = bpipe_dpktf_datap( bpipe[ i ], data[ i ], name, type );

      for ( string::size_type j = 0; j < xfer_tokens.size( ); j++ ) {

	if ( 0 == xfer_tokens[ j ].compare( name ) ) {

	  BPMatrix*   matrix = datafield->get_matrix( );

	  // The index to the call to bpipe_hdrf_data must always be 0
	  // since there could only be one field with the given name.
	  bpipe_data_copy( dst, bpipe_hdrf_data( bpipe[ i ], name, 0 ),
			   matrix, type );

	}

      }

      datafield->set_input_data( i, dst );
	  
    }

  return EXIT_SUCCESS;

}
