// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of mergerays
//
// mergerays is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// mergerays is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <stdlib.h>
#include <unistd.h>

#include <suplibxx/str.h>

#include "BpipeOutput.h"
#include "DataField.h"

int addDataField( void* node, void* udata ) {

  DataField* data_field = (DataField*) node;
  BPipe*      bpipe = (BPipe*) udata;

  char* name = data_field->get_name( );

  BPDataType type = data_field->get_type( );
  
  BPMatrix* matrix = data_field->get_matrix( );
  
  if ( bpipe_dpktf_add( bpipe, name, type, matrix ) )
    return !0;

  return 0;

}

int printDataField( void* d ) {
  DataField* ptr = (DataField*) d;
  cout << ptr << endl;
  return 0;
}

int cmpDataField( const void* k, const void* b ) {

  DataField* key = (DataField*) k;
  DataField* node = (DataField*) b;

  return strcmp( key->get_name( ), node->get_name( ) );

}

BpipeOutput::~BpipeOutput( ) {

  if ( output_data_fields ) {
    size_t j = 0;
    DataField* datafield = NULL;
    while( ( datafield = output_data_fields[ j++ ] ) )
      delete datafield;
    delete [] output_data_fields;
    output_data_fields = NULL;
  }

  if ( bpipe ) {
    bpipe_delete( bpipe ); 
    bpipe = NULL;
  }

}

BpipeOutput::BpipeOutput( BpipeInput* in, ParFile& parfile, string& xfer,
			  int& status, ostream& os ) : bpipe( NULL ),
  bpo( NULL ), input( in ), data( NULL ), output_data_fields( NULL ) {

  status = EXIT_FAILURE;

  try {

    parfile.pget( "output", filename );

    bpipe = bpipe_new( );
    if ( NULL == bpipe ) {
      os << "BpipeOutput::BpipeOutput( BpipeInput*, ParFile&, int&, ";
      os << "ostream& ) : Unable to allocate memory for bpipe" << endl;
      return;
    }

    SLinkList data_fields = sll_new( cmpDataField );
    if ( NULL == data_fields ) {
      os << "BpipeOutput::BpipeOutput( BpipeInput*, ParFile&, int&, ";
      os << "ostream& ) : Unable to allocate memory for data_fields" << endl;
      return;
    }

    //
    // loop through all the input bpipes to write the output header 
    // fields which are not requested to be transferred to the data field.
    //
    for ( size_t i = 0; i < input->get_num( ); i++ ) {
      BPipe* bpipe_input = input->get_bpipe( i, os );
      if ( EXIT_FAILURE == 
	   add_header_fields( bpipe_input, xfer, data_fields, os ) )
	return;
    }

    // loop thru all the input bpipes to extract the data fields.
    for ( size_t j = 0; j < input->get_num( ); j++ ) {
      BPipe* bpipe_input = input->get_bpipe( j, os );
      if ( EXIT_FAILURE == 
	   select_data_fields( bpipe_input, xfer, data_fields, os ) )
	return;

      /*
	os << "j = " << j << endl;
	if ( 0 != sll_traverse( data_fields, printDataField ) ) {
	os << "sll_traverse returned with an error" << endl;
	return;
	}
	*/

    }


    if ( EXIT_FAILURE == init_bpipe_output( data_fields, os ) )
      return;

    if ( EXIT_FAILURE == get_data_field_ptr( data_fields, os ) )
      return;

    if ( EXIT_FAILURE == input->resize_core( data_fields, xfer, os ) )
      return;

    if ( EXIT_FAILURE == set_output_data_fields( data_fields, os ) )
      return;

    sll_delete( data_fields, SLL_NULL_DELETE );

    status = EXIT_SUCCESS;

  } catch( ParFileException& pfe ) {

    os << pfe << endl;
    status = EXIT_FAILURE;

  }

}

int BpipeOutput::add_header_fields( BPipe* bpipe_input, string& xfer, 
				    SLinkList data_fields,
				    ostream& os ) {

  // last is used by bpipe_hdrf_next to keep track of where it is
  void* last = NULL;	             // initialize it for first time through
  char* name = NULL;
  size_t index;

  vector< string > xfer_token;
  suplib::tok( xfer_token, xfer, "," );

  // For a given bpipe, loop until there are no more header fields available
  while ( bpipe_hdrf_next( bpipe_input, &name, &index, &last ) ) {

    //
    // extract the data type and a pointer to the data.  these should
    // be valid since we've been handed name and index by bpipe_hdrf_next
    //
    BPDataType type = bpipe_hdrf_type( bpipe_input, name, index );
    void*  thedata = bpipe_hdrf_data( bpipe_input, name, index );

    BPMatrix* matrix = bpipe_hdrf_matrix ( bpipe_input, name, index );

    if ( NULL == matrix ) {
      os << "BpipeOutput::add_header_fields( BPipe*, " << xfer;
      os << ", ostream& ) : " << bpipe_strerror( bpipe_errno ) << endl;
      return EXIT_FAILURE;
    }

    if ( BPNOERROR != bpipe_errno ) {
      os << "BpipeOutput::add_header_fields( BPipe*, " << xfer;
      os << ", ostream& ) : " << bpipe_strerror( bpipe_errno ) << endl;
      return EXIT_FAILURE;
    }

    // loop thru all the parameters which have
    // to be transferred to the data fields.
    for ( string::size_type i = 0, once = 1; i < xfer_token.size( ); i++ ) {

      //
      // Add to the output header fields if it is not
      // requested to be transferred to the data field.
      //
      if ( 0 != xfer_token[ i ].compare( name ) ) {

	if ( 1 == once ) {

	  //
	  // The header field is not in the xfer parameter
	  // so write the input stream header field to the
	  // output header.
	  //
	  if ( bpipe_hdrf_add( bpipe, name, type, matrix,
			       thedata, index ) ) {
	    os << "BpipeOutput::add_header_fields( BPipe*, ";
	    os << xfer << ", ostream& ) : ";
	    os << bpipe_strerror( bpipe_errno ) << endl;
	    return EXIT_FAILURE;
	  }

	  // To make sure that the header field
	  // is only written out once per stream
	  once++;

	}

      } else {

	//
	// The header field is to be transferred to the data field.
	//
	if ( EXIT_FAILURE == 
	     select_data_fields( name, type, matrix, data_fields, os ) )
	  return EXIT_FAILURE;

      }

    }

  }

  return EXIT_SUCCESS;

}

size_t BpipeOutput::copy_data( size_t i, ostream& os ) {

  size_t n = input->read_dpkts( i, os );
  if ( 0 == n )
    return 0;

  size_t j = 0;
  DataField* datafield = NULL;
  while( ( datafield = output_data_fields[ j++ ] ) )
    datafield->copy_data( i );

  return n;

}


int BpipeOutput::get_data_field_ptr( SLinkList data_fields, ostream& os ) {

  for ( SLLNode node = sll_head_node( data_fields );
	NULL != node; node = sll_next_node( data_fields, node ) ) {

    DataField* datafield = (DataField*) sll_node_get_data( node );

    char* name = datafield->get_name( );
    BPDataType type = datafield->get_type( );

    void* ptr = bpipe_dpktf_datap( bpipe, data, name, type );
    if ( NULL == ptr ) {
      os << "BpipeOutput::get_data_field_ptr( ostream& ) : NULL == ptr";
      os << endl;
      return EXIT_FAILURE;
    }

    datafield->set_output_data( ptr );

  }

  return EXIT_SUCCESS;

}

int BpipeOutput::init_bpipe_output( SLinkList data_fields, ostream& os ) {

  //
  // delete any node which is not the present in all the input streams.
  //
  for ( SLLNode node = sll_head_node( data_fields );
	NULL != node; node = sll_next_node( data_fields, node ) ) {

    DataField* datafield = (DataField*) sll_node_get_data( node );

    // One of the streams do not have the same data type.
    if ( input->get_num( ) != datafield->get_count( ) ) {
      delete datafield;
      sll_destroy_node( data_fields, node );
    }

  }

  //
  // apply a bpipe_dpktf_add to all the nodes.
  //
  if ( 0 != sll_utraverse( data_fields, addDataField, (void*) bpipe ) ) {
    os << "BpipeOutput::init_bpipe_output( SLinkList, ostream& ) : ";
    os << "Unable to add data field" << endl;
    return EXIT_FAILURE;
  }
  
  bpo = bpipe_output( bpipe, const_cast< char* >( filename.c_str( ) ) );
  if ( NULL == bpo ) {
    os << "BpipeOutput::init_bpipe_output( SLinkList, ostream& ) : ";
    os << "Unable to allocate memory for bpo" << endl;
    return EXIT_FAILURE;
  }

  // map data field fields
  if ( NULL == ( data = bpipe_map_alloc( bpipe, 1, NULL ) ) &&
       bpipe_errno != BPNOERROR ) {
    os << "BpipeOutput::init_bpipe_output( SLinkList, ostream& ) : ";
    os << bpipe_strerror( bpipe_errno ) << "\tbpipe_errno = ";
    os << bpipe_errno << endl;
    return EXIT_FAILURE;
  }

  /* output header. must do this after calling bpipe_map */
  if ( bpipe_write_hdr( bpipe ) ) {
    os << "BpipeOutput::init_bpipe_output( SLinkList, ostream& ) : ";
    os << bpipe_strerror( bpipe_errno );
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;

}

int BpipeOutput::select_data_fields( BPipe* bpipe_input, string& xfer, 
				     SLinkList data_fields, ostream& os ) {

  void*  last = NULL;     // initialize it for the first time through
  DpktField *dpktf=NULL;

  vector< string > xfer_tok;
  suplib::tok( xfer_tok, xfer, "," );

  // For each input bpipe, loop thru the 
  while ( (dpktf=bpipe_dpktf_next( bpipe_input, &last )) ) {

    const char* name = bpipe_dpktf_name( dpktf );

    for ( string::size_type i = 0; i < xfer_tok.size( ); i++ ) {

      if ( xfer_tok[ i ].compare( name ) ) {

	BPDataType type = bpipe_dpktf_type( dpktf );

	BPMatrix* dpkt_matrix = 
	  bpipe_dpktf_matrix( dpktf, BPDSite_CORE, NULL );
	if ( NULL == dpkt_matrix ) {
	  os << "BpipeOutput::select_data_fields( Bpipe*, " << xfer;
	  os << ", ostream& ) : " << bpipe_strerror( bpipe_errno ) << endl;
	  return EXIT_FAILURE;
	}

	if ( EXIT_FAILURE == 
	     select_data_fields( (char*) name, type, dpkt_matrix, data_fields, os ) )
	  return EXIT_FAILURE;

      }
	     
    }

  }

  return EXIT_SUCCESS;

}

//
// If a data field is not within the linklist data_fields then
// insert it, else update to take the minimum size of the matrix.
//
int BpipeOutput::select_data_fields( char* nom, BPDataType t, 
				     BPMatrix* matrix,
				     SLinkList data_fields,
				     ostream& os ) {

  int status;
  int num = input->get_num( );

  DataField* datafield = new DataField( nom, t, matrix, num, status, os );
  if ( EXIT_FAILURE == status )
    return EXIT_FAILURE;

  DataField* ptr = (DataField*) sll_search( data_fields, datafield,
					      cmpDataField );

  if ( NULL == ptr ) {

    if ( sll_insert( data_fields, datafield ) ) {
      os << "BpipeOutput::select_data_fields( " << nom << ", " << t << ", ";
      os << "BPMatrix*, ostream& ) : Unable to insert data field" << endl;
      return EXIT_FAILURE;
    }

  } else {

    if ( EXIT_FAILURE == ptr->update( datafield, os ) )
      return EXIT_FAILURE;

    bpipe_matrix_delete( matrix );

    delete datafield;

  }

  return EXIT_SUCCESS;

}

int BpipeOutput::write_dpkts( ostream& os ) {

  for ( size_t i = 0; i < input->get_num( ); i++ )
    while ( copy_data( i, os ) )
      bpipe_write_dpkt( bpipe, data, bpo );

  return EXIT_SUCCESS;

}

int BpipeOutput::set_output_data_fields( SLinkList data_fields, 
					  ostream& os ) {

  size_t count = sll_count( data_fields );
  
  // allocate one more then neccessary for trailing NULL.
  output_data_fields = new DataField* [ count + 1 ];
  if ( NULL == output_data_fields ) {
    os << "BpipeOutput::set_output_data_fields( SLinkList, ostream& ) : ";
    os << "Unable to allocate memory for output_data_fields" << endl;
    return EXIT_FAILURE;
  }

  size_t i = 0;
  for ( SLLNode node = sll_head_node( data_fields );
	NULL != node; node = sll_next_node( data_fields, node ) ) {
    output_data_fields[ i ] = (DataField*) sll_node_get_data( node );
    i++;
  }

  output_data_fields[ count ] = NULL;

  return EXIT_SUCCESS;

}
