#ifndef DATAFIELD_H
#define DATAFIELD_H

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of mergerays
//
// mergerays is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// mergerays is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <string>
#include <limits>
#include <float.h>
#include <cstdlib>

#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>

class DataField {

  friend ostream& operator << ( ostream& os, DataField& a ) {
    a.print( os );
    return os;
  }

  friend ostream& operator << ( ostream& os, DataField* a ) {
    os << *a;
    return os;
  }

public:

  ~DataField( ) { 
    if ( input_data )
      delete [] input_data;
    matrix = NULL;
  }

  DataField( ) : count( 1 ), type( _BPDType_NOTYPE ), matrix( NULL ), 
    input_data( NULL ), output_data( NULL ), header_data( NULL ) { }

  DataField( char* n, BPDataType t, BPMatrix* mat, int nbpipe, int& status, 
	      ostream& os ) : count( 1 ), type( t ), name( n ), 
    matrix( mat ), input_data( NULL ), output_data( NULL ),
    header_data( NULL ) {
      
    status = init_DataField( n, t, nbpipe, os );

  }

  int copy_data( int i ) {
    bpipe_data_copy( output_data, input_data[ i ], matrix, type );
    return EXIT_SUCCESS;
  }
  //: copy the data from the ith stream for the `name' datafield.

  size_t     get_count( ) { return count; }

  BPMatrix*  get_matrix( ) { return matrix; }

  char*      get_name( ) { return const_cast<char*>( name.c_str( ) ); }

  void*      get_input_data( int i ) { return input_data[ i ]; }

  BPDataType get_type( ) { return type; }

  void       set_input_data( int i, void* ptr ) { input_data[ i ] = ptr; }

  void       set_output_data( void* ptr ) { output_data = ptr; }

  int init_DataField( char* nom, BPDataType t, int nbpipe, ostream& os ) {

    name = nom;
    type = t;

    input_data = new void*[ nbpipe ];
    if ( NULL == input_data ) {
      os << "DataField::init_DataField( " << nom << ", " << t;
      os << "BPMatrix*, " << nbpipe << ", ostream& ) : Unable to ";
      os << "allocate inpu_data" << endl;
      return EXIT_FAILURE;
    }
    for ( int i = 0; i < nbpipe; i++ )
      input_data[ i ] = NULL;

    return EXIT_SUCCESS;

  }

  void print( ostream& os ) {
    os << "name = " << name;
    os << "\ncount = " << count;
    os << "\ntype = " << type;
    os << "\nne = " << matrix->ne;
    os << "\nnd = " << matrix->nd << endl;
    for ( size_t i = 0; i < matrix->nd; i++ ) {
      os << "extent[" << i << "] = " << matrix->extent[ i ];
      if ( i == matrix->nd - 1 )
	os << endl;
      else
	os << '\t';
    }

    if ( input_data )
      for ( size_t j = 0; j < count; j++ )
	if ( input_data[ j ] )
	  os << name << "[" << j << "]=" << *((double*) (input_data[ j ])) << endl;

  }

  size_t min_time( size_t* empty_stream ) {

    size_t bin = 0;
    double mintime = DBL_MAX;

    for ( size_t i = 0; i < count; i++ ) {
      if ( 0 == empty_stream[ i ] ) {
	double time_i = *( (double*) input_data[ i ] );
	if ( time_i < mintime ) {
	  mintime = time_i;
	  bin = i;
	}
      }
    }

    return bin;
  }
  //: Find which stream has the minimum time.
  //: It is up to the user to make sure that 
  //: the name of the data field is `time'

  int update( DataField* datafield, ostream& os ) { 

    if ( 0 != name.compare( datafield->get_name( ) ) &&
	 type != datafield->get_type( ) )
      return EXIT_SUCCESS;
    // If the name and the type do not match then don't update.

    count++;

    BPMatrix* this_matrix = matrix;

    BPMatrix* datafield_matrix = datafield->get_matrix( );

    matrix = bpipe_matrix_min( this_matrix, datafield_matrix );
    if ( NULL == matrix ) {
      os << "DataField::update( DataField*, ostream& ) : ";
      os << "Unable to allocate matrix matrix" << endl;
      return EXIT_FAILURE;
    }

    bpipe_matrix_delete( this_matrix );

    return EXIT_SUCCESS;

  }

protected:

  size_t     count;     
  //: The number of streams containing data field with the 
  //: following: type, name, and matrix properties.

  BPDataType type;
  //: The datatype for the data field.

  string       name;
  //: The name of the data field.

  BPMatrix*  matrix;
  //: The dimension of the `name' data field.

  void**     input_data;
  //: The pointer to the input data for the datafield `name'.
  //  This is an array of pointers since the pointer must are
  //  different for the various input streams.

  void*      output_data;
  //: The pointer to the output data.

  void*      header_data;
  //: The pointer to the header data.

};

#endif
