// process.cc

/****************************************************************************
 * Author     : Terry Gaetz
 */

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of multilayer_reflect
//
// multilayer_reflect is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// multilayer_reflect is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <fstream>
#include <iostream>

#include <paramxx/ParFile.h>
#include <rl_raylib/rl_Ray.h>                // rl_Ray
#include <rl_raysuplib/rl_Multilayer_rdb.h>  // rl_Multilayer_rdb
#include <rl_raylib/rl_MultilayerSurface.h>  // rl_MultilayerSurface

extern "C" {
#include <bpipe/bpipe.h>                     // BPipe typedefs
#include <bpipe/datatypes.h>                 // DVector3 DCVector2
#include <bpipe/bpipe_axaf.h>                // misc. #defines
}

#include <tracefctxx/TraceFct.h>             // TraceFct
#include <tracefct/exiterrvals.h>
#include <suplib/debug.h>

#include <cstdio>

// Some bit macros
typedef unsigned int bitmask_t;
  
#define BITMASK(b) (1 << (b))
#define BITSET(a,b) ((a) |= BITMASK(b))
#define BITCLEAR(a,b) ((a) &= ~BITMASK(b))
#define BITTEST(a,b) (!(!((a) & BITMASK(b))))

#ifdef  CHECK_REFLECTIVITY
  #include <cstdio>
  static FILE* dbg_reflectivity = NULL;
#endif

using namespace std;

static struct Ray
{
  /// ray position
  DVector3     *pos;
  /// ray direction vector (direction cosines)
  DVector3     *dir;
  /// surf normal, most recent reflection
  DVector3     *norm;
  /// statistical weight of ray
  double       *wt;
  /// photon energy
  double       *energy;
  /// polarization amplitude (cosine component)
  DCVector2    *c2;
  /// polarization amplitude (sine component)
  DCVector2    *s2;
  /// polarization amplitude (sine component)
  unsigned int *id;
  /// surface missed flags
  bitmask_t    *missed;
} s_ray;

long         n_ray;
/// number of rays reflected
long         n_rfl;
/// number of ghost rays
long         n_ghosts;
/// number of rays violating \f$n \cdot v \ge 0\f$
long         n_bad_ndotv;

/// flag indicating whether to dump out information on bad 
/// \f$n \cdot v \ge 0\f$ rays
int          dump_bad_ndotv;
/// flag indicating whether to print out a summary
int          do_summary;
  
/// flags indicating whether the ray intercepted the surface
unsigned int surf_missed_mask;

BPipe*       bpi;
BPipeOutput* bpo;
void*        data;

/* Routine to get the data pointer for a named data packet field */

static char*
dpktf_data( BPipe* bpipe, char *name, void *data )
{
  TraceFct tf("dpktf_data");

  DpktField *field = bpipe_dpktf(bpipe, name);

  if (field == NULL)
  {
    tf.exit(ExitERR_misc, "%s isn't in input data stream", name);
  }

  return bpipe_dpktf_data(field, data);
}

/*---------------------------------------------------------- 
 * set up the bpipes 
 */
static void
init_bpipe( char const instream[], char const outstream[] )
{
  TraceFct tf("rl_BPipe::init");

  /*-------------------------------------------------------------------
   * acquire & set up input & output bpipes;
   * set up pointers to data fields - attach BPipe fields to ray struct...
   */
  
  /* Setup bpipe input and output streams */
  
  if ((bpi = bpipe_new()) == NULL) 
  {
    tf.exit(ExitERR_misc, "couldn't open bpipe: %s",
            bpipe_strerror(bpipe_errno));
  }

  if ( bpipe_input(bpi, const_cast<char*>(instream)) )
  {
    tf.exit(ExitERR_misc, "couldn't open input file `%s': %s", instream,
            bpipe_strerror(bpipe_errno));
  }

  bpo = bpipe_output(bpi, const_cast<char*>(outstream));
  if (! bpo )
  {
    tf.exit(ExitERR_misc, "couldn't open input file `%s': %s", outstream,
            bpipe_strerror(bpipe_errno));
  }

  // it's not certain that there'll be a surf_missed flag set, so create
  // one if necessary
  switch ( bpipe_dpktf_add( bpi, BP_AXAF_surf_missed, BPDType_uint, NULL ) )
  { 
  case -1:     // error
               tf.exit( ExitERR_misc, "couldn't add surf_missed field: %s",
                        bpipe_strerror(bpipe_errno));
    
  case  1:     // already exists
               break;

  case  0:     // doesn't exist, so remove from output image
  
               switch (bpipe_dpktf_delete( bpi, BP_AXAF_surf_missed,
                                                  BPDSite_OUTPUT, bpo ) )
               {
               case -1:   tf.exit( ExitERR_misc,
                                   "couldn't add surf_missed field: %s",
                                   bpipe_strerror(bpipe_errno));
                          break;

               case 1:    break;

               default:   tf_assert( NULL ); 
                          break;
               }
               break;
  
  default:     tf_assert( NULL );
  }

  if ( NULL == (data = bpipe_map_alloc(bpi, 1, NULL)) )
  {
    tf.exit(ExitERR_alloc, "unable to allocate memory for data packet: %s",
            bpipe_strerror(bpipe_errno));
  }

  if ( bpipe_hdrf_data( bpi, const_cast<char*>("srfno"), 0 ) )
  {
    unsigned int surf_no =
      * ((unsigned int *) bpipe_hdrf_data( bpi,
                                           const_cast<char*>("srfno"), 0 ) );
    surf_missed_mask = 1<<(surf_no-1);
  }
  else
  {
    surf_missed_mask = 0;
  }

  s_ray.pos    = (DVector3*) dpktf_data(bpi,
                                        const_cast<char*>(BP_AXAF_phot_pos),
                                        data);

  s_ray.dir    = (DVector3*) dpktf_data(bpi,
                                        const_cast<char*>(BP_AXAF_phot_dir),
                                        data);

  s_ray.norm   = (DVector3*) dpktf_data(bpi,
                                        const_cast<char*>(BP_AXAF_surf_norm),
                                        data);

  s_ray.energy = (double*) dpktf_data(bpi,
                                      const_cast<char*>(BP_AXAF_phot_energy),
                                      data);

  s_ray.wt     = (double*) dpktf_data(bpi,
                                      const_cast<char*>(BP_AXAF_phot_wt),
                                      data);

  s_ray.c2     = (DCVector2*) dpktf_data(bpi,
                                         const_cast<char*>(BP_AXAF_pol_cos),
                                         data);

  s_ray.s2     = (DCVector2*) dpktf_data(bpi,
                                         const_cast<char*>(BP_AXAF_pol_sin),
                                         data);

  s_ray.id     = (unsigned int*) dpktf_data(bpi,
                                            const_cast<char*>(BP_AXAF_phot_id),
                                            data);

  s_ray.missed = (unsigned int*) dpktf_data(bpi,
                                       const_cast<char*>(BP_AXAF_surf_missed),
                                       data);

  if (bpipe_write_hdr(bpi))
  {
    tf.exit(ExitERR_misc, "bpipe_write_hdr: %s", bpipe_strerror(bpipe_errno));
  }
}

/*---------------------------------------------------------- 
 * clean up when done
 */
static void
cleanup_bpipe()
{
  TraceFct tf("~rl_BPipe()");

  if ( bpipe_errno != BPNOERROR )
  {
    tf.exit( ExitERR_misc, bpipe_strerror(bpipe_errno) );
  }

  if ( do_summary )
  {
    tf.message( "summary: reflected:                  %d\n", n_rfl       );
    tf.message( "summary: missed surface:             %d\n", n_ghosts    );
    tf.message( "summary: hit wrong side of surface:  %d\n", n_bad_ndotv );
  }
  bpipe_delete(bpi);
}

/*---------------------------------------------------------- 
 * get a ray from the bpipe
 */
static int
get_ray( rl_Ray& ray, rl_MultilayerSurface& surface )
{
  TraceFct tf("rl_BPipe::read");

  int ray_missed = 1;
  int rc = 1;

  while ( rc && ray_missed )  // skip over rays which missed surface...
  {
    rc = bpipe_read_dpkts(bpi, data, 1);
    if ( rc != 1 )
    {
      if (bpipe_errno != BPNOERROR)
      {
        tf.exit(ExitERR_misc, bpipe_strerror(bpipe_errno));
      }
      break;
    }

    // if it missed the last surface, write it and continue
    ray_missed = surf_missed_mask & *s_ray.missed;

    if ( ray_missed )
    {
      ++n_ghosts;
      if (bpipe_write_dpkt(bpi, data, bpo))
      {
        tf.exit(ExitERR_misc, "(%u): error writing photon", *s_ray.id);
      }
      continue;
    }
  }

  // At this point we either have a valid ray (rc == 1) or we have
  // reached the end of the input bpipe stream (rc == 0).
  //
  // Process the ray if we have one.

  if ( rc != 0 )
  {
    // check whether ray is incident from exterior of the surface
    double ndotv = s_ray.dir->x * s_ray.norm->x
                 + s_ray.dir->y * s_ray.norm->y
                 + s_ray.dir->z * s_ray.norm->z;

    if ( ndotv < 0.0 )
    {
      ++n_bad_ndotv;
      if ( dump_bad_ndotv )
      {
        tf.message( "%u: ray hits wrong side of surface (n . v = %f)\n",
             *s_ray.id, ndotv );
      }
    }
    // copy ray/surface data over into ray and surface...  

    ray.position(0)  =  s_ray.pos->x;
    ray.position(1)  =  s_ray.pos->y;
    ray.position(2)  =  s_ray.pos->z;
    ray.direction(0) =  s_ray.dir->x;
    ray.direction(1) =  s_ray.dir->y;
    ray.direction(2) =  s_ray.dir->z;
    ray.energy()     = *s_ray.energy;
    ray.set_id( *s_ray.id );

    rl_PolCSPOD cs;
    cs.c2_[0] = rl_Traits::complex( s_ray.c2->q1.r, s_ray.c2->q1.i );
    cs.c2_[1] = rl_Traits::complex( s_ray.c2->q2.r, s_ray.c2->q2.i );
    cs.s2_[0] = rl_Traits::complex( s_ray.s2->q1.r, s_ray.s2->q1.i );
    cs.s2_[1] = rl_Traits::complex( s_ray.s2->q2.r, s_ray.s2->q2.i );
    ray.set_polarization( cs );

    dvm3_Vector norm( s_ray.norm->x, s_ray.norm->y, s_ray.norm->z );
    surface.set_normal( norm );
  }

  return rc;
}

/*---------------------------------------------------------- 
 * put a ray to the output bpipe
 */
static int 
put_ray( rl_Ray const& ray )
{
  // copy ray/surface data back from ray and surface...  

  s_ray.pos->x = ray.position(0);
  s_ray.pos->y = ray.position(1);
  s_ray.pos->z = ray.position(2);

  s_ray.dir->x = ray.direction(0);
  s_ray.dir->y = ray.direction(1);
  s_ray.dir->z = ray.direction(2);

  rl_PolCSPOD cs;
  ray.get_PolCSPOD( cs );
  s_ray.c2->q1.r = cs.c2_[0].real();
  s_ray.c2->q1.i = cs.c2_[0].imag();
  s_ray.c2->q2.r = cs.c2_[1].real();
  s_ray.c2->q2.i = cs.c2_[1].imag();
  s_ray.s2->q1.r = cs.s2_[0].real();
  s_ray.s2->q1.i = cs.s2_[0].imag();
  s_ray.s2->q2.r = cs.s2_[1].real();
  s_ray.s2->q2.i = cs.s2_[1].imag();

  *s_ray.wt = cs.intensity();

  // push the ray into the output pipe
  int rc = bpipe_write_dpkt(bpi, data, bpo);
  if ( rc )
  {
    TraceFct tf("rl_BPipe::write");
    tf.exit(ExitERR_misc, "(%u): error writing photon", *s_ray.id);
  }
  ++n_rfl;

  return !rc;
}

/*---------------------------------------------------------- 
 * parst the transforms string to find out how to do the interpolation
 */
rl_Traits::EInterpMode
parsexfrms( const string& reflectance_xfrm, const string& energy_xfrm ) 
{
  TraceFct tf( "parsexfrms" );

  if ( string("lin") == reflectance_xfrm && string("lin") == energy_xfrm )
  {
    return rl_Traits::ELinLin;
  } 
  else if ( string("lin") == reflectance_xfrm && string("log") == energy_xfrm )
  {
    return rl_Traits::ELinLog;
  }
  else if ( string("log") == reflectance_xfrm && string("lin") == energy_xfrm ) 
  {
    return rl_Traits::ELogLin;
  }
  else if ( string("log") == reflectance_xfrm && string("log") == energy_xfrm ) 
  {
    return rl_Traits::ELogLog;
  } 
    
  tf.exit( 1, string("invalid reflectance/energy transformation type(s): ") + reflectance_xfrm + "/" + energy_xfrm );
}

/*---------------------------------------------------------- 
 * process the rays
 */
void
process( const ParFile& pf )
{
  TraceFct tf( "process" );

  ofstream logfile( pf.pgetstring( "logfile" ).c_str() );
  if (! logfile )
  {
    tf.exit( 1, "Unable to open logfile: %s\n", pf.pgetstring( "logfile" ).c_str() );
  }
  #ifdef    CHECK_REFLECTIVITY
    if (! dbg_reflectivity )
    {
      dbg_reflectivity = fopen("reflectivity_cmp.dbg", "w");
      if (! dbg_reflectivity )
      {
        tf.exit(1, "ml_reflect():  file could not be opened\n");
      }
    }
    fprintf( dbg_reflectivity, "energy\treflectivity\nN\tN\n" );
  #endif // CHECK_REFLECTIVITY

  rl_Traits::EInterpMode interp_mode = parsexfrms( pf.pgetstring( "reflectance_xfrm" ), pf.pgetstring( "energy_xfrm" ) );

  rl_Multilayer_rdb ml( pf.pgetstring( "layerdata_rdb" ).c_str(), interp_mode );
  dvm3_Vector       norm;
  rl_MultilayerSurface surface( ml, norm );
  rl_Ray            ray;

  // set up the BPipe

  init_bpipe( pf.pgetstring( "input" ).c_str(), 
              pf.pgetstring( "output" ).c_str() );

  dump_bad_ndotv = (rl_Traits::Bool)dbflag("bad_ndotv");
  do_summary     = (rl_Traits::Bool)dbflag("summary");

  while ( get_ray( ray, surface ) )
  {
    #ifdef CHECK_REFLECTIVITY
      double weight_in = ray.intensity();
    #endif

    int rc = surface.reflect( ray );

    if ( rc )
    {
      logfile << "reflect returned an error at layer " << rc-1 
              << " (" << surface.layer(rc-1).layer_name() << ")" << endl;
      logfile << "The ray energy was: " << ray.energy() << endl;
      logfile << "The valid energy range for layer " << rc-1 
              << " is [" << surface.layer(rc-1).energy_min() << ", "
                         << surface.layer(rc-1).energy_max() << "]"  << endl;

      tf.message( "reflect returned an error at layer %d (%s)\n",
                  rc-1, surface.layer(rc-1).layer_name() );
      tf.message( "The ray energy was: %e\n", ray.energy() );
      tf.exit( 1, 
               "The valid energy range for layer %d is [%e, %e]\n",
               rc-1, surface.layer(rc-1).energy_min(), 
                     surface.layer(rc-1).energy_max() );
    }

    #ifdef CHECK_REFLECTIVITY
      fprintf( dbg_reflectivity, "%20.15f\t%20.15f\n", 
	       ray.energy( ),
               ray.intensity() / weight_in );
    #endif

    put_ray( ray );
  }
  cleanup_bpipe();
}
