#include <iostream>
#include <fstream>
#include <cstdlib>

#include <paramxx/ParFile.h>

using namespace std;

/*
  PROCEDURE
     setp - set one or more parameters in an IRAF-style parameter
     file

  SYNOPSIS
     setp paramfile [ passign1...  ]

  DESCRIPTION
     setp sets parameters in the specified paramfile. The passign
     arguments  can be either positional or absolute assignments.
     Positional assignments consist of the value  alone  and  are
     assigned  one to one to the corresponding non-hidden parame-
     ter in the param file.  The absolute or keyword form  is  an
     assignment,  parameter  =  value, where the parameter can be
     any parameter of the program being run.  Here, order is  not
     relevant.  (See the param man page for more information).

     If no parameter  assignments  are  specified,  the  user  is
     prompted for a new value for each parameter in the parameter
     file.  Both hidden and non-hidden  parameters  are  prompted
     for.   The  default  value  is displayed in parenthesis. The
     user can input a new value or a carriage return to keep  the
     default value.

     If a parameter name is input in the  command  line  followed
     immediately by the "!" character (no spaces), that parameter
     is prompted for.  Again, the default value is  displayed  in
     parenthesis.  The  user  can input a new value or a carriage
     return to keep the default value.

     Thus,

          example% setp src

     will prompt for each parameter in the src parameter file:
          QPOE file (/pros/xray_old/data/snr.qp):
          region (c 512 512 100): c 512 512 50
          mode (ql):

     The region parameter can also be changed directly:

          example% setp src region="c 512 512 50"

     Finally, we can prompt specifically for the region parameter
     alone:

          example% setp src region!

          region (c 512 512 100): c 512 512 50

     In searching for the parameter file, .par extension will  be
     added  to  the paramfile name, if no extension is specified.

     If the file is not located in the directory specified by the
     paramfile, the directory pointed to by the UPARM environment
     variable will be searched.

  OPTIONS
     none

  ENVIRONMENT
    The environment variable PFILES specifies the list of directories,
    delimited by colon(:), in which to look for the parameter file.

     The environment variable UPARM specifies  the  directory  in
     which to look for the parameter file, if the pathname of the
     file is not part of the paramfile specification.
*/
static int parset( int argc, char** argv ) {

  try {

    ParFile pfile( argc, argv, argv[ 1 ] );

    ofstream parfile( pfile.get_filename( ).c_str( ), ios::out );
    if ( !parfile ) {
      cerr << "Unable to open `" << pfile.get_filename( ).c_str( ) << "'\n";
      return EXIT_FAILURE;
    }

    // It looks kind of odd that the file gets opened then somethings get
    // written out to it.  This is ok since the constructor of the ParFile
    // class takes care of merging the command line arguments.
    parfile << pfile;

    parfile.close( );

    return EXIT_SUCCESS;

  } catch ( ParFileException& pfe ) {

    cerr << pfe << endl;
    return EXIT_FAILURE;
    
  } catch ( Exception& e ) {

    cerr << e << endl;
    return EXIT_FAILURE;

  }

}

int main( int argc, char** argv ) {

  if ( argc < 2 ) {
    cerr << "Usage:\t" << argv[0] << " pfile par=val [par1=val1]\n";
    return !EXIT_SUCCESS;
  }

  return parset( argc, argv );

}
