/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of quef
 *
 * quef is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * quef is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <suplib/debug.h>


#include "quef_def.h"
#include "inquef.h"
#include "diagnostic.h"

#define AVGENERGY(elem) ((elem.emax + elem.emin) / 2)

#define QECLIP(lut,elem) \
  do { \
	if ( elem < lut->qemin ) \
	  elem = lut->qemin; \
	if ( elem > lut->qemax ) \
	  elem = lut->qemax; } \
  while(0)


void interpolate_qe_table( QE* lookuptable ) {

  size_t num = lookuptable->num;
  QuantumEfficiency* table = lookuptable->qe;

  size_t idx;

  tf_enter( "interpolate_qe_table" );

  for ( idx = 0 ; idx < num - 1 ; idx++ )
  {
    /* average energies in this bin and the next */
    double e0 = AVGENERGY(table[idx]);
    double e1 = AVGENERGY(table[idx+1]);

    /* qe at those average energies. low and high don't have a meaning
       yet */
    double qe0 = table[idx].qemid;
    double qe1 = table[idx+1].qemid;

    /* derive the parameters for the linear fit */
    double m = ( qe1 - qe0 ) / ( e1 - e0 );
    double b = qe1 - m * e1;

    /* now calculate the QE at the bin boundaries.  we have

         e0 < table[idx].emax < e1
	 e0 < table[idx+1].emin < e1

      so we interpolate table[idx].qemax and table[idx+1].qemin.

      there's no guarantee that table[idx].emax == table[idx+1].emin,
      either.
    */
    
    table[idx].qemax   = m * table[idx].emax + b;
    table[idx+1].qemin = m * table[idx+1].emin + b;

    if ( lookuptable->qeclip )
    {
      QECLIP(lookuptable,table[idx].qemax);
      QECLIP(lookuptable,table[idx+1].qemin);
    }

    /* but, we need to extrapolate for the outer bounds at idx=0 and
       idx=num-1, so watch for those as well. make sure we don't
       extrapolate below zero*/
    if ( idx == 0 )
    {
      table[idx].qemin  = m * table[idx].emin  + b;
      
      if ( lookuptable->qeclip )
	QECLIP(lookuptable,table[idx].qemin);
    }

    if ( idx+1 == num-1 )
    {
      table[idx+1].qemax  = m * table[idx+1].emax  + b;
      if ( lookuptable->qeclip )
	QECLIP(lookuptable,table[idx+1].qemax);
    }
  }

  if ( dbflag( "dump-qe-all") || dbflag( "dump-qe-interp" ) )
    print_qe( lookuptable,
	      "Sorted QE Table after interpolation to bin boundaries");


  /*
   * Set up for more efficient lin-log, log-lin, log-log interpolation; 
   * take logs now rather than for each ray interpolation.
   * (Assumes that number of rays is much larger than number of table
   *  entries.)
   */
  switch( lookuptable->abcissa ) {
    int i;

  case InterpLIN: /* purposely do nothing */
    break;

  case InterpLOG:
    for ( i = 0; i < lookuptable->num; i++ ) {
      if ( lookuptable->qe[ i ].emin <= 0 ||
	   lookuptable->qe[ i ].emax <= 0 )
	tf_exit( ExitERR_illegalArg, "one or more of the energy limits for bin %d [%f,%f) are not positive: cannot take logarithm of energy\n", i, lookuptable->qe[ i ].emin,
		 lookuptable->qe[ i ].emax);
      lookuptable->qe[ i ].emin = log( lookuptable->qe[ i ].emin );
      lookuptable->qe[ i ].emax = log( lookuptable->qe[ i ].emax );
    }
    break;

  default:
    tf_exit( ExitERR_illegalArg, "internal error" );
    break;
  }

  switch( lookuptable->ordinate ) {
    int i;

  case InterpLIN: /* purposely do nothing */ 
    break;

  case InterpLOG:
    for ( i = 0; i < lookuptable->num; i++ ) {
      lookuptable->qe[ i ].qemin = log( lookuptable->qe[ i ].qemin );
      lookuptable->qe[ i ].qemax = log( lookuptable->qe[ i ].qemax );
    }
    break;

  default:
    tf_exit( ExitERR_illegalArg, "internal error" );
    break;
  }

  if ( dbflag( "dump-qe-all") || dbflag( "dump-qe-xfrm" ) )
    print_qe( lookuptable, "Sorted QE Table after transform" );

  tf_leave();
}
