/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of quef
 *
 * quef is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * quef is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <stdio.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>


#include "quef_def.h"
#include "quef.h"

#include "readbp.h"
#include "diagnostic.h"

/*
  PROCEDURE
    compare two nodes based on a numeric ordering of the energy field

  RETURN
    an integer
*/
static int
cmp_energy( const void *key, const void *element )
{

  double *key_ptr = (double*) key;
  QuantumEfficiency* element_ptr = (QuantumEfficiency*) element;

  if ( *key_ptr < element_ptr->emin )
    return -1;
  else if ( *key_ptr >= element_ptr->emax )
    return 1;
  else
    return 0;

}

static double
lookup_table( QE* table /*The quantum efficiency lookuptable*/,
	      Photon* photon  /* The photon structure */,
	      int* in_table  /* whether the energy was in a bin in the
				table */
	      )
{

  double w, efficiency;
  double key = photon->energy[0];
  QuantumEfficiency* ptr;
  
  *in_table = 1;

  if ( table->abcissa == InterpLOG )
    key = log( key );
  
  ptr = bsearch( &key, table->qe, table->num, 
		 sizeof( QuantumEfficiency ), cmp_energy );

  if ( NULL == ptr )
  {
    *in_table = 0;
    return 0.0;
  }

  /*
   * log interpolation in energy
   */
  w     = (key - ptr->emin) / (ptr->emax - ptr->emin );
  efficiency = (1.0 - w) * ptr->qemin + w * ptr->qemax;

  if ( table->ordinate == InterpLOG )
  {
    efficiency = exp( efficiency );
  }

  return efficiency;
}


/*
  PROCEDURE
    Initialize the binary pipe for reading and writing.
    Loop thru the photon from the binary pipe.
    Do a binary search of the photon energy within the lookup table.
    If the energy of the photon is not within the quantum efficiency lookup 
    then a diagnostic message is printed (and the photon weight is set to 
    zero). Otherwise the photon's weight is multiplied by the quantum 
    efficiency as found in the lookup table.  Write to the binary pipe.

  RETURN
    void
*/
void 
quantum_efficiency(
  QE* lookuptable,	/* The quantum efficiency lookup table */
  InOut* io,		/* Where to read/write the input/output */
  int verbose
  )
{
  Bpipe pipe;
  Photon photon;

  int nphotons = 0;
  int noutside = 0;
  int notfirst = 0;

  tf_enter( "quantum_efficiency" );

  /* init bpipe for reading and writing */
  init_bpipe( io, &pipe, &photon );


  /* loop thru photons from bpipe */
  while( bpipe_read_dpkts( pipe.bpipe, pipe.data, 1 ) ) {

    int in_table;

    nphotons++;

    photon.weight[0] *= lookup_table( lookuptable, &photon, &in_table );

    /* 
     * if the photon is not in the lookup table, it's weight will now
     * have been set to zero.  if requested, don't output it
     */
    if ( ! in_table )
    {
      noutside++;

      if ( lookuptable->kill )
	continue;

      if ( verbose )
      {
	if ( ! notfirst++ )
	{
	  tf_message( "# rays with out of bounds energies\n" );
	  tf_message( "id\tenergy\n" );
	  tf_message( "N\tN\n" );
	}

	tf_message( "%u\t%f\n", *photon.id, *photon.energy );
      }

    }

    if ( bpipe_write_dpkt( pipe.bpipe, pipe.data, pipe.outstream ) )
      tf_exit( ExitERR_misc, "error writing photon %d", nphotons); 

  }
  if ( bpipe_errno != BPNOERROR )
    tf_exit( ExitERR_misc, "error reading photon %d: %s", nphotons,
             bpipe_strerror( bpipe_errno) );

  if  ( noutside )
    tf_message( "%d out of %d photons were outside QE table limits\n",
		noutside, nphotons );

  bpipe_delete( pipe.bpipe );

  tf_leave();
}
