/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of quef
 *
 * quef is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * quef is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <linklist/linklist.h>

#include <tparam/tparam.h>

#include <suplib/str.h>
#include <suplib/debug.h>


#include <mst_rdb/mst_rdb.h>

#include "quef_def.h"
#include "util.h"
#include "inquef.h"
#include "diagnostic.h"
#include "param_utils.h"

#include "gparam_def.h"

#define NUM_FIELDS      3

static TokListToken
interp_type[] =
{
  { "lin", InterpLIN },
  { "log", InterpLOG }
};
TokList interp_types = GenTokList(interp_type);


/*
  PROCEDURE
    A compare function (double) in ascending order.

  RETURN
    
*/
static int
  ascending_order_in_energy( const void *key, const void *element )
{
  double e0 = ((QuantumEfficiency*) key    )->emin;
  double e1 = ((QuantumEfficiency*) element)->emin;
  double diff = e0 - e1;

  return
    ( diff > 0.0 ) ?  1 :
    ( diff < 0.0 ) ? -1 :
                      0;
}


/*
  PROCEDURE
    Sort the lookup table to make sure that it is in ascending order.
    Verify that the entries of the lookup table are correct, namely that the
    ith energy is greater or equal to the energy of the (i+1)th entry.

  RETURN
    void
*/
static void sort_verify( QE* lookuptable, SLinkList ll )
{
  size_t i, num;
  QuantumEfficiency* table;
  QuantumEfficiency* tablei;
  SLLNode node;

  tf_enter( "sort_verify" );

  lookuptable->num = num = sll_count( ll );

  tablei = table =
    (QuantumEfficiency*) malloc( (num)*sizeof(QuantumEfficiency));
  if ( NULL == table )
    tf_exit( ExitERR_alloc,
	     "Unable to allocate QuantumEfficiency table of size %d", num );

  for ( node = sll_head_node( ll );  node ; node = sll_next_node( ll, node ) )
  {
    QuantumEfficiency* node_data = (QuantumEfficiency*)
                                            sll_node_get_data( node );
    *tablei++ = *node_data;
  }

  /* verify that bin limits are ordered correctly */
  for ( i = 0 ; i < num ; i++ )
  {
    if ( table[i].emin > table[i].emax )
      tf_exit( ExitERR_misc,
	       "out of order bin limits bin[%d].emin(%f) > bin[%d].emax(%f)",
	       i, table[i].emin, i, table[i].emax );
  }


  /* 
   * The table is usually already sorted so hopefully qsort randomizes
   * the input
   */
  qsort( table, num, sizeof( QuantumEfficiency ), ascending_order_in_energy );


  /* verify that bins don't overlap */
  for ( i = 0 ; i < num-1 ; i++ )
  {
    if ( table[i].emax > table[i+1].emin )
      tf_exit( ExitERR_misc,
	       "overlapping bins: bin[%d].emax(%f) > bin[%d].emin(%f)",
	       i, table[i].emax, i+1, table[i+1].emin );
  }


  lookuptable->qe = table;

  tf_leave();
}


/*
  PROCEDURE
    Get the user's requested /rdb header names to be read in.

  RETURN
    The array of character strings containing the names of the columns to
    be read in.
*/
static char** 
get_field_names( paramfile pf )
{
  char** string_argv;

  tf_enter( "get_field_names" );

  if( NULL == (string_argv=(char **) calloc( NUM_FIELDS+1, sizeof(char *) )) )
    tf_exit( ExitERR_alloc, "Unable to allocate `fields'\n" );

  string_argv[0] = strpget( pf, BENERGY );
  string_argv[1] = strpget( pf, EENERGY );
  string_argv[2] = strpget( pf, QEFFICIENCY );
  string_argv[ NUM_FIELDS ] = NULL;

  tf_leave();
  return string_argv;
}


static void read_rdb_table( paramfile pf, QE* lookuptable )
{
  char** column_names;
  size_t num = 0;
    
  DataColumnMap_st   *map;
  rdbHeader *header;
  RDB_Type types[] = { RDB_Num, RDB_Num, RDB_Num };
  size_t* offset;

  SLinkList ll;
  QuantumEfficiency data;
  QuantumEfficiency* data_ptr;

  FILE* fp;

  tf_enter( "read_rdb_table" );

  {
    char* filename = strpget( pf, LOOKUP );
    fp = open_file_with_mode( filename, "r" );
    free( filename );
  }

  column_names = get_field_names( pf );

  /* count the number of fields */
  while( column_names[ num ] )
    num++;

  offset = (size_t*) malloc( num*sizeof(size_t) );
  if ( NULL == offset )
    tf_exit( ExitERR_alloc, "Unable to allocate offset" );

  offset[ 0 ] = offsetof( QuantumEfficiency, emin );
  offset[ 1 ] = offsetof( QuantumEfficiency, emax );
  offset[ 2 ] = offsetof( QuantumEfficiency, qemid );

  header = rdb_rd_hdr( fp );

  map = rdb_map_cols_arst( header, num, column_names, types, offset );

  if ( ( ll = sll_new( NULL )) == NULL )
    tf_exit( ExitERR_alloc, "Unable to allocate `ll'" );

  data.qemin = data.qemax = 0;

  while( rdb_col_read_st( fp, header, map, &data ) ) {

    data_ptr = (QuantumEfficiency*) malloc( sizeof(QuantumEfficiency) ); 
    if ( NULL == data_ptr )
      tf_exit( ExitERR_alloc, "Unable to allocate `data_ptr'" );

    *data_ptr = data;

    sll_insert_tail( ll, data_ptr );
  }

  rdb_free_hdr( header );   /* free up rdbHeader dynamic allocation */
  free( map->map );
  free( map );
  free( offset );
  free_tokens( column_names );
  fclose( fp );


  /* cause we wanna flatten out the end points */
  sort_verify( lookuptable, ll );
  sll_delete( ll, free );

  tf_leave();
}



static InterpType get_interp_type( paramfile pf, const char* var )
{
  InterpType type;
  char dummy[1024];
  
  tf_enter( "get_interp_type" );

  tpgetstr( pf, var, dummy, MAX_LENGTH, 1 );

  type = (InterpType) tokmatch( dummy, &interp_types );
  if ( -1 == type )
    tf_exit( ExitERR_illegalArg, "%s: unknown interpolation type `%s'",
	     var, dummy );

  tf_leave();

  return type;
}

/*
  PROCEDURE
    Read the quantum efficiency table from an /rdb table.

  RETURN 
    void
 */
void get_lookup_table( paramfile pf, QE* lookuptable )
{
  tf_enter( "get_lookup_table" );

  lookuptable->kill = pgetb( pf, DOORDIE );

  lookuptable->abcissa = get_interp_type( pf, XTYPE );
  lookuptable->ordinate = get_interp_type( pf, YTYPE );

  read_rdb_table( pf, lookuptable );

  if ( dbflag( "dump-qe-all" ) || dbflag( "dump-qe-raw" ) )
    print_qe( lookuptable, "Sorted QE Table, as read" );

  interpolate_qe_table( lookuptable );

  tf_leave();
}
