/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <string.h>

#include <randlib_c/randlib.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <mathconst/mathconst.h>

#include <lua.h>
#include <luasup/luasup.h>
#include <suplib/str.h>
#include <suplib/debug.h>

#include "main.h"
#include "args.h" 
#include "lua_interface.h"
#include "misc.h"

#include "dist.h"

#define DIST	 "dist"

typedef struct
{
  double	radius;		/* radius of disk */
  double	z;		/* z coord of disk */

  ImgDist	idist;

  double	theta;		/* off axis angle of image */
  double	phi;		/* off axis angle of image */
  double	sin_theta0;
  double	cos_theta0;
  double	sin_phi0;
  double	cos_phi0;
} DiskData;

static void	define(void);
static int	init(PositionGeneratorData *data);
static int	generate( void *data, DVector3 *direction, double *z );
static void	cleanup(void *data);

Module disk_module = { "disk", define };

static PositionGeneratorSpec spec =
{
  "disk",
  init,
  generate,
  cleanup
};

static void
define( void )
{
  DiskData *data;
  char *name, *radius_spec;
  double x, y, z;
  lua_Object optargs = LUA_NOOBJECT;
  char err_pfx[1024];

  tf_enter( "disk" );

  slua_args( define_error, "disk",
	     "name",		SLua_STRING,	&name,
	     "radius",		SLua_STRING,	&radius_spec,
	     "optargs", 	SLua_TABLE | SLua_OPTIONAL, 	&optargs,
	     NULL
	     );

  /* generate error message prefix */
  sprintf( err_pfx, "disk `%s'", name );

  if ( NULL == ( data = (DiskData *) malloc( sizeof(DiskData) ) ) )
    define_error( "%s: error allocating memory", err_pfx );

  /* copy optargs, so can delete elements from it as we go along */
  optargs = 
    optargs == LUA_NOOBJECT ? lua_createtable() : slua_tbl_copy( optargs );

  /* Axial Source Position */
  get_src_z( err_pfx, optargs, &data->z );

  /* Angular Source Position */
  get_src_pos( err_pfx, optargs, &data->theta, &data->phi );

  /* determine  distribution */
  get_ImgDist( err_pfx, &data->idist, optargs, data->z );

  /* generate positions on unit sphere of center of image in (x,y,z)
     from original (theta, phi) (in tilted lat,long) */
  x = sin( data->theta ) * cos( data->phi );
  y = sin( data->theta ) * sin( data->phi );
  z = cos( data->theta );

  /* generate sin and cos of (lat,long) from (x,y,z) */
  data->sin_theta0 = y;
  data->cos_theta0 = sqrt( x*x + z*z );
  data->sin_phi0 = x / data->cos_theta0;
  data->cos_phi0 = z / data->cos_theta0;

  /* we should have read (and deleted) all relevant optargs.  Those
     that remain should be flagged as errors */
  verify_empty_optargs( err_pfx, optargs );

  /* now that we know the axial location, process the height and
     width, converting into angular units if necessary */
  parse_extent( err_pfx, "radius",  radius_spec,  
		data->z, &data->radius,  &z );

  if ( data->radius <= 0.0 )
    define_error( "%s: radius must be positive", err_pfx );

  new_pos_gen( &spec, name, data );

  free( name );

  tf_leave( );
}


static int
init( PositionGeneratorData *data )
{
  DiskData *disk = (DiskData *) data->data;

  data->max_extent	= disk->radius + disk->theta;
  data->z		= disk->z;
  data->factor		= 1.0;

  return 0;
}

static int
generate( void *data, DVector3 *direction, double *z )
{
  DiskData *disk = (DiskData *) data;
  double r2;
  double theta, phi, cos_theta, sin_theta, cos_phi, sin_phi;
  double cos_theta_p, sin_theta_p, cos_phi_p, sin_phi_p;

  *z		= disk->z;

  r2 = disk->radius * disk->radius;
  do
  {
    gen_ImgDist( &disk->idist,
		 &phi,   -disk->radius, disk->radius,
		 &theta, -disk->radius, disk->radius );

  }
  while( phi * phi + theta * theta > r2 );
  
  cos_theta = cos( theta );
  sin_theta = sin( theta );

  cos_phi = cos( phi );
  sin_phi = sin( phi );

  /* use trignometric identities to get values for theta0 + theta and
     phi0 + phi */

  cos_theta_p = disk->cos_theta0 * cos_theta - disk->sin_theta0 * sin_theta;
  cos_phi_p   = disk->cos_phi0   * cos_phi   - disk->sin_phi0   * sin_phi;

  sin_theta_p = disk->sin_theta0 * cos_theta + disk->cos_theta0 * sin_theta;
  sin_phi_p   = disk->sin_phi0   * cos_phi   + disk->cos_phi0   * sin_phi;

  direction->x = - cos_theta_p * sin_phi_p;
  direction->y = - sin_theta_p;
  direction->z =   cos_theta_p * cos_phi_p;

  return 0;
}

static void
cleanup( void *data )
{
  free ( data );
}
