/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <math.h>

#include <randlib_c/randlib.h>
#include <lua.h>
#include <luasup/luasup.h>
#include <suplib/str.h>
#include <suplib/debug.h>

#include <tracefct/tracefct.h>


#include "dist.h"
#include "lua_interface.h"
#include "misc.h"

static TokListToken
dist[] =
{
  "gaussian", DT_GAUSSIAN,
  "uniform",  DT_UNIFORM
};
static TokList dists = GenTokList( dist );


#define DISTTYPE "type"
#define FWHM 	 "fwhm"
#define SIGMA	 "sigma"
#define CENTER   "center"

#define DIST2D	 "dist"
#define XDIST	 "x"
#define YDIST	 "y"


static void
init_ImgDist( ImgDist *idist )
{
  idist->twoD = 0;
  idist->x.type = DT_UNIFORM;
  idist->y.type = DT_UNIFORM;
}

static void
parse_gauss( const char *err_pfx, Dist *dist,
	     int parse_center, lua_Object dist_tbl, double z )
{
  char *spec;
  double raw;
  
  if ( parse_center &&
       NULL != ( spec = slua_tbl_el_si( dist_tbl, CENTER ) ) )
  {
    slua_tbl_el_delete( dist_tbl, CENTER );
    parse_extent( err_pfx, SIGMA, spec, z, &dist->p[1], &raw );
  }
  else
    dist->p[0] = 0.0;

  if ( NULL != ( spec = slua_tbl_el_si( dist_tbl, SIGMA ) ) )
  {
    slua_tbl_el_delete( dist_tbl, SIGMA );
    parse_extent( err_pfx, SIGMA, spec, z, &dist->p[1], &raw );

    if ( dist->p[1] <= 0.0 )
      define_error( "%s: `%s' must be positive", err_pfx, SIGMA );
  }

  else if ( NULL != ( spec = slua_tbl_el_si( dist_tbl, FWHM ) ) )
  {
    slua_tbl_el_delete( dist_tbl, FWHM );
    parse_extent( err_pfx, SIGMA, spec, z, &dist->p[1], &raw );

    dist->p[1]  /= 2.354820054504;
    if ( dist->p[1] <= 0.0 )
      define_error( "%s: `%s' must be positive", err_pfx, FWHM );
  }
  else
    define_error( "must provide either `%s' or `%s' for Gaussian distribution",
		  SIGMA, FWHM );
}

static void
parse_dist( const char *err_pfx, Dist *dist, const char *coord, int twoD,
	  lua_Object optargs, double z )
{
  char *type;

  if ( slua_tbl_el_exists( optargs, coord ) )
  {
    lua_Object dist_tbl = slua_tbl_el_tbl( optargs, coord);

    if ( ! twoD )
    {
      if ( NULL != ( type = slua_tbl_el_si( dist_tbl, DISTTYPE ) ) )
      {
	slua_tbl_el_delete( dist_tbl, DISTTYPE );
	dist->type = tokmatch( type, &dists );
      }
      else
	dist->type = DT_UNIFORM;
    }

    switch ( dist->type )
    {
      case DT_GAUSSIAN:
	parse_gauss( err_pfx, dist, 1, dist_tbl, z );

      case DT_UNIFORM:
	break;

      default:
	define_error( "%s: unknown distribution: `%s'", err_pfx, type );
    }
    verify_empty_optargs( err_pfx, dist_tbl );
    slua_tbl_el_delete( optargs, coord );
  }
}

void
get_ImgDist( const char *err_pfx, ImgDist *idist, lua_Object optargs,
	     double z )
{
  init_ImgDist( idist );
  
  /* check for twoD distribution type */
  if ( slua_tbl_el_exists( optargs, DIST2D ) )
  {
    char *type;
    Dist dist;

    lua_Object dist_tbl = slua_tbl_el_tbl( optargs, DIST2D);

    idist->twoD = 1;

    if ( NULL != ( type = slua_tbl_el_si( dist_tbl, DISTTYPE ) ) )
    {
      slua_tbl_el_delete( dist_tbl, DISTTYPE );
      idist->x.type = idist->y.type = tokmatch( type, &dists );
    }
    else
    {
      idist->x.type = idist->y.type =  DT_UNIFORM;
    }

    switch( idist->x.type )
    {
    case DT_GAUSSIAN:
      idist->x.p[0] = idist->y.p[0] = 0.0;
      parse_gauss( err_pfx, &dist, 0, dist_tbl, z );
      idist->x.p[1] = idist->y.p[1] = dist.p[1] / sqrt(2.0);

      {
	double meanv[2]; 
	double covm[4];

	meanv[0] = idist->x.p[0];
	meanv[1] = idist->y.p[0];

	covm[0] = idist->x.p[1] * idist->x.p[1];
	covm[3] = idist->y.p[1] * idist->y.p[1];
	covm[1] = covm[2] = 0.0;

	setgmn( meanv, covm, 2, idist->gparm );
      }
      break;

    case DT_UNIFORM:
      break;

    default:
      define_error( "%s: unknown distribution: `%s'", err_pfx, idist->x.type );
    }

    verify_empty_optargs( err_pfx, dist_tbl );
    slua_tbl_el_delete( optargs, DIST2D );
  }
  
  parse_dist( err_pfx, &idist->x, XDIST, idist->twoD, optargs, z );
  parse_dist( err_pfx, &idist->y, YDIST, idist->twoD, optargs, z );
}

static double
gen_dist( Dist *dist, double min, double max )
{
  double val;
  
  switch( dist->type )
  {
  case DT_UNIFORM:
    val = genunf( min, max );
    break;

  case DT_GAUSSIAN:
    do
      val = gennor( dist->p[0], dist->p[1] );
    while ( val < min || val > max );
    break;

  default:
    tf_assert( 0 );
    break;
  }

  return val;
}

void
gen_ImgDist( ImgDist *idist,
	     double *x, double xmin, double xmax,
	     double *y,double ymin, double ymax )
{
  if ( idist-> twoD )
  {
    double p[2];
    double work[2];

    switch( idist->x.type )
    {
    case DT_UNIFORM:
      *x = genunf( xmin, xmax );
      *y = genunf( ymin, ymax );
      break;

    case DT_GAUSSIAN:
      do
	genmn( idist->gparm, p, work );
      while ( p[0] < xmin || p[0] > xmax ||
	      p[1] < ymin || p[1] > xmax );
      *x = p[0];
      *y = p[1];
      break;

    default:
      tf_assert(0);
      break;
    }
  }

  else
  {
    *x = gen_dist( &idist->x, xmin, xmax );
    *y = gen_dist( &idist->y, ymin, ymax );
  }

}
