/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>

#include <mathconst/mathconst.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <randlib_c/randlib.h>
#include <suplib/str.h>

#include "main.h"
#include "common.h"
#include "lua_interface.h"

/*
  the list of entrance apertures created by the lua routines.  This
  is eventually replaced by the "ea" array in the System structure.
*/
static SLinkList ea_list;

void
ea_init( void )
{
  if ( NULL == ( ea_list = sll_new( SLL_NULL_CMP ) ) )
    tf_exit( ExitERR_alloc, "unable to initialize entrance ap list" );

  /* create the entrance apertures */
  sys.ea_area = 0;
  process_lua( sys.ea_spec, sys.ea_override );
  if ( lua_dostring( "entrance_aperture()" ) )
    tf_exit( ExitERR_misc, "error creating entrance apertures" );

  sys.n_ea = sll_count( ea_list );
  /* make sure that there's at least one entrance aperture out there */
  if ( 0 == sys.n_ea )
    tf_exit( ExitERR_misc, "no entrance apertures were specified" );
  
  /* now, create an array of entrance apertures */
  if ( NULL ==
       ( sys.ea = (EntranceAperture *)
	 malloc( sys.n_ea * sizeof(EntranceAperture ) ) ) )
       tf_exit( ExitERR_alloc, "unable to create entrance ap array" );

  {
    SLLNode node;
    EntranceAperture *bin;
    double cum_ea = 0;

    for ( node = sll_head_node( ea_list ),
	    bin = sys.ea ;
	  node ;
	  node = sll_next_node( ea_list, node ),
	    bin++ )
    {
      EntranceAperture *ea = (EntranceAperture *) sll_node_get_data(node );
      *bin = *ea;
      bin->area_min = cum_ea / sys.ea_area;
      cum_ea += bin->area;
      bin->area_max = cum_ea / sys.ea_area;
    }

    /* free the list of EntranceApertures, but don't free the
       per-aperture data, which contains the aperture specific specs */
    sll_delete( ea_list, free );
  }
}

/* compare a probability against an EntranceAperture structure */
static int
ea_cmp( const void *p, const void *s )
{
  double prob = * (double *) p;
  const EntranceAperture *ea = (EntranceAperture *) s;

  return
    prob < ea->area_min ? -1 :
    prob >= ea->area_max ? 1 :
    0;
}

void
entrance_aperture( Photon *photon, double z )
{
  double prob, x_node_at_ea, y_node_at_ea, distance_from_node, distance_from_ea, radius;
  EntranceAperture *ea;

  /* which entrance aperture to use? */
  prob = ranf( );
  ea = (EntranceAperture *)
    bsearch( &prob, sys.ea, sys.n_ea, sizeof(EntranceAperture), ea_cmp );

  /* if we don't get one, this is a real problem. */
  tf_assert( NULL != ea );

  /* get position on entrance aperture */
  ea->pos( ea, &photon->position );

  /* adjust photon direction to account for it hitting away from the
     optical axis */
  {
    /* project the node back along the inital direction vector to the
       plane of the ea */
    distance_from_node =
      ( photon->position.z - sys.node ) / photon->direction.z;
    x_node_at_ea = distance_from_node * photon->direction.x;
    y_node_at_ea = distance_from_node * photon->direction.y;
    
    /* distance along Z axis times direction vector to get actual distance
       between source and ea */
    distance_from_ea = ( photon->position.z - z ) / photon->direction.z;
    photon->direction.x +=
      ( photon->position.x - x_node_at_ea ) / distance_from_ea;
    photon->direction.y +=
      ( photon->position.y - y_node_at_ea ) / distance_from_ea;
  }

  /* normalize it */
  radius = sqrt( photon->direction.x * photon->direction.x +
		 photon->direction.y * photon->direction.y +
		 photon->direction.z * photon->direction.z );

  photon->direction.x /= radius;
  photon->direction.y /= radius;
  photon->direction.z /= radius;
}


/***************************************************************/
/* Lua methods */


static void	circle_define( EntranceAperture *ea, const char *name,
			       int *next_arg );
static void	annulus_define( EntranceAperture *ea, const char *name,
				int *next_arg );
static void	wedge_define( EntranceAperture *ea, const char *name,
				int *next_arg );

static TokListToken
ea_type[] =
{
  { "annulus",		EA_Annulus	},
  { "circle",		EA_Circle	},
  { "wedge",		EA_Wedge	},
};
static TokList ea_types = GenTokList(ea_type);

void
ea_define_lua( void )
{
  EntranceAperture *ea;
  char *name;
  char *ea_type;
  int next_arg;

  tf_enter( "ea_define" );

  get_lua_args_n( "ea", 1,
		"name",		Lua_STRING,	&name,
		"type",		Lua_STRING,	&ea_type,
		NULL
		);

  next_arg = 3;

  if ( NULL ==
       ( ea = (EntranceAperture *) malloc( sizeof(EntranceAperture) ) ) )
    define_error(
      "ea: error allocating memory for entrance aperture `%s'", name );

  switch( ea->type = tokmatch( ea_type, &ea_types ))
  {
    case EA_Annulus:
      annulus_define( ea, name, &next_arg );
    break;

    case EA_Circle:
      circle_define( ea, name, &next_arg );
    break;

    case EA_Wedge:
      wedge_define( ea, name, &next_arg );
    break;
    
    default:
      define_error( "ea: illegal entrance aperture type: `%s'", ea_type );
  }

  check_n_lua_arg( "ea", next_arg - 1 );

  if ( sll_insert_head( ea_list, ea ) )
    define_error( "ea: error installing entrance aperture '%s'", name );

  free( ea_type );
  free( name );

  sys.ea_area += ea->area;

  tf_leave( );
}


/****************************************************************/

typedef struct Circle
{
  double r2;
  double area;
  DVector3 origin;
} Circle;

static void
circle_random_pos( EntranceAperture *ea, DVector3 *pos )
{
  Circle *circ = (Circle *) ea->data;
  double r = sqrt( circ->r2 * ranf() );
  double phi = M_2PI * ranf();

  pos->x = r * cos( phi ) + circ->origin.x;
  pos->y = r * sin( phi ) + circ->origin.y;
  pos->z = circ->origin.z;
}

static void
circle_define( EntranceAperture *ea, const char *name, int *next_arg )
{
  double r;
  
  Circle *circ = (Circle *) malloc( sizeof(Circle) );
  if ( NULL == circ )
    define_error("ea circle: error allocating memory for `%s'", name );

  get_lua_args_n( "ea circle", *next_arg,
		  "x",	Lua_NUMBER,	&circ->origin.x,
		  "y",	Lua_NUMBER,	&circ->origin.y,
		  "z",	Lua_NUMBER,	&circ->origin.z,
		  "r",	Lua_NUMBER,	&r,
		  NULL
		  );
  *next_arg += 4;

  if ( r <= 0.0 )
    define_error( "ea circle: r must be non-negative for '%s'", name );
  circ->r2 = r * r;

  switch( sys.ray_dist )
  {
  case RayDist_RANDOM:
    ea->pos = circle_random_pos;
    break;

  default:
    tf_assert( 0 );
  }
  ea->data = circ;
  /* convert effective area from mm^2 to cm^2 */
  ea->area = M_PI * circ->r2 / 100;
}

static void
circle_cleanup( Circle *data )
{
  free( data );
}

/****************************************************************/

typedef struct Annulus
{
  double ri2;		/* square of the inner radius */
  double dr2;		/* difference of the squared inner and outer
			   radii */
  double area;
  DVector3 origin;
} Annulus;

static void
annulus_random_pos( EntranceAperture *ea, DVector3 *pos )
{
  Annulus *ann = (Annulus *) ea->data;
  double r = sqrt( ann->ri2 + ann->dr2 * ranf() );
  double phi = M_2PI * ranf();

  pos->x = r * cos( phi ) + ann->origin.x;
  pos->y = r * sin( phi ) + ann->origin.y;
  pos->z = ann->origin.z;
}

static void
annulus_define( EntranceAperture *ea, const char *name, int *next_arg )
{
  double ri, ro;

  Annulus *ann = (Annulus *) malloc( sizeof(Annulus) );
  if ( NULL == ann )
    define_error("ea annulus: error allocating memory for `%s'", name );

  get_lua_args_n( "ea annulus", *next_arg,
		  "x",	Lua_NUMBER,	&ann->origin.x,
		  "y",	Lua_NUMBER,	&ann->origin.y,
		  "z",	Lua_NUMBER,	&ann->origin.z,
		  "ri",	Lua_NUMBER,	&ri,
		  "ro",	Lua_NUMBER,	&ro,
		  NULL
		  );

  *next_arg += 5;
  if ( ri < 0.0 )
    define_error( "ea annulus: ri must be non-negative for '%s'", name );

  if ( ro <= 0.0 )
    define_error( "ea annulus: ro must be positive for '%s'", name );
  ann->ri2 = ri * ri;
  ann->dr2 = ro * ro - ann->ri2;

  switch( sys.ray_dist )
  {
  case RayDist_RANDOM:
    ea->pos = annulus_random_pos;
    break;

  default:
    tf_assert( 0 );
  }
  ea->data = ann;
  /* convert effective area from mm^2 to cm^2 */
  ea->area = M_PI * ann->dr2 / 100;
}

static void
annulus_cleanup( Annulus * data )
{
  free( data );
}


/****************************************************************/

typedef struct Wedge
{
  double theta_min;
  double theta_max;

  double ri2;
  double dr2;
  
  double area;
  DVector3 origin;
} Wedge;

static void
wedge_random_pos( EntranceAperture *ea, DVector3 *pos )
{
  Wedge *wedg = (Wedge *) ea->data;
  double r = sqrt( wedg->ri2 + wedg->dr2 * ranf() );
  double phi = ( wedg->theta_max - wedg->theta_min ) * ranf() + wedg->theta_min;

  pos->x = r * cos( phi ) + wedg->origin.x;
  pos->y = r * sin( phi ) + wedg->origin.y;
  pos->z = wedg->origin.z;
}

static void
wedge_define( EntranceAperture *ea, const char *name, int *next_arg )
{
  double theta_min, theta_max, ri, ro;
  
  Wedge *wedg = (Wedge *) malloc( sizeof(Wedge) );
  if ( NULL == wedg )
    define_error("ea wedge: error allocating memory for `%s'", name );

  get_lua_args_n( "ea wedge", *next_arg,
		  "x",	        Lua_NUMBER,	&wedg->origin.x,
		  "y",	        Lua_NUMBER,	&wedg->origin.y,
		  "z",   	Lua_NUMBER,	&wedg->origin.z,
		  "ri",         Lua_NUMBER,     &ri,
		  "ro",         Lua_NUMBER,     &ro,
		  "theta_min",	Lua_NUMBER,	&theta_min,
		  "theta_max",  Lua_NUMBER,     &theta_max,
		  NULL
		  );
  *next_arg += 7;

  if ( ri < 0.0 )
    define_error( "ea wedge: ri must be non-negative for '%s'", name );
  
  if ( ro <= 0.0 )
    define_error( "ea wedge: ro must be positive for '%s'", name );
 
  wedg->ri2 = ri * ri;
  wedg->dr2 = ro * ro - wedg->ri2;
  wedg->theta_min = theta_min;
  wedg->theta_max = theta_max;
  
  /* remove 2 PI degeneracy */
  wedg->theta_min = fmod( wedg->theta_min, M_2PI );
  wedg->theta_max = fmod( wedg->theta_max, M_2PI );

  /* if the wedge straddles the discontinuity, make it continuous */
  if ( wedg->theta_min > wedg->theta_max ) wedg->theta_max += M_2PI;

  switch( sys.ray_dist )
  {
  case RayDist_RANDOM:
    ea->pos = wedge_random_pos;
    break;

  default:
    tf_assert( 0 );
  }
  ea->data = wedg;
  /* convert effective area from mm^2 to cm^2 */
  ea->area = ( wedg->theta_max - wedg->theta_min ) / M_PI * wedg->dr2 / 100;
}

static void
wedge_cleanup( Wedge *data )
{
  free( data );
}


void
ea_cleanup( void )
{
  if ( sys.ea )
  {
    int i;
    for ( i = 0 ; i < sys.n_ea ; i++ )
    {
      switch( sys.ea[i].type )
      {
      case EA_Annulus:
	annulus_cleanup( sys.ea[i].data );
	break;

      case EA_Circle:
	circle_cleanup( sys.ea[i].data );
	break;

      case EA_Wedge:
	wedge_cleanup( sys.ea[i].data );
	break;

      default:
	tf_assert( 0 );
      }
    }
    free( sys.ea );
  }
}

