/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <linklist/linklist.h>

#include "main.h"
#include "common.h"

#define SERVICE void (*)(void *, double, void *, void *)
void
register_event(
  char *	mod_name,
  EventType	event_type,
  void *	module,			/* the associated module */
  double	event_time,		/* the time (in seconds) of the
					   event */
  double	current_time,		/* the current time */
  void *	module_data,		/* module specific data */
  void *	event_data,		/* event specific data */
  void (*service)(void *module, double time,
		  void *module_data,
		  void *event_data),	/* the event servicing
					   routine */
  void (*del) (void *event_data ),	/* delete event data */
  void (*cleanup)( Event *event ),	/* module specific event cleanup
					   routine */
  SLinkList *	queue			/* the event queue */
)
{
  Event *event;

  if ( sll_count( sys.spare_events ) )
  {
    SLLNode node  = sll_detach_head_node( sys.spare_events );

    event = (Event *) sll_node_get_data( node );

    event->time = event_time;
    event->type = event_type;
    event->module = module;
    event->service = service;
    event->del = del;
    event->cleanup = cleanup;
    event->event_data = event_data;
    event->module_data = module_data;
    sll_insert_node( queue, node );
  }
  else
  {
    if ( NULL == (event = (Event *) malloc( sizeof(Event) )) )
      tf_exit( ExitERR_alloc, "%s: error creating event at time %d",
                            mod_name, current_time );

    event->time = event_time;
    event->type = event_type;
    event->module = module;
    event->service = service;
    event->del = del;
    event->cleanup = cleanup;
    event->event_data = event_data;
    event->module_data = module_data;

    if ( sll_insert( queue, event ) )
      tf_exit( ExitERR_alloc, "%s: error creating event at time %d",
                            mod_name, current_time );
  }
}

void
delete_event( void *event )
{
  ((Event *)event)->del( ((Event *)event)->event_data );
  free( event );
}

void
register_flux_gen_event(
  FluxGenerator *	gen,		/* the associated flux generator */
  double		time,		/* the time (in seconds) of the
					   event */
  void *		data,		/* event specific data */
  void (*service)(void *gen, double time,
		  FluxGeneratorData *module_data,
		  void *event_data),	/* the event servicing
					   routine */
  void (*del) (void *event_data )	/* delete event data */
)
{
  register_event( gen->name, Event_FLUX, gen, time, gen->time, gen->data.data,
		  data, (SERVICE) service, del,
		  cleanup_flux_gen_event, gen->events );
}

void
register_flux_modifier_event(
  FluxModifier *mod,			/* the associated flux modifier */
  double time,				/* the time (in seconds) of the
					   event */
  void *data,				/* event specific data */
  void (*service)(void *mod, double time,
		  void *module_data,
		  void *event_data),	/* the event servicing
					   routine */
  void (*del) (void *event_data )	/* delete event data */
)
{
  register_event( mod->name, Event_FLUX, mod, time, mod->gen->time, mod->data,
		  data, service, del, cleanup_flux_modifier_event,
		  mod->gen->events );
}

void
register_extent_event(
  PositionGenerator *gen,		/* the associated position generator */
  double time,				/* the time (in seconds) of the
					   event */
  void *data,				/* event specific data */
  void (*service)(void *gen, double time,
		  PositionGenerator *module_data,
		  void *event_data),	/* the event servicing
					   routine */
  void (*del) (void *event_data )	/* delete event data */
)
{
  SLLNode node;

  /* register event in source extent event list */
  register_event( gen->name, Event_EXTENT, gen, time, gen->source->time,
		  &gen->data, data, (SERVICE) service, del,
		  cleanup_extent_event,
		  gen->source->extent_events );

  /* register the event in each of the source's flux generator event
     lists as well */
  for ( node = sll_head_node( gen->source->flux_gen ) ;
        node ;
        node = sll_next_node( gen->source->flux_gen, node )
      )
    register_event( gen->name, Event_EXTENT, gen, time, gen->source->time,
		    gen->data.data, data, (SERVICE) service, del,
		    cleanup_extent_event,
		    ((FluxGenerator *) sll_node_get_data( node ))->events );
}

int
event_cmp(
  const void *event1,
  const void *event2
)
{
  if      ( ((Event *) event1)->time > ((Event *) event2)->time )
    return 1;
  else if ( ((Event *) event1)->time < ((Event *) event2)->time )
    return -1;

  return 0;
}
