/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <suplib/str.h>
#include <linklist/linklist.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

void
new_flux_gen(
  FluxGeneratorSpec *spec,
  char *	name,
  void *	data
)
{
  FluxGenerator *gen;
  Source *source = sys.current_source;

  tf_enter( "new_flux_gen" );

  /* ensure that we're creating a source */
  if ( NULL == sys.current_source )
    tf_exit( ExitERR_misc, "flux_gen `%s': not part of a source", name );

  if ( NULL == (gen = (FluxGenerator *) malloc( sizeof(FluxGenerator) )) )
    tf_exit( ExitERR_alloc, "%s: unable to create flux generator",
				sys.current_source->name );

  memcpy( &gen->spec, spec, sizeof(FluxGeneratorSpec) );
  gen->spec.name = str_dup( spec->name );
  gen->name = str_join( "/", 2, sys.current_source->name, name );

  gen->source		= source;
  gen->n_mods		= 0;
  gen->data.data	= data;
  gen->data.iter_lim	= ULONG_MAX;
  gen->data.iter_warn	= ULONG_MAX;

  gen->flux_factors	= NULL;

  /* create a list of flux modifiers */
  if ( NULL == (gen->mods = sll_new( SLL_NULL_CMP )) )
    tf_exit( ExitERR_alloc, "%s: unable to allocate flux modifier list", gen->name );

  /* create a list of events */
  if ( NULL == ( gen->events = sll_new( event_cmp ) ) )
    tf_exit( ExitERR_alloc, "%s: couldn't create flux generator event queue" , gen->name );

  /* install generator in the current source's list */
  if ( sll_insert_head( source->flux_gen, gen ) )
    tf_exit( ExitERR_alloc, "%s: error installing flux generator", gen->name );

  /* install generator in the current source's invalid photon list */
  if ( sll_insert_head( source->phot_invalid, gen ) )
    tf_exit( ExitERR_alloc, "%s: error installing photon", gen->name );

  gen->time = 0.0;
  gen->id = ++source->n_gen;

  sys.current_fgen = gen;

  tf_leave( );
}

void
delete_flux_gen( void *data )
{
  FluxGenerator *gen = (FluxGenerator *) data;

  if ( gen->n_mods )
    free( gen->flux_factors );

  sll_delete( gen->mods, delete_flux_modifier );

  sll_delete( gen->events, delete_event );

  free( gen->spec.name );
  free( gen->name );

  gen->spec.cleanup( gen->data.data );

  free( gen );
}

int
init_flux_gen( void *data )
{
  FluxGenerator *gen = (FluxGenerator *) data;

  tf_enter( "init_flux_gen" );

  if ( gen->spec.init( &gen->data ) )
    tf_exit( ExitERR_misc, "%s: error initializing flux generator",
	     gen->name );

  if ( gen->n_mods )
  {
    if ( NULL == (gen->flux_factors =
		    (double *) malloc( gen->n_mods * sizeof(double) )) )
      tf_exit( ExitERR_misc, "%s: error creating flux modifier array",
	       gen->name );

    gen->flux_factors_lim = gen->flux_factors + gen->n_mods;

    sll_traverse( gen->mods, (SLL_ACTION) init_flux_modifier );
  }

  calculate_eff_flux( gen );

  tf_leave( );
  return 0;
}

void
calculate_eff_flux(
  FluxGenerator *gen
)
{
  double *flux_factor;

  gen->flux_factor = gen->source->pos_gen->data.factor;

  if ( gen->flux_factors )
    for ( flux_factor = gen->flux_factors ;
	  flux_factor < gen->flux_factors_lim ;
	  flux_factor++ )
      gen->flux_factor *= *flux_factor;

  gen->eff_flux = gen->data.base_flux * gen->flux_factor;
}

void
cleanup_flux_gen_event(
  Event *event
)
{
  FluxGenerator *gen = (FluxGenerator *) event->module;

  gen->eff_flux = gen->data.base_flux * gen->flux_factor;
}

int
set_flux_gen_time(
  void *data,		/* a pointer to a flux generator */
  void *udata		/* the time */
)
{
  ((FluxGenerator *) data)->time = * (double *) udata;
  return 0;
}

int
flux_gen_time_cmp(
  const void *gen1,
  const void *gen2
)
{
  if      ( ((FluxGenerator *) gen1)->time > ((FluxGenerator *) gen2)->time )
    return 1;
  else if ( ((FluxGenerator *) gen1)->time < ((FluxGenerator *) gen2)->time )
    return -1;

  return 0;
}
