/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <string.h>

#include <suplib/str.h>
#include <suplib/debug.h>

#include <tparam/tparam.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

typedef enum
{
  LT_ksec,
  LT_sec,
  LT_rmm2,
  LT_rcm2,
  LT_krays,
  LT_Mrays,
  LT_rays
} InputLimitType;

/* limits */
static TokListToken
limit_type[] =
{
  { "Mrays",	LT_Mrays},
  { "krays",	LT_krays},
  { "ksec",	LT_ksec	},
  { "r/cm2",	LT_rcm2  },
  { "r/mm2",	LT_rmm2	},
  { "rays",	LT_rays },
  { "sec",	LT_sec	},
};
static TokList limit_types = GenTokList(limit_type);

/* ray distribution at entrance aperture */
static TokListToken
ray_dist_type[] =
{
  { "random",	RayDist_RANDOM	},
};
static TokList ray_dist_types = GenTokList(ray_dist_type);

static char *
getstr(paramfile pf, char *pname, int abort_if_empty)
{
  char string[NAME_LENGTH+1];
  char *copy;
  tpgetstr( pf, pname, string, NAME_LENGTH, abort_if_empty );
  if ( NULL == ( copy = str_dup( string ) ) )
    tf_exit( ExitERR_alloc, "couldn't allocate space for %s", pname );

  return copy;
}

int
gparam(int argc, char *argv[])
{
  paramfile pf;
  char dummy[NAME_LENGTH+1];

  tf_enter( "gparam" );

  pf = tparamopen( NULL, argv, argc, "r" );

  /* only get other stuff if version isn't set */
  if ( tpgetb( pf, "version" ) )
    return 1;

  /* only get other stuff if version isn't set */
  if ( paccess(pf, "help" ) && tpgetb( pf, "help" ) )
    return 2;

  /* get debug flags */
  tpgetstr( pf, "debug", debug, NAME_LENGTH, 0 );
  debug_init( debug );

  /* hardwire the mirror info for now */
  /*
    sys.mirror.z_f   = 59.9186;
    sys.mirror.z_a   = 902.1106;
    sys.mirror.rho_f = 598.5083;
    sys.mirror.rho_a = 560.6541;
    sys.mirror.l     = 842.1920;
    sys.mirror.node  = 24.9555;
  */
  /* hardwire the mirror info for now */
  
  sys.source_spec = getstr( pf, "source", 1 );
  sys.source_override = getstr( pf, "source_override",  0 );
  str_trim( sys.source_override );

  sys.ea_spec = getstr( pf, "ea", 1 );
  sys.ea_override = getstr( pf, "ea_override",  0 );
  str_trim( sys.ea_override );

  sys.output = getstr( pf, "output", 1 );

  sys.logfile = getstr( pf, "logfile", 0 );

  sys.limit = tpgetd( pf, "limit" );

  tpgetstr( pf, "limit_type", dummy, NAME_LENGTH, 1 );
  switch( tokmatch( dummy, &limit_types ) )
  {
  case LT_ksec:
    sys.limit_type = Limit_TIME;
    sys.limit *= 1000;
    break;
    
  case LT_sec:
    sys.limit_type = Limit_TIME;
    break;

  case LT_rmm2:
    sys.limit_type = Limit_DENSITY;
    sys.limit *= 100;	/* convert to rays / cm^2 */
    break;

  case LT_rcm2:
    sys.limit_type = Limit_DENSITY;
    break;

  case LT_Mrays:
    sys.limit_type = Limit_COUNTS;
    sys.limit *= 1e6;
    break;

  case LT_krays:
    sys.limit_type = Limit_COUNTS;
    sys.limit *= 1000;
    break;

  case LT_rays:
    sys.limit_type = Limit_COUNTS;
    break;

  default:
    tf_exit( ExitERR_illegalArg, "illegal limit type: `%s'", dummy );
  }

  tpgetstr( pf, "ray_dist", dummy, NAME_LENGTH, 1 );
  if ( -1 == (sys.ray_dist = tokmatch( dummy, &ray_dist_types )) )
    tf_exit( ExitERR_illegalArg, "illegal ray distribution type: `%s'", dummy );

  sys.node = tpgetd( pf, "node" );
  
  sys.seed1 = tpgeti(pf, "seed1");
  if (sys.seed1 < 1 || sys.seed1 > 2147483562)
    tf_exit(ExitERR_illegalArg, "seed1 must have range [1,2147483562]");

  sys.seed2 = tpgeti(pf, "seed2");
  if (sys.seed2 < 1 || sys.seed2 > 2147483398)
    tf_exit(ExitERR_illegalArg, "seed2 must have range [1,2147483398]");

  sys.block = tpgeti(pf, "block");
  if (sys.block < 0 || sys.block > 1048575)
    tf_exit(ExitERR_illegalArg, "block must have range [0,1048575]");

  paramclose( pf );
  tf_leave( );

  return 0;
}


void
cleanup_params( void )
{
  free( sys.source_spec );
  free( sys.source_override );
  free( sys.ea_spec );
  free( sys.ea_override );
  free( sys.output );
  free( sys.logfile );

}
