/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <float.h>
#include <limits.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <mathconst/mathconst.h>
#include <randlib_c/randlib.h>

#include <suplib/debug.h>
#include <suplib/str.h>
#include <suplib/lists.h>
#include <suplib/units.h>
#include <suplib/units_linear.h>
#include <suplib/units_angular.h>

#include <luasup/luasup.h>

#include "main.h"
#include "args.h"
#include "lua_interface.h"
#include "image.h"
#include "misc.h"


typedef enum
{
  DF_ERROR,
  DF_PBM,
  DF_PGM,
  DF_RAW_ASCII,
  DF_RAW_FLOAT,
  DF_RAW_DOUBLE,
  DF_FITS
} DataFormat;

static TokListToken
format_type[] =
{
  { "ascii",		DF_RAW_ASCII	},
  { "double",		DF_RAW_DOUBLE	},
  { "fits",		DF_FITS 	},
  { "float",		DF_RAW_FLOAT	},
  { "pbm",		DF_PBM		},
  { "pgm",		DF_PGM		},
};
static TokList format_types = GenTokList(format_type);

typedef enum
{
  FLIP_X  = 0x01,
  FLIP_Y  = 0x10,
  FLIP_XY = 0x11
} FlipWhat;

static TokListToken
flip_type[] =
{
  { "x",	FLIP_X },
  { "y",	FLIP_Y },
  { "xy",	FLIP_XY },
};
static TokList flip_types = GenTokList(flip_type);


typedef struct
{
  double x;		/* x coordinate of pixel */
  double y;		/* y coordinate of pixel */
  double p_min;		/* minimum scaled probability of pixel */
} Pixel;

typedef struct
{
  char *	file;		/* the file to read */
  char *	name;		/* the name of the position generator */
  double	clip;		/* minimum value of a pixel */
  DataFormat	fmt;		/* format of the data */
  double	z;		/* z coord of rect */
  Pixel	*	pixel;		/* the data */
  Axis		x;
  Axis		y;
  size_t	npix;		/* number of pixels */
  double	theta;		/* off axis angle of image */
  double	phi;		/* off axis angle of image */

  double	sin_theta0;
  double	cos_theta0;
  double	sin_phi0;
  double	cos_phi0;
} ImageData;


static void	define(void);
static int	init(PositionGeneratorData *data);
static int	generate(void *data, DVector3 *position, double *z);
static void	cleanup(void *data);

Module image_module = { "image", define };

static void	init_Axis( Axis *axis );
static DataFormat	guess_format( const char *image, int *nx, int *ny );
static int	parse_pixsz( const char *err_pfx, lua_Object optargs, const char *which,
			     double z, double *pixsz, double *pixsz_raw );

static PositionGeneratorSpec spec =
{
  "image",
  init,
  generate,
  cleanup
};

static void
define( void )
{
  lua_Object optargs = LUA_NOOBJECT;
  ImageData *data;
  char *name;
  char err_pfx[1024];

  tf_enter( "image" );

  slua_args( define_error, "image",
	     "name",		SLua_STRING,	&name,
	     "optargs",	SLua_TABLE | SLua_OPTIONAL,	&optargs,
	     NULL
	     );

  /* generate error message prefix */
  sprintf( err_pfx, "image `%s'", name );

  /* allocate space for the image information, and initialize it */
  if ( NULL == ( data = (ImageData *) malloc( sizeof(ImageData) ) ) )
    define_error("%s: error allocating memory", err_pfx );

  data->pixel	= NULL;
  data->name	= name;
  init_Axis( &data->x );
  init_Axis( &data->y );

  /* copy optargs, so can delete elements from it as we go along */
  optargs = optargs == LUA_NOOBJECT ? lua_createtable() : slua_tbl_copy( optargs );

  /**********************************************************************
    Image File Name

    For now, this is required.  later one will be able to generate or
    reuse internal images

  */
  switch( slua_tbl_el_s( optargs, FILENAME, &data->file ) )
  {
  case 0:
    slua_tbl_el_delete( optargs, FILENAME );
    break;

  case -1:
    define_error( "%s: out of memory allocating file name", err_pfx );
    /* NOT REACHED */
    break;

  case 1:
    define_error( "%s: `" FILENAME "' attribute must be specified", err_pfx );
    /* NOT REACHED */
    break;
  }

  /* generate new error message prefix, include file name */
  sprintf( err_pfx, "image `%s': file `%s':", name, data->file );

  /**********************************************************************
   Image size (format dependent)

   This need only be specified as an attribute if it's not in the actual
   image file.  We may need this information to guess at the image file
   type.

   Normlly we'd delete the from the optargs list, but it's an error
   to specify it as an attribute as well as in the image header.  We need
   to keep it around so that when processing image headers we can flag
   the error.

  */
  if ( slua_tbl_el_exists( optargs, NPIX_X ) )
  {
    double npix;
    slua_tbl_el_n( optargs, NPIX_X, &npix );
    data->x.n = npix;
  }

  if ( slua_tbl_el_exists( optargs, NPIX_Y ) )
  {
    double npix;
    slua_tbl_el_n( optargs, NPIX_Y, &npix );
    data->y.n = npix;
  }

  /**********************************************************************
   Image Format
  */

  {
    char *format = slua_tbl_el_si( optargs, FORMAT );
    /* if not specified, see if can auto detect it */
    if ( format )
    {
      if ( -1 == (data->fmt = tokmatch( format, &format_types )) )
	define_error( "%s: illegal format type: `%s'",
		      err_pfx, format );

      slua_tbl_el_delete( optargs, FORMAT );
    }

    /* didn't find it */
    else
    {
      if ( DF_ERROR ==
	   ( data->fmt = guess_format( data-> file, &data->x.n, &data->y.n ) ) )
	define_error( "%s: unable to determine image format; please specify",
		      err_pfx );
    }
  }

  /**********************************************************************
   Axial Source Position

   It must be determined before the pixel size is calculated below.
  */

  get_src_z( err_pfx, optargs, &data->z );

  /**********************************************************************
   Read Image header for info

  */
  switch( data->fmt )
  {
  case DF_PBM:
  case DF_PGM:
    if ( slua_tbl_el_exists( optargs, NPIX_X ) ||
	 slua_tbl_el_exists( optargs, NPIX_Y ) )
      define_error( "%s: don't specify image size in optargs for PGM/PBM images",
		    err_pfx );
    rd_image_hdr_pxm( err_pfx, data->file, &data->x.n, &data->y.n );
    break;

  case DF_FITS:
    if ( slua_tbl_el_exists( optargs, NPIX_X ) ||
	 slua_tbl_el_exists( optargs, NPIX_Y ) )
      define_error( "%s: don't specify image size in optargs for FITS images",
		    err_pfx );
    rd_image_hdr_fits( err_pfx, data->file, data->z, &data->x, &data->y );
    break;
    
  case DF_RAW_ASCII:
  case DF_RAW_FLOAT:
  case DF_RAW_DOUBLE:
    if ( data->x.n == INT_MIN || data->y.n == INT_MIN )
      define_error( "%s: must specify image size for RAW image formats",
		    err_pfx );
    break;

  default:
    define_error( "%s: internal error at %s: %d", err_pfx, __FILE__, __LINE__ );
  }

  slua_tbl_el_delete( optargs, NPIX_X );
  slua_tbl_el_delete( optargs, NPIX_Y );

  if ( data->x.n <= 0 || data->y.n <= 0 )
    define_error( "%s: image dimensions  (%d, %d) must be positive",
		  err_pfx, data->x.n, data->y.n );

  /**********************************************************************
   Angular Source Position
  */
  get_src_pos( err_pfx, optargs, &data->theta, &data->phi );
  {
    double x, y, z;
    /* generate positions on unit sphere of center of image in (x,y,z)
       from original (theta, phi) (in tilted lat,long) */
    x = sin( data->theta ) * cos( data->phi );
    y = sin( data->theta ) * sin( data->phi );
    z = cos( data->theta );

    /* generate sin and cos of (lat,long) from (x,y,z) */
    data->sin_theta0 = y;
    data->cos_theta0 = sqrt( x*x + z*z );
    data->sin_phi0 = x / data->cos_theta0;
    data->cos_phi0 = z / data->cos_theta0;
  }
  
  /**********************************************************************
   Flip Image?
  */
  {
    char *flip = slua_tbl_el_si( optargs, FLIP );
    if ( flip )
    {
      int flip_type;
      if ( -1 == (flip_type = tokmatch( flip, &flip_types )) )
	define_error( "%s: illegal image flip type: `%s'",
		      err_pfx, flip );
      slua_tbl_el_delete( optargs, FLIP );

      
      /* be careful here about how flipping is done.  the image format
	 (in particular FITS images with RA&Dec) may have already had
	 a hidden flip; a user request to flip the image should flip
	 the flip, not set it */
      if ( flip_type & FLIP_X ) data->x.flip = 1 - data->x.flip;
      if ( flip_type & FLIP_Y ) data->y.flip = 1 - data->y.flip;
    }
  }
  
  /**********************************************************************
   Clipping Value
  */
  data->clip	= 0.0;
  slua_tbl_el_n( optargs, CLIP_LIM, &data->clip );
  slua_tbl_el_delete( optargs, CLIP_LIM );
  if ( data->clip < 0 )
    define_error( "image: clip value must be non-negative" );

  
  /**********************************************************************
   Pixel sizes (format dependent)

    while PBM, PGM and the RAW formats all must have a pixel size
    specified, allow FITS images to have their pixel sizes over-ridden
  */
  parse_pixsz( err_pfx, optargs, PIXSZ_X, data->z, &data->x.pix_sz, &data->x.pix_sz_raw );
  parse_pixsz( err_pfx, optargs, PIXSZ_Y, data->z, &data->y.pix_sz, &data->y.pix_sz_raw );
  if ( parse_pixsz( err_pfx, optargs, PIXSZ, data->z, &data->x.pix_sz,
		    &data->x.pix_sz_raw ) )
  {
    data->y.pix_sz = data->x.pix_sz;
    data->y.pix_sz_raw = data->x.pix_sz_raw;
  }

  data->x.pix_delta = data->x.pix_sz;
  data->y.pix_delta = data->y.pix_sz;

  if ( data->x.flip ) data->x.pix_sz *= -1;
  if ( data->y.flip ) data->y.pix_sz *= -1;

  /* check pixel size */
  if ( -DBL_MAX == data->x.pix_sz || -DBL_MAX == data->y.pix_sz ) 
    define_error( "%s: pixel dimensions must be specified", err_pfx );

  if ( 0 == data->x.pix_sz || 0 == data->y.pix_sz ) 
    define_error( "%s: pixel dimensions (%f,%f) must be non-zero", err_pfx,
		  data->x.pix_sz, data->y.pix_sz);

  
  /**********************************************************************
   The position in the image corresponding to the specified Angular
   Source Position

   The user may specify either the pixel index (which may be fractional)
   or the world coordinates of the center.  In the latter case,
   the relationship of the world coordinates must also be specified,
   either via the world coordinate of the pixel
   with indices (1,1), or the pixel with world coordinate (0,0).
  */

  {
    int t;

    /* first check if it's been specified directly */
    if ( (t = slua_tbl_el_exists( optargs, POINT_IDX_X ) +
	  slua_tbl_el_exists( optargs, POINT_IDX_Y ))
	 )
    {
      if ( 2 != t )
	define_error( "%s: must specify both `%s' and `%s'",
		      err_pfx, POINT_IDX_X, POINT_IDX_Y );
      else
      {
	slua_tbl_el_n( optargs, POINT_IDX_X, &data->x.pnt_idx);
	slua_tbl_el_n( optargs, POINT_IDX_Y, &data->y.pnt_idx);
	slua_tbl_el_delete( optargs, POINT_IDX_X );
	slua_tbl_el_delete( optargs, POINT_IDX_Y );
      }
    }

    /* maybe the wcs coordinate is specified? */
    else  if ( ( t = slua_tbl_el_exists( optargs, POINT_WCS_X ) +
		 slua_tbl_el_exists( optargs, POINT_WCS_Y ) )
	       )
    {
      if ( 2 != t )
	define_error( "%s: must specify both `%s' and `%s'",
		      err_pfx, POINT_WCS_X, POINT_WCS_Y );

      slua_tbl_el_n( optargs, POINT_WCS_X, &data->x.pnt_wcs);
      slua_tbl_el_n( optargs, POINT_WCS_Y, &data->y.pnt_wcs);
      slua_tbl_el_delete( optargs, POINT_WCS_X );
      slua_tbl_el_delete( optargs, POINT_WCS_Y );

      /* must also specify mapping from WCS to pixel indices */
      if ( ( t = slua_tbl_el_exists( optargs, REF_WCS_X ) +
	     slua_tbl_el_exists( optargs, REF_WCS_Y ) )
	   )
      {
	if ( 2 != t )
	  define_error( "%s: must specify both `%s' and `%s'",
			err_pfx, REF_WCS_X, REF_WCS_Y );
	  
	slua_tbl_el_n( optargs, REF_WCS_X, &data->x.ref_wcs);
	slua_tbl_el_n( optargs, REF_WCS_Y, &data->y.ref_wcs);
	slua_tbl_el_delete( optargs, REF_WCS_X );
	slua_tbl_el_delete( optargs, REF_WCS_Y );
      }

      if ( (t = slua_tbl_el_exists( optargs, REF_IDX_X ) +
	    slua_tbl_el_exists( optargs, REF_IDX_Y ) )
	   )
      {
	if ( 2 != t )
	  define_error( "%s: must specify both `%s' and `%s'",
			err_pfx, REF_IDX_X, REF_IDX_Y );
	slua_tbl_el_n( optargs, REF_IDX_X, &data->x.ref_idx);
	slua_tbl_el_n( optargs, REF_IDX_Y, &data->y.ref_idx);
	slua_tbl_el_delete( optargs, REF_IDX_X );
	slua_tbl_el_delete( optargs, REF_IDX_Y );
      }

      /* now check that we've enough info.  it's possible that
	 some of this came from the image header, so we can't
	 assume that if it's not in the optargs list it wasn't
	 specified */

      if ( -DBL_MAX == data->x.ref_wcs  || -DBL_MAX == data->y.ref_wcs )
	define_error( "%s: must specify both %s and %s as well as %s and %s",
		      err_pfx, REF_WCS_X, REF_WCS_Y, POINT_WCS_X, POINT_WCS_Y );
	
      if ( -DBL_MAX == data->x.ref_idx  || -DBL_MAX == data->y.ref_idx )
	define_error( "%s: must specify both %s and %s as well as %s and %s",
		      err_pfx, REF_IDX_X, REF_IDX_Y, POINT_WCS_X, POINT_WCS_Y );


      /* now, calculate what we really want, which is POINT_IDX_X
	 and POINT_IDX_Y */
      data->x.pnt_idx = data->x.ref_idx +
	( data->x.pnt_wcs - data->x.ref_wcs ) / data->x.pix_sz_raw;

      data->y.pnt_idx = data->y.ref_idx +
	( data->y.pnt_wcs - data->y.ref_wcs ) / data->y.pix_sz_raw;
    }

    /*
      nothing was specified, so just use the center of the image.  note that the index of the center
      of the lower left pixel is (1,1)
     */
    else
    {
      data->x.ref_idx = 1.0;
      data->x.ref_wcs = 1.0;
      data->x.pnt_idx = (data->x.n + 1) / 2;

      data->y.ref_idx = 1.0;
      data->y.ref_wcs = 1.0;
      data->y.pnt_idx = (data->y.n + 1) / 2;
    }
  }  


  /* we should have read (and deleted) all relevant optargs.  Those
     that remain should be flagged as errors */
  verify_empty_optargs( err_pfx, optargs );

  /* inform the ray generator of the new position generator
     available */
  new_pos_gen( &spec, data->name, data );

  tf_leave( );
}

static int
parse_pixsz( const char *err_pfx, lua_Object optargs, const char *which,
	   double z, double *pixsz, double *pixsz_raw )
{
  char *ptr;

  if ( (ptr = slua_tbl_el_si( optargs, which )) ) 
  {
    parse_extent( err_pfx, which, ptr, z, pixsz, pixsz_raw );
    slua_tbl_el_delete( optargs, which );
    return 1;
  }
  return 0;
}

static void
dump_Axis( const char*which, Axis *axis )
{
  tf_message( " %s.n: %d\n", which, axis->n );
  tf_message( " %s.pix_sz_raw: %g\n", which, axis->pix_sz_raw );
  tf_message( " %s.pix_sz: %g\n", which, axis->pix_sz );
  tf_message( " %s.pix_delta: %g\n", which, axis->pix_delta );
  tf_message( " %s.flip: %d\n", which, axis->flip );
  tf_message( " %s.ref_idx: %g\n", which, axis->ref_idx );
  tf_message( " %s.ref_wcs: %g\n", which, axis->ref_wcs );
  tf_message( " %s.pnt_idx: %g\n", which, axis->pnt_idx );
  tf_message( " %s.pnt_wcs: %g\n", which, axis->pnt_wcs );
}

static void
dump_ImageData( ImageData *image )
{
  tf_message( "Image %s:\n", image->name );
  tf_message( " file: %s\n", image->file );
  tf_message( " clip: %f\n", image->clip );
  tf_message( " fmt : %d\n", image->fmt );
  tf_message( " z   : %f\n", image->z );
  tf_message( " npix: %d\n", image->npix );
  tf_message( " theta: %f\n", image->theta );
  tf_message( " phi : %f\n", image->phi );
  tf_message( " sin_theta0: %f\n", image->sin_theta0 );
  tf_message( " sin_phi0 : %f\n", image->sin_phi0 );
  tf_message( " cos_theta0: %f\n", image->cos_theta0 );
  tf_message( " cos_phi0 : %f\n", image->cos_phi0 );
  dump_Axis( "x", &image->x );
  dump_Axis( "y", &image->y );
  
}


static int
init( PositionGeneratorData *pgdata )
{
  int x, y;
  double *datap, *data, *data_lim;
  char err_pfx[1024];

  ImageData *image = (ImageData *) pgdata->data;

  double p_sum, p_min;
  Pixel *pixel;

  sprintf( err_pfx, "image `%s': file `%s'", image->name, image->file );

  image->npix = image->x.n * image->y.n;

  switch( image->fmt )
  {
  case DF_RAW_ASCII:
    data = rd_image_data_ascii( err_pfx, image->file, image->npix );
    break;

  case DF_RAW_FLOAT:
    data = rd_image_data_float( err_pfx, image->file, image->npix );
    break;

  case DF_RAW_DOUBLE:
    data = rd_image_data_double( err_pfx, image->file, image->npix );
    break;

  case DF_PBM:
  case DF_PGM:
    data = rd_image_data_pxm( err_pfx, image->file );
    break;

  case DF_FITS:
    data = rd_image_data_fits( err_pfx, image->file );
    break;

  default:
    tf_enter( "image init" );
    tf_assert( 0 );
    /* NOT REACHED */
  }

  if ( NULL == data )
    return 1;
  
  
  /* warn if the pixel size is a bit small. limit is 1/1000 arcsec */
#define TOO_SMALL_PIX (M_PI / ( 180 * 60 * 60 ) * 0.001)

  if ( fabs(image->x.pix_sz) < TOO_SMALL_PIX )
    tf_message( "%s: x pixel size (%e rad) is less than 0.001 arc seconds\n",
		err_pfx, image->x.pix_sz );

  if ( fabs(image->y.pix_sz) < TOO_SMALL_PIX )
    tf_message( "%s: y pixel size (%e rad) is less than 0.001 arc seconds\n",
		err_pfx, image->y.pix_sz );

  /*
    loop through the data, determine the number of pixels above the
    clip limit
   */

  p_sum = 0.0;
  data_lim = data + image->npix;
  for ( datap = data ; datap < data_lim ; datap++ )
    if ( *datap <= image->clip )
      image->npix--;
    else
      p_sum += *datap;

  if ( NULL ==
       (image->pixel = (Pixel *) malloc( image->npix * sizeof(Pixel) )) )
  {
    tf_message( "%s: error allocating pixel data (%d bytes)\n", err_pfx, 
		image->npix * sizeof(Pixel) );
    return 1;
  }

  /*
    scan through the image, storing "useful" pixels into a list,
    along with their positions and probabilities. convert from
    pixel indices to radians; the position is the left or bottom
    of the pixel, not the center.  the index is 1 at the center
    of the pixel.
  */

  datap = data;
  p_min = 0.0;
  pixel = image->pixel;

  for ( y = 1 ; y <= image->y.n ; y++ )
    for ( x = 1 ; x <= image->x.n ; x++, datap++ )
      if (*datap > image->clip)
      {
        pixel->x = ( x - image->x.pnt_idx - 0.5 ) * image->x.pix_sz;
        pixel->y = ( y - image->y.pnt_idx - 0.5 ) * image->y.pix_sz;
        pixel->p_min = p_min;
	if ( dbflag( "dump_pixel" ) )
	  tf_message( "(%f,%f)\n", pixel->x, pixel->y );
        p_min += *datap / p_sum;
        ++pixel;
      }

  free( data );

  /* prior to this, the magnitude of pix_sz referred to the pixel
     size; it's sign gave the orientation of the image in the data.
     Since the latter is now encoded in the image pixel list, pix_sz
     is now just the pixel size */
  image->x.pix_sz = fabs( image->x.pix_sz );
  image->y.pix_sz = fabs( image->y.pix_sz );

  pgdata->max_extent	= sqrt( (double)
				(image->x.n * image->x.n *
				 image->x.pix_sz * image->x.pix_sz +
				 image->y.n * image->y.n *
				 image->y.pix_sz * image->y.pix_sz
				 ) ) / 2.0 + image->theta;
  pgdata->z		= image->z;
  pgdata->factor	= 1.0;

  if ( dbflag( "dump_image" ) )
    dump_ImageData( image );

  return 0;
}

static int
comp( const void *key, const void *data )
{
  return *(double *) key < ((Pixel *)data)->p_min ? -1 :
         *(double *) key > ((Pixel *)data)->p_min ?  1 :
         0;
}

static int
generate( void *data, DVector3 *direction, double *z )
{
  ImageData *image = (ImageData *) data;
  Pixel *pixel, *hi_bnd;
  double theta, phi, p, cos_theta, sin_theta, cos_phi, sin_phi;
  double cos_theta_p, sin_theta_p, cos_phi_p, sin_phi_p;

  /* grab a pixel */
  p = ranf();

  bnd_bsearch( &p, image->pixel, image->npix, sizeof(Pixel),
				 &pixel, &hi_bnd, comp );
  tf_assert5( NULL != pixel );

  theta	= pixel->y + genunf( 0.0, image->y.pix_sz );
  phi	= pixel->x + genunf( 0.0, image->y.pix_sz );

  cos_theta = cos( theta );
  sin_theta = sin( theta );

  cos_phi = cos( phi );
  sin_phi = sin( phi );

  /* use trignometric identities to get values for theta0 + theta and
     phi0 + phi */

  cos_theta_p = image->cos_theta0 * cos_theta - image->sin_theta0 * sin_theta;
  cos_phi_p   = image->cos_phi0   * cos_phi   - image->sin_phi0   * sin_phi;

  sin_theta_p = image->sin_theta0 * cos_theta + image->cos_theta0 * sin_theta;
  sin_phi_p   = image->sin_phi0   * cos_phi   + image->cos_phi0   * sin_phi;

  direction->x = - cos_theta_p * sin_phi_p;
  direction->y = - sin_theta_p;
  direction->z =   cos_theta_p * cos_phi_p;
  *z	       = image->z;

  return 0;
}

static void
cleanup( void *data )
{
  ImageData *image = (ImageData *) data;
  if ( image->pixel )
    free( image->pixel );
  free( image->name );
  free( image->file );
  free ( data );
}


static DataFormat
guess_format( const char *image, int *nx, int *ny )
{
  size_t nc, i;
  char buf[80];
  FILE *file = fopen( image, "rb" );

  if ( NULL == file )
    return -1;

  /* read in the first 80 characters */
  nc = fread( buf, 1, 80, file );
  fclose( file );

  if ( 0 == nc )
    return -1;

  if ( 0 == strncasecmp( buf, "simple", 6 ) )
    return DF_FITS;

  if ( 0 == strncasecmp( buf, "P1", 2 ) )
    return DF_PBM;
  
  if ( 0 == strncasecmp( buf, "P2", 2 ) ||
       0 == strncasecmp( buf, "P5", 2) )
    return DF_PGM;

  /* check for legal characters in ASCII */
  for (i = 0;
       i < nc &&
	 (isspace(buf[i]) || isdigit(buf[i]) || buf[i] == '.' )
	 ; i++ )
    ;

  if ( i == nc )
    return DF_RAW_ASCII;

  /* last hope is to use the file size and the (hopefully specified)
     nx and ny */

  {
    double ipart;
    struct stat buf;

    if ( stat( image, &buf ) )
      return -1;

    /* if the axis lengths weren't specified, assume a square image */
    if ( *nx == 0 && *ny == 0 )
    {
      if ( 0.0 == modf( sqrt( buf.st_size / sizeof(float) ), &ipart ) )
      {
	*nx = *ny = sqrt( buf.st_size / sizeof(float) );
	return DF_RAW_FLOAT;
      }
      else if ( 0.0 == modf( sqrt( buf.st_size / sizeof(double) ), &ipart ) )
      {
	*nx = *ny = sqrt( buf.st_size / sizeof(float) );
	return DF_RAW_DOUBLE;
      }
    }

    /* if one axis length was specified, but the other wasn't, can't do much here! */
    else if ( *nx == 0 || *ny == 0 )
      return -1;

    /* last try; both were specified */
    else if ( 0.0 == modf( buf.st_size / sizeof(float) / *nx / *ny , &ipart ) )
      return DF_RAW_FLOAT;

    else if ( 0.0 == modf( buf.st_size / sizeof(double) / *nx / *ny , &ipart ) )
      return DF_RAW_DOUBLE;

  }

  return DF_ERROR;
}

static void
init_Axis( Axis *axis )
{
  axis->n	= INT_MIN;
  axis->pix_sz  = -DBL_MAX;
  axis->ref_idx = -DBL_MAX;
  axis->ref_wcs = -DBL_MAX;
  axis->pnt_idx = -DBL_MAX;
  axis->pnt_wcs = -DBL_MAX;
  axis->flip    = 0;

}
