/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include <lua.h>
#include <lualib.h>
#include <luasup/luasup.h>

#include <suplib/str.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"
#include "lua_interface.h"

extern Module *modules[];

static int get_lua_args_va( const char *func, int n, va_list ap );


typedef struct
{
  char *name;
  lua_CFunction func;
} LuaCFunction;

/* the internal lua functions provided by raygen */
static LuaCFunction lua_cfunctions[] =
{
 { "begin_source",	begin_source_lua	},
 { "end_source",	end_source_lua		},
 { "ea_define",		ea_define_lua 		},
 { "tf_lua_error",	tf_lua_error 		},
 { NULL,		NULL			}
};

void
tf_lua_error( void )
{
  tf_assert( lua_getparam( 1 ) != LUA_NOOBJECT );
  tf_exit( ExitERR_misc, "%s", lua_getstring( lua_getparam( 1 ) ) );
}

void
define_error( const char *fmt, ... )
{
  va_list ap;
  char buffer[1024];

  va_start( ap, fmt );

  vsprintf( buffer, fmt, ap );

  va_end( ap );

  lua_error( buffer );
}

void
lua_init( void )
{
  Module **mod;
  LuaCFunction *func;

  tf_enter( "init_lua" );

  /* attach the standard lua libraries */
  lua_iolibopen( );
  lua_strlibopen( );
  lua_mathlibopen( );

  /* register standard functions */
  for ( func = lua_cfunctions ; NULL != func->name ; func++ )
    lua_register( func->name, func->func );

  /* register aperture definition functions */
  for (mod = modules ; *mod ; mod++ )
    lua_register((*mod)->name, (*mod)->define);

  /* turn debugging mode on */
  tf_assert( lua_dostring( "$debug" ) == 0 );

  lua_pushobject(lua_getglobal("tf_lua_error"));
  lua_seterrormethod( );

  tf_leave();
}

void
process_lua( char* script_file, char* override_script)
{
  tf_enter( "process_lua" );

  /* if no override script provided, set up empty function */
  if ( '\0' == override_script[0] )
  {
    tf_assert( 0 == lua_dostring( "function override()\n end" ) );
  }

  /* override script is in a file */
  else if ( '@' == override_script[0] )
  {
    if ( lua_dofile( &override_script[1] ) )
      tf_exit( ExitERR_misc,
              "Error while parsing override function in file `%s'",
               &override_script[1] );
  }

  /* the script is in the passed string.  */
  else
  {
    const char *start = "function override()\n";
    const char *end   = "\nend";

    char *script = (char *) malloc( strlen(start) + strlen(end) +
				    strlen(override_script ) + 1 );

    if ( NULL == script )
      tf_exit( ExitERR_alloc, "couldn't allocate override function" );

    strcat( strcat( strcpy( script, start ), override_script ), end );

    if ( lua_dostring( script ) )
      tf_exit( ExitERR_misc, "error while parsing override function" );

    free( script );
  }

  if ( lua_dofile( script_file ) )
    tf_exit(ExitERR_misc,
	    "Error while executing source specification file `%s'",
	    script_file );

  tf_leave( );
}


void
get_lua_args( const char *func, ... )
{
  va_list	ap;
  int		n;

  va_start( ap, func );

  n = get_lua_args_va( func, 1, ap );
  
  if ( lua_getparam(n) != LUA_NOOBJECT )
    define_error( "too many arguments in call to lua function `%s'", func);
}

void
get_lua_args_n( const char *func, int n, ... )
{
  va_list	ap;

  va_start( ap, n );

  get_lua_args_va( func, n, ap );
}

static int
get_lua_args_va( const char *func, int n, va_list ap )
{
  lua_Object	obj;
  char *	name;

  for ( ; (name = (char *) va_arg( ap, char *)) && name[0] != '\0' ; n++)
  {
    LuaArgType	type = (LuaArgType) va_arg( ap, LuaArgType );
    void *	dest = (void *)     va_arg( ap, void * );

    if ( LUA_NOOBJECT == ( obj = lua_getparam( n ) ) )
    {
      while ( NULL != (char *) va_arg( ap, char *) )
      {
        type = (LuaArgType) va_arg( ap, LuaArgType );
        dest = (void *)     va_arg( ap, void * );
      }
      va_end( ap );
      define_error("too few arguments in call to lua function `%s' " 
		   "at argument `%s'", func, name );
    }

    if ( type == Lua_NUMBER && lua_isnumber( obj ) )
      *((double *) dest) = lua_getnumber( obj );
    else if ( type == Lua_STRING && lua_isstring( obj ) )
      *(( char **) dest) = str_dup( lua_getstring( obj ) );
    else
    {
      while ( NULL != (char *) va_arg( ap, char *) )
      {
        type = (LuaArgType) va_arg( ap, LuaArgType );
        dest = (void *)     va_arg( ap, void * );
      }
      va_end( ap );
      define_error( "incorrect argument type in call to lua function `%s' "
		    "for argument `%s': expected a %s", func, name,
		    type == Lua_NUMBER ? "number" : "string" );
    }
  }

  va_end( ap );

  return n;
}

char *
get_lua_str_arg_p(
	const char *funcname,	/* the name of the lua function */
	const char *argname	/* the name of the argument that's expected */
	)
{
  lua_Object obj;

  if ( LUA_NOOBJECT == ( obj = lua_getparam(1) ) )
    return NULL;

  if ( lua_getparam(2) != LUA_NOOBJECT )
    tf_exit(ExitERR_illegalArg,
	    "extra arguments in call to lua function %s", funcname);

  if ( !lua_isstring(obj) )
    tf_exit( ExitERR_illegalArg,
	    "incorrect type for argument %s in call to lua function %s",
	    argname, funcname);

  return str_dup( lua_getstring(obj ) );
}


void
check_n_lua_arg(
  const char *	func,		/* the name of the lua function */
  int		n		/* the number of arguments expected */
)
{
  if ( lua_getparam( n+1 ) != LUA_NOOBJECT )
    define_error( "too many arguments in call to lua function `%s'", func);
}


/*

  verify_empty_table - determine if a table is empty.  if not,
  it dies.
  
*/

void
verify_empty_optargs(
  const char *	err_pfx,	/* some sort of prefix to print */
  lua_Object	table		/* the optargs table to examine */
  )
{
  char msg[1024];
  lua_Object index = LUA_NOOBJECT;
  int nelem = 0;

  sprintf( msg, "%s: unknown optarg(s): \n", err_pfx );
  while( LUA_NOOBJECT != slua_tbl_next( table, &index ) )
  {
    nelem++;
    strcat( msg, "  " );
    strcat( msg, lua_getstring( index ) );
    strcat( msg, "\n" );
  }

  if ( nelem )
    define_error( msg );
}

char *
get_string(
  const char *err_pfx,
  lua_Object optargs,
  const char *index,
  const char *dflt
  )
{
  char *ptr = NULL;

  if ( 1 == slua_tbl_el_s( optargs, index, &ptr ) )
    ptr = str_dup( dflt );

  if ( NULL == ptr )
    define_error( "%s: out of memory copying value of %s", err_pfx, index );

  slua_tbl_el_delete( optargs, index );
  return ptr;
}
