#ifndef MAIN_H
#define MAIN_H

/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>

#include <lua.h>
#include <linklist/linklist.h>
#include <bpipe/datatypes.h>

#define NAME_LENGTH 255

/* the names of possible image attributes */

/* the angular position of the source */
#define THETA	"theta"
#define PHI	"phi"

/* the axial position of the source */
#define SRC_Z	"z"

/*
  the type of module that the source specification script
  is in the midst of building
 */
typedef enum
{
  Mod_SOURCE,	/* source */
  Mod_FGEN,	/* flux generator */
  Mod_FMOD,	/* flux modifier */
  Mod_EGEN,	/* extent generator */
  Mod_DGEN,	/* direction generator */
  Mod_PGEN	/* position generator */
} ModuleType;

/* tags for lua types */
typedef enum
{
  Lua_STRING, Lua_NUMBER
} LuaArgType;

/* the limit on the number of photons that are generated */
typedef enum
{
  Limit_COUNTS,
  Limit_DENSITY,
  Limit_TIME
} LimitType;

/* the ray distribution at the entrance aperture */
typedef enum
{
  RayDist_RANDOM
} RayDistType;


typedef struct
{
  double energy;
  double time;
  DVector3 position;
  DVector3 direction;
} Photon;

typedef struct
{
  char *	name;
  lua_CFunction	define;
} Module;

typedef struct
{
  double	max_extent;		/* maximum radial extent of the source
					   from the optical axis [radians] */

  double	factor;			/* the extent factor by which the
  					   effective flux is multiplied */

  double	z;			/* z coord of source */

  void *	data;			/* generator specific data */
} PositionGeneratorData;

typedef struct
{
  char *	name;		/* the name of the flux generator */
  int		(*init)( PositionGeneratorData *data );
  int		(*generate)( void *data, DVector3 *pos, double *z );
  void		(*cleanup)( void *data );
} PositionGeneratorSpec;

typedef struct
{
  struct Source	*	source;		/* the parent source of the
					   generator */
  PositionGeneratorSpec spec;
  PositionGeneratorData data;
  char *	name;

} PositionGenerator;

typedef struct
{
  size_t	iter_lim;		/* maximum number of iterations to
  					   satisfy modifier events */
  size_t	iter_warn;		/* number of iterations
					   between modifier warnings */
  void *	data;			/* flux generator specific data */
} DirectionGeneratorData;

typedef struct
{
  char *	name;		/* the name of the flux generator */
  int		(*init)   ( DirectionGeneratorData * );
  int		(*check)( void *data, DVector3 *direction );
  void		(*cleanup)( void * );
} DirectionGeneratorSpec;

typedef struct
{
  struct Source	*	source;		/* the parent source of the
					   generator */
  DirectionGeneratorSpec spec;
  DirectionGeneratorData data;
  char *	name;
} DirectionGenerator;


typedef struct Source
{
  char *	name;			/* the name of the source */

  size_t	n_gen;			/* the number of flux generators */
  SLinkList *	flux_gen;		/* array of flux generators */

  SLinkList *	phot_valid;		/* a list of generators with valid
					   photons, sorted by arrival time */
  SLinkList *	phot_invalid;		/* a list of generators with invalid
					   photons */
  SLinkList *	phot_temp;		/* temporary list to ease swapping
  					   valid and invalid lists */

  PositionGenerator *	pos_gen;
  DirectionGenerator *	dir_gen;

  double	theta;			/* off axis angle of source */


  double	dist;			/* distance to source [mm] */

  double	time;			/* clock */

  SLinkList *	extent_events;		/* list of extent generator events */
  size_t	extent_iter_lim;	/* maximum number of iterations to
  					   satisfy extent events */
  size_t	extent_iter_warn;	/* number of iterations
  					   between extent warnings */
} Source;

typedef struct
{
  char *	name;			/* the generic name of the flux
					   modifier */
  int		(*init)   ( void *data );
  void		(*cleanup)( void *);
} FluxModifierSpec;

typedef struct
{
  char *	name;			/* the specific label for this
					   modifier */
  struct FluxGenerator *gen;		/* the generator to modify */
  struct Source	*	source;		/* the parent source of the
					   generator */
  size_t	id;			/* modifier id */

  double *	factor;			/* a pointer to this modifier's
					   factor slot in the parent
					   generator's list. */
  FluxModifierSpec spec;		/* modifier specific specification */
  void *	data;			/* modifier specific data */
} FluxModifier;

typedef struct
{
  double	base_flux;		/* base flux, in photons/s/cm^2 */
  size_t	iter_lim;		/* maximum number of iterations to
  					   satisfy modifier events */
  size_t	iter_warn;		/* number of iterations
					   between modifier warnings */
  void *	data;			/* flux generator specific data */
} FluxGeneratorData;

typedef struct
{
  char *	name;		/* the name of the flux generator */
  int		(*init)   ( FluxGeneratorData *data );
  int		(*generate)( void *data, double *energy );
  void		(*cleanup)( void *data );
} FluxGeneratorSpec;

typedef struct FluxGenerator
{
  Source *	source;			/* the source that this generator
  					   belongs to */

  size_t	id;			/* generator id.  used as offset
					   in source's fluxgen array */

  size_t	n_mods;			/* the number of modifiers */
  SLinkList 	mods;			/* list of modifiers */

  double *	flux_factors;		/* array of factors which
					   modify the base flux */
  double *	flux_factors_lim;	/* a pointer to one past the
					   array of flux factors */
  double	flux_factor;		/* product of all of the flux
					   factors */
  double	eff_flux;		/* effective flux */

  SLinkList 	events;			/* event queue for generator and
  					   modifiers */

  double	time;			/* the time of arrival of the next
  					   photon from this generator */
  FluxGeneratorSpec spec;
  FluxGeneratorData data;		/* generator specific data */
  char *	name;			/* the generator specific name */

} FluxGenerator;

typedef enum
{
  Event_FLUX,
  Event_EXTENT
} EventType;

typedef struct Event
{
  double	time;		/* time of the event */
  EventType	type;		/* type of event */
  void *	module;		/* module that caused the event */
  void *	module_data;	/* module specific data */
  void *	event_data;	/* event specific data */
  void		(*service)(void *module, double time,
  			   void *module_data,
			   void *event_data);	/* service routine */
  void		(*del)(void *event_data);	/* delete event data */
  void		(*cleanup)(struct Event *event);	/* cleanup routine */
} Event;

typedef enum
{
  EA_Annulus,
  EA_Circle,
  EA_Wedge
} EA_Type;

typedef struct EntranceAperture
{
  void *	data;
  EA_Type	type;
  double	area;
  double	area_min;
  double	area_max;
  void		(*pos)(struct EntranceAperture *ea, DVector3 *pos);
} EntranceAperture;

typedef struct Mirror {
  double rho_f;
  double rho_a;
  double z_f;
  double z_a;
  double l;
  double node;
} Mirror;

typedef struct
{
  char *	source_spec;		/* the lua source specification file */
  char *	source_override;	/* source spec override */

  char *	ea_spec;		/* the lua source specification file */
  char *	ea_override;		/* source spec override */

  char *	output;			/* the name of the output stream */
  char *	logfile;		/* logfile */

  double	max_open_angle;		/* the maximum opening angle
  					   to back of paraboloid for sources.
  					   set in init_source. set in
					   init_ system */

  double	t_open_angle;		/* tangent of the maximum opening angle
  					   set in init_source */

  long seed1, seed2, block;		/* random number seeds and starting
					   block */

  double	limit;			/* the limit on the current run */
  LimitType	limit_type;		/* the type of limit on the current
					   run */

  double	time;			/* the time of the last photon
					   emitted */

  size_t	counts;			/* the number of photons produced */

  RayDistType	ray_dist;		/* the type of ray distribution at the
					   entrance aperture */

  double	z;			/* where the rays will be left */

  SLinkList	sources;		/* a list of the sources */
  SLinkList	gen;			/* a list of flux generators, one per
  					   source, representing the next photon
  					   from that source */

  EntranceAperture *ea;			/* array of entrance apertures */
  int		n_ea;			/* number of entrance apertures */
  double	ea_area;		/* total area of entrance aperture(s)*/

  Source *	current_source;		/* the current source.  in use
  					   during the creation of sources.
  					   changed by the begin_source and
  					   end_source routines */

  FluxGenerator *current_fgen;		/* the current flux generator.  in
  					   use during the addition of flux
  					   modifiers.  changed by the
  					   begin_flux_gen and
  					   end_flux_gen routines */

  SLinkList	spare_events;		/* spare event nodes stuck here by
  					   update_photon_time when it handles
  					   an event */

  double node;                           /* node Z position from which
					    to measure source angle */
} System;

/* dir_gen.c */
void new_dir_gen(DirectionGeneratorSpec *spec, char *name, void *data);
int delete_dir_gen(void *data);

/* ea.c */
void ea_init(void);
void ea_cleanup(void);
void ea_define_lua( void );
void entrance_aperture( Photon *photon, double z );

/* events.c */
void register_event(char *mod_name, EventType event_type, void *module,
		    double event_time, double current_time,
		    void *module_data, void *event_data,
		    void (*service)(void *module, double time,
				    void *module_data, void *event_data),
		    void (*del)(void *event_data),
		    void (*cleanup)(Event *event), SLinkList *queue);
void delete_event(void *event);
void register_flux_gen_event(FluxGenerator *gen, double time, void *data,
			     void (*service)(void *gen, double time,
					     FluxGeneratorData *module_data,
					     void *event_data),
			     void (*del)(void *event_data));
void register_flux_modifier_event(FluxModifier *mod, double time, void *data, void (*service)(void *mod, double time, void *module_data, void *event_data), void (*del)(void *event_data));
void register_extent_event(PositionGenerator *gen, double time, void *data, void (*service)(void *gen, double time, PositionGenerator *module_data, void *event_data), void (*del)(void *event_data));
int event_cmp(const void *event1, const void *event2);

/* flux_gen.c */
void new_flux_gen(FluxGeneratorSpec *spec, char *name, void *data);
void delete_flux_gen(void *data);
int init_flux_gen(void *data);
void calculate_eff_flux(FluxGenerator *gen);
void cleanup_flux_gen_event(Event *event);
int set_flux_gen_time(void *data, void *udata);
int flux_gen_time_cmp(const void *gen1, const void *gen2);

/* flux_mod.c */
void new_flux_modifier(FluxModifierSpec *spec, char *name, void *data);
void delete_flux_modifier(void *data);
int init_flux_modifier(FluxModifier *mod);
void cleanup_flux_modifier_event(Event *event);

/* gparam.c */
int gparam(int argc, char *argv[]);
void cleanup_params( void );

/* main.c */
int main(int argc, char *argv[]);

/* output.c */
void output_init(void);
void output_cleanup(void);
void output(Photon *photon);

/* pos_gen.c */
void new_pos_gen(PositionGeneratorSpec *spec, char *name, void *data);
int delete_pos_gen(void *data);
void init_pos_gen(PositionGenerator *gen);
void cleanup_extent_event(Event *event);

/* raygen.c */
void raygen_init(void);
void raygen_cleanup(void);

/* photon.c */
void photon_loop(void);
void generate_photon(Source *source, FluxGenerator *gen);
void entrance_pupil(Source *source, Photon *photon, double z);
FluxGenerator *nextPhoton_source(Source *source);
void init_photon_times(Source *source);
int update_photon_time(FluxGenerator *gen);

/* source.c */
void sources_create( void );
void sources_init( void );
void sources_cleanup( void );
void begin_source_lua(void);
void end_source_lua(void);

/* system.c */
void system_init(void);
void system_cleanup(void);

/* test.c */
void set_eapad_lua(void);

/* usage.c */
void usage( void );

#endif /* ! MAIN_H */
