/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>

#include <math.h>
#include <mathconst/mathconst.h>

#include <suplib/str.h>
#include <suplib/units.h>
#include <suplib/units_angular.h>
#include <suplib/units_linear.h>

#include "main.h"
#include "lua_interface.h"

/*

  parse_angle - parse a angle spec. upon error calls define_error.
  converts to radians.  If no unit is specified, it assumes it's
  radians

 */

double
parse_angle(
  const char *error_pfx,	/* the name of the calling function
				   (for errors )*/
  const char *spec,		/* the angular spec to parse */
  const char *var_name		/* the name of the variable to receive
				   the value (for errors) */
  )
{
  UnitVal uv;
  uv.unit = UNIT_radian;

  switch( units_parse( spec, &uv, &UnitsAngular_def ) )
  {
  case UNITS_BADNUM:
    define_error( "%s: unparseable number for %s: %s",
		  error_pfx, var_name, spec );
    break;

  case UNITS_BADSPEC:
    define_error( "%s: unknown angular unit for %s: %s",
		  error_pfx, var_name, spec );
    break;
  }

  return uv.value * units_cvt( &UnitsAngular_def, uv.unit, UNIT_radian );
}

/*
  PROCEDURE
  parse_extent - parse a linear or angular extent, converting it to an 
  angular extent in radians.

  DESCRIPTION
  It calls @code{define_error()} upon error.

  AUTHOR
  Diab Jerius
*/

void
parse_extent(
  const char *err_pfx,		/* the error string prefix */
  const char *which,		/* a string of meaning to the user to
				   describe the extent (usually a
				   variable name) */
  const char *extent,		/* the string containing the extent
				   description */
  double z,			/* the axial position of the object */
  double *final_extent,		/* the final angular extent, in radians */
  double *raw_extent		/* the extent as determined directly
				   from *extent */
  )
{
  UnitVal uv;

  /* first check if it's an angle */
  switch( units_parse( extent, &uv, &UnitsAngular_def ) )
  {
  case UNITS_NOSPEC:
    uv.unit = UNIT_radian;
    /* fall through */

  case UNITS_OK:
    *raw_extent = uv.value;
    *final_extent = uv.value * units_cvt( &UnitsAngular_def, uv.unit, UNIT_radian );
    break;

  case UNITS_BADNUM:
    define_error( "%s: bad value for `%s': %s", err_pfx, which, extent );

    /* wasn't an angle; check if it's a linear dimension */
  case UNITS_BADSPEC:
    switch( units_parse( extent, &uv, &UnitsLinear_def ) )
    {
    case UNITS_OK:
      *raw_extent = uv.value;
      *final_extent = atan2( uv.value * units_cvt( &UnitsLinear_def, uv.unit,
					    UNIT_millimeter ), fabs(z));
      break;

    case UNITS_BADNUM:
      define_error( "%s: bad value for `%s': %s", err_pfx, which, extent );

    case UNITS_BADSPEC:
      define_error( "%s: bad units for `%s': %s", err_pfx, which, extent );
    }
  }
}


void
ffree( void *stuff )
{
  if ( stuff )
    free( stuff );
}
