/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <stdlib.h>

#include <bpipe/bpipe.h>
#include <bpipe/bpipe_axaf.h>
#include <bpipe/datatypes.h>
#include <mathconst/mathconst.h>

#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

#include "polarize.h"

BPipe *		bpipe;
BPipeOutput *	bpo;
static void *	dpkt = NULL;
DVector3 *	position;
DVector3 *	direction;
DCVector2 *	pol_sin;
DCVector2 *	pol_cos;
double *	weight;
double *	energy;
double *	time;
unsigned int *	id;
  


/* simple routine to get the data pointer for a named data packet field */
static char *
dpktf_data( BPipe *bpipe, char *name, void *data )
{
  DpktField *field = bpipe_dpktf( bpipe, name );

  if ( field == NULL )
    tf_exit( ExitERR_misc, "unable to retrieve info for field `%s': %s",
	     name, bpipe_strerror( bpipe_errno) );

  return bpipe_dpktf_data( field, data );
}

void
output_init( void )
{
  size_t dpkt_size;

  tf_enter( "output_init" );

  if ( NULL == ( bpipe = bpipe_new( ) ) )
    tf_exit( ExitERR_misc, bpipe_strerror( bpipe_errno ) );

  /* open up output stream */
  if ( NULL == ( bpo = bpipe_output( bpipe, sys.output ) ) )
    tf_exit( ExitERR_misc, "error opening output stream `%s': %s",
	     sys.output, bpipe_strerror( bpipe_errno ) );

#if 0
  /* create some header fields */
  if ( bpipe_hdrf_add( bpipe, "shell", BPDType_int, NULL, &sys.shell, 0 )
       )
    tf_exit( ExitERR_misc, "error creating header fields: %s",
	     bpipe_strerror( bpipe_errno ) );
#endif

  /* create some data packet fields */

  if ( bpipe_dpktf_add( bpipe, BP_AXAF_phot_pos, BPDType_DVector3, NULL )
       ||
       bpipe_dpktf_add( bpipe, BP_AXAF_phot_dir, BPDType_DVector3, NULL )
       ||
       bpipe_dpktf_add( bpipe, BP_AXAF_phot_wt,  BPDType_double,   NULL )
       ||
       bpipe_dpktf_add( bpipe, BP_AXAF_phot_energy,  BPDType_double,   NULL )
       ||
       bpipe_dpktf_add( bpipe, BP_AXAF_time,     BPDType_double,   NULL )
       ||
       bpipe_dpktf_add( bpipe, BP_AXAF_phot_id,  BPDType_uint,   NULL )
       ||
       bpipe_dpktf_add( bpipe, BP_AXAF_pol_sin,  BPDType_DCVector2,   NULL )
       ||
       bpipe_dpktf_add( bpipe, BP_AXAF_pol_cos,  BPDType_DCVector2,   NULL )
     )
    tf_exit( ExitERR_misc, "error creating data packet fields: %s",
	     bpipe_strerror( bpipe_errno ) );

  if ( 0 == ( dpkt_size = bpipe_map( bpipe ) ) )
    tf_exit(ExitERR_misc, "error mapping bpipe: %s",
	    bpipe_strerror( bpipe_errno ) );

  if ( NULL == ( dpkt = malloc( dpkt_size ) ) )
    tf_exit( ExitERR_misc, "error allocating data packet" );

  position	= (DVector3 *)	dpktf_data( bpipe, BP_AXAF_phot_pos, dpkt );
  direction	= (DVector3 *)  dpktf_data( bpipe, BP_AXAF_phot_dir, dpkt );
  pol_sin	= (DCVector2 *)	dpktf_data( bpipe, BP_AXAF_pol_sin, dpkt );
  pol_cos	= (DCVector2 *)	dpktf_data( bpipe, BP_AXAF_pol_cos, dpkt );
  weight	= (double *)    dpktf_data( bpipe, BP_AXAF_phot_wt, dpkt );
  energy	= (double *)    dpktf_data( bpipe, BP_AXAF_phot_energy, dpkt );
  time		= (double *)    dpktf_data( bpipe, BP_AXAF_time, dpkt );
  id		= (unsigned int *) dpktf_data( bpipe, BP_AXAF_phot_id, dpkt );


  /* output header */
  if ( bpipe_write_hdr(bpipe) )
    tf_exit( ExitERR_misc, "error writing bpipe header: %s",
	     bpipe_strerror( bpipe_errno ) );

#if 0
  /* deal with polarization state */
    if ( ! gi->ipolar ) {
        pt->ecc = 1.0e0;
        pt->cpsi = 1.0e0;
        pt->spsi = 0.0e0;
        pt->sqre = sqrt(2.0);
    } else {
        pt->ecc = value[6];
        pt->cpsi = cos(value[5]*deg2rad);
        pt->spsi = sin(value[5]*deg2rad);
        pt->sqre = sqrt(1.0 + value[6]*value[6]);
    }

  /*--- initialize ray polarization ----*/
  xtmp         =  p_ptsrc->cpsi / p_ptsrc->sqre;
  p_ray->c2[0] =  dComplex ( xtmp, (double)0.0 );

  xtmp         =  p_ptsrc->spsi / p_ptsrc->sqre;
  p_ray->c2[1] =  dComplex ( xtmp, (double)0.0 );

  xtmp         = -p_ptsrc->spsi / p_ptsrc->sqre;
  ctmp         =  dComplex( xtmp, (double)0.0 );
  p_ray->s2[0] =  dRCmul( p_ptsrc->ecc, ctmp );

  xtmp         =  p_ptsrc->cpsi / p_ptsrc->sqre;
  ctmp         =  dComplex( xtmp, (double)0.0 );
  p_ray->s2[1] =  dRCmul( p_ptsrc->ecc, ctmp );

  /*--- initialize ray weight ----------*/
  ctmp = dCmul( p_ray->c2[0], p_ray->c2[0] );
  ctmp = dCadd( ctmp, dCmul( p_ray->c2[1], p_ray->c2[1] ) );
  ctmp = dCadd( ctmp, dCmul( p_ray->s2[0], p_ray->s2[0] ) );
  ctmp = dCadd( ctmp, dCmul( p_ray->s2[1], p_ray->s2[1] ) );
  p_ray->wt = ctmp.r;

#endif

  setup_polarization( pol_cos, pol_sin, weight);

  tf_leave( );
}

void
output_cleanup( void )
{
  if ( dpkt )
    free( dpkt );

  bpipe_delete( bpipe );
}

void
output( Photon *photon )
{
  *position	= photon->position;
  *direction	= photon->direction;
  *energy	= photon->energy;
  *time		= photon->time;
  *id		= sys.counts;

  /* write the data packet out */
  if ( bpipe_write_dpkt( bpipe, dpkt, bpo ) )
  {
    tf_enter( "output" );
    tf_exit( ExitERR_misc, "error writing photon (id = %d, time = %f): %s",
	     *id, *time, bpipe_strerror( bpipe_errno ) );
  }
}
