/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <string.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <lua.h>
#include <luasup/luasup.h>
#include <suplib/str.h>

#include "main.h"
#include "args.h" 
#include "dist.h"
#include "lua_interface.h"
#include "misc.h"

typedef struct
{
  double	half_width;	/* half width of rect */
  double	half_height;	/* half height of rect */
  double	z;		/* z coord of rect */

  DVector2	xlim;
  DVector2	ylim;

  ImgDist	idist;

  double	theta;		/* off axis angle of image */
  double	phi;		/* off axis angle of image */
  double	sin_theta0;
  double	cos_theta0;
  double	sin_phi0;
  double	cos_phi0;
} RectData;

static void	define(void);
static int	init(PositionGeneratorData *data);
static int	generate( void *data, DVector3 *direction, double *z );
static void	cleanup(void *data);

Module rect_module = { "rect", define };

static PositionGeneratorSpec spec =
{
  "rect",
  init,
  generate,
  cleanup
};

static void
define( void )
{
  RectData *data;
  char *name, *width_spec, *height_spec;
  double x, y, z;
  lua_Object optargs = LUA_NOOBJECT;
  char err_pfx[1024];

  tf_enter( "rect" );

  slua_args( define_error, "rect",
	     "name",		SLua_STRING,	&name,
	     "width",		SLua_STRING,	&width_spec,
	     "height",		SLua_STRING,	&height_spec,
	     "optargs", 	SLua_TABLE | SLua_OPTIONAL, 	&optargs,
	     NULL
	     );

  /* generate error message prefix */
  sprintf( err_pfx, "rect `%s'", name );

  if ( NULL == ( data = (RectData *) malloc( sizeof(RectData) ) ) )
    define_error( "%s: error allocating memory", err_pfx );


  /* copy optargs, so can delete elements from it as we go along */
  optargs = optargs == LUA_NOOBJECT ? lua_createtable() : slua_tbl_copy( optargs );

  /* Axial Source Position */
  get_src_z( name, optargs, &data->z );

  /* Angular Source Position */
  get_src_pos( name, optargs, &data->theta, &data->phi );

  /* determine density distribution */
  get_ImgDist( err_pfx, &data->idist, optargs, data->z );

  /* generate positions on unit sphere of center of image in (x,y,z)
     from original (theta, phi) (in tilted lat,long) */
  x = sin( data->theta ) * cos( data->phi );
  y = sin( data->theta ) * sin( data->phi );
  z = cos( data->theta );

  /* generate sin and cos of (lat,long) from (x,y,z) */
  data->sin_theta0 = y;
  data->cos_theta0 = sqrt( x*x + z*z );
  data->sin_phi0 = x / data->cos_theta0;
  data->cos_phi0 = z / data->cos_theta0;

  /* we should have read (and deleted) all relevant optargs.  Those
     that remain should be flagged as errors */
  verify_empty_optargs( err_pfx, optargs );

  /* now that we know the axial location, process the height and
     width, converting into angular units if necessary */
  parse_extent( err_pfx, "width",  width_spec,  data->z, &data->half_width,  &z );
  parse_extent( err_pfx, "height", height_spec, data->z, &data->half_height, &z );

  if ( data->half_width < 0.0 || data->half_height < 0.0 )
    define_error( "%s: height and width must be non-negative", err_pfx );

  data->half_width	/= 2.0;
  data->half_height	/= 2.0;

  new_pos_gen( &spec, name, data );

  free( name );

  tf_leave( );
}

static int
init( PositionGeneratorData *data )
{
  RectData *rect = (RectData *) data->data;

  data->max_extent	= sqrt( rect->half_width * rect->half_width +
  				rect->half_height * rect->half_height )
    			  + rect->theta;
  data->z		= rect->z;
  data->factor		= 1.0;

  return 0;
}

static int
generate( void *data, DVector3 *direction, double *z )
{
  RectData *rect = (RectData *) data;
  double theta, phi, cos_theta, sin_theta, cos_phi, sin_phi;
  double cos_theta_p, sin_theta_p, cos_phi_p, sin_phi_p;

  *z		= rect->z;

  gen_ImgDist( &rect->idist,
	       &phi, -rect->half_width, rect->half_width,
	       &theta, -rect->half_height, rect->half_height );

  cos_theta = cos( theta );
  sin_theta = sin( theta );

  cos_phi = cos( phi );
  sin_phi = sin( phi );

  /* use trignometric identities to get values for theta0 + theta and
     phi0 + phi */

  cos_theta_p = rect->cos_theta0 * cos_theta - rect->sin_theta0 * sin_theta;
  cos_phi_p   = rect->cos_phi0   * cos_phi   - rect->sin_phi0   * sin_phi;

  sin_theta_p = rect->sin_theta0 * cos_theta + rect->cos_theta0 * sin_theta;
  sin_phi_p   = rect->sin_phi0   * cos_phi   + rect->cos_phi0   * sin_phi;

  direction->x = - cos_theta_p * sin_phi_p;
  direction->y = - sin_theta_p;
  direction->z =   cos_theta_p * cos_phi_p;

  return 0;
}

static void
cleanup( void *data )
{
  free ( data );
}
