/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <string.h>
#include <limits.h>

#include <lua.h>
#include <linklist/linklist.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"
#include "lua_interface.h"

static void	delete_source(void *data);
static Source *	new_source(char *name);
static int	init_source( void *data );

void
sources_create( void )
{
  if ( NULL == ( sys.sources = sll_new( SLL_NULL_CMP ) ) )
    tf_exit( ExitERR_alloc, "unable to initialize source list" );
  
  /* BEGIN NODE FIX */
#if 0

  sys.t_open_angle = 0.0;
  sys.max_open_angle = ( 1.0 / 180.0 ) * M_PI;	/* 1 degree */

  /* determine maximum opening angle for source */
  t_open_angle =
     ( sys.mirror.rho_a + (sys.mirror.node - source->pos_gen->data.z) *
              tan(source->pos_gen->data.max_extent) ) /
			   ( sys.mirror.z_a - source->pos_gen->data.z);

  if ( sys.t_open_angle < t_open_angle )
    sys.t_open_angle = t_open_angle;

#endif
  /* END NODE FIX */

  /* create the sources from the lua source script */
  process_lua( sys.source_spec, sys.source_override );

  /* generate the sources via the source() function.
     first ensure that it has been defined */
  {
    lua_Object source = lua_getglobal( "source" );

    if ( lua_isnil( source ) )
      tf_exit( ExitERR_misc, "%s: the function source() is not defined", sys.source_spec );

    if ( ! lua_isfunction( source ) )
      tf_exit( ExitERR_misc, "%s: the symbol `source' is not a function", sys.source_spec );

    if ( lua_callfunction( source ) )
      tf_exit( ExitERR_misc, "%s: error creating sources", sys.source_spec );
  }

  /* make sure that there's at least one source out there */
  if ( 0 == sll_count( sys.sources ) )
    tf_exit( ExitERR_misc, "error: no sources were specified" );

  /* ensure that the last source was correctly ended */
  if ( NULL != sys.current_source )
    tf_exit( ExitERR_misc,
	     "%s: specification not ended with call to end_source( \"%s\" )",
	     sys.current_source->name, sys.current_source->name );
}

void
sources_init( void )
{
  /* initialize sources */
  sll_traverse( sys.sources, (SLL_ACTION) init_source );

  /* BEGIN NODE FIX */
#if 0
  /* determine limiting entrance aperture */
  tf_message( "tan open angle\t= %f\n", sys.t_open_angle );
  tf_message( "tan max open angle\t= %f\n", tan( sys.max_open_angle ));

  if ( sys.t_open_angle > tan( sys.max_open_angle ) )
    sys.t_open_angle = tan( sys.max_open_angle );
#endif
  /* END NODE FIX */
}

void
sources_cleanup( void )
{
  sll_delete( sys.sources, delete_source );
}

void
begin_source_lua( void )
{
  char *name;

  tf_enter( "begin_source" );

  get_lua_args( "begin_source",
                "source_name", Lua_STRING, &name,
                NULL
              );

  if ( NULL != sys.current_source )
    tf_exit( ExitERR_misc, "source `%s': attempt to nest sources",
	     name );

  sys.current_source = new_source( name );

  tf_leave( );
}

void
end_source_lua( void )
{
  char *name;
  int  source_error = 0;

  tf_enter( "begin_source" );

  get_lua_args( "end_source",
                "source_name", Lua_STRING, &name,
                NULL
              );

  if ( NULL == sys.current_source )
    tf_exit( ExitERR_misc, "source `%s': source end outside of source specification?",
    				name );

  if ( strcmp( sys.current_source->name, name ) )
    tf_exit( ExitERR_misc,
	     "attempt to end source `%s' while creating source `%s'",
	     name, sys.current_source->name );

  /* check source to ensure that everything was specified */
  if ( 0 == sys.current_source->n_gen )
  {
    tf_message( "source `%s': no flux generator\n", sys.current_source->name );
    ++source_error;
  }

  if ( NULL == sys.current_source->pos_gen )
  {
    tf_message( "source `%s': no position generator\n",
    		 sys.current_source->name );
    ++source_error;
  }

  if ( source_error )
    tf_exit( ExitERR_misc, "source `%s': incomplete specification",
    				sys.current_source->name );

  sys.current_source = NULL;
  sys.current_fgen   = NULL;

  free( name );

  tf_leave( );
}

static Source *
new_source(
  char *name
)
{
  Source *source;

  tf_enter( "new_source" );

  if ( NULL == ( source			= (Source *) malloc( sizeof(Source) ))
       ||
       NULL == ( source->phot_temp	= sll_new( flux_gen_time_cmp ) )
       ||
       NULL == ( source->phot_valid	= sll_new( flux_gen_time_cmp ) )
       ||
       NULL == ( source->phot_invalid	= sll_new( flux_gen_time_cmp ) )
       ||
       NULL == ( source->extent_events	= sll_new( event_cmp ) )
       ||
       NULL == ( source->flux_gen	= sll_new( SLL_NULL_CMP ) )
     )
    tf_exit( ExitERR_alloc, "%s: unable to create source" );

  source->name = name;

  source->n_gen			= 0;
  source->pos_gen		= NULL;
  source->dir_gen		= NULL;
  source->extent_iter_lim	= ULONG_MAX;
  source->extent_iter_warn	= ULONG_MAX;

  if ( sll_insert_head( sys.sources, source ) )
    tf_exit( ExitERR_misc, "%s: error inserting source into system list",
	     source->name );

  tf_leave( );

  return source;
}

static void
delete_source(
  void *data
)
{
  Source *source = data;

  free( source->name );

  if ( source->flux_gen )
    sll_delete( source->flux_gen, delete_flux_gen );

  /* the following lists contain flux generators,
     which have already been deleted by the preceding
     sll_delete() call.  we just need to delete the
     list framework */
  
  if ( source->phot_valid )
    sll_delete( source->phot_valid, NULL );

  if ( source->phot_valid )
    sll_delete( source->phot_invalid, NULL );

  if ( source->phot_valid )
    sll_delete( source->phot_temp, NULL );

  delete_pos_gen( source->pos_gen );

  delete_dir_gen( source->dir_gen );

  sll_delete( source->extent_events, delete_event );

  free( source );
}


/*

  PROCEDURE
  init_source

  RETURNS
  It exits upon error, else it returns zero.

 */
static int
init_source( void *data )
{
  Source *source = (Source *) data;

  tf_enter( "init_source" );

  init_pos_gen( source->pos_gen );

  if ( source->dir_gen )
    source->dir_gen->spec.init( &source->dir_gen->data );

  sll_traverse( source->flux_gen, init_flux_gen );

  init_photon_times( source );

  if ( sll_insert( sys.gen, sll_head( source->phot_valid ) ) )
    tf_exit( ExitERR_alloc, "%s: error attaching to system photon list",
	     source->name );

  tf_leave( );
  return 0;
}
