/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of raygen
 *
 * raygen is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * raygen is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <float.h>

#include <mst_rdb/mst_rdb.h>
#include <linklist/linklist.h>

#include <luasup/luasup.h>

#include <randlib_c/randlib.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include <suplib/str.h>

#include "main.h"
#include "lua_interface.h"
#include "misc.h"

/* the names of possible spectrum optargs */

#define FORMAT	"format"
#define UNITS	"units"
#define SCALE	"scale"
#define EMIN	"emincol"
#define EMAX	"emaxcol"
#define FLUXVAL	"fluxcol"
#define ITERLIM "iterlim"
#define ITERWARN "iterwarn"
#define FILENAME "file"

/* default values for column names */

#define DEF_EMIN "emin"
#define DEF_EMAX "emax"
#define DEF_FLUXVAL "flux"


typedef enum
{
  DF_RDB
} DataFormat;

static TokListToken
format_type[] =
{
  { "rdb",		DF_RDB		},
};
static TokList format_types = GenTokList(format_type);

typedef enum
{
  DU_Photons,
  DU_IntensityErgs,
  DU_IntensityKeV
} DataUnits;

static TokListToken
unit_type[] =
{
  { "ergs/s/cm2/kev",	DU_IntensityErgs },
  { "kev/s/cm2/kev",	DU_IntensityKeV  },
  { "photons/s/cm2",	DU_Photons	 }
};
static TokList unit_types = GenTokList(unit_type);


typedef struct
{
  double e_min;		/* minimum energy in bin */
  double e_max;		/* maximum energy in bin */
  double p_min;		/* minimum scaled probability of pixel */
  double p_max;		/* maximum scaled probability of pixel */
  double binval;	/* input weight of the bin */
} SpectrumBin;

typedef struct
{
  char *e_min;	/* name of column which contains minimum
		   energy of bins */
  char *e_max;	/* name of column which contains maximum
		   energy of bins */
  char *binval;	/* name of column which contains value of bins */
} RDB_Info;

typedef struct
{
  char *	file;
  char *	name;
  DataUnits	units;		/* units of the data */
  DataFormat	format;		/* format of the data file */
  void *	info;		/* format specific information */
  void		(*info_cleanup)(void *);
  SpectrumBin *	bins;		/* the data */
  double	scale;	/* the scale factor */
  long		nbins;		/* number of bins */
  
} SpectrumData;

static void	define(void);
static int	init(FluxGeneratorData *data);
static int	generate(void *data, double *energy);
static void	cleanup(void *data);
static int	comp( const void *key, const void *data );
static int	rd_rdb_spectrum( const char *err_pfx, SpectrumData *
				 spectrum );

static void	rdb_info_cleanup( void *info );

Module spectrum_file_module = { "spectrum", define };

static FluxGeneratorSpec spec =
{
  "spectrum file",
  init,
  generate,
  cleanup
};

static void
define( void )
{
  lua_Object optargs = LUA_NOOBJECT;
  SpectrumData *data;
  char *name, *misc;
  char err_pfx[1024];

  tf_enter( "spectrum file define" );

  slua_args( define_error, "spectrum",
	     "name",		SLua_STRING,	&name,
	     "optargs",	SLua_TABLE | SLua_OPTIONAL,	&optargs,
	     NULL
	     );

  /* generate error message prefix */
  sprintf( err_pfx, "spectrum `%s'", name );

  if ( NULL == ( data = (SpectrumData *) malloc( sizeof(SpectrumData) ) ) )
    define_error("%s: error allocating memory", err_pfx );

  data->units = DU_Photons;
  data->scale = 1.0;
  data->format = DF_RDB;
  data->name = name;

  /* copy optargs, so can delete elements from it as we go along */
  optargs = optargs == LUA_NOOBJECT ? lua_createtable() : slua_tbl_copy( optargs );

  /**********************************************************************
    Spectrum File Name

    For now, this is required.  later one will be able to generate or
    reuse internal spectra

  */
  switch( slua_tbl_el_s( optargs, FILENAME, &data->file ) )
  {
  case 0:
    slua_tbl_el_delete( optargs, FILENAME );
    break;

  case -1:
    define_error( "%s: out of memory allocating file name", err_pfx );
    /* NOT REACHED */
    break;

  case 1:
    define_error( "%s: `file' attribute must be specified", err_pfx );
    /* NOT REACHED */
    break;
  }

  /* generate new error message prefix, include file name */
  sprintf( err_pfx, "spectrum `%s': file `%s':", name, data->file );

  /**********************************************************************
   Spectrum File Format
  */
  if ( (misc = slua_tbl_el_si( optargs, FORMAT )) )
  {
    if ( -1 == (data->format = tokmatch( misc, &format_types )) )
      define_error( "%s: illegal file format: `%s'", err_pfx, misc );

    slua_tbl_el_delete( optargs, FORMAT );
  }
  
  /**********************************************************************
   Spectrum File Units
  */
  if ( (misc = slua_tbl_el_si( optargs, UNITS )) )
  {
    if ( -1 == (data->units = tokmatch( misc, &unit_types )) )
      define_error( "%s: illegal units: `%s'", err_pfx, misc );

    slua_tbl_el_delete( optargs, UNITS );
  }

  /**********************************************************************
   Scale Factor
  */
  slua_tbl_el_n( optargs, SCALE, &data->scale );
  slua_tbl_el_delete( optargs, SCALE );
 
  /**********************************************************************
   Format Specific stuff
  */

  switch ( data->format )
  {
  case DF_RDB:
    {
      RDB_Info *info;
      if ( NULL == (data->info = (void *) malloc( sizeof(RDB_Info) )))
	define_error( err_pfx, "error allocating file info structure" );
      info = (RDB_Info*) data->info;
      info->e_min = get_string( err_pfx, optargs, EMIN, DEF_EMIN );
      info->e_max = get_string( err_pfx, optargs, EMAX, DEF_EMAX );
      info->binval = get_string( err_pfx, optargs, FLUXVAL, DEF_FLUXVAL );
      data->info_cleanup = rdb_info_cleanup;
      break;
    }

  default:
    tf_assert( 0 );
  }

  /* we should have read (and deleted) all relevant optargs.  Those
     that remain should be flagged as errors */
  verify_empty_optargs( err_pfx, optargs );

  new_flux_gen( &spec, name, data );

  tf_leave( );
}

static int
init( FluxGeneratorData *data )
{
  double p_sum, p_min;
  SpectrumBin *binp;
  char err_pfx[1024];

  SpectrumBin *bin;
  SpectrumBin *bin_lim;
  SpectrumData *spectrum = (SpectrumData *) data->data;

  sprintf( err_pfx, "spectrum `%s': file `%s'", spectrum->name,
	   spectrum->file );

  tf_enter( "spectrum file init" );

  data->iter_lim = 100;
  data->iter_warn = 10;

  /* read in the data */
  switch ( spectrum->format )
  {
  case DF_RDB:
      
    if ( rd_rdb_spectrum( err_pfx, spectrum ) )
      return 1;
    break;

  default:
    tf_assert( 0 );
    break;
  }


  /*
    loop through the data, get the normalized probabilities set up
    and determine the average counts/s/cm2
   */

  bin = spectrum->bins;
  bin_lim = spectrum->bins + spectrum->nbins;
  p_sum = 0.0;
  for ( binp = bin ; binp < bin_lim ; binp++ )
  {
    double e_ave = (binp->e_max + binp->e_min) / 2;
    double e_width = binp->e_max - binp->e_min;

    /* convert data to counts/s/cm2 */
    switch( spectrum->units )
    {
    case DU_Photons:
      break;

    case DU_IntensityErgs:
      /* convert from ergs/s/cm2/keV to counts/s/cm2 */
      binp->binval *= e_width / ( e_ave * 1.6021892e-9 ); /* = ergs/keV */
      break;

    case DU_IntensityKeV:
      /* convert from keV/s/cm2/keV to counts/s/cm2 */
      binp->binval *= e_width / e_ave;
      break;

    default:
      tf_assert( 0 );
    }
    binp->binval *= spectrum->scale;
    p_sum += binp->binval;
  }

  p_min = 0.0;
  data->base_flux = 0;
  for ( binp = bin ; binp < bin_lim ; binp++ )
  {
    binp->p_min = p_min;
    binp->p_max = binp->binval / p_sum + p_min;
    p_min = binp->p_max;

    data->base_flux += binp->binval;
  }

  tf_leave( );
  return 0;
}

static int
generate( void *data, double *energy )
{
  SpectrumData *spectrum = (SpectrumData *) data;
  SpectrumBin *bin;
  double p;

  /* grab a bin */
  p = ranf();

  bin = (SpectrumBin *) bsearch( &p, spectrum->bins, spectrum->nbins,
  			         sizeof(SpectrumBin), comp );
  tf_assert5( NULL != bin );

  *energy = genunf(bin->e_min, bin->e_max);

  return 0;
}

static void
cleanup( void *data )
{
  SpectrumData *spectrum = (SpectrumData *) data;

  if ( ! data )
    return;

  ffree( spectrum->bins );
  ffree( spectrum->file );
  ffree( spectrum->name );
  spectrum->info_cleanup( spectrum->info ); 
  ffree( spectrum );
}

static void
rdb_info_cleanup( void *info )
{
  if ( info )
  {
    RDB_Info *ptr = (RDB_Info *) info;
    ffree( ptr->e_min );
    ffree( ptr->e_max );
    ffree( ptr->binval );
    ffree( ptr );
  }
}

static int
rd_rdb_spectrum(
  const char *err_pfx,
  SpectrumData * spectrum
)
{
  rdbHeader *rdb = NULL;
  DataColumnMap_st *map = NULL;
  SpectrumBin bin, *bin_p, *bin_lim;
  FILE *fin = NULL;
  RDB_Info *info = (RDB_Info *) spectrum->info;

  static RDBFieldStInfo rdb_fields[] =
  {
    { NULL,	RDB_Num,	offsetof( SpectrumBin, e_min ) },
    { NULL,	RDB_Num,	offsetof( SpectrumBin, e_max ) },
    { NULL,	RDB_Num,	offsetof( SpectrumBin, binval ) }
  };
#define NFIELDS ( sizeof(rdb_fields) / sizeof(RDBFieldStInfo) )

  tf_enter( "rd_spectrum_rdb" );

  rdb_fields[0].name = info->e_min;
  rdb_fields[1].name = info->e_max;
  rdb_fields[2].name = info->binval;

  spectrum->bins = NULL;
  
  if ( NULL == (fin = fopen(spectrum->file, "r" )) )
  {
    tf_message( "%s: error opening file\n", err_pfx );
    goto error;
  }

  if ( NULL == ( rdb = rdb_rd_hdr( fin ) ) )
  {
    tf_message( "%s: error reading rdb header\n", err_pfx );
    goto error;
  }

  map = rdb_map_cols_stst( rdb, NFIELDS, rdb_fields );

  /* determine number of bins */
  spectrum->nbins = 0;
  while( rdb_col_read_st( fin, rdb, map, &bin ) )
    if ( bin.binval > 0.0 )
      spectrum->nbins++;

  if ( ferror( fin ) )
  {
    tf_message( "%s: error reading data\n", err_pfx );
    goto error;
  }

  if ( 0 == spectrum->nbins )
  {
    tf_message( "%s: no bins had a positive value\n", err_pfx );
    goto error;
  }

  if ( NULL ==
       (spectrum->bins = (SpectrumBin *) malloc( spectrum->nbins * sizeof(SpectrumBin) )) )
  {
    tf_message( "%s: unable to allocate bin array of %d elements\n",
		err_pfx, spectrum->nbins );
    goto error;
  }

  /* now, read in bins */
  if ( rdb_rewind( fin, rdb ) )
  {
    tf_message( "%s: unable to rewind. is this a real file?\n", err_pfx );
    goto error;
  }

  bin_p = spectrum->bins;
  bin_lim = bin_p + spectrum->nbins;
  while( bin_p < bin_lim && rdb_col_read_st( fin, rdb, map, bin_p ) )
  {
    tf_message( "%g\n", bin_p->binval );
    if ( bin_p->binval <= 0.0 )
      continue;
    bin_p++;
  }

  if ( bin_p != bin_lim )
  {
    tf_message( "%s: unexpected EOF\n", err_pfx );
    goto error;
  }
  
  fclose( fin );
  rdb_free_map( map );
  rdb_free_hdr( rdb );

  tf_leave( );
  return 0;

  error:
  if ( fin ) fclose( fin );
  if ( map ) rdb_free_map( map );
  if ( rdb ) rdb_free_hdr( rdb );
  if ( spectrum->bins )
  {
    free( spectrum->bins );
    spectrum->bins = NULL;
    spectrum->nbins = 0;
  }

  tf_leave( );
  return 1;
}

static int
comp( const void *key, const void *data )
{
  return *(double *) key < ((SpectrumBin *)data)->p_min ? -1 :
         *(double *) key > ((SpectrumBin *)data)->p_max ?  1 :
         0;
}

