// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of RDB
//
// RDB is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// RDB is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <cstdlib>
#include "RDBColumn.h"

/**
 * \param is input stream.
 * \param col RDBColumn to fill.
 *
 * \return A reference to the input stream.
 *
 * Reads one data element into this object.
 *
 */
istream&
operator>>( 
  istream& is, 
  RDBColumn& col 
  ) throw ( RDBErr ) {

  try { 
    col.read( is );
    
  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "operator>>(istream&,RDBColumn&):  " );
    throw( rdberr );

  } catch ( ... ) {
    throw( RDBErr( "operator>>(istream&,RDBColumn&):  unexpected exception caught" ) );

  }

  return is;

}

/**
 * \param is input stream.
 * \param col RDBColumn to fill.
 *
 * \return A reference to the input stream.
 *
 * Reads one data element into this object.
 *
 */
istream&
operator>>( 
  istream& is, 
  RDBColumn* col 
  ) throw ( RDBErr ) {

  try { 
    col->read( is );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "operator>>(istream&,RDBColumn*):  " );
    throw( rdberr );

  } catch ( ... ) {
    throw( RDBErr( "operator>>(istream&,RDBColumn*):  unexpected exception caught" ) );

  }

  return is;

}

/**
 * \param os output stream.
 * \param col RDBColumn to fill.
 *
 * \return A reference to the output stream.
 *
 * Writes one data element from this object.
 *
 */
ostream&
operator<<( 
  ostream& os, 
  const RDBColumn& col 
  ) {

  return col.write( os );

}

/**
 * \param os output stream.
 * \param col RDBColumn to fill.
 *
 * \return A reference to the output stream.
 *
 * Writes one data element from this object.
 *
 */
ostream&
operator<<( 
  ostream& os, 
  const RDBColumn* col 
  ) {

  return col->write( os );

}
  
/**
 * \param name the column name.
 * \param def the column definition.
 *
 * Initializes the RDBColumn name and definition.  Sets the exception
 * throwing behavior to true.  
 *
 */
RDBColumn::RDBColumn( 
  const string& name,
  const string& def
  ) throw ( RDBErr )
  : _name(name), _def(def), _width(-1), _type(STRING), _just(LEFT),
    _desc(""), _changed(false), _throw(true), _errno(0),
    _precision(DBL_DIG), _group(false), _initgroup(true) {

  if ( !_def.empty( ) ) {
    try { 
      setDef( _def );

    } catch ( RDBErr& rdberr ) {
      rdberr.set_message( "RDBColumn::RDBColumn(string&,string&):  " );
      throw( rdberr );

    } catch ( ... ) { 
      throw( RDBErr( "RDBColumn::RDBColumn(string&,string&):  unexpected exception caught" ) );

    }
  }
}

/**
 * \param col the RDBColumn object to copy.
 *
 * Copies the argument. 
 *
 */
RDBColumn::RDBColumn( 
  const RDBColumn& col
  ) 
  : _name(col._name), _def(col._def), _width(col._width),
    _type(col._type), _just(col._just), _desc(col._desc),
    _changed(col._changed), _throw(col._throw), _errno(col._errno),
    _precision(col._precision), _group(col._group),
    _initgroup(col._initgroup) {

}

/**
 * Nothing to do.
 *
 */
RDBColumn::~RDBColumn( 
  void 
  ) {

}

/**
 * \param col the RDBColumn object to copy.
 *
 */
RDBColumn& 
RDBColumn::operator=( 
  const RDBColumn& col 
  ) {

  if ( this != &col ) {
    _name      = col._name;
    _def       = col._def;
    _width     = col._width;
    _type      = col._type;
    _just      = col._just;
    _desc      = col._desc;
    _throw     = col._throw;
    _errno     = col._errno;
    _precision = col._precision;
    _changed   = col._changed;
    _group     = col._group;

  }

  return *this;

}

/**
 * \param group incidcates whether or not this is a group column.
 *
 */
void
RDBColumn::setGroup( 
  bool group
  ) {

  _group = group;

}

/**
 * \return Whether or not this is a group column.
 *
 */
bool
RDBColumn::getGroup( 
  void
  ) const {

  return _group;

}

/**
 * \param name the RDBColumn name to use.
 *
 * Sets the RDBColumn name.
 *
 */
void
RDBColumn::setName( 
  const string& name
  ) {

  _name = name;

}

/**
 * \param def the RDBColumn definition to use.
 *
 * \exception RDBErrColFmt error if the definition is an unrecognized
 * format. 
 * 
 * Sets the RDBColumn definition, as well as the width, type,
 * justification, and description.
 *
 */
void
RDBColumn::setDef(
  const string& def
  ) throw ( RDBErr ) {

  _def = def;
  
  _width = -1;
  _type  = STRING;
  _just  = LEFT;
  _desc.erase( );

  size_t defsize = def.size( );
  char* cdef = new char[defsize+1];
  strcpy( cdef, def.c_str() );
  //  cdef[defsize] = '\0';
  int n = 0;

  if ( isdigit( cdef[0] ) ) {
    char* ptr = NULL;
    _width = strtol( cdef, &ptr, 10 );
    if ( ptr ) {
      n = ptr - cdef;

    }
  }

  if ( NUMERIC == cdef[n] || STRING == cdef[n] || MONTH == cdef[n] ) {
    _type = Type(cdef[n++]);

  } else if ( '-' == cdef[n] ) {
    _type = STRING;
    while ( '-' == cdef[n] ) {
      n++;

    }
  } else {
    delete [] cdef;
    throw( RDBErr( string("RDBColumn::setDef(string&):  Bad type in column defintion '") + def + "' for '" + _name + "'" ) );

  }  

  if ( '\0' != cdef[n] && ( LEFT == cdef[n] || RIGHT == cdef[n] ) ) {
    _just = Just(cdef[n++]);

  }

  if ( '\0' != cdef[n] ) {
    if ( ' ' == cdef[n] ) {
      while ( ' ' == cdef[n] ) { 
	n++; 

      }
      _desc = &cdef[n];
    }
    else {
      delete [] cdef;
      throw( RDBErr( string("RDBColumn::setDef(string&):  Bad just in column defintion '") + def + "' for '" + _name + "'" ) );

    }
  }

  _changed = false;
  delete [] cdef;

}

/**
 * \param width the width of the RDBColumn.
 *
 * The width has no effect on the storage or printing of the data.
 * The RDB docs indicate that it is used only by the ptbl command.
 *
 */
void
RDBColumn::setWidth(
  const long width
  ) {

  _width   = width;
  _changed = true;

}

/**
 * \param type of the RDBColumn.
 *
 * Possible values are 'M' for month, 'N' for numeric, 'S' for
 * string.  String is the default type.
 *
 */
void
RDBColumn::setType( 
  const RDBColumn::Type type
  ) {

  _type    = type;
  _changed = true;

}

/**
 * \param just the justification of the data when printing.
 *
 * Possible values are '<' for left justification and '>' for right
 * justification.  Default is for months and strings to be left
 * justified and numeric data to be right justified.
 *
 */
void
RDBColumn::setJust( 
  const RDBColumn::Just just
  ) {

  _just    = just;
  _changed = true;

}

/**
 * \param desc is the column documentation.
 *
 * Sets the RDBColumn description.
 *
 */
void
RDBColumn::setDesc( 
  const string& desc 
  ) {

  _desc    = desc;
  _changed = true;

}

/**
 * \param precision
 *
 * Sets the RDBColumn precision which controls the output precision of
 * floating point data.
 *
 */
void
RDBColumn::setPrecision( 
  const int precision
  ) {

  _precision = precision;

}

/**
 * \param t state of the exception throwing behavior.
 *
 * Sets the RDBColumn exception throwing behavior.
 *
 */
void
RDBColumn::setThrow( 
  const bool t
  ) {

  _throw = t;

}

/**
 * \param no
 *
 * Sets the RDBColumn error status.
 *
 */
void
RDBColumn::setErrNo(
  const int no
  ) {

  _errno = no;

}

/// Maps data to user-supplied memory, if possible.
void
RDBColumn::mapData( 
  double data[], 
  const size_t nelems 
  ) throw ( RDBErr ) {

  throw( RDBErr( string("RDBColumn::mapData(double[],size_t):  Wrong data type:  double[] in column '") + _name + "'" ) );

}

/// Maps data to user-supplied memory, if possible.
void
RDBColumn::mapData( 
  long data[], 
  const size_t nelems
  ) throw ( RDBErr ) {

  throw( RDBErr( string("RDBColumn::mapData(long[],size_t):  Wrong data type:  long[] in column '") + _name + "'" ) );

}

/// Maps data to user-supplied memory, if possible.
void
RDBColumn::mapData( 
  string data[], 
  const size_t nelems
  ) throw ( RDBErr ) {

  throw( RDBErr( string("RDBColumn::mapData(stringdouble[],size_t):  Wrong data type:  string[] in column '") + _name + "'" ) );

}

/**
 * Returns the RDBColumn name.
 *
 */
string 
RDBColumn::getName( 
  void
  ) const {

  return _name;

}

/**
 * Returns the RDBColumn definition, reconstructing it from its
 * constituent parts if any of them have changed.
 *
 */
string
RDBColumn::getDef(
  void
  ) {

  if ( _changed ) {
    _def.erase( );
    _strstrm.seekp( 0 );
    if ( ! _strstrm.good( ) )
      _strstrm.clear( );
    if ( -1 != _width )
      _strstrm << _width;

    if ( RDBColumn::NUMERIC == _type )
      _strstrm << 'N';
    else if ( RDBColumn::STRING == _type ) 
      _strstrm << 'S';
    else 
      _strstrm << 'M';

    if ( RDBColumn::LEFT == _just )
      _strstrm << '<';
    else 
      _strstrm << '>';

    if ( 0 != _desc.length( ) )
      _strstrm << ' ' << _desc;

    _strstrm << '\0';
    _def     = _strstrm.str( ).c_str( );
    _changed = false;

  }

  return _def;
  
}

/**
 * Returns the RDBColumn width.
 *
 */
long 
RDBColumn::getWidth( 
  void
  ) const {

  return _width;

}

/**
 * Returns the RDBColumn type.
 *
 */
RDBColumn::Type
RDBColumn::getType(
  void
  ) const {

  return _type;

}

/**
 * Returns the RDBColumn justification.
 *
 */
RDBColumn::Just
RDBColumn::getJust(
  void
  ) const {

  return _just;

}

/**
 * Returns the RDBColumn description.
 *
 */
string
RDBColumn::getDesc(
  void
  ) const {

  return _desc;
  
}

/**
 * Returns the RDBColumn precision used for stream output.
 *
 */
int
RDBColumn::getPrecision(
  void
  ) const {

  return _precision;
  
}

/**
 * Returns the RDBColumn exception throwing flag.
 *
 */
bool
RDBColumn::getThrow(
  void
  ) const {

  return _throw;
  
}

/**
 * Returns a string describing the RDBColumn error flag.
 *
 */
char*
RDBColumn::getErr(
  void
  ) const {

  return strerror( _errno );

}

/**
 * Returns the RDBColumn error flag.
 *
 */
int
RDBColumn::getErrNo(
  void
  ) const {

  return _errno;  

}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * Handles the trivial conversion for double to double for child classes. 
 *
 */
void
RDBColumn::convert(
  const double& idata,
  double& odata
  ) throw ( RDBErr ) {

  odata = idata;

}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * Handles the trivial conversion for double to long for child classes. 
 *
 */
void
RDBColumn::convert(
  const double& idata,
  long& odata
  ) throw ( RDBErr ) {

  odata = (long) idata;

}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * Handles the non-trivial conversion for double to string for child
 * classes.
 *
 * \warning This is slow.  Like dirt.  Don't do it often.
 *
 */
void
RDBColumn::convert(
  const double& idata,
  string& odata
  ) throw ( RDBErr ) {

  _strstrm.seekp( 0 );
  if ( ! _strstrm.good( ) )
    _strstrm.clear( );
  _strstrm << setprecision( _precision ) << idata << '\0';
  odata = _strstrm.str( ).c_str( );

}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * Handles the trivial conversion for long to double for child classes. 
 *
 */
void
RDBColumn::convert(
  const long& idata,
  double& odata
  ) throw ( RDBErr ) {

  odata = (double) idata;  

}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * Handles the trivial conversion for long to long for child classes. 
 *
 */
void
RDBColumn::convert(
  const long& idata,
  long& odata
  ) throw ( RDBErr ) {
  
  odata = idata;

}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * Handles the non-trivial conversion for long to string for child classes. 
 *
 * \warning This is slow.  Like dirt.  Don't do it often.
 *
 */
void
RDBColumn::convert(
  const long& idata,
  string& odata
  ) throw ( RDBErr ) {

  _strstrm.seekp( 0 );
  if ( ! _strstrm.good( ) )
    _strstrm.clear( );
  _strstrm << idata << '\0';
  odata = _strstrm.str( ).c_str( );  

}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * \exception RDBErr error if idata is non-numeric.
 *
 * Handles the non-trivial conversion for string to double for child classes. 
 *
 */
void
RDBColumn::convert(
  const string& idata,
  double& odata
  ) throw ( RDBErr ) {

  errno = 0;
  char* endptr = NULL;
  odata = strtod( idata.length( ) ? idata.c_str( ) : "NaN", &endptr );
  _errno = errno;
  if ( ( endptr && ( *endptr != '\0' ) ) ) {
    if ( _throw ) {
      throw( RDBErr( string("RDBColumn::convert(string&,double&):  Non numeric data in column '") + _name + "'  at '" + endptr + "'" ) );
        
    } else {
      _errno = RDBColumn::OUTOFRANGE;
        
    }
  }

}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * \exception RDBErr error if idata is non-numeric.
 * \exception RDBErr error if idata represents a floating point number and precision is lost converting to integer number.
 *
 * Handles the non-trivial conversion for string to long for child classes. 
 *
 */
void
RDBColumn::convert(
  const string& idata,
  long& odata
  ) throw ( RDBErr ) {

  errno = 0;
  char* endptr = NULL;
  odata = strtol( idata.length( ) ? idata.c_str( ) : "NaN", &endptr, 10 );
  _errno = errno;
  if ( ( endptr && *endptr != '\0' ) ) {
    if ( *endptr != '.' ) {
      if ( _throw ) {
	throw( RDBErr( string("RDBColumn::convert(string&,long&):  Non numeric data in column '") + _name + "' at '" + endptr + "'" ) );
	  
      } else {
	_errno = RDBColumn::NONNUMERIC;
	  
      }
    } else {
      char* beginptr = endptr+1;
      endptr = NULL;
      strtol( beginptr, &endptr, 10 );
      _errno = errno;
      if ( ( endptr && *endptr != '\0' ) ) {
	if ( _throw ) {
	  throw( RDBErr( string("RDBColumn::convert(string&,long&):  Non numeric data in column '") + _name + "'at '" + endptr + "'" ) );
	  
	} else {
	  _errno = RDBColumn::NONNUMERIC;
	  
	}
      } else {
	if ( _throw ) {
	  throw( RDBErr( string("RDBColumn::convert(string&,long&):  Lost precision in column '") + _name + "'" ) );
	  
	} else {
	  _errno = RDBColumn::LOSTPRECISION;
	  
	}
      }
    }
  }
}

/**
 * \param idata input data.
 * \param odata output data.
 *
 * Handles the trivial conversion for string to string for child classes. 
 *
 */
void
RDBColumn::convert(
  const string& idata,
  string& odata
  ) throw ( RDBErr ) {

  odata = idata;

}

/**
 * \param is input string.
 * \param data double data.
 *
 * \return istream the input stream.
 *
 * Extracts a double from the input stream.
 *
 */
istream& 
RDBColumn::extract( 
  istream& is,
  double& data
  ) throw ( RDBErr ) {

  is >> data;
  if ( ! is.good( ) ) {
    throw( RDBErr( string("RDBColumn::extract(istream&,double&):  Non numeric data in column '") + _name + "'" ) );
  }

  return is;

}

/**
 * \param is input string.
 * \param data long data.
 *
 * \return istream the input stream.
 *
 * Extracts a long from the input stream.
 *
 */
istream& 
RDBColumn::extract( 
  istream& is,
  long& data
  ) throw ( RDBErr ) {

  is >> data;
  if ( ! is.good( ) ) {
    throw( RDBErr( string("RDBColumn::extract(istream&,long&):  Non numeric data in column '") + _name + "'" ) );

  }

  return is;

}

/**
 * \param is input string.
 * \param data string data.
 *
 * \return istream the input stream.
 *
 * Extracts a string from the input stream.  Extraction stops at the
 * first tab or newline character.
 *
 */
istream& 
RDBColumn::extract( 
  istream& is,
  string& data
  ) throw ( RDBErr ) {

  string tmp;
  is >> data;
  if ( ! is.good( ) ) {
    throw( RDBErr( string("RDBColumn::extract(stream&,string&):  Row format error in column '") + _name + "'" ) );

  }

  while ( ' ' == is.peek( ) ) {
    is >> tmp;
    if ( ! is.good( ) ) {
      throw( RDBErr( string("RDBColumn::extract(stream&,string&):  Row format error in column '") + _name + "'" ) );

    }
    data += ' ' + tmp;

  }

  return is;

}

/**
 * \param os output string.
 * \param data double data.
 *
 * \return ostream the output stream.
 *
 * Inserts a double into the output stream.
 *
 */
ostream& 
RDBColumn::insert( 
  ostream& os,
  double& data
  ) const {

  return os << setprecision( _precision ) << data;

}

/**
 * \param os output string.
 * \param data double data.
 *
 * \return ostream the output stream.
 *
 * Inserts a long into the output stream.
 *
 */
ostream& 
RDBColumn::insert( 
  ostream& os,
  long& data
  ) const {

  return os << setprecision( _precision ) << data;

}

/**
 * \param os output string.
 * \param data double data.
 *
 * \return ostream the output stream.
 *
 * Inserts a string into the output stream.
 *
 */
ostream& 
RDBColumn::insert( 
  ostream& os,
  string& data
  ) const {

  return os << setprecision( _precision ) << data;

}
