// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of RDB
//
// RDB is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// RDB is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include "RDBComment.h"

/**
 * \param is the input stream.
 * \param rdbcomment the comment to fill.
 *
 * \return A reference to the input stream.
 *
 * If the line is an RDB comment line, i.e. it starts with a '#', then
 * fill the comment.  Otherwise, set ios::failbit.
 *
 */
istream&
operator>>( 
  istream& is, 
  RDBComment& rdbcomment 
  ) throw ( RDBErr ) {

  string line;
  getline( is, line, '\n' );
  try { 
    rdbcomment.setComment( line );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "operator>>(istream&,RDBComment&):  " );
    throw( rdberr );

  } catch ( ... ) {
    throw( RDBErr( "operator>>(istream&,RDBComment&):  unexpected exception caught" ) );

  }

  return is;

}

/**
 * \param os the output stream.
 * \param rdbcomment the comment to print.
 *
 * \return A reference to the output stream
 *
 * Places the comment on the output stream.  Appends a '#' character
 * if its a plain comment or a "#:" string if its a keyword/value
 * comment. 
 *
 */
ostream& 
operator<<( 
  ostream& os, 
  const RDBComment& rdbcomment 
  ) {

  os <<
    (
      (
	0 != rdbcomment._keyword.size( ) 
	|| 
	0 != rdbcomment._value.size( )
      )
      ?
      ("#:")
      :
      ("# ")
    ) << rdbcomment._comment;

  return os;

}

/**
 * \param comment the comment string.
 *
 * Assigns the string comment to the RDBComment's data member.  Parses
 * for keyword=value pairs.
 *
 */
RDBComment::RDBComment( 
  const string& comment 
  ) throw ( RDBErr ) {

  try {
    setCommentText( comment );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "RDBComment::RDBComment(string&)" );
    throw( rdberr );

  } catch ( ... ) {
    throw( RDBErr( "RDBComment::RDBComment(string&):  unexpected exception caught" ) );

  }
}

/**
 * \param rdbcomment a reference to the RDBComment to copy.
 *
 * Copies the RDBComment into this object.
 *
 */
RDBComment::RDBComment( 
  const RDBComment& rdbcomment
  ) {

  operator=( rdbcomment );

}

/**
 * Nothing to do.
 *
 */
RDBComment::~RDBComment(
  void 
  ){

}

/**
 * \param rdbcomment a reference to the RDBComment to copy.
 *
 * \returns A reference to this.
 *
 */
const RDBComment&
RDBComment::operator=( 
  const RDBComment& rdbcomment
  ) {

  if ( this != &rdbcomment ) {
    _comment = rdbcomment._comment;
    _keyword = rdbcomment._keyword;
    _value   = rdbcomment._value;

  }

  return *this;

}

/**
 * \param comment a string comment to copy.
 *
 * \returns A reference to this.
 *
 */
const RDBComment&
RDBComment::operator=( 
  const string& comment
  ) {

  setCommentText( comment );

  return *this;

}

/**
 * \param comment
 *
 * Strips leading '#' character if present.  If next character is ':',
 * then it searches for a keyword=value pair.
 *
 */
void
RDBComment::setComment( 
  const string& comment
  ) throw ( RDBErr ) {
  
  size_t pos = comment.find_first_not_of( " " );
  if ( 0 == pos ) {
    if ( '#' != comment[0] || ' ' != comment[1] ) {
      throw( RDBErr( "RDBComment::setComment(string&):  Leading characters not `# '" ) );

    }
  } else if ( comment.npos != pos ) { 
    if ( '#' != comment[pos] ) {
      throw( RDBErr( "RDBComment::setComment(string&):  First non-space character not `#'" ) );

    }
  } else {
    throw( RDBErr( "RDBComment::setComment(string&):  No comment found" ) );

  }

  setCommentText( comment );

}

/**
 * \param comment
 *
 * Strips leading '#' character if present.  If next character is ':',
 * then it searches for a keyword=value pair.
 *
 */
void
RDBComment::setCommentText( 
  const string& comment
  ) {

  size_t pos = comment.find_first_not_of( " " );
  _comment = comment;
  _keyword.erase( );
  _value.erase( );
  
  // if the line has a leading '#' character, remove it...
  // users creating comments need not include leading '#'.
  if ( '#' == _comment[pos] ) {
    _comment.erase( 0, pos + 1 );

  }
  
  // see if its a keyword value pair...
  if ( ':' == _comment[0] ) {
    _comment.erase( 0, 1 );
    
    pos = _comment.find( "=", 0 );
    if ( pos > _comment.size( ) )
      pos = _comment.size( );
    
    _keyword = _comment;
    _keyword.erase( pos );
    
    size_t wspos = _keyword.find_first_not_of( " " );
    _keyword.erase( 0, wspos );
    wspos = _keyword.find_last_not_of( " " );
    _keyword.erase( wspos + 1 );
    
    _value = _comment;
    _value.replace( 0, pos+1, "" );
    
    wspos = _value.find_first_not_of( " " );
    _value.erase( 0, wspos );
    wspos = _value.find_last_not_of( " " );
    _value.erase( wspos + 1 );
    
  } else {
    pos = _comment.find_first_not_of( "\t " );
    if ( _comment.npos != pos ) {
      _comment.erase( 0, pos );

    }
  }
}

/**
 * \param keyword the keyword part of a keyword=value pair.
 *
 * Updates just the keyword half of a keyword=value pair.
 *
 */
void
RDBComment::setKeyword( 
  const string& keyword
  ) throw ( RDBErr ) {

  try {
    setComment( string("#:") + keyword + " = " + _value );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "RDBComment::setKeyword(string&):  " );
    throw( rdberr );

  } catch ( ... ) {
    throw( RDBErr( "RDBComment::setKeyword(string&):  unexpected exception caught" ) );

  }
}

/**
 * \param value the value part of a keyword=value pair.
 *
 * Updates just the value half of a keyword=value pair.
 *
 */
void
RDBComment::setValue(
  const string& value
  ) throw ( RDBErr ) { 

  try {
    setComment( string("#:") + _keyword + " = " + value );

  } catch ( RDBErr& rdberr ) {
    rdberr.set_message( "RDBComment::setValue(string&):  " );
    throw( rdberr );

  } catch ( ... ) {
    throw( RDBErr( "RDBComment::setValue(string&):  unexpected exception caught" ) );

  }
}

/**
 * \return The comment string.
 *
 */
string
RDBComment::getComment(
  void
  ) const {

  return 
    (
      (0 != _keyword.size( ) || 0 != _value.size( ))
      ?
      ("#:")
      :
      ("# ")
      +
      _comment
      );
  
}

/**
 * \return The comment string.
 *
 */
string
RDBComment::getCommentText(
  void
  ) const {

  return _comment;

}

/**
 * \return The keyword half of a keyword=value pair.
 *
 */
string
RDBComment::getKeyword(
  void
  ) const {

  return _keyword;

}

/**
 * \return The value half of a keyword=value pair.
 *
 */
string
RDBComment::getValue(
  void
  ) const {

  return _value;

}
