// File:   t_rl_BasicRay.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rl_ray
//
// rl_ray is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_ray is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <rl_basicray/rl_BasicRay.h>

#include <mathconst/mathconst.h>
#include <cstdlib>
#include <iostream>
#include <testlib/testlib.h>
#include <cfloat>

using namespace std;

static
int rl_BasicRay_test01()    /* test ctor, dtor, intensity, operator<< */
{
  double psi = 1.0;
  dvm3_Vector pos( 1.2, 503, 17.7 );
  dvm3_Vector dir( sin(psi * M_DEG2RAD), 0.0, cos(psi * M_DEG2RAD) );
  double      energy = 1.49;
  long int    id = 1402;

  rl_BasicRay ray( pos, dir, energy, id );

  int rc0 = !fuzzy_compare(ray.position()[0],    1.2)
         && !fuzzy_compare(ray.position()[1],  503.0)
         && !fuzzy_compare(ray.position()[2],   17.7);
  int rc1 = !fuzzy_compare(ray.direction()[0], sin(psi * M_DEG2RAD))
         && !fuzzy_compare(ray.direction()[1], 0.0                 )
         && !fuzzy_compare(ray.direction()[2], cos(psi * M_DEG2RAD));
  int rc2 = !fuzzy_compare(ray.energy(),       double(1.49));
  int rc3 = (ray.id() == (long int)1402);
  return  rc0 && rc1 && rc2 && rc3;
}

static
int rl_BasicRay_test02()    /* test copy ctor */
{
  double psi = 1.0;
  dvm3_Vector pos( 1.2, 503, 17.7 );
  dvm3_Vector dir( sin(psi * M_DEG2RAD), 0.0, cos(psi * M_DEG2RAD) );
  double      energy = 1.49;
  long int    id = 1402;

  rl_BasicRay foo( pos, dir, energy, id );
  rl_BasicRay ray( foo );

  int rc0 = !fuzzy_compare(ray.position()[0],    1.2)
         && !fuzzy_compare(ray.position()[1],  503.0)
         && !fuzzy_compare(ray.position()[2],   17.7);
  int rc1 = !fuzzy_compare(ray.direction()[0], sin(psi * M_DEG2RAD))
         && !fuzzy_compare(ray.direction()[1], 0.0                 )
         && !fuzzy_compare(ray.direction()[2], cos(psi * M_DEG2RAD));
  int rc2 = !fuzzy_compare(ray.energy(),       double(1.49));
  int rc3 = (ray.id() == (long int)1402);
  return  rc0 && rc1 && rc2 && rc3;
}

static
int rl_BasicRay_test03()    /* test reflect_direction_vector */
{
  double psi = 1.0;
  dvm3_Vector pos( 1.2, 503, 17.7 );
  dvm3_Vector dir( sin(psi * M_DEG2RAD), 0.0, cos(psi * M_DEG2RAD) );
  double      energy = 1.49;
  long int    id = 1402;

  rl_BasicRay ray( pos, dir, energy, id );

  dvm3_Vector norm( 1.0, 0.0, 0.0 );

  double sg = ray.reflect_direction_vector( norm );

  int rc0 = !fuzzy_compare(ray.position()[0],    1.2)
         && !fuzzy_compare(ray.position()[1],  503.0)
         && !fuzzy_compare(ray.position()[2],   17.7);
  int rc1 = !fuzzy_compare(ray.direction()[0], -sin(psi * M_DEG2RAD))
         && !fuzzy_compare(ray.direction()[1],  0.0                 )
         && !fuzzy_compare(ray.direction()[2],  cos(psi * M_DEG2RAD));
  int rc2 = !fuzzy_compare(sg,                  sin(psi * M_DEG2RAD));
  return  rc0 && rc1 && rc2;
}

static
int rl_BasicRay_test04()    /* test translate_rotate/derotate_detranslate */
{
  double psi = 1.0;
  dvm3_Vector pos( 1.2, 503.0, 17.7 );
  dvm3_Vector dir( sin(psi * M_DEG2RAD), 0.0, cos(psi * M_DEG2RAD) );
  double      energy = 1.49;
  long int    id = 1402;

  rl_BasicRay ray( pos, dir, energy, id );

  // note:  test matrix from equation 4-46 of Goldstein, "Classical Mechanics",
  //        2nd edition.
  double theta = 0.1;
  double phi   = 0.3;
  double psi1  = 0.2;
  dvm3_Vector v0(  cos(psi1)*cos(phi) - cos(theta)*sin(psi1)*sin(phi),
                   cos(psi1)*sin(phi) + cos(theta)*sin(psi1)*cos(phi),
                                        sin(theta)*sin(psi1)
                 );
  dvm3_Vector v1( -sin(psi1)*cos(phi) - cos(theta)*cos(psi1)*sin(phi),
                  -sin(psi1)*sin(phi) + cos(theta)*cos(psi1)*cos(phi),
                                        sin(theta)*cos(psi1)
                 );
  dvm3_Vector v2(                       sin(theta)*sin(phi),
                                       -sin(theta)*cos(phi),
                                        cos(theta)
                 );
  dvm3_RotMat m1( v0, v1, v2 );
  dvm3_Vector trn( 1.0, 2.0, 3.0 );
  ray.translate_rotate( trn, m1 );
  ray.derotate_detranslate( trn, m1 );
  double const tol = 1000.0 * DBL_EPSILON;
  
  int rc0 = !Test_fuzzy_compare(ray.position()[0],   1.2, tol)
         && !Test_fuzzy_compare(ray.position()[1], 503.0, tol)
         && !Test_fuzzy_compare(ray.position()[2],  17.7, tol);
  int rc1 = !Test_fuzzy_compare(ray.direction()[0], sin(psi * M_DEG2RAD), tol)
         && !Test_fuzzy_compare(ray.direction()[1], 0.0                 , tol)
         && !Test_fuzzy_compare(ray.direction()[2], cos(psi * M_DEG2RAD), tol);
  int rc2 = !fuzzy_compare(ray.position()[0],   1.2)
         && !fuzzy_compare(ray.position()[1], 503.0)
         && !fuzzy_compare(ray.position()[2],  17.7);
  int rc3 = !fuzzy_compare(ray.direction()[0], sin(psi * M_DEG2RAD)) 
         && !fuzzy_compare(ray.direction()[1], 0.0                 )
         && !fuzzy_compare(ray.direction()[2], cos(psi * M_DEG2RAD));

  return  rc0 && rc1 && rc2 && rc3;
}

int
main() 
{
  int rc = 1;
  int rc1;

  rc1 = rl_BasicRay_test01();
  TEST( "ctor, dtor, const accessors", rc1, 1 );
  rc1 &= rc;

  rc1 = rl_BasicRay_test02();
  TEST( "copy ctor", rc1, 1 );
  rc1 &= rc;

  rc1 = rl_BasicRay_test03();
  TEST( "reflect_direction_vector", rc1, 1 );
  rc1 &= rc;

  rc1 = rl_BasicRay_test04();
  TEST( "translate_rotate/derotate_...", rc1, 1 );
  rc1 &= rc;

  return !rc;
}

