// File:   rl_DielectricData.cc
// Author: Terry Gaetz
//
// --8<--8<--8<--8<--
//
// Copyright (C) 2006, 2007 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

// Description: organize dielectric constant information 
//              into a bsearchable array
//
// History
//--------
// 0.2.0  tjg  1998-Jun-23 convert to C++
// 0.1.0  tjg  1998-Jun-17 rename to dielectricdata.c; extensive revisions.
// 0.0.1  tjg  1995-May-18 original

#include "rl_DielectricData.h"

#include <cstdlib>     // system includes
#include <cstring>
#include <cmath>
#include <cfloat>      // DBL_MAX

using namespace std;

#define EFFECTIVELY_ZERO 1.0e-30

//=========================================================================
// dtor, ctors...

//-------------------------------------------------------------------------
rl_DielectricData::
~rl_DielectricData()
{
  if ( decrement_ ) { delete [] decrement_; }
}

//-------------------------------------------------------------------------
rl_DielectricData::
rl_DielectricData( rl_DielectricData const& other )
  throw ( rl_Exception )
  : nelts_(       other.nelts_       )
  , decrement_(   new rl_DielectricDataBinPOD[other.nelts_] )
  , emin_(        other.emin_        )
  , emax_(        other.emax_        )
  , bulkdensity_( other.bulkdensity_ )
  , interp_mode_( other.interp_mode_ )
{ 
  memcpy( decrement_, other.decrement_, 
          nelts_ * sizeof( rl_DielectricDataBinPOD ) ); 
}

//-------------------------------------------------------------------------
rl_DielectricData::
rl_DielectricData( rl_Traits::rl_DielectricPOD const* diel,
                   size_t                             num_pts,
                   rl_Traits::EInterpMode             interp_mode,
                   double                             bulk_density )
  throw ( rl_Exception )
{ 
  init( diel, num_pts, interp_mode, bulk_density );
}

//-------------------------------------------------------------------------
void rl_DielectricData::
init( rl_Traits::rl_DielectricPOD const* diel,
      size_t                             num_pts,
      rl_Traits::EInterpMode             interp_mode,
      double                             bulk_density )
  throw ( rl_Exception )
{
  nelts_       = num_pts - 1;
  interp_mode_ = interp_mode;

  emin_        = diel[0].energy_;
  emax_        = diel[num_pts-1].energy_;

  bulkdensity_ = bulk_density;

  try {

    decrement_   = new rl_DielectricDataBinPOD[nelts_];

  } catch ( rl_Exception& e ) {

    e.set_rethrow_message( __FILE__, __LINE__ );
    throw e;
  }

  // read reflectance data into rl_DielectricBinPOD array

  size_t n;
  for ( n = 0 ; n < nelts_ ; ++n )
  {
    decrement_[n].lo_.energy_ = diel[n].energy_;
    decrement_[n].lo_.alpha_  = diel[n].alpha_ * bulk_density;;
    decrement_[n].lo_.gamma_  = diel[n].gamma_ * bulk_density;;

    decrement_[n].hi_.energy_ = diel[n+1].energy_;
    decrement_[n].hi_.alpha_  = diel[n+1].alpha_ * bulk_density;;
    decrement_[n].hi_.gamma_  = diel[n+1].gamma_ * bulk_density;;
  }

  // Set up for more efficient lin-log, log-lin, log-log interpolation; 
  // take logs now rather than for each ray interpolation.
  // (Assumes that number of rays is much larger than number of table
  //  entries.)

  switch ( interp_mode_ ) {

  case rl_Traits::ELinLin:   
    break;

  case rl_Traits::ELinLog:   
    for ( n = 0 ; n < nelts_ ; ++n )
    {
      decrement_[n].lo_.energy_ = log(decrement_[n].lo_.energy_);
      decrement_[n].hi_.energy_ = log(decrement_[n].hi_.energy_);
    }
    break;

  case rl_Traits::ELogLin:   
    for ( n = 0 ; n < nelts_ ; ++n )
    {
      decrement_[n].lo_.alpha_  = log(decrement_[n].lo_.alpha_);
      decrement_[n].hi_.alpha_  = log(decrement_[n].hi_.alpha_);
      decrement_[n].lo_.gamma_  = log(decrement_[n].lo_.gamma_);
      decrement_[n].hi_.gamma_  = log(decrement_[n].hi_.gamma_);
    }
    break;

  case rl_Traits::ELogLog:  
    for ( n = 0 ; n < nelts_ ; ++n )
    {
      decrement_[n].lo_.energy_ = log(decrement_[n].lo_.energy_);
      decrement_[n].hi_.energy_ = log(decrement_[n].hi_.energy_);
      decrement_[n].lo_.alpha_  = log(decrement_[n].lo_.alpha_);
      decrement_[n].hi_.alpha_  = log(decrement_[n].hi_.alpha_);
      decrement_[n].lo_.gamma_  = log(decrement_[n].lo_.gamma_);
      decrement_[n].hi_.gamma_  = log(decrement_[n].hi_.gamma_);
    }
    break;
  }
}

//=========================================================================
// accessors...

/* forward declarations */
extern "C" {
static int diel_compare( void const* ck, void const* ce );
}

//-------------------------------------------------------------------------
int rl_DielectricData::
alpha_gamma( double energy, double& alpha, double& gamma )
{
  int rc = 0;
  if ( decrement_ )            // not a vacuum...
  {
    double               w, en;

    switch ( interp_mode_ )
    {
    case rl_Traits::ELinLin:   // intentional fall-through
    case rl_Traits::ELogLin:   en = energy;
                               break;
  
    case rl_Traits::ELinLog:   // intentional fall-through
    case rl_Traits::ELogLog:   en = log(energy);
                               break;
    }
    if ( en < decrement_[0].lo_.energy_ )
    { return -1; }
  
    if ( en > decrement_[nelts_-1].hi_.energy_ )
    { return +1; }
    
    rl_DielectricDataBinPOD* diel;
    if ( en < decrement_[nelts_-1].hi_.energy_ )
    {
      diel = (rl_DielectricDataBinPOD*)
                  bsearch( &en,                              // key
                           decrement_,                       // array base
                           nelts_,                           // num elts
                           sizeof(rl_DielectricDataBinPOD),  // sizeof elt
                           diel_compare                      // cmp func
                         );
      if (! diel )
      { rc = 1; }   // didn't find requested energy...
    }
    else
    {
      diel = &decrement_[nelts_-1];
    }
  
    // log interpolation in energy
    w     = (en - diel->lo_.energy_) / (diel->hi_.energy_ - diel->lo_.energy_);

    alpha = (1.0 - w) * diel->lo_.alpha_ + w * diel->hi_.alpha_;
    gamma = (1.0 - w) * diel->lo_.gamma_ + w * diel->hi_.gamma_;

    #if defined (TEST_ALPHA_GAMMA)
      fprintf( stdout, "___ ALPHA_GAMMA ___\n" );
      fprintf( stdout, "E                     %.15e\n", en );
      fprintf( stdout, "E alpha gamma lo      %.15e %.15e %.15e\n",
	       diel->lo_.energy_, diel->lo_.alpha_, diel->lo_.gamma_ );
      fprintf( stdout, "E alpha gamma hi      %.15e %.15e %.15e\n",
	       diel->hi_.energy_, diel->hi_.alpha_, diel->hi_.gamma_ );
      fprintf( stdout, "w                     %.15e\n", w );
      fprintf( stdout, "alpha gamma           %.15e %.15e\n",
	       alpha, gamma );
    #endif

    switch ( interp_mode_ )
    {
    case rl_Traits::ELinLin:   // intentional fall-through
    case rl_Traits::ELinLog:   break;

    case rl_Traits::ELogLin:   // intentional fall-through
    case rl_Traits::ELogLog:   alpha = exp(alpha);
                               gamma = exp(gamma);
                               break;
    }
  }
  else
  {
    alpha = 0.0;       // vacuum case...
    gamma = 0.0;
  }

  return rc;
}

//-------------------------------------------------------------------------
//
// compare two nodes based on a numeric ordering of the energy field

int diel_compare( void const* ck,   // search key
                  void const* ce    // array element
                )
{
  double *pk = (double*) ck;

  rl_DielectricData::rl_DielectricDataBinPOD *pe 
       = (rl_DielectricData::rl_DielectricDataBinPOD*)ce;

  if ( *pk < pe->lo_.energy_ )
  {
    return -1;
  }
  else if ( *pk >= pe->hi_.energy_ )
  {
    return +1;
  }
  else
  {
    return 0;
  }
}
