// File:   rl_DielectricLayer.cc
// Author: Terry Gaetz
//
// --8<--8<--8<--8<--
//
// Copyright (C) 2006, 2007 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

// #define DEBUG_DIELECTRICLAYER
#define DIEL_GAMMA_NEGATIVE   1

#include <rl_DielectricLayer.h>     // rl_DielectricLayer

#include <cstring>                  // strlen strcpy
#include <cstdio>                   // stdout
// #include <cfloat>                   // DBL_MAX
#include <mathconst/mathconst.h>    // M_PI

#include <iomanip>

using namespace std;

//=========================================================================
// dtor, ctors...

//-------------------------------------------------------------------------
rl_DielectricLayer::
~rl_DielectricLayer()
{ 
  if ( name_ ) { delete [] name_; } 
}

//-------------------------------------------------------------------------
// default constructor:  initialize as an infinitely thick vacuum layer.

rl_DielectricLayer::
rl_DielectricLayer( char const layer_name[] )
  throw ( rl_Exception )
  : diel_info_()
  , thickness_(-1.0)
  , zcoat_(-1.0)
  , alpha_(0.0)
  , gamma_(0.0)
  , rtype_(rl_Traits::ERoughNone)
  , srough_(0.0)
  , lambda_(826.5)
  , prop_( complex(1.0,0.0) )
  , kt_perp_( complex(0.0,0.0) )
  , phase_factor_(0.0)
  , rflcoef_()
  , t_ij_()
  , t_ji_()
  , is_substrate_(rl_Traits::False)
{
  try {

    if ( layer_name )
    { 
      name_ = new char[strlen(layer_name)+1];
    }
    else
    {
      name_ = new char[1];
      *name_ = '\0';
    }

  } catch ( std::exception& bad_alloc ) {

    char msg[] = "rl_DielectricLayer(layername): "
                 "unable to allocate space for the layer name";

    throw rl_Exception( msg );
  }
}

//-------------------------------------------------------------------------
// DEEP Copy Constructor.

rl_DielectricLayer::
rl_DielectricLayer( rl_DielectricLayer const& other )
  throw ( rl_Exception )
  : diel_info_()
  , thickness_(-1.0)
  , zcoat_(-1.0)
  , alpha_(0.0)
  , gamma_(0.0)
  , rtype_(rl_Traits::ERoughNone)
  , srough_(0.0)
  , lambda_(826.5)
  , prop_( complex(1.0,0.0) )
  , kt_perp_( complex(0.0,0.0) )
  , phase_factor_(0.0)
  , rflcoef_()
  , t_ij_()
  , t_ji_()
  , is_substrate_(rl_Traits::False)
{
  try {

    if ( other.name_ )
    {
      name_ = new char[strlen(other.name_)+1];
      strcpy( name_, other.name_ );
    }
    else
    {
      name_ = new char[1];
      *name_ = '\0';
    }

  } catch ( std::exception& bad_alloc ) {

    char msg[] = "rl_DielectricLayer copy constructor: "
                 "unable to allocate space for the layer name";

    throw rl_Exception( msg );
  }
}

//-------------------------------------------------------------------------

rl_DielectricLayer::
rl_DielectricLayer( rl_Traits::rl_DielectricPOD const* diel,
                    std::size_t                        ndielpts,
                    double                             layer_thickness,
                    double                             roughness,
                    rl_Traits::ERoughType              roughness_type,
                    rl_Traits::EInterpMode             interp_mode,
                    double                             bulkdensity,
                    char const*                        layer_name,
                    rl_Traits::Bool                    is_substrate )
  throw ( rl_Exception )
{ 
  init( diel, ndielpts, layer_thickness, roughness, roughness_type, 
        interp_mode, bulkdensity, layer_name, is_substrate ); 
}
  
//-------------------------------------------------------------------------

void rl_DielectricLayer::
init( rl_Traits::rl_DielectricPOD const* diel,
      std::size_t                        ndielpts,
      double                             layer_thickness,
      double                             roughness,
      rl_Traits::ERoughType              roughness_type,
      rl_Traits::EInterpMode             interp_mode,
      double                             bulkdensity,
      char const*                        layer_name,
      rl_Traits::Bool                    is_substrate )
  throw ( rl_Exception )
{
  diel_info_.init( diel, ndielpts, interp_mode, bulkdensity );

  try {

    if ( layer_name )
    {
      name_ = new char[strlen(layer_name)+1];
      strcpy( name_, layer_name );
    }
    else
    {
      name_ = new char[1];
      *name_ = '\0';
    }

  } catch ( std::exception& bad_alloc ) {

    char msg[256] = "rl_DielectricLayer::init(...): "
                    "unable to allocate space for layer name";

    throw rl_Exception( msg );
  }

  thickness_    = layer_thickness;  // layer thickness (Angstroms)
  rtype_        = roughness_type;   // layer roughness type
  srough_       = roughness;        // roughness parameter (Angstroms)
  is_substrate_ = is_substrate;     // 0;  1, if this is the substrate,

  switch ( rtype_ )
  {
  case rl_Traits::ERoughDebyeWaller_RSAO    :           // fall through
  case rl_Traits::ERoughDebyeWaller_CSAO    :           // fall through
  case rl_Traits::ERoughDebyeWaller_Spiller :           // fall through
  case rl_Traits::ERoughNevotCroce          :           // fall through
  case rl_Traits::ERoughModifiedDebyeWaller :           // fall through
  case rl_Traits::ERoughNone                :  break;     

  default :
    { char msg[256];
      char const format[] = "rl_DielectricPODArray::init(...): "
                            "Invalid roughness type was %d\n";
      sprintf( msg, format, rtype_ );
      throw rl_Exception( msg );
    }
  } /* end rl_Traits::ERoughType cases */

}

//=========================================================================
// mutators...

//-------------------------------------------------------------------------
int rl_DielectricLayer::
setup_for( double energy, double sinphi )
{
  lambda_ = 12.39854 / energy;       // wavelength (Angstrom)

  if ( rtype_ != rl_Traits::ERoughNone )
  {
    phase_factor_ = 2.0 * M_PI * srough_ / lambda_;
  }
  else
  {
    phase_factor_ = 0.0;
  }

  int rc = diel_info_.alpha_gamma( energy, alpha_, gamma_ );

  //-------------------
  // kt_perp_
  //
  // This is the component of the complex wave vector perpendicular
  // to the interface divided by k0, the magnitude of the wave vector
  // in the vacuum.
  //
  // The boundary condition at the interface requires that the component
  // of the wave vector parallel to the interface is the same for each
  // layer.  The perpendicular component is then
  //     k_{\perp,j} = \sqrt{ \sin^2 \phi_0 + \Delta \epsilon }
  // where \phi_0 is the graze angle in the vacuum and
  // \Delta\epsilon is the dielectric decrement, i.e.,
  // complex( -alpha, -gamma ).
  // Note that in obtaining k_{\perp,j} we take a complex square
  // root.  The imaginary part of the principal root is obtained
  // by multiplying b by the sign of gamma_.

  double const difab = sinphi * sinphi - alpha_;
  double const sumab = sqrt( difab * difab + gamma_ * gamma_ );

  // Force b to be evaluated to avoid sqrt(-tiny) when i86
  // extended FP registers in use.  Is there a g++ compiler option
  // to force this?  There's supposed to be a way to set the FP mode...

  double const a   = sqrt( (sumab + difab) / 2.0 );
  double const tmp =       (sumab - difab) / 2.0;
  double const b   = (tmp >= 0.0) ? sqrt( tmp ) : sqrt( -tmp );

  // the following ifdefs allow hardwiring the branch cut for efficiency
  #ifdef DIEL_GAMMA_GENERAL
    kt_perp_ = complex( a, ((gamma_ > 0.0) ? 1.0 : -1.0) * b );
  #else
    #ifdef DIEL_GAMMA_NEGATIVE
      kt_perp_ = complex( a, -b );
    #endif
    #ifdef DIEL_GAMMA_POSITIVE
      kt_perp_ = complex( a,  b );
    #endif
  #endif

  #if defined(DEBUG_DIELECTRICLAYER)
    fprintf(stdout, "\nDEBUG: ___SETUP_FOR___ %s ---------------------\n",
	             name_ );
    fprintf(stdout, "kt_perp_          (%.15e  %.15e)\n",
                     kt_perp_.real(),
                     kt_perp_.imag() );
    fprintf(stdout, "alpha_ gamma_     (%.15e  %.15e)\n",
                     alpha_, gamma_ );
    fprintf(stdout, "energy, sinphi    (%.15e  %.15e)\n",
                     energy, sinphi );
    fprintf(stdout, "difab sumab  tmp  (%.15e  %.15e  %.15e)\n",
                     difab, sumab, tmp );
    fprintf(stdout, "a, b              (%.15e  %.15e)\n",
                     a, b );
  #endif

  //-------------------
  // evaluate layer propagator
  // the layer propagator is exp( phase_factor )

  if (! rc )
  {
    if (! is_vacuum() && ! is_substrate() )
    {
      zcoat_ = 4.0 * M_PI * thickness_ / lambda_;  
         // dimensionless thickness of layer

      prop_ = exp( complex( -b * zcoat_, -a * zcoat_ ) );
    #if defined(DEBUG_DIELECTRICLAYER)
      fprintf(stdout, "zcoat_             %.15e\n",
                       zcoat_ );
      fprintf(stdout, "prop_ (r,i)       (%.15e  %.15e)\n",
                       prop_.real(), prop_.imag() );
    #endif
    }
  }
  return rc;  // 0 if successful
}

//-------------------------------------------------------------------------
// calculate complex reflection coefficients at interface between 
// this layer and layer immediately above it.

void rl_DielectricLayer::
reflect_amp( rl_DielectricLayer const& upper, double sinphi )
{
  double  alpha_jm1 = upper.alpha_;
  double  gamma_jm1 = upper.gamma_;
  double  alpha_j   =       alpha_;
  double  gamma_j   =       gamma_;

  complex upper_delta_eps( -alpha_jm1, -gamma_jm1 );
  complex       delta_eps( -alpha_j,   -gamma_j   );

  complex eps_upper( upper_delta_eps + 1.0 );
  complex eps(             delta_eps + 1.0 );

  complex num_perp    =  upper.kt_perp_ - kt_perp_;
  complex denom_perp  =  upper.kt_perp_ + kt_perp_;

  rflcoef_.perp()     =  num_perp / denom_perp;

  #if defined(DEBUG_DIELECTRICLAYER)
    fprintf(stdout, "\nDEBUG: ___REFLECT_AMP___ ___AAA0___ --------\n");
    fprintf(stdout, "upper, this layer   (%s  %s)\n", 
	             upper.name_, name_);
    fprintf(stdout, "alpha, gamma (j-1): (%.15e, %.15e)\n",
                     alpha_jm1, gamma_jm1);
    fprintf(stdout, "alpha, gamma (j):   (%.15e, %.15e)\n",
                     alpha_j, gamma_j);
    fprintf(stdout, "rflcoef (perp)      (%.15e  %.15e)\n",
                     rflcoef_.perp().real(),
                     rflcoef_.perp().imag() );
    fprintf(stdout, "num_perp            (%.15e  %.15e)\n",
                     num_perp.real(),
                     num_perp.imag() );
    fprintf(stdout, "denom_perp          (%.15e  %.15e)\n",
                     denom_perp.real(),
                     denom_perp.imag() );
    fprintf(stdout, "upper.kt_perp_      (%.15e  %.15e)\n",
                     upper.kt_perp_.real(),
                     upper.kt_perp_.imag() );
    fprintf(stdout, "kt_perp_             %.15e  %.15e)\n",
                     kt_perp_.real(),
                     kt_perp_.imag() );
  #endif

  complex num_para    =        delta_eps * upper.kt_perp_;
          num_para   -=  upper_delta_eps *       kt_perp_;
          num_para   +=        num_perp;

  complex denom_para  =        delta_eps * upper.kt_perp_;
          denom_para +=  upper_delta_eps *       kt_perp_;
          denom_para +=        denom_perp;

  rflcoef_.para()     =  num_para / denom_para;

  double eps_epsupper_r  = alpha_jm1 * alpha_j - gamma_jm1 * gamma_j;
         eps_epsupper_r -= alpha_jm1 + alpha_j;
         eps_epsupper_r += 1.0;
   
  double eps_epsupper_i  = alpha_jm1 * gamma_j + gamma_jm1 * alpha_j;
         eps_epsupper_i -= gamma_jm1 + gamma_j;

  eps_epsupper_ = complex( eps_epsupper_r, eps_epsupper_i );

  #if defined(DEBUG_DIELECTRICLAYER)
    fprintf(stdout, "eps_epsupper (r,i): (%.15e, %.15e)\n",
                     eps_epsupper_r, eps_epsupper_i);
    fprintf(stdout, "num_para            (%.15e  %.15e)\n",
                     num_para.real(),
                     num_para.imag() );
    fprintf(stdout, "denom_perp          (%.15e  %.15e)\n",
                     denom_para.real(),
                     denom_para.imag() );
    fprintf(stdout, "rflcoef (para)      (%.15e  %.15e)\n",
                     rflcoef_.para().real(),
                     rflcoef_.para().imag() );
  #endif

  if (! is_substrate() )
  {
    t_ij_.perp() = 2.0 * upper.kt_perp_ / denom_perp;
    t_ji_.perp() = 2.0 *       kt_perp_ / denom_perp;

    t_ij_.para() = 2.0 * ( eps       * upper.kt_perp_ ) / denom_para;
    t_ji_.para() = 2.0 * ( eps_upper *       kt_perp_ ) / denom_para;

    tij_tji_perp_ = t_ij_.perp() * t_ji_.perp();
    tij_tji_para_ = t_ij_.para() * t_ji_.para();

    r_perp_          = rflcoef_.perp();
    r_para_          = rflcoef_.para();
    r2_perp_         = rflcoef_.perp() * rflcoef_.perp();
    r2_para_         = rflcoef_.para() * rflcoef_.para();
    r2_tij_tji_perp_ = r2_perp_ + tij_tji_perp_;
    r2_tij_tji_para_ = r2_para_ + tij_tji_para_;

    #if defined(DEBUG_DIELECTRICLAYER)
      fprintf(stdout, "t_ij_ (para)      (%.15e  %.15e)\n",
                       t_ij_.para().real(),
                       t_ij_.para().imag() );
      fprintf(stdout, "t_ji_ (para)      (%.15e  %.15e)\n",
                       t_ji_.para().real(),
                       t_ji_.para().imag() );
      fprintf(stdout, "t_ij_ (perp)      (%.15e  %.15e)\n",
                       t_ij_.perp().real(),
                       t_ij_.perp().imag() );
      fprintf(stdout, "t_ji_ (perp)      (%.15e  %.15e)\n",
                       t_ji_.perp().real(),
                       t_ji_.perp().imag() );
      fprintf(stdout, "tij_tji_perp_     (%.15e  %.15e)\n",
                       tij_tji_perp_.real(),
                       tij_tji_perp_.imag() );
      fprintf(stdout, "tij_tji_para_     (%.15e  %.15e)\n",
                       tij_tji_para_.real(),
                       tij_tji_para_.imag() );
      fprintf(stdout, "r_perp_           (%.15e  %.15e)\n",
                       r_perp_.real(),
                       r_perp_.imag() );
      fprintf(stdout, "r_para_           (%.15e  %.15e)\n",
                       r_para_.real(),
                       r_para_.imag() );
      fprintf(stdout, "r2_perp_          (%.15e  %.15e)\n",
                       r2_perp_.real(),
                       r2_perp_.imag() );
      fprintf(stdout, "r2_para_          (%.15e  %.15e)\n",
                       r2_para_.real(),
                       r2_para_.imag() );
      fprintf(stdout, "r2_tij_tji_perp_  (%.15e  %.15e)\n",
                       r2_tij_tji_perp_.real(),
                       r2_tij_tji_perp_.imag() );
      fprintf(stdout, "r2_tij_tji_para_  (%.15e  %.15e)\n",
                       r2_tij_tji_para_.real(),
                       r2_tij_tji_para_.imag() );
  #endif

  }

  switch ( rtype_ )
  {
  case rl_Traits::ERoughDebyeWaller_RSAO :

         apply_DebyeWaller_RSAO_factor( upper, sinphi );
         break;

  case rl_Traits::ERoughDebyeWaller_CSAO :

         apply_DebyeWaller_CSAO_factor( upper, sinphi );
         break;

  case rl_Traits::ERoughDebyeWaller_Spiller :

         apply_DebyeWaller_Spiller_factor( sinphi );
         break;

  case rl_Traits::ERoughNevotCroce :

         apply_NevotCroce_factor( upper, sinphi );
         break;

  case rl_Traits::ERoughModifiedDebyeWaller :

         apply_ModifiedDW_factor( upper, sinphi );
         break;

  case rl_Traits::ERoughNone :

         /* do nothing */
         break;

  default :

         cerr << "Invalid roughness type in rl_DielectricLayer::reflect_amp"
              << endl;
         cerr << "Roughness type was " << rtype_ << endl;
         exit(1);

  } /* end rl_Traits::ERoughType cases */
}

//-------------------------------------------------------------------------
// c   Calculate n-layer reflectivity using recurrence relation.
// c
//       subroutine reflect_nlayer ( ca, cprop, n, refl )
//       implicit complex (c)
//       dimension ca(15), cprop(14)
// c
//       ceps = ca( n )
//       do m=n-1, 1, -1
//          ceps = (ca(m) + cprop(m)*ceps) / (1.0 + cprop(m)*ca(m)*ceps)
//       end do
// c
//       rmod = abs( ceps )
//       refl = rmod*rmod
// c
//       return
//       end

void rl_DielectricLayer::
reflect_nlayer( rl_DielectricLayer layer[],
                int                num )
{
  complex eps_para( layer[num-1].rflcoef_.para() );
  complex eps_perp( layer[num-1].rflcoef_.perp() );
  const complex one( 1.0 );
  #if defined(DEBUG_DIELECTRICLAYER)
    fprintf(stdout, "\nDEBUG: ___REFLECT_NLAYER___ ___AAA0___ -----\n");
    fprintf(stdout, "\n==> layer %5d -- ", num-1);
    fprintf(stdout, "layer[n-1] (%5d  %.s)\n", 
	             num-1, layer[num-1].name_);
    fprintf(stdout, "rl_DielectricLayer::reflect_nlayer ");
    fprintf(stdout, "======================\n");
    fprintf(stdout, "eps (para) %5d (%.15e  %.15e)\n",
                     num-1, eps_para.real(), eps_para.imag());
    fprintf(stdout, "eps (perp) %5d (%.15e  %.15e)\n",
                     num-1, eps_perp.real(), eps_perp.imag());
  #endif

  for ( int n = num-2; n >= 0; --n )
  {
    rl_Traits::complex prop_eps_para( layer[n].propagator() * eps_para );
    eps_para = (layer[n].rflcoef_.para() 
              + layer[n].r2_tij_tji_para_ * prop_eps_para)
             / (one + layer[n].rflcoef_.para() * prop_eps_para);

    rl_Traits::complex prop_eps_perp( layer[n].propagator() * eps_perp );
    eps_perp = (layer[n].rflcoef_.perp() 
              + layer[n].r2_tij_tji_perp_ * prop_eps_perp)
             / (one + layer[n].rflcoef_.perp() * prop_eps_perp);

  #if defined(DEBUG_DIELECTRICLAYER)
    fprintf(stdout, "\nDEBUG: ___REFLECT_NLAYER___ ___AAA1___ -----\n");
    fprintf(stdout, "\n==> layer %5d -- ", n);
    fprintf(stdout, "rl_DielectricLayer::reflect_nlayer ");
    fprintf(stdout, "======================\n");
    cprint_constraints_on( stdout );
    fprintf(stdout, "rflcoef (para) %5d, (%.15e  %.15e)\n",
                     n, layer[n].rflcoef_.para().real(),
                        layer[n].rflcoef_.para().imag() );

    fprintf(stdout, "rflcoef (perp) %5d, (%.15e  %.15e)\n",
                     n, layer[n].rflcoef_.perp().real(),
                        layer[n].rflcoef_.perp().imag() );

    fprintf(stdout, "r2_tt   (para) %5d, (%.15e  %.15e)\n",
                     n, layer[n].r2_tij_tji_para_.real(), 
                        layer[n].r2_tij_tji_para_.imag() );

    fprintf(stdout, "r2_tt   (perp) %5d, (%.15e  %.15e)\n",
                     n, layer[n].r2_tij_tji_perp_.real(), 
                        layer[n].r2_tij_tji_perp_.imag() );

    fprintf(stdout, "eps     (para) %5d, (%.15e  %.15e)\n",
                     n, eps_para.real(), eps_para.imag());
    fprintf(stdout, "eps     (perp) %5d, (%.15e  %.15e)\n",
                     n, eps_perp.real(), eps_perp.imag());
  #endif
  }
  layer[0].rflcoef_.para() = eps_para;
  layer[0].rflcoef_.perp() = eps_perp;

  #if defined(DEBUG_DIELECTRICLAYER)
    fprintf(stdout, "\n");
  #endif

#if 0
  refl  = abs( eps );
  refl *= refl;
#endif
}

//=========================================================================

//-------------------------------------------------------------------------
// Routine to evaluate Debye-Waller roughness factor
// <code>phase_factor</code> is <code>2 pi srough / lambda</code>
//
// Implemented in the IDL code as dw_csao: SAO variant on Windt's DW, 
// using complex factor
//      exp(double(-2d*kc*kc*sigma_t(j)*sigma_t(j)*c_t(*,*,j)*c_t(*,*,j)))

void rl_DielectricLayer::
apply_DebyeWaller_CSAO_factor( rl_DielectricLayer const& upper, // upper layer
                               double sinphi             // sin(graze angle)
                             )
{
  double phase_factor = -2 * phase_factor_ * phase_factor_;
  double sinphi_sq    = sinphi * sinphi;

  complex sinterm_0( sinphi_sq - upper.alpha_, -upper.gamma_ );
  sinterm_0 /= complex( 1.0 - upper.alpha_, -upper.gamma_ );

  complex ckphase( sinterm_0 );
  ckphase *= phase_factor;
  ckphase  = exp(ckphase);

  rflcoef_.para() *= ckphase;
  rflcoef_.perp() *= ckphase;
}

//-------------------------------------------------------------------------
// Routine to evaluate Debye-Waller roughness factor
// <code>phase_factor</code> is <code>2 pi srough / lambda</code>
//
// Implemented in the IDL code as dw_csao: SAO variant on Windt's DW,
// taking only the real part of the factor
//      exp(double(-2d*kc*kc*sigma_t(j)*sigma_t(j)*c_t(*,*,j)*c_t(*,*,j)))

void rl_DielectricLayer::
apply_DebyeWaller_RSAO_factor( rl_DielectricLayer const& upper, // upper layer
                               double sinphi             // sin(graze angle)
                             )
{
  double phase_factor = -2.0 * phase_factor_ * phase_factor_;
  double sinphi_sq    = sinphi * sinphi;

  complex sinterm_0( sinphi_sq - upper.alpha_, -upper.gamma_ );
  sinterm_0 /= complex( 1.0 - upper.alpha_, -upper.gamma_ );

  complex ckphase( sinterm_0 );
  ckphase *= phase_factor;
  double real_phase_factor = ckphase.real();
  real_phase_factor = exp( real_phase_factor );

  rflcoef_.para() *= real_phase_factor;
  rflcoef_.perp() *= real_phase_factor;
}

//-------------------------------------------------------------------------
// Routine to evaluate Debye-Waller roughness factor
// <code>phase_factor</code> is <code>2 pi srough / lambda</code>
//
// Implemented in the IDL code as dw_spiller: Variant on the DW rising 
// from a paper by Spiller
//      exp(-2d*kc*kc*sigma_t(j)*sigma_t(j)*c_t(*,*,j)*c_t(*,*,j)*
//          n(*,*,j)*n(*,*,j))

void rl_DielectricLayer::
apply_DebyeWaller_Spiller_factor( double sinphi        // sin(graze angle)
                        )
{
  double phase_factor = phase_factor_;

  phase_factor *= sinphi;
  phase_factor *= phase_factor;
  phase_factor *= 2.0;

  complex ckphase( phase_factor * (alpha_ - 1.0),
                   phase_factor *  gamma_ );

  complex cexp_factor = exp(ckphase);

  rflcoef_.para() *= cexp_factor;
  rflcoef_.perp() *= cexp_factor;
}

//-------------------------------------------------------------------------
// Routine to evaluate "modified Debye-Waller" roughness factor
// <code>phase_factor</code> is <code>2 pi srough / lambda</code>

void rl_DielectricLayer::
apply_ModifiedDW_factor( rl_DielectricLayer const& upper,  // upper layer
                         double                    sinphi  // sin(graze angle)
                       )
{
  double phase_factor = phase_factor_;
  double sinphi_sq    = sinphi * sinphi;

  #if defined(DEBUG_DIELECTRICLAYER)
    fprintf(stdout, "\nDEBUG: ___MDW___ ---------------------------\n");
    cerr << "===================================================================" 
         << endl;
    cerr << "Layer j-1:" << endl;
    upper.cdump_on( stdout );
    cerr << "-------------------------------------------------------------------" 
         << endl;
    cerr << "Layer j:" << endl;
    cdump_on( stdout );
    cerr << "-------------------------------------------------------------------" 
         << endl;
  #endif

  phase_factor     *= -2.0 * phase_factor;

  #if defined(DEBUG_DIELECTRICLAYER)
    cerr << "sinphi:                  " << sinphi << endl;
    cerr << "phi:                     " << 60.0 * asin(sinphi) / M_DEG2RAD << endl;
    cerr << "-2 (2pi sigma/lambda)^2: " << phase_factor << endl;
  #endif

  complex sinterm_0( sinphi_sq - upper.alpha_, -upper.gamma_ );
  complex nopt0  = sqrt( complex( 1.0 - upper.alpha_, -upper.gamma_ ));
  complex cos2_0 = sqrt( sinterm_0 ) / nopt0;

  #if defined(DEBUG_DIELECTRICLAYER)
    cerr << "cos_{j-1}:               " << cos2_0 << endl;
  #endif

  complex sinterm_1( sinphi_sq - alpha_, -gamma_ );
  complex nopt1  =  sqrt( complex( 1.0 - alpha_, -gamma_ ) );
  complex cos2_1 =  sqrt( sinterm_1 ) / nopt1;

  #if defined(DEBUG_DIELECTRICLAYER)
    complex cosij = cos2_0 * cos2_1;
    complex cosj = cos2_0 * nopt0 / nopt1;
    complex sin2_0 = sqrt( complex(1.0) - cos2_0 * cos2_0 );
    complex sin2_1 = sqrt( complex(1.0) - cos2_1 * cos2_1 );
    cerr << "cos_{j}:                  " << cos2_1 << endl;
    cerr << "cos_{j-1} n_{j-1} / n_{j} " << cos2_0 * nopt0 / nopt1 << endl;
    cerr << "n_{j-1} sin_{j-1}         " << sin2_0 * nopt0 << endl;
    cerr << "n_{j}   sin_{j}           " << sin2_1 * nopt1 << endl;
    cerr << "cos_{j} cos_{j-1}         " << cosij << endl;
  #endif

  complex cexp_factor( cos2_0 );
  cexp_factor *= cos2_1;
  cexp_factor *= phase_factor;
  cexp_factor  = exp( cexp_factor );

  rflcoef_.para() *= cexp_factor;
  rflcoef_.perp() *= cexp_factor;

  #if defined(DEBUG_DIELECTRICLAYER)
    cerr << "===================================================================" 
         << endl;
  #endif
}

//-------------------------------------------------------------------------
// Routine to evaluate Nevot-Croce factor
// <code>phase_factor</code> is <code>2 pi srough / lambda</code>
//
//<code>
//  \NC_\mathit{j-1,j}
//     =  \exp\left[ -2 \k_0^2
//           \left[ \left( \sin^2 \tilde\alpha_0 + \Delta\eps_{j}   \right)
//                  \left( \sin^2 \tilde\alpha_0 + \Delta\eps_{j-1} \right)
//           \right]^{1\over 2}
//        \sigma_\j^2 \right]
//</code>

void rl_DielectricLayer::
apply_NevotCroce_factor( rl_DielectricLayer const& upper,  // upper layer
                         double                    sinphi  // sin(graze angle)
                       )
{
  double phase_factor = phase_factor_;
  double sinphi_sq    = sinphi * sinphi;

  phase_factor  *=  phase_factor;
  phase_factor  *= -2.0;

  complex sinterm_0( sinphi_sq - upper.alpha_, -upper.gamma_ );
  complex sinterm_1( sinphi_sq -       alpha_,       -gamma_ );
  complex ckphase( sinterm_0 );

  ckphase *= sinterm_1;
  ckphase  = sqrt( ckphase );
  ckphase *= phase_factor;
  ckphase  = exp(ckphase);

  rflcoef_.para() *= ckphase;
  rflcoef_.perp() *= ckphase;
}

//=========================================================================
// i/o

//-------------------------------------------------------------------------
std::ostream& rl_DielectricLayer::
dump_on( std::ostream& os, char const pre[], char const pst[] ) const
{
  if ( strlen(pre) )
  { os << pre; }

  os << "Layer:                  " << name_         << endl;
  os.setf( ios::scientific, ios::floatfield );
  os.precision(12);

  os << "minimum energy (keV):   " << diel_info_.energy_min() << endl;
  os << "maximum energy (keV):   " << diel_info_.energy_max() << endl;
  os << "bulk density scale:     " << diel_info_.bulk_density_factor() << endl;

  os << "thickness (Angstrom):   " << thickness_    << endl;
  if (! is_vacuum() )
  {
    os << "reflection coefficient: " << rflcoef_      << endl;
  }
  os << "alpha:                  " << alpha_        << endl;
  os << "gamma:                  " << gamma_        << endl;
  os << "refractive index:       " 
     << sqrt( complex( 1.0 - alpha_, -gamma_ )) << endl;
  if (! is_vacuum() )
  {
    os << "roughness type:         " << rtype_        << endl;
    os << "sigma_rough:            " << srough_       << endl;
    os << "wavelength:             " << lambda_       << endl;
    os << "zcoat:                  " << zcoat_        << endl;
    os << "propagator:             " << prop_         << endl;
    os << "kt_perp_:               " << kt_perp_      << endl;
  }
  os << "is_vacuum:              " << is_vacuum()   << endl;
  os << "is_substrate:           " << is_substrate_ << endl;

  if ( strlen(pst) )
  { os << pst; }

  return os;
}

//=========================================================================
// i/o

//-------------------------------------------------------------------------
void rl_DielectricLayer::
cdump_on( std::FILE* of, char const pre[], char const pst[] ) const
{
  if ( strlen(pre) )
  { fprintf(of, "%s", pre); }

  fprintf(of, "Layer:                  %s\n", name_);

  fprintf(of, "minimum energy (keV):   %.15e\n", diel_info_.energy_min());
  fprintf(of, "maximum energy (keV):   %.15e\n", diel_info_.energy_max());
  fprintf(of, "bulk density scale:     %.15e\n", 
               diel_info_.bulk_density_factor());

  fprintf(of, "thickness (Angstrom):   %.15e\n", thickness_);
  if (! is_vacuum() )
  {
    rflcoef_.cprint_on(of);
  }
  fprintf( of, "alpha:                  %.15e\n", alpha_);
  fprintf( of, "gamma:                  %.15e\n", gamma_);

  complex srt = sqrt( complex( 1.0 - alpha_, -gamma_ ) );
  fprintf(of, "refractive index:       (%.15e, %.15e)\n", 
               srt.real(), srt.imag());

  if (! is_vacuum() )
  {
    fprintf(of, "roughness type:         %d\n",    rtype_);
    fprintf(of, "sigma_rough:            %.15e\n", srough_);
    fprintf(of, "wavelength:             %.15e\n", lambda_);
    fprintf(of, "zcoat:                  %.15e\n", zcoat_);
    fprintf(of, "propagator:             (%.15e, %.15e)\n",
                prop_.real(), prop_.imag());
    fprintf(of, "kt_perp_:               (%.15e, %.15e)\n",
                kt_perp_.real(), kt_perp_.imag());
  }
  fprintf(of, "is_vacuum:              %d\n", is_vacuum());
  fprintf(of, "is_substrate:           %d\n", is_substrate_);

  if ( strlen(pst) )
  { fprintf(of, "%s", pst); }
}

//-------------------------------------------------------------------------
void rl_DielectricLayer::
cprint_constraints_on( std::FILE* of, 
                       char const pre[], char const pst[] ) const
{
  if ( strlen(pre) )
  { fprintf(of, "%s", pre); }

  fprintf(of, "Layer:                  %s\n", name_);

  complex identperp = t_ij_.perp() - rflcoef_.perp();
  complex identpara = t_ij_.para() - rflcoef_.para();

  fprintf(of, "minimum energy (keV):   %.14e\n", diel_info_.energy_min());
  fprintf(of, "maximum energy (keV):   %.14e\n", diel_info_.energy_max());
  fprintf(of, "bulk density scale:     %.14e\n", 
               diel_info_.bulk_density_factor());

  fprintf(of, "thickness (Angstrom):   %.14e\n", thickness_);
  fprintf( of, "alpha:                  %.14e\n", alpha_);
  fprintf( of, "gamma:                  %.14e\n", gamma_);

  complex srt = sqrt( complex( 1.0 - alpha_, -gamma_ ) );
  fprintf(of, "refractive index:       (%.14e, %.14e)\n", 
               srt.real(), srt.imag());

  if (! is_vacuum() )
  {
    fprintf(of, "roughness type:         %d\n",    rtype_);
    fprintf(of, "sigma_rough:            %.14e\n", srough_);
    fprintf(of, "wavelength:             %.14e\n", lambda_);
    fprintf(of, "zcoat:                  %.14e\n", zcoat_);
    fprintf(of, "propagator:             (%.14e, %.14e)\n",
                 prop_.real(), prop_.imag());
    fprintf(of, "kt_perp_:               (%.14e, %.14e)\n",
                 kt_perp_.real(), kt_perp_.imag());
    fprintf(of, "\n");

    fprintf(of, "Tji_perp:               (%.14e, %.14e)\n",
                 t_ji_.perp().real(), t_ji_.perp().imag());
    fprintf(of, "Tij_perp:               (%.14e, %.14e)\n",
                 t_ij_.perp().real(), t_ij_.perp().imag());
    fprintf(of, "Rij_perp:               (%.14e, %.14e)\n",
                 rflcoef_.perp().real(), rflcoef_.perp().imag());
    fprintf(of, "Rij^2 (perp):           (%.14e, %.14e)\n",
                 r2_perp_.real(), r2_perp_.imag());
    fprintf(of, "Tij Tji (perp):         (%.14e, %.14e)\n",
                 tij_tji_perp_.real(), tij_tji_perp_.imag());
    double tprpi = (fabs(identperp.imag()) >= 1e-15)
                  ? identperp.imag() : 0.0;
    fprintf(of, "Tij_perp-Rij_perp:      (%.14e, %.14e)\n",
                 identperp.real(), tprpi);
    fprintf(of, "Rij^2+Tij Tji (perp):   (%.14e, %.14e)\n",
                 r2_tij_tji_perp_.real(), r2_tij_tji_perp_.imag());
    fprintf(of, "\n");

    fprintf(of, "Tji_para:               (%.14e, %.14e)\n",
                 t_ji_.para().real(), t_ji_.para().imag());
    fprintf(of, "Tij_para:               (%.14e, %.14e)\n",
                 t_ij_.para().real(), t_ij_.para().imag());
    fprintf(of, "Rij_para:               (%.14e, %.14e)\n",
                 rflcoef_.para().real(), rflcoef_.para().imag());
    fprintf(of, "Rij^2 (para):           (%.14e, %.14e)\n",
                 r2_para_.real(), r2_para_.imag());
    fprintf(of, "Tij Tji (para):         (%.14e, %.14e)\n",
                 tij_tji_para_.real(), tij_tji_para_.imag());
    fprintf(of, "Tij_para-Rij_para:      (%.14e, %.14e)\n",
                 identpara.real(), identpara.imag());
    fprintf(of, "Rij^2+Tij Tji (para):   (%.14e, %.14e)\n",
                 r2_tij_tji_para_.real(), r2_tij_tji_para_.imag());
    fprintf(of, "\n");
  }
  fprintf(of, "is_vacuum:              %d\n", is_vacuum());
  fprintf(of, "is_substrate:           %d\n", is_substrate_);

  if ( strlen(pst) )
  { fprintf(of, "%s", pst); }
}
