// File:   rl_DielectricPODArray.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006, 2007 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

// Description: read in dielectric constant information from rdb table; 
//              organize into bsearchable array
//
// History
//--------
// 0.3.0  tjg  1998-Jul-08 split out into rl_raylibsup
// 0.2.0  tjg  1998-Jun-23 convert to C++
// 0.1.0  tjg  1998-Jun-17 rename to dielectricdata.c; extensive revisions.
// 0.0.1  tjg  1995-May-18 original

#include <rl_DielectricPODArray.h>      // rl_DielectricPODArray

//#include <rl_Traits.h>

#include <cstdio>                     // fopen
#include <cstdlib>                    // qsort
#include <cstring>
#include <cmath>

using namespace std;

typedef enum
{
  Ealpha_gamma
} ERfldataType;

// forward declaration
extern "C" {
  int diel_compare( const void      *ck,   // search key
                    const void      *ce    // array element
                  );
}

//-------------------------------------------------------------------------
// dtor, ctors...

rl_DielectricPODArray::
~rl_DielectricPODArray()
{ if ( data_ ) { delete [] data_; } }

rl_DielectricPODArray::
rl_DielectricPODArray( )
  : nelts_(0), data_(0)
{}

rl_DielectricPODArray::
rl_DielectricPODArray( size_t        nelts,
                       double const* energy,
                       double const* alpha,
                       double const* gamma )
    throw ( rl_Exception )
  : nelts_( nelts )
{
  /*
   * allocate array to hold data
   */

  try {

    data_ = new rl_Traits::rl_DielectricPOD[ nelts_ ];

  } catch ( std::exception& bad_alloc ) {

    char msg[256] = "rl_DielectricPODArray(nelts,energy,alpha,gamma): "
                    "unable to allocate rl_DielectricPOD array";
    
    throw rl_Exception( msg );
  }

  init( nelts, energy, alpha, gamma );
}

rl_DielectricPODArray::
rl_DielectricPODArray( size_t                       nelts,
                       rl_Traits::rl_DielectricPOD* diel )
    throw ( rl_Exception )
  : nelts_( nelts )
{
  try {

    data_ = new rl_Traits::rl_DielectricPOD[ nelts_ ];

  } catch ( std::exception& bad_alloc ) {

    char msg[] = "rl_DielectricPODArray(nelts,diel): "
                 "unable to allocate rl_DielectricPOD array";

    throw rl_Exception( msg );
  }

  init( nelts, diel );
}

void rl_DielectricPODArray::
init( size_t        nelts,
      double const* energy,
      double const* alpha,
      double const* gamma )
{
  for ( size_t n = 0 ; n < nelts ; ++n )
  {
    data_[n].energy_ = energy[n];
    data_[n].alpha_  = alpha[n];
    data_[n].gamma_  = gamma[n];
  }
  /* 
   * sort on energy...
   */
  qsort( data_, nelts_, sizeof( rl_Traits::rl_DielectricPOD ), diel_compare );
}

void rl_DielectricPODArray::
init( size_t                       nelts,
      rl_Traits::rl_DielectricPOD* diel )
{
  for ( size_t n = 0 ; n < nelts ; ++n )
  {
    data_[n].energy_ = diel[n].energy_;
    data_[n].alpha_  = diel[n].alpha_;
    data_[n].gamma_  = diel[n].gamma_;
  }
  /* 
   * sort on energy...
   */
  qsort( data_, nelts_, sizeof( rl_Traits::rl_DielectricPOD ), diel_compare );
}

void rl_DielectricPODArray::
cprint_on( std::FILE* of, char const pre[], char const pst[] ) const
{
  if ( std::strlen(pre) ) { std::fprintf(of, "%s", pre); }
  for ( size_t i = 0 ; i < nelts_ ; ++i )
  {
    std::fprintf(of, "%.15e %.15e %.15e\n",
       data_[i].energy_, data_[i].alpha_, data_[i].gamma_);
  }
  if ( std::strlen(pst) ) { std::fprintf(of, "%s", pst); }
}

int diel_compare( const void      *ck,   // search key
                  const void      *ce    // array element
                )
{
  double                      *pk = (double*)                     ck;
  rl_Traits::rl_DielectricPOD *pe = (rl_Traits::rl_DielectricPOD*)ce;

  if ( *pk < pe->energy_ )
  { return -1; }
  else if ( *pk > pe->energy_ )
  { return +1; }
  else
  { return 0; }
}

