// File:   rl_Multilayer.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006, 2007 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <rl_Multilayer.h>         // rl_Multilayer

#include <cstdio>
#include <cstddef>
#include <cmath>
#include <mathconst/mathconst.h>       // M_PI

#include <rl_DielectricData.h>      // rl_DielectricData
#include <rl_DielectricLayer.h>     // rl_DielectricLayer
#include <rl_ReflectionCoefPOD.h>   // rl_ReflectionCoefPOD

using namespace std;

//=========================================================================
// dtor, ctors

//-------------------------------------------------------------------------
rl_Multilayer::
~rl_Multilayer()
{
  if ( own_data_ && layer_ ) { delete [] layer_; }
}

//-------------------------------------------------------------------------
rl_Multilayer::
rl_Multilayer( int                 num_layers,
               rl_DielectricLayer* layers,
               rl_Traits::Bool     adopt_data )
  : num_layers_(num_layers),
    vacuum_( "vacuum" ),
    layer_(layers),
    own_data_(adopt_data)
{}

//-------------------------------------------------------------------------
void rl_Multilayer::
init( int                 num_layers,
      rl_DielectricLayer* layers,
      rl_Traits::Bool     adopt_data )
{
  if ( own_data_ && layer_ ) { delete [] layer_; }
  num_layers_ = num_layers;
  layer_      = layers;
  own_data_   = adopt_data;
}

//=========================================================================
// mutators

//-------------------------------------------------------------------------
// C The thicknesses in Angstrom are in zcoat0.  The variable zcoat
// C is 4*pi*zcoat0 / wavelength (in Angstrom).
// c
// C REFLECT computes reflectivities for polarizations perpendicular and
// C parallel to the plane of incidence.  This version can handle 1200 layers.
// [The C version has no such limit.  TJG]
//
//    subroutine reflect( nlayer, alpha, gamma, zcoat, sinphi, rcosphi,
//   a                    fourpi, lambda, srough, rpara, rperp )

int
rl_Multilayer::
multilayer_reflect_coef( rl_ReflectionCoefPOD& rfl,
                         double                energy,
                         double                sinphi )
{
  vacuum_.setup_for( energy, sinphi );

  int rc = 0;
  int n;
  for ( n = 0 ; n < num_layers_ ; ++n )
  {
    rc = layer_[n].setup_for( energy, sinphi );
    if ( rc )
    {
      rc = n+1;
      break;
    }
  }

  if (! rc )
  {
    //----------------------------------
    // Calculate complex reflectance amplitudes
    // for first interface [between vacuum and material 1].

    layer_[0].reflect_amp( vacuum_, sinphi );

    for ( n = 1 ; n < num_layers_ ; ++n )
    {
      layer_[n].reflect_amp( layer_[n-1], sinphi );
    }

    //----------------------------------
    // Calculate reflectivity from n layers for polarizations

    layer_->reflect_nlayer( layer_, num_layers_ );
  }
  else
  {
    cerr << "Problem setting up layer " << (rc-1) << endl;
  }
  rfl = layer_[0].reflection_coef();

  return rc;
}

int rl_Multilayer::
multilayer_reflectivity(
        double& rfl,                 // out: multilayer reflectivity
        double  energy,              // in:  energy of ray
        double  sg,                  // in:  sine graze angle
        double  polarization_factor  // in:  polarization_factor
)
{ 
  rl_ReflectionCoefPOD rflcoef;
  rflcoef.init();
  int rc = multilayer_reflect_coef( rflcoef, energy, sg );
  rfl    = rflcoef.reflectivity( polarization_factor );

  return rc;
}

//=========================================================================
// accessors 

//-------------------------------------------------------------------------
rl_DielectricLayer const& rl_Multilayer::
layer( int layer_no ) const
{ return layer_[layer_no]; }

//=========================================================================
// i/o

//-------------------------------------------------------------------------
std::ostream& rl_Multilayer::
dump_on( std::ostream& os, int layer_no,
                           char const pre[], 
                           char const pst[] ) const
{
  if ( strlen(pre) )
  { os << pre; }

  layer_[layer_no].dump_on( os, pre, pst );

  if ( strlen(pst) )
  { os << pst; }

  return os;
}

//-------------------------------------------------------------------------
void rl_Multilayer::
cdump_on( std::FILE* of, int layer_no,
                         char const pre[], 
                         char const pst[] ) const
{
  if ( strlen(pre) )
  { fprintf(of, "%s", pre); }

  layer_[layer_no].cdump_on( of, "", "" );

  if ( strlen(pst) )
  { fprintf(of, "%s", pst); }
}
