// File:   rl_Polarization.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006, 2007 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <rl_Polarization.h>                 // rl_Polarization rl_PolCSPOD

#include <cmath>                             // sqrt
#include <cfloat>                            // DBL_EPSILON
#include <mathconst/mathconst.h>             // M_DEG2RAD
#include <iostream>                          // ostream

#include <rl_raylib/rl_ReflectionCoefPOD.h>  // rl_ReflectionCoefPOD

using namespace std;

/* mean graze angles for shells */
#ifdef CONST_SINGRAZE
  static double alpha[] = {51.2, 46.2, 41.2, 36.4, 31.7, 27.0};
#endif

#ifdef    POLREFL_DEBUG
  static FILE *dbgw; /* --- tjg debug --- */
#endif /* POLREFL_DEBUG */

//#########################################################################
// forward declaration of statics...
//#########################################################################

//-------------------------------------------------------------------------
static void
make_Triad( dvm3_Vector&       x,
            dvm3_Vector&       y,
            dvm3_Vector const& s );

//-------------------------------------------------------------------------
static void
pol_perp_par(
        dvm3_Vector&       perp,  // out: perpendicular direction vector
        dvm3_Vector&       parv1, // out: parallel direction vector ("v1")
        dvm3_Vector&       parv2, // out: parallel direction vector ("v2")
        dvm3_Vector const& norm,  // in:  normal vector
        dvm3_Vector const& v1,    // in:  initial  ray direction vector
        dvm3_Vector const& v2     // in:  final    ray direction vector
);


//#########################################################################
// rl_PolCSPOD methods
//#########################################################################

//-------------------------------------------------------------------------
void rl_PolCSPOD::
init( double  b_over_a,  // polarization ellipse minor/major
      double  psi        // polarization ellipse angle
    )
{
  //--- initialize ray polarization ----

  double cpsi = cos(psi * M_DEG2RAD);
  double spsi = sin(psi * M_DEG2RAD);

  double sqre = sqrt(1.0 + b_over_a * b_over_a );

  c2_[0] =  rl_Traits::complex( cpsi / sqre, (double)0.0 );
  c2_[1] =  rl_Traits::complex( spsi / sqre, (double)0.0 );

  rl_Traits::complex ctmp( -spsi / sqre, (double)0.0 );
  s2_[0] =  b_over_a * ctmp;

  ctmp   =  rl_Traits::complex( cpsi / sqre, (double)0.0 );
  s2_[1] =  b_over_a * ctmp;
}

//-------------------------------------------------------------------------
double rl_PolCSPOD::
intensity() const
{
  rl_Traits::complex crossterms( c2_[0] * conj( s2_[0] )
                               + c2_[1] * conj( s2_[1] ) );
  double rfl = norm(c2_[0]) + norm(s2_[0]) + norm(c2_[1]) + norm(s2_[1])
             + 2.0 * crossterms.real();

  #ifdef OSAC_WEIGHT_KLUDGE
    /*
     * for exact agreement with OSAC - degrade weight to 8 places.
     *
     * 'Make w agree with its integer-equivalent output value'
     *    (see vflect.f)
     */
    return 1.e-8 * ((int)(1.0e8 * rfl + 0.5));
  #else
     return rfl;
  #endif
}

//-------------------------------------------------------------------------
void rl_PolCSPOD::
attenuate( double factor )
{
  const double fac = sqrt(factor);
  c2_[0] *= fac;
  c2_[1] *= fac;
  s2_[0] *= fac;
  s2_[1] *= fac;
}

//-------------------------------------------------------------------------
// write polarization amplitude to stream os.  The state is
// written as "[(r,i)(r,i)][(r,i)(r,i)]".

std::ostream& rl_PolCSPOD::
print_on( std::ostream& os ) const
{
  os << "[(" << c2_[0].real() << "," << c2_[0].imag() << ")("
             << c2_[1].real() << "," << c2_[1].imag() << ")][("
             << s2_[0].real() << "," << s2_[0].imag() << ")("
             << s2_[1].real() << "," << s2_[1].imag() << ")]";
  return os;
}

//-------------------------------------------------------------------------
// write polarization amplitude to stream os.  The state is
// written as "[(r,i)(r,i)]\n[(r,i)(r,i)]\nintensity".

void rl_PolCSPOD::
cprint_on( FILE* of ) const
{
  fprintf(of, "\nc2: [(%.15e, %.15e)(%.15e, %.15e)]",
              c2_[0].real(), c2_[0].imag(),
              c2_[1].real(), c2_[1].imag() );
  fprintf(of, "\ns2: [(%.15e, %.15e)(%.15e, %.15e)]",
              s2_[0].real(), s2_[0].imag(),
              s2_[1].real(), s2_[1].imag() );
  fprintf(of, "\n%.15e\n", intensity());
}



//#########################################################################
// rl_Polarization methods
//#########################################################################

//=========================================================================
// ctors/dtors...

//-------------------------------------------------------------------------
void rl_Polarization::
init( rl_PolCSPOD const& cs, dvm3_Vector const& dir )
{
  dvm3_Vector x1;
  dvm3_Vector y1;

  // define x1 and y1 unit vectors...
  //
  // x1 is the normalized y1 <cross-product> cs,
  // y1 is v <cross-product> x1
  //
  // the set [x1, y1, dir] forms an orthonormal basis.

  make_Triad( x1, y1, dir );

  // Construct real & imag parts of C and S vectors from
  // c2_.q1, c2_.q2, s1_.q1, and s2_.q2 amplitudes.
  //
  // Note that c2_ and s2_ are complex coefficients
  // multiplying the (real) polarization unit vectors.
  //
  // C_r_ = Re[ c2_.q1 * x1 + c2_.q2 * y1 ]
  // C_i_ = Im[ c2_.q1 * x1 + c2_.q2 * y1 ]
  //
  // S_r_ = Re[ s2_.q1 * x1 + s2_.q2 * y1 ]
  // S_i_ = Im[ s2_.q1 * x1 + s2_.q2 * y1 ]

  C_r_.lincomb( cs.c2_[0].real(),  x1,  cs.c2_[1].real(),  y1 );
  C_i_.lincomb( cs.c2_[0].imag(),  x1,  cs.c2_[1].imag(),  y1 );
  S_r_.lincomb( cs.s2_[0].real(),  x1,  cs.s2_[1].real(),  y1 );
  S_i_.lincomb( cs.s2_[0].imag(),  x1,  cs.s2_[1].imag(),  y1 );
}

//=========================================================================
// accessors...

//-------------------------------------------------------------------------
double rl_Polarization::
intensity() const
{
  return dot(C_r_, C_r_) + dot(C_i_, C_i_) + dot(S_r_, S_r_) + dot(S_i_, S_i_)
                 + 2.0 * ( dot(C_i_, S_r_) - dot(C_r_, S_i_) );
}

//-------------------------------------------------------------------------

void rl_Polarization::
get_PolCSPOD( rl_PolCSPOD& cs, dvm3_Vector const& dir ) const
{
  // define x2 and y2 unit vectors...
  // x2 is the normalized y2 <cross-product> cs,
  // y2 is v <cross-product> x2

  dvm3_Vector x2;
  dvm3_Vector y2;
  make_Triad( x2, y2, dir );

  // Convert the C, S complex vectors back into c2_ and s2_
  // complex amplitudes

  double d1 = dot( x2, C_r_);
  double d2 = dot( x2, C_i_);
  cs.c2_[0] = rl_Traits::complex( d1, d2 );
  cs.c2_[1] = rl_Traits::complex( dot( y2, C_r_), dot( y2, C_i_) );
  cs.s2_[0] = rl_Traits::complex( dot( x2, S_r_), dot( x2, S_i_) );
  cs.s2_[1] = rl_Traits::complex( dot( y2, S_r_), dot( y2, S_i_) );
}

//=========================================================================
// mutators...

//-------------------------------------------------------------------------
void rl_Polarization::
attenuate( double factor )
{
  const double fac = sqrt(factor);
  C_r_ *= fac;
  C_i_ *= fac;
  S_r_ *= fac;
  S_i_ *= fac;
}

//-------------------------------------------------------------------------
// Description: Reflect the polarization state of a ray:
//
// Given surface normal, and incident and reflected ray direction vectors,
//  and the parallel and perpendicular reflection coefficients,
//  evaluate the reflected polarization vectors.
//
// Notes      : based in part on OSAC routine "pzrefl"

void rl_Polarization::
reflect(
         dvm3_Vector          const& normal,
         dvm3_Vector          const& dir_in,
         dvm3_Vector          const& dir_out,
         rl_ReflectionCoefPOD const& rflcoef
       )
{
  // calculate sine of the grazing angle
  // double sg = dot( dir_in, normal );

  #ifdef CONST_SINGRAZE
    sg = sin(alpha[CONST_SINGRAZE-1] * 0.0002908882);
  #endif

  //x // reflect the direction vector:
  //x //    v_o = (1 - 2 norm <diadprod> norm) <dotprod> v_i
  //x 
  //x dir_out.lincomb( 1.0, dir_in, -2.0 * sg, &srf->norm );

  // form the incoming/outgoing polarization basis vectors 

  dvm3_Vector  perp;        // perpendicular to vo-vn plane
  dvm3_Vector  parvi;       // normal to vi, in vi-vn plane
  dvm3_Vector  parvo;       // normal to vo, in vo-vn plane

  pol_perp_par( perp, parvi, parvo, normal, 
                dir_in, dir_out );
  
  // build components of T matrix:
  //
  //   T_par = parvo <dyadproduct> parvi
  //   T_prp = prpo  <dyadproduct> prpi
  //
  //   Note that (for now) prpo = prpi = perp

  dvm3_Matrix  T_prp;       // projector for perp vector
  dvm3_Matrix  T_par;       // projector for par  vector

  T_par.dyad_product( parvo, parvi );
  T_prp.dyad_product( perp,  perp  );

  // combine with complex reflectivities; form real and imag parts of T 

  dvm3_Matrix  T_re;        // real part of T matrix
  dvm3_Matrix  T_im;        // imag part of T matrix

  T_re.lincomb( rflcoef.para().real(), T_par, 
                rflcoef.perp().real(), T_prp );
  T_im.lincomb( rflcoef.para().imag(), T_par, 
                rflcoef.perp().imag(), T_prp );

  /* 
   * Apply Re, Im part of T matrix to Re, Im polarization vectors.
   */ 
  dvm3_Vector  Tr_Cr;    T_re.mvmult( Tr_Cr, C_r_ );              // T_re . C_r_
  dvm3_Vector  Tr_Ci;    T_re.mvmult( Tr_Ci, C_i_ );              // T_re . C_i_
  dvm3_Vector  Ti_Cr;    T_im.mvmult( Ti_Cr, C_r_ );              // T_im . C_r_
  dvm3_Vector  Ti_Ci;    T_im.mvmult( Ti_Ci, C_i_ );              // T_im . C_i_

  dvm3_Vector  Tr_Sr;    T_re.mvmult( Tr_Sr, S_r_ );              // T_re . S_r_
  dvm3_Vector  Tr_Si;    T_re.mvmult( Tr_Si, S_i_ );              // T_re . S_i_
  dvm3_Vector  Ti_Sr;    T_im.mvmult( Ti_Sr, S_r_ );              // T_im . S_r_
  dvm3_Vector  Ti_Si;    T_im.mvmult( Ti_Si, S_i_ );              // T_im . S_i_

  // collect and reassemble into outgoing PolVec... 

  C_r_ = Tr_Cr - Ti_Ci;
  C_i_ = Tr_Ci + Ti_Cr;
  S_r_ = Tr_Sr - Ti_Si;
  S_i_ = Tr_Si + Ti_Sr;
}

//-------------------------------------------------------------------------
// write polarization vectors to stream os.  The state is
// written as "[x y z][x y z][x y z][x y z]".

std::ostream& rl_Polarization::
print_on( std::ostream& os ) const
{
  os << C_r_ << C_i_ << S_r_ << S_i_;
  return os;
}

//-------------------------------------------------------------------------
// write polarization vectors to stream os.  The state is
// written as "[x y z][x y z][x y z][x y z]".

void rl_Polarization::
cprint_on( FILE* of ) const
{
  double* vec = new double[3];
  C_r_.copy_to_cvec( vec );
  fprintf(of, "\nC_r_: [%.15e, %.15e %.15e]\n",
              vec[0], vec[1], vec[2]);
  C_i_.copy_to_cvec( vec );
  fprintf(of, "\nC_i_: [%.15e, %.15e %.15e]\n",
              vec[0], vec[1], vec[2]);
  S_r_.copy_to_cvec( vec );
  fprintf(of, "\nS_r_: [%.15e, %.15e %.15e]\n",
              vec[0], vec[1], vec[2]);
  S_i_.copy_to_cvec( vec );
  fprintf(of, "\nS_i_: [%.15e, %.15e %.15e]\n",
              vec[0], vec[1], vec[2]);
  delete [] vec;
}

#define TINY_VALUE (100.0 * DBL_EPSILON)

//#########################################################################
// file static definitions...
//#########################################################################

//-------------------------------------------------------------------------
void 
pol_perp_par(
        dvm3_Vector&       perp,  // out: perpendicular direction vector
        dvm3_Vector&       parv1, // out: parallel direction vector ("v1")
        dvm3_Vector&       parv2, // out: parallel direction vector ("v2")
        dvm3_Vector const& norm,  // in:  normal vector
        dvm3_Vector const& v1,    // in:  initial  ray direction vector
        dvm3_Vector const& v2     // in:  final    ray direction vector
)
{
  // find perpendicular unit vector

  if ( fabs(dot(v1, v2)) < 0.99999 )
  {
    perp.cross( v2, v1 );       // use v2 cross v1
  }
  else
  {
    if ( fabs(dot(norm,v1)) < 0.99999 )
    {
      perp.cross( norm, v1 );   // use norm cross v1
    }
    else
    {
      if ( fabs(1.0 - fabs( norm[0] )) < TINY_VALUE )
      {
        perp.init( norm[2], 0.0, -norm[0] );    // use norm cross e_y
      }
      else
      {
        perp.init( 0.0, -norm[2], -norm[1] );   // use norm cross e_x
      }
    }
  }
  perp.unitize();

  // now find the parallel unit vectors

  parv1.cross( v1, perp );
  parv2.cross( v2, perp );
}

//-------------------------------------------------------------------------
void
make_Triad( dvm3_Vector&       x,
            dvm3_Vector&       y,
            dvm3_Vector const& s )
{
  // form the "x" and "y" unit vectors

  // make "x" unit vector

  if ( fabs(1.0 - fabs( s[1] )) > TINY_VALUE )
  {
    x.init( s[2], 0.0, -s[0] );       // x = e_x <crossprod> s
  }
  else
  {
    x.init( 1.0, 0.0, 0.0 );          // x = e_x is close enough
  }
  x.unitize();                        // make x a unit vector

  // y = s <crossprod> x; s and x are orthogonal
  // unit vectors so normalization isn't needed.

  y.cross( s, x );                    // y = s <crossprod> x

  // we now have an orthonormal triad.
}
#undef TINY_VALUE

#undef CONST_SINGRAZE
