// File:   t_mrl_mml.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2007 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include "rl_raylib/rl_Multilayer.h"
#include "rl_raylib/rl_ReflectionCoefPOD.h"

#include <rl_raylib/rl_DielectricLayer.h>
#include <rl_raylib/rl_DielectricPODArray.h>

#include <mathconst/mathconst.h>
#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>

using namespace std;

extern
rl_DielectricPODArray*
new_dielectric_pod_array( char const srcfile[] );

static
int rl_mml(int                    nlayer2,
           double                 energy,
           double                 angle,
           double                 thick0,
           double                 thick1,
           char const*            outfile,
           rl_DielectricPODArray* diel0,
           rl_DielectricPODArray* diel1 )
{
#if defined( DEBUG_MRL_MML )
  fprintf( stderr, "---------------\n" );
  fprintf( stderr, "nlayer2        %d\n",    nlayer2 );
  fprintf( stderr, "energy         %.15e\n", energy  );
  fprintf( stderr, "angle          %.15e\n", angle   );
  fprintf( stderr, "thick0         %.15e\n", thick0  );
  fprintf( stderr, "thick1         %.15e\n", thick1  );
#endif

  rl_DielectricLayer* diel = new rl_DielectricLayer[ 2*nlayer2 + 1 ];

  double sg     = sin( angle * M_DEG2RAD );
  
#if defined( DEBUG_MRL_MML )
  fprintf( stderr, "singraze       %.15e\n", sg);
#endif

  int j = 0;
  for ( int i = 0 ; i < nlayer2-1 ; ++i )
  {
    diel[j].init( diel0->const_data_ptr(), diel0->num_elts(),
                  thick0, 0.0, rl_Traits::ERoughNone,
                  rl_Traits::ELinLin, 1.0, "ch2" );
    diel[j].setup_for( energy, sg );
#if defined( DEBUG_MRL_MML )
    fprintf( stderr, "___aa0___ ch2    >%d< >%d<\n", j, i);
#endif

    ++j;

    diel[j].init( diel1->const_data_ptr(), diel1->num_elts(),
                    thick1, 0.0, rl_Traits::ERoughNone,
                    rl_Traits::ELinLin, 1.0, "Ir" );
    diel[j].setup_for( energy, sg );
#if defined( DEBUG_MRL_MML )
    fprintf( stderr, "___aa1___ Ir     >%d< >%d<\n", j, i);
#endif

    ++j;
  }
  diel[j].init( diel0->const_data_ptr(), diel0->num_elts(),
                thick0, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "ch2" );
  diel[j].setup_for( energy, sg );
#if defined( DEBUG_MRL_MML )
  fprintf( stderr, "___aa2___ ch2    >%d<\n", j);
#endif

  ++j;

  diel[j].init( diel1->const_data_ptr(), diel1->num_elts(),
                1e30, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Ir", rl_Traits::True );
  diel[j].setup_for( energy, sg );
#if defined( DEBUG_MRL_MML )
  fprintf( stderr, "___aa3___ Ir(s)  >%d<\n", j);
#endif
  
  int nml = ++j;

  rl_Multilayer ml( nml, diel, rl_Traits::False );

  FILE* out = fopen(outfile, "w");
  if (! out )
  {
    fprintf(stderr, "Failed to open output file %s\n", outfile);
    exit(EXIT_FAILURE);
  }
  fprintf( out, "%d %.16e\n", ml.num_layers(), energy );
  for ( int i = 0 ; i < ml.num_layers() ; ++i )
  {
    fprintf(out, "%.16e %.16e %.16e\n", 
            diel[i].alpha(), diel[i].gamma(), diel[i].thickness() );
  }
  fclose( out );

  rl_ReflectionCoefPOD rflcoef;
  int rc = ml.multilayer_reflect_coef( rflcoef, energy, sg );

  double rflct;
  ml.multilayer_reflectivity( rflct, energy, sg );

  if ( rc )
  {
    ml.cdump_on( stderr, rc-1,
                 static_cast<char const*>("ERROR: "),
                 static_cast<char const*>("in multilayer_reflect_coef") );
  }

  int rc1 = 0;
  double rfl = 0.0;
  rc1 = (ml.multilayer_reflectivity( rfl, energy, sg ));
  rc &= !rc1;
  if ( rc1 )
  {
    ml.cdump_on( stderr, rc1-1,
                 static_cast<char const*>("ERROR: "),
                 static_cast<char const*>("in multilayer_reflectivity") );
  }

  fprintf( stdout, "%.15e\n", rfl );

  delete [] diel;

  return !rc;
}

int
main( int argc, char* argv[] ) 
{
  if ( argc != 8 )
  {
    cerr << "Usage: " << argv[0] 
         << " nlayer2 energy(keV) angle(arcmin) "
	 << "thickness0(Angstrom) thickness1(Angstrom) "
	 << "outfilename srcdir\n";
    return EXIT_FAILURE;
  }

  int    nlayer2 = atoi( (++argv)[0] );
  double energy  = atof( (++argv)[0] );
  double angle   = atof( (++argv)[0] );
  double thick0  = atof( (++argv)[0] );
  double thick1  = atof( (++argv)[0] );

  char const* outfile = (++argv)[0];

  int srclen = strlen((++argv)[0]);

#if defined( DEBUG_MRL_MML )
  fprintf( stderr, "--------------------------------------\n" );
  fprintf( stderr, "nlayer2        %d\n",    nlayer2 );
  fprintf( stderr, "energy         %.15e\n", energy  );
  fprintf( stderr, "angle          %.15e\n", angle   );
  fprintf( stderr, "thick0         %.15e\n", thick0  );
  fprintf( stderr, "thick1         %.15e\n", thick1  );
  fprintf( stderr, "outfile        %s\n",    outfile );
  fprintf( stderr, "dir            %s\n",    argv[0] );
#endif

  angle /= 60.0;

  char f0[]   = "/tests/inputs/ch2_stripped.rdb";
  char f1[]   = "/tests/inputs/all_dbs_fit_i_stripped.rdb";

  char* src0 = new char[static_cast<size_t>(srclen + sizeof(f0) + 1)];
  char* src1 = new char[static_cast<size_t>(srclen + sizeof(f1) + 1)];

  strncpy( src0, argv[0], srclen+1 );
  strncpy( src1, argv[0], srclen+1 );

  strncat( src0, f0, sizeof(f0)+1 );
  strncat( src1, f1, sizeof(f1)+1 );

#if defined( DEBUG_MRL_MML )
  fprintf( stderr, "src0           %s\n",    src0 );
  fprintf( stderr, "src1           %s\n",    src1 );
#endif

  rl_DielectricPODArray* diel0 = new_dielectric_pod_array(src0);
  rl_DielectricPODArray* diel1 = new_dielectric_pod_array(src1);

  rl_mml(nlayer2, energy, angle, thick0, thick1, outfile, diel0, diel1);

  delete    diel0;
  delete    diel1;

  delete [] src0;
  delete [] src1;

  return EXIT_SUCCESS;
}
