// File:   t_rl_DielectricData.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <rl_raylib/rl_DielectricData.h>
#include <rl_raylib/rl_DielectricPODArray.h>

#include <mathconst/mathconst.h>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <testlib/testlib.h>

using namespace std;

extern
rl_DielectricPODArray*
new_dielectric_pod_array( char const srcfile[] );

static
int rl_DielectricData_test01( rl_DielectricPODArray* diel )    
  /* test ctor, dtor */
{
  rl_DielectricData  ml( diel->const_data_ptr(), 
                         diel->num_elts(), 
                         rl_Traits::ELinLin, 1.0 );

  return 1;
}

static
int rl_DielectricData_test02( double                 energy,
                              rl_DielectricPODArray* diel0,
                              rl_DielectricPODArray* diel1,
                              rl_DielectricPODArray* diel2,
                              rl_DielectricPODArray* diel3 )
{
  rl_DielectricData ml0( diel0->const_data_ptr(), 
                         diel0->num_elts(), 
                         rl_Traits::ELinLin, 1.0 );
  rl_DielectricData ml1( diel1->const_data_ptr(), 
                         diel1->num_elts(), 
                         rl_Traits::ELinLin, 1.0 );
  rl_DielectricData ml2( diel2->const_data_ptr(), 
                         diel2->num_elts(), 
                         rl_Traits::ELinLin, 1.0 );
  rl_DielectricData ml3( diel3->const_data_ptr(), 
                         diel3->num_elts(), 
                         rl_Traits::ELinLin, 1.0 );

  double alpha0, alpha1, alpha2, alpha3;
  double gamma0, gamma1, gamma2, gamma3;

  ml0.alpha_gamma( energy, alpha0, gamma0 );
  ml1.alpha_gamma( energy, alpha1, gamma1 );
  ml2.alpha_gamma( energy, alpha2, gamma2 );
  ml3.alpha_gamma( energy, alpha3, gamma3 );

#if defined (DEBUG_DIELECTRIC_DATA)
  fprintf(stderr, "LAYER, E, ALPHA, GAMMA\n");
  fprintf(stderr, "0 %.14e %.14e %.14e\n", energy, alpha0, gamma0);
  fprintf(stderr, "1 %.14e %.14e %.14e\n", energy, alpha1, gamma1);
  fprintf(stderr, "2 %.14e %.14e %.14e\n", energy, alpha2, gamma2);
  fprintf(stderr, "3 %.14e %.14e %.14e\n", energy, alpha3, gamma3);
#endif

  return !fuzzy_compare(alpha0, 2.16766835256698e-04 ) &&
         !fuzzy_compare(gamma0, 7.89114358264237e-06 ) &&
	 !fuzzy_compare(alpha1, 2.18106737650038e-03 ) &&
         !fuzzy_compare(gamma1, 5.77889009473482e-04 ) &&
	 !fuzzy_compare(alpha2, 1.22126231552795e-03 ) &&
         !fuzzy_compare(gamma2, 2.55168199130435e-04 ) &&
	 !fuzzy_compare(alpha3, 4.33590612422360e-04 ) &&
         !fuzzy_compare(gamma3, 3.83633780621118e-05 ) ;
}

int
main( int argc, char** argv )
{
  if ( argc != 2 )
  {
    cerr << "Usage: " << argv[0] << " srcdir\n";
    return EXIT_FAILURE;
  }

  char f0[]   = "/tests/inputs/ch2_stripped.rdb";
  char f1[]   = "/tests/inputs/all_dbs_fit_i_stripped.rdb";
  char f2[]   = "/tests/inputs/Cr_Gullikson95_stripped.rdb";
  char f3[]   = "/tests/inputs/Zerodur_Gullikson95_stripped.rdb";

  int srclen = strlen((++argv)[0]);

  char* src0 = new char[static_cast<size_t>(srclen + sizeof(f0) + 1)];
  char* src1 = new char[static_cast<size_t>(srclen + sizeof(f1) + 1)];
  char* src2 = new char[static_cast<size_t>(srclen + sizeof(f2) + 1)];
  char* src3 = new char[static_cast<size_t>(srclen + sizeof(f3) + 1)];

  START("Testing rl_DielectricData class; reading of dielectric data");

  strncpy( src0, argv[0], srclen+1 );
  strncpy( src1, argv[0], srclen+1 );
  strncpy( src2, argv[0], srclen+1 );
  strncpy( src3, argv[0], srclen+1 );

  strncat( src0, f0, sizeof(f0)+1 );
  strncat( src1, f1, sizeof(f1)+1 );
  strncat( src2, f2, sizeof(f2)+1 );
  strncat( src3, f3, sizeof(f3)+1 );

  rl_DielectricPODArray* diel0 = new_dielectric_pod_array(src0);
  rl_DielectricPODArray* diel1 = new_dielectric_pod_array(src1);
  rl_DielectricPODArray* diel2 = new_dielectric_pod_array(src2);
  rl_DielectricPODArray* diel3 = new_dielectric_pod_array(src3);

  int rc = 1;
  int rc1;

  rc1 = rl_DielectricData_test01( diel0 );
  TEST( "ctor, dtor", rc1, 1 );
  rc &= rc1;

  rc1 = rl_DielectricData_test02( 1.5, diel0, diel1, diel2, diel3 );
  TEST( "alpha_gamma", rc1, 1 );
  rc &= rc1;

  SUMMARY();

  delete diel0;
  delete diel1;
  delete diel2;
  delete diel3;

  delete [] src0;
  delete [] src1;
  delete [] src2;
  delete [] src3;

  return !rc;
}

