// File:   t_rl_DielectricLayer.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <rl_raylib/rl_DielectricLayer.h>
#include <rl_raylib/rl_DielectricPODArray.h>

#include <mathconst/mathconst.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <testlib/testlib.h>

using namespace std;

extern
rl_DielectricPODArray*
new_dielectric_pod_array( char const srcfile[] );

static
int rl_DielectricLayer_test01( rl_DielectricPODArray* diel0,
                               rl_DielectricPODArray* diel1,
                               rl_DielectricPODArray* diel2,
                               rl_DielectricPODArray* diel3 )
                             /* test ctor, dtor */
{
  rl_DielectricLayer* diel = new rl_DielectricLayer[4];

#ifdef DEBUG_DIELECTRICLAYER
  FILE* out = fopen("dl_test01.out", "w");
  if (! out )
  {
    fprintf(stderr, "Failed to open output file dl_test01.out");
    exit(EXIT_FAILURE);
  }
  fprintf(out, ">>>>>>>>>>-------------------------------\n");
  diel[0].cdump_on( out, 
                    static_cast<char const*>("diel[0]"), 
                    static_cast<char const*>("") ); 
  diel[1].cdump_on( out, 
                    static_cast<char const*>("diel[1]"), 
                    static_cast<char const*>("") ); 
  diel[2].cdump_on( out, 
                    static_cast<char const*>("diel[2]"), 
                    static_cast<char const*>("") ); 
  diel[3].cdump_on( out, 
                    static_cast<char const*>("diel[3]"), 
                    static_cast<char const*>("") ); 
  fprintf(out, "<<<<<<<<<<-------------------------------\n");
#endif

  diel[0].init( diel0->const_data_ptr(), diel0->num_elts(), 
                10.0, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "ch2" );
  diel[1].init( diel1->const_data_ptr(), diel1->num_elts(), 
                297.0, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Ir" );
  diel[2].init( diel2->const_data_ptr(), diel2->num_elts(), 
                93.0, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Cr" );
  diel[3].init( diel3->const_data_ptr(), diel3->num_elts(), 
                1e30, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Zerodur" );

#ifdef DEBUG_DIELECTRICLAYER
  fprintf(out, ">>>>>>>>>>-------------------------------\n");
  diel[0].cdump_on( out, 
                    static_cast<char const*>("diel[0]"), 
                    static_cast<char const*>("") ); 

  diel[1].cdump_on( out, 
                    static_cast<char const*>("diel[1]"), 
                    static_cast<char const*>("") ); 

  diel[2].cdump_on( out, 
                    static_cast<char const*>("diel[2]"), 
                    static_cast<char const*>("") ); 

  diel[3].cdump_on( out, 
                    static_cast<char const*>("diel[3]"), 
                    static_cast<char const*>("") ); 
  fprintf(out, "<<<<<<<<<<-------------------------------\n");
  fclose(out);
#endif

  return 1;
}

int
main( int argc, char* argv[] ) 
{
  if ( argc != 2 )
  {
    cerr << "Usage: " << argv[0] << " srcdir\n";
    return EXIT_FAILURE;
  }
  char f0[]   = "/tests/inputs/ch2_stripped.rdb";
  char f1[]   = "/tests/inputs/all_dbs_fit_i_stripped.rdb";
  char f2[]   = "/tests/inputs/Cr_Gullikson95_stripped.rdb";
  char f3[]   = "/tests/inputs/Zerodur_Gullikson95_stripped.rdb";

  int srclen = strlen((++argv)[0]);

  char* src0 = new char[static_cast<size_t>(srclen + sizeof(f0) + 1)];
  char* src1 = new char[static_cast<size_t>(srclen + sizeof(f1) + 1)];
  char* src2 = new char[static_cast<size_t>(srclen + sizeof(f2) + 1)];
  char* src3 = new char[static_cast<size_t>(srclen + sizeof(f3) + 1)];

  strncpy( src0, argv[0], srclen+1 );
  strncpy( src1, argv[0], srclen+1 );
  strncpy( src2, argv[0], srclen+1 );
  strncpy( src3, argv[0], srclen+1 );

  strncat( src0, f0, sizeof(f0)+1 );
  strncat( src1, f1, sizeof(f1)+1 );
  strncat( src2, f2, sizeof(f2)+1 );
  strncat( src3, f3, sizeof(f3)+1 );

  rl_DielectricPODArray* diel0 = new_dielectric_pod_array(src0);
  rl_DielectricPODArray* diel1 = new_dielectric_pod_array(src1);
  rl_DielectricPODArray* diel2 = new_dielectric_pod_array(src2);
  rl_DielectricPODArray* diel3 = new_dielectric_pod_array(src3);

  int rc = rl_DielectricLayer_test01(diel0, diel1, diel2, diel3);

  delete    diel0;
  delete    diel1;
  delete    diel2;
  delete    diel3;

  delete [] src2;
  delete [] src3;

  TEST( "ctor, dtor", rc, 1 );

  return !rc;
}
