// File:   t_rl_Multilayer.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include "rl_raylib/rl_Multilayer.h"
#include "rl_raylib/rl_ReflectionCoefPOD.h"

#include <rl_raylib/rl_DielectricLayer.h>
#include <rl_raylib/rl_DielectricPODArray.h>

#include <mathconst/mathconst.h>
#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>

// #define DEBUG_MULTILAYER

using namespace std;

extern
rl_DielectricPODArray*
new_dielectric_pod_array( char const srcfile[] );

static
int rl_Multilayer_test_ol(double                 energy,
                          double                 angle,
                          double                 ol,
                          double                 srough,
                          char const*            outfile,
                          rl_DielectricPODArray* diel0,
                          rl_DielectricPODArray* diel1,
                          rl_DielectricPODArray* diel2,
                          rl_DielectricPODArray* diel3 )
{
  rl_DielectricLayer diel[4];
  
  diel[0].init( diel0->const_data_ptr(), diel0->num_elts(),
                ol, srough, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "ch2" );
  diel[1].init( diel1->const_data_ptr(), diel1->num_elts(),
                327.0, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Ir" );
  diel[2].init( diel2->const_data_ptr(), diel2->num_elts(),
                93.0, srough, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Cr" );
  diel[3].init( diel3->const_data_ptr(), diel3->num_elts(),
                1.e30, srough, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Zerodur", rl_Traits::True );
  
  rl_Multilayer ml( 4, diel, rl_Traits::False );

  double angle_deg = angle / 60.0;

  double sg     = sin( angle_deg * M_DEG2RAD );

  FILE* out = fopen(outfile, "w");
  if (! out )
  {
    fprintf(stderr, "Failed to open output file %s\n", outfile);
    exit(EXIT_FAILURE);
  }

  fprintf( out, "With %.3f arcmin graze angle\n", angle);
  fprintf( out, "With %.3f A overlayer\n", ol);
  fprintf( out, "With %.3f A roughness\n", srough);
  fprintf( out, "NUM_LAYERS: %d\n", ml.num_layers() );

  rl_ReflectionCoefPOD rflcoef;
  int rc = ml.multilayer_reflect_coef( rflcoef, energy, sg );
  if ( rc )
  {
    ml.cdump_on( out, rc-1,
                 static_cast<char const*>("ERROR: "),
                 static_cast<char const*>("in multilayer_reflect_coef") );
  }

  int rc1 = 0;
  double rfl = 0.0;
  rc1 = (ml.multilayer_reflectivity( rfl, energy, sg ));
  rc &= !rc1;
  if ( rc1 )
  {
    ml.cdump_on( out, rc1-1,
                 static_cast<char const*>("ERROR: "),
                 static_cast<char const*>("in multilayer_reflectivity") );
  }

  for ( int i = 0 ; i < ml.num_layers() ; ++i )
  {
    fprintf(out, "Layer: %d, Energy %.3f\n", i, energy);
    fprintf(out, "Layer: %d, reflectivity %.7e\n", i, diel[i].reflectivity() );
#if defined( DEBUG_MULTILAYER )
    diel[i].cdump_on( out,
                      static_cast<char const*>(""),
                      static_cast<char const*>("") );
    diel[i].cprint_constraints_on( 
                      out,
                      static_cast<char const*>(""),
                        static_cast<char const*>("") );
#endif
  }

  fclose(out);

  return !rc;
}

static
int rl_Multilayer_test_nool(double                 energy,
                            double                 angle,
                            double                 srough,
                            char const*            outfile,
                            rl_DielectricPODArray* diel0,
                            rl_DielectricPODArray* diel1,
                            rl_DielectricPODArray* diel2 )
{
  rl_DielectricLayer diel[3];
  
  diel[0].init( diel0->const_data_ptr(), diel0->num_elts(),
                327.0, srough, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Ir" );
  diel[1].init( diel1->const_data_ptr(), diel1->num_elts(),
                93.0, srough, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Cr" );
  diel[2].init( diel2->const_data_ptr(), diel2->num_elts(),
                1.e30, srough, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Zerodur", rl_Traits::True );
  
  rl_Multilayer ml( 3, diel, rl_Traits::False );

  // double alpha  = 0.5;
  //
  double angle_deg = angle / 60.0;

  double sg     = sin( angle_deg * M_DEG2RAD );

  FILE* out = fopen(outfile, "w");
  if (! out )
  {
    fprintf(stderr, "Failed to open output file %s\n", outfile);
    exit(EXIT_FAILURE);
  }

  fprintf( out, "NO overlayer\n");
  fprintf( out, "NUM_LAYERS: %d\n", ml.num_layers() );

  rl_ReflectionCoefPOD rflcoef;
  int rc = ml.multilayer_reflect_coef( rflcoef, energy, sg );
  if ( rc )
  {
    ml.cdump_on( out, rc-1,
                 static_cast<char const*>("ERROR: "),
                 static_cast<char const*>("in multilayer_reflect_coef") );
  }

  int rc1 = 0;
  double rfl = 0.0;
  rc1 = (ml.multilayer_reflectivity( rfl, energy, sg ));
  rc &= !rc1;
  if ( rc1 )
  {
    ml.cdump_on( out, rc1-1,
                 static_cast<char const*>("ERROR: "),
                 static_cast<char const*>("in multilayer_reflectivity") );
  }

  for ( int i = 0 ; i < ml.num_layers() ; ++i )
  {
    fprintf(out, "Layer: %d, Energy %.3f\n", i, energy);
    fprintf(out, "Layer: %d, reflectivity %.6e\n", i, diel[i].reflectivity() );
#if defined( DEBUG_MULTILAYER )
    diel[i].cdump_on( out,
                      static_cast<char const*>(""),
                      static_cast<char const*>("") );
    diel[i].cprint_constraints_on( 
                     out,
                     static_cast<char const*>(""),
                     static_cast<char const*>("") );
#endif
  }

  fclose(out);

  return !rc;
}

int
main( int argc, char* argv[] ) 
{
  if ( argc != 8 )
  {
    cerr << "Usage: " << argv[0] 
         << " energy(keV) angle(arcmin) olthickness(Angstrom) "
         << "srough(Angstrom) haveoverlayer outfilename srcdir\n";
    return EXIT_FAILURE;
  }

  double energy = atof( (++argv)[0] );
  double angle  = atof( (++argv)[0] );
  double thick  = atof( (++argv)[0] );
  double srough = atof( (++argv)[0] );
  int    haveol = atoi( (++argv)[0] );
  char const* outfile = (++argv)[0];
  int srclen = strlen((++argv)[0]);

#if defined( DEBUG_MULTILAYER_TEST )
  fprintf( stderr, "energy    >%.15e<\n", energy  );
  fprintf( stderr, "angle     >%.15e<\n", angle   );
  fprintf( stderr, "thick     >%.15e<\n", thick   );
  fprintf( stderr, "srough    >%.15e<\n", srough  );
  fprintf( stderr, "haveol    >%d<\n",    haveol  );
  fprintf( stderr, "outfile   >%s<\n",    outfile );
#endif

  char f0[]   = "/tests/inputs/ch2_stripped.rdb";
  char f1[]   = "/tests/inputs/all_dbs_fit_i_stripped.rdb";
  char f2[]   = "/tests/inputs/Cr_Gullikson95_stripped.rdb";
  char f3[]   = "/tests/inputs/Zerodur_Gullikson95_stripped.rdb";

  char* src0 = new char[static_cast<size_t>(srclen + sizeof(f0) + 1)];
  char* src1 = new char[static_cast<size_t>(srclen + sizeof(f1) + 1)];
  char* src2 = new char[static_cast<size_t>(srclen + sizeof(f2) + 1)];
  char* src3 = new char[static_cast<size_t>(srclen + sizeof(f3) + 1)];

  strncpy( src0, argv[0], srclen+1 );
  strncpy( src1, argv[0], srclen+1 );
  strncpy( src2, argv[0], srclen+1 );
  strncpy( src3, argv[0], srclen+1 );

  strncat( src0, f0, sizeof(f0)+1 );
  strncat( src1, f1, sizeof(f1)+1 );
  strncat( src2, f2, sizeof(f2)+1 );
  strncat( src3, f3, sizeof(f3)+1 );

  rl_DielectricPODArray* diel0 = new_dielectric_pod_array(src0);
  rl_DielectricPODArray* diel1 = new_dielectric_pod_array(src1);
  rl_DielectricPODArray* diel2 = new_dielectric_pod_array(src2);
  rl_DielectricPODArray* diel3 = new_dielectric_pod_array(src3);

  int rc  = 1;
  int rc1;

  if ( haveol )
  {
    rc1 = rl_Multilayer_test_ol(  energy, angle, thick, srough,
                                  outfile,
                                  diel0, diel1, diel2, diel3);
  } else {
    rc1 = rl_Multilayer_test_nool(energy, angle,        srough,
                                  outfile,
                                         diel1, diel2, diel3);
  }

  delete    diel0;
  delete    diel1;
  delete    diel2;
  delete    diel3;

  delete [] src0;
  delete [] src1;
  delete [] src2;
  delete [] src3;

  return !rc;
}
