// File:   t_rl_MultilayerSurface.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include "rl_raylib/rl_MultilayerSurface.h"
#include "rl_raylib/rl_ReflectionCoefPOD.h"
#include "rl_raylib/rl_Ray.h"
#include "rl_raylib/rl_Polarization.h"

#include <rl_raylib/rl_DielectricPODArray.h>
#include <rl_raylib/rl_DielectricLayer.h>

#include <mathconst/mathconst.h>
#include <cstdlib>
#include <iostream>
#include <testlib/testlib.h>

const double TOL = 500.0 * DBL_EPSILON;

const double GRAZEANGLE = 52.0 / 60.0  * M_DEG2RAD;

extern 
rl_DielectricPODArray*
new_dielectric_pod_array( char const srcfile[] );

static
int rl_MultilayerSurface_test01( rl_DielectricPODArray* diel0
                                 )
                                 /* Ir substrate only */
{
  rl_DielectricLayer diel[1];

  diel[0].init( diel0->const_data_ptr(), diel0->num_elts(),
                1e30, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "ol" );

  rl_Multilayer ml( 1, diel, rl_Traits::False );

  dvm3_Vector   norm( cos( GRAZEANGLE ), 0.0, sin( GRAZEANGLE ) );

  rl_MultilayerSurface surf( ml, norm );
  dvm3_Vector   pos( 100.0, 0.0, 0.0 );
  dvm3_Vector   dir(   0.0, 0.0, 1.0 );
  rl_BasicRay   bray( pos, dir, 1.5, 12 );
  rl_PolCSPOD   cs;
  cs.init();
  rl_Ray        ray(  bray, cs );

  surf.reflect( ray );

  rl_ReflectionCoefPOD rfl = surf.reflection_coefs();

  int rc = 1;
  int rc1;

  rc1 = !Test_fuzzy_compare(rfl.para().real(), -8.815071158052333e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.para().imag(),  3.517246540337252e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.perp().real(), -8.813154338444567e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.perp().imag(),  3.525061583151267e-01, TOL);
  rc &= rc1;

  rc1 = !Test_fuzzy_compare(rfl.reflectivity(), 9.008712565264683e-01, TOL);
  rc &= rc1;

  ray.get_PolCSPOD( cs );

#if defined( DEBUG_MULTILAYERSURFACE )
  fprintf( stderr, "rl_MultilayerSurface_test01:  REFLECTIVITY:  %.12e\n",
                   rfl.reflectivity() );
  fprintf( stderr, "rl_MultilayerSurface_test01:  INTENSITY:     %.12e\n",
                   cs.intensity() );
#endif

  rc1 = !Test_fuzzy_compare(rfl.reflectivity(), cs.intensity(), TOL);
  rc &= rc1;

  return !rc;
}

static
int rl_MultilayerSurface_test02( rl_DielectricPODArray* diel0,
                                 rl_DielectricPODArray* diel1
				 )
                                 /* ch2 overlayer + Ir substrate */
{
  rl_DielectricLayer diel[2];

  diel[0].init( diel0->const_data_ptr(), diel0->num_elts(),
                22.0, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "ol" );
  diel[1].init( diel1->const_data_ptr(), diel1->num_elts(),
                1e30, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Ir" );

  rl_Multilayer ml( 2, diel, rl_Traits::False );

  dvm3_Vector   norm( cos( GRAZEANGLE ), 0.0, sin( GRAZEANGLE ) );
 
  rl_MultilayerSurface surf( ml, norm );
  dvm3_Vector   pos( 100.0, 0.0, 0.0 );
  dvm3_Vector   dir(   0.0, 0.0, 1.0 );
  rl_BasicRay   bray( pos, dir, 1.49, 12 );
  rl_PolCSPOD   cs;
  cs.init();
  rl_Ray        ray(  bray, cs );

  surf.reflect( ray );

  rl_ReflectionCoefPOD rfl = surf.reflection_coefs();

  int rc = 1;
  int rc1;

  rc1 = !Test_fuzzy_compare(rfl.para().real(), -8.815071158052333e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.para().imag(),  3.517246540337252e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.perp().real(), -8.813154338444567e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.perp().imag(),  3.525061583151267e-01, TOL);
  rc &= rc1;

  rc1 = !Test_fuzzy_compare(rfl.reflectivity(), 9.008712565264683e-01, TOL);
  rc &= rc1;

  ray.get_PolCSPOD( cs );

#if defined( DEBUG_MULTILAYERSURFACE )
  fprintf( stderr, "rl_MultilayerSurface_test02:  REFLECTIVITY:  %.12e\n",
                 rfl.reflectivity() );
  fprintf( stderr, "rl_MultilayerSurface_test02:  INTENSITY:     %.12e\n",
                 cs.intensity() );
#endif

  rc1 = !Test_fuzzy_compare(rfl.reflectivity(), cs.intensity(), TOL);
  rc &= rc1;

  return !rc;
}

static
int rl_MultilayerSurface_test03( rl_DielectricPODArray* diel0,
                                 rl_DielectricPODArray* diel1,
                                 rl_DielectricPODArray* diel2,
                                 rl_DielectricPODArray* diel3
                                 )
                                 /* ch2 + Ir + Cr + Zerodur */
{
  rl_DielectricLayer diel[4];

  diel[0].init( diel0->const_data_ptr(), diel0->num_elts(),
                22.0, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "ol" );
  diel[1].init( diel1->const_data_ptr(), diel1->num_elts(),
                327.8525, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Ir" );
  diel[2].init( diel2->const_data_ptr(), diel2->num_elts(),
                97.22625, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Cr" );
  diel[3].init( diel3->const_data_ptr(), diel3->num_elts(),
                1.e30, 0.0, rl_Traits::ERoughNone,
                rl_Traits::ELinLin, 1.0, "Zerodur", rl_Traits::True );

  rl_Multilayer ml( 4, diel, rl_Traits::False );

  dvm3_Vector   norm( cos( GRAZEANGLE ), 0.0, sin( GRAZEANGLE ) );

  rl_MultilayerSurface surf( ml, norm );
  dvm3_Vector   pos( 100.0, 0.0, 0.0 );
  dvm3_Vector   dir(   0.0, 0.0, 1.0 );
  rl_BasicRay   bray( pos, dir, 1.50, 12 );
  rl_PolCSPOD   cs;
  cs.init();
  rl_Ray        ray(  bray, cs );

  surf.reflect( ray );

  rl_ReflectionCoefPOD rfl = surf.reflection_coefs();

  int rc = 1;
  int rc1;

  rc1 = !Test_fuzzy_compare(rfl.para().real(), -8.815071158052333e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.para().imag(),  3.517246540337252e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.perp().real(), -8.813154338444567e-01, TOL);
  rc &= rc1;
  rc1 = !Test_fuzzy_compare(rfl.perp().imag(),  3.525061583151267e-01, TOL);
  rc &= rc1;

  rc1 = !Test_fuzzy_compare(rfl.reflectivity(), 9.008712565264683e-01, TOL);
  rc &= rc1;

  ray.get_PolCSPOD( cs );

#if defined( DEBUG_MULTILAYERSURFACE )
  fprintf( stderr, "rl_MultilayerSurface_test03:  REFLECTIVITY:  %.12e\n",
                   rfl.reflectivity() );
  fprintf( stderr, "rl_MultilayerSurface_test03:  INTENSITY:     %.12e\n",
                   cs.intensity() );
#endif

  rc1 = !Test_fuzzy_compare(rfl.reflectivity(), cs.intensity(), TOL);
  rc &= rc1;

  return !rc;
}


int
main( int argc, char* argv[] )
{
  if ( argc != 2 )
  {
    cerr << "Usage: " << argv[0] << " srcdir\n";
    return EXIT_FAILURE;
  }
  char f0[]   = "/tests/inputs/ch2_stripped.rdb";
  char f1[]   = "/tests/inputs/all_dbs_fit_i_stripped.rdb";
  char f2[]   = "/tests/inputs/Cr_Gullikson95_stripped.rdb";
  char f3[]   = "/tests/inputs/Zerodur_Gullikson95_stripped.rdb";

  int srclen = strlen((++argv)[0]);

  char* src0 = new char[static_cast<size_t>(srclen + sizeof(f0) + 1)];
  char* src1 = new char[static_cast<size_t>(srclen + sizeof(f1) + 1)];
  char* src2 = new char[static_cast<size_t>(srclen + sizeof(f2) + 1)];
  char* src3 = new char[static_cast<size_t>(srclen + sizeof(f3) + 1)];

  strncpy( src0, argv[0], srclen+1 );
  strncpy( src1, argv[0], srclen+1 );
  strncpy( src2, argv[0], srclen+1 );
  strncpy( src3, argv[0], srclen+1 );

  strncat( src0, f0, sizeof(f0)+1 );
  strncat( src1, f1, sizeof(f1)+1 );
  strncat( src2, f2, sizeof(f2)+1 );
  strncat( src3, f3, sizeof(f3)+1 );

  rl_DielectricPODArray* diel0 = new_dielectric_pod_array(src0);
  rl_DielectricPODArray* diel1 = new_dielectric_pod_array(src1);
  rl_DielectricPODArray* diel2 = new_dielectric_pod_array(src2);
  rl_DielectricPODArray* diel3 = new_dielectric_pod_array(src3);

  int rc  = 1;
  int rc1;

  START("Testing the rl_MultilayerSurface class");

  rc1 = rl_MultilayerSurface_test01(diel1);
  TEST( "Ir Substrate",  rc1, 1 );
  rc &= rc1;

  rc1 = rl_MultilayerSurface_test02(diel0, diel1);
  TEST( "ch2 overlayer on Ir substrate",  rc1, 1 );
  rc &= rc1;

  rc1 = rl_MultilayerSurface_test03(diel0, diel1, diel2, diel3);
  TEST( "ch2 overlayer on Ir, Cr, Zerodur substrate",  rc1, 1 );
  rc &= rc1;

  SUMMARY();

  delete    diel0;
  delete    diel1;
  delete    diel2;
  delete    diel3;

  delete [] src0;
  delete [] src1;
  delete [] src2;
  delete [] src3;

  return !rc;
}
