// File;   t_rl_Polarization.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of rl_raylib
//
// rl_raylib is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// rl_raylib is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <rl_raylib/rl_Polarization.h>
#include <rl_raylib/rl_ReflectionCoefPOD.h>

#include <mathconst/mathconst.h>
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <testlib/testlib.h>

using namespace std;

const double TOL = 500.0 * DBL_EPSILON;

static
int rl_Polarization_test01()    /* test ctor, dtor, intensity */
{
  rl_PolCSPOD foo;
  foo.init();
 
  return !Test_fuzzy_compare(foo.c2_[0].real(), 7.071067811865e-01, TOL)  &&
         !Test_fuzzy_compare(foo.c2_[0].imag(), 0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.c2_[1].real(), 0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.c2_[1].imag(), 0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.s2_[0].real(), 0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.s2_[0].imag(), 0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.s2_[1].real(), 7.071067811865e-01, TOL)  &&
         !Test_fuzzy_compare(foo.s2_[1].imag(), 0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.intensity(),   1.000000000000e+00, TOL);
}

static
int rl_Polarization_test02()    /* test ctor arguments */
{
  rl_PolCSPOD foo;
  foo.init( 0.5, 30 );  /* b_over_a, angle */

  return !Test_fuzzy_compare(foo.c2_[0].real(),  7.745966692415e-01, TOL)  &&
         !Test_fuzzy_compare(foo.c2_[0].imag(),  0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.c2_[1].real(),  4.472135955000e-01, TOL)  &&
         !Test_fuzzy_compare(foo.c2_[1].imag(),  0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.s2_[0].real(), -2.236067977500e-01, TOL)  &&
         !Test_fuzzy_compare(foo.s2_[0].imag(),  0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.s2_[1].real(),  3.872983346207e-01, TOL)  &&
         !Test_fuzzy_compare(foo.s2_[1].imag(),  0.000000000000e+00, TOL)  &&
         !Test_fuzzy_compare(foo.intensity(),    1.000000000000e+00, TOL);
}

double safe_div( const double x, const double y ) {

  /*
    return x / y;
  */
  return (
    ( y < 1 && x > y * DBL_MAX )
    ?
    DBL_MAX 
    :
    (
      ( y > 1 && x < y * DBL_MIN || x == 0 )
      ?
      0 
      :
      x / y
      )
    );
}

static
int rl_Polarization_test03()    /* test attenuate */
{
  rl_PolCSPOD foo;
  foo.init( 0.5, 30 );
  foo.attenuate( 0.5 );

  int rc  = 0;

  rc  = !Test_fuzzy_compare(foo.c2_[0].real(),  5.477225575051662e-01, TOL)
     && !Test_fuzzy_compare(foo.c2_[0].imag(),  0.000000000000000e+00, TOL)
     && !Test_fuzzy_compare(foo.c2_[1].real(),  3.162277660168379e-01, TOL)
     && !Test_fuzzy_compare(foo.c2_[1].imag(),  0.000000000000000e+00, TOL)
     && !Test_fuzzy_compare(foo.s2_[0].real(), -1.581138830084189e-01, TOL)
     && !Test_fuzzy_compare(foo.s2_[0].imag(),  0.000000000000000e+00, TOL)
     && !Test_fuzzy_compare(foo.s2_[1].real(),  2.738612787525831e-01, TOL)
     && !Test_fuzzy_compare(foo.s2_[1].imag(),  0.000000000000000e+00, TOL)
     && !Test_fuzzy_compare(foo.intensity(),    5.000000000000001e-01, TOL);

  return rc;
}

static
int rl_Polarization_test04()    /* */
{
  rl_PolCSPOD foo;
  foo.init();
  double psi = 1.0;
  dvm3_Vector dir( sin(psi * M_DEG2RAD), 0.0, cos(psi * M_DEG2RAD) );
  rl_Polarization pol( foo, dir );

  int rc = !Test_fuzzy_compare(foo.c2_[0].real(),  7.071067811865475e-01, TOL)
        && !Test_fuzzy_compare(foo.c2_[0].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.c2_[1].real(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.c2_[1].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[0].real(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[0].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[1].real(),  7.071067811865475e-01, TOL)
        && !Test_fuzzy_compare(foo.s2_[1].imag(),  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(pol.C_r()[0],  7.069990853988242e-01, TOL)
        && !Test_fuzzy_compare(pol.C_r()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_r()[2], -1.234071493982692e-02, TOL)
        && !Test_fuzzy_compare(pol.C_i()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_i()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_i()[2],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_r()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_r()[1],  7.071067811865475e-01, TOL)
        && !Test_fuzzy_compare(pol.S_r()[2],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[2],  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(dir[0],  1.745240643728351e-02, TOL)
        && !Test_fuzzy_compare(dir[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(dir[2],  9.998476951563913e-01, TOL);

  rc    &= !Test_fuzzy_compare(foo.intensity(),    1.000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.intensity(),    1.000000000000e+00, TOL);

  return rc;
}

static
int rl_Polarization_test05()    /* */
{
  rl_PolCSPOD foo;
  foo.init( 0.5, 30 );
  double psi = 1.0;
  dvm3_Vector dir( sin(psi * M_DEG2RAD), 0.0, cos(psi * M_DEG2RAD) );
  rl_Polarization pol( foo, dir );

  int rc = !Test_fuzzy_compare(foo.c2_[0].real(),  7.745966692414834e-01, TOL)
        && !Test_fuzzy_compare(foo.c2_[0].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.c2_[1].real(),  4.472135954999579e-01, TOL)
        && !Test_fuzzy_compare(foo.c2_[1].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[0].real(), -2.236067977499789e-01, TOL)
        && !Test_fuzzy_compare(foo.s2_[0].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[1].real(),  3.872983346207417e-01, TOL)
        && !Test_fuzzy_compare(foo.s2_[1].imag(),  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(pol.C_r()[0],  7.744786944169147e-01, TOL)
        && !Test_fuzzy_compare(pol.C_r()[1],  4.472135954999579e-01, TOL)
        && !Test_fuzzy_compare(pol.C_r()[2], -1.351857589656843e-02, TOL)
        && !Test_fuzzy_compare(pol.C_i()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_i()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_i()[2],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_r()[0], -2.235727413516178e-01, TOL)
        && !Test_fuzzy_compare(pol.S_r()[1],  3.872983346207417e-01, TOL)
        && !Test_fuzzy_compare(pol.S_r()[2],  3.902476716472085e-03, TOL)
        && !Test_fuzzy_compare(pol.S_i()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[2],  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(dir[0],  1.745240643728351e-02, TOL)
        && !Test_fuzzy_compare(dir[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(dir[2],  9.998476951563913e-01, TOL);

  rc    &= !Test_fuzzy_compare(foo.intensity(), 1.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.intensity(), 1.000000000000000e+00, TOL); 

  return rc;
}

static
int rl_Polarization_test06()    /* */
{
  rl_PolCSPOD foo;
  foo.init( 0.5, 30 );
  double psi = 1.0;
  dvm3_Vector dir( sin(psi * M_DEG2RAD), 0.0, cos(psi * M_DEG2RAD) );
  rl_Polarization pol( foo, dir );
  pol.attenuate( 0.5 );

  int rc = !Test_fuzzy_compare(foo.c2_[0].real(),  7.745966692414834e-01, TOL)
        && !Test_fuzzy_compare(foo.c2_[0].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.c2_[1].real(),  4.472135954999579e-01, TOL)
        && !Test_fuzzy_compare(foo.c2_[1].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[0].real(), -2.236067977499789e-01, TOL)
        && !Test_fuzzy_compare(foo.s2_[0].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[1].real(),  3.872983346207417e-01, TOL)
        && !Test_fuzzy_compare(foo.s2_[1].imag(),  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(pol.C_r()[0],  5.476391367067044e-01, TOL)
        && !Test_fuzzy_compare(pol.C_r()[1],  3.162277660168379e-01, TOL)
        && !Test_fuzzy_compare(pol.C_r()[2], -9.559076688448550e-03, TOL)
        && !Test_fuzzy_compare(pol.C_i()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_i()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_i()[2],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_r()[0], -1.580898014981950e-01, TOL)
        && !Test_fuzzy_compare(pol.S_r()[1],  2.738612787525831e-01, TOL)
        && !Test_fuzzy_compare(pol.S_r()[2],  2.759467749640023e-03, TOL)
        && !Test_fuzzy_compare(pol.S_i()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[2],  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(dir[0],  1.745240643728351e-02, TOL)
        && !Test_fuzzy_compare(dir[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(dir[2],  9.998476951563913e-01, TOL);

  rc    &= !Test_fuzzy_compare(foo.intensity(),    1.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.intensity(),    5.000000000000000e-01, TOL);

  return rc;
}

static
int rl_Polarization_test07()    /* */
{
  rl_PolCSPOD foo;
  foo.init( 0.5, 30 );
  foo.attenuate( 0.5 );
  double psi = 1.0;
  dvm3_Vector dir( sin(psi * M_DEG2RAD), 0.0, cos(psi * M_DEG2RAD) );
  rl_Polarization pol( foo, dir );

  int rc = !Test_fuzzy_compare(foo.c2_[0].real(),  5.477225575051662e-01, TOL)
        && !Test_fuzzy_compare(foo.c2_[0].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.c2_[1].real(),  3.162277660168379e-01, TOL)
        && !Test_fuzzy_compare(foo.c2_[1].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[0].real(), -1.581138830084189e-01, TOL)
        && !Test_fuzzy_compare(foo.s2_[0].imag(),  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(foo.s2_[1].real(),  2.738612787525831e-01, TOL)
        && !Test_fuzzy_compare(foo.s2_[1].imag(),  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(pol.C_r()[0],  5.476391367067044e-01, TOL)
        && !Test_fuzzy_compare(pol.C_r()[1],  3.162277660168379e-01, TOL)
        && !Test_fuzzy_compare(pol.C_r()[2], -9.559076688448551e-03, TOL)
        && !Test_fuzzy_compare(pol.C_i()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_i()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.C_i()[2],  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(pol.S_r()[0], -1.580898014981950e-01, TOL)
        && !Test_fuzzy_compare(pol.S_r()[1],  2.738612787525831e-01, TOL)
        && !Test_fuzzy_compare(pol.S_r()[2],  2.759467749640023e-03, TOL)
        && !Test_fuzzy_compare(pol.S_i()[0],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[1],  0.000000000000000e+00, TOL)
        && !Test_fuzzy_compare(pol.S_i()[2],  0.000000000000000e+00, TOL);

  rc    &= !Test_fuzzy_compare(dir[0],  1.745240643728351e-02, TOL);
  rc    &= !Test_fuzzy_compare(dir[1],  0.000000000000000e+00, TOL);
  rc    &= !Test_fuzzy_compare(dir[2],  9.998476951563913e-01, TOL);

  rc    &= !Test_fuzzy_compare(foo.intensity(),    5.000000000000000e-01, TOL)
        && !Test_fuzzy_compare(pol.intensity(),    5.000000000000000e-01, TOL);

  return rc;
}

int
main() 
{
  int rc  = 1;
  int rc1;

  START("Testing the rl_Polarization class");

  rc1 = rl_Polarization_test01();
  TEST("rl_PolCSPOD: ctor, dtor, intensity", rc1, 1);
  rc &= rc1;

  rc1 = rl_Polarization_test02();
  TEST("rl_PolCSPOD: ctor arguments",        rc1, 1);
  rc &= rc1;

  rc1 = rl_Polarization_test03();
  TEST("rl_PolCSPOD: attenuate",             rc1, 1);
  rc &= rc1;

  rc1 = rl_Polarization_test04();
  TEST("rl_Polarization: ctor, dtor",        rc1, 1);
  rc &= rc1;

  rc1 = rl_Polarization_test05();
  TEST("attenuate1",                         rc1, 1);
  rc &= rc1;

  rc1 = rl_Polarization_test06();
  TEST("attenuate2",                         rc1, 1);
  rc &= rc1;

  rc1 = rl_Polarization_test07();
  TEST("attenuate3",                         rc1, 1);
  rc &= rc1;

  SUMMARY();

  return !rc;
}
