/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <float.h>
#include <math.h>

#include "imagefcts/imagefcts.h"

/*

  PROCEDURE
  weightpos - determine the mean weighted position of a group of objects

  DESCRIPTION
  determine the mean weighted position of a group of objects.
  requires a user supplied function which extracts an object's
  position and weight.  designed to be run over sub-groups of the list
  (for low memory situations).  optimized for both weighted and
  unweighted (weight = 1) data.

  RETURNS
  It returns the mean weighted position of the group.

  AUTHOR
  Diab Jerius

*/
double
weightpos(
  void *objs,			/*  list of objects whose mean
				    position is to be determined */
  unsigned long n_tot,		/*  total number of objects to process */
  double tot_wt,		/*  total weight of objects. if zero,
				    each object is assumed to have a weight of 1. */
  size_t s_obj,			/*  size of an object */
  double (*get_x)(const void *obj, double *x) /*  function which
						  returns the position
						  and weight of an
						  object */
  )

{
  double wt_ave = 0.0;
  double x = 0.0;

  char *obj = objs;
  char *obj_lim = obj + n_tot * s_obj;

  if (tot_wt)
  {
    for (; obj < obj_lim ; obj += s_obj)
    {
      double xo;
      double w = get_x(obj, &xo);

      xo *= w;
      /* prevent overflow */
      if ( DBL_MAX - fabs(x) < fabs((double) xo) )
      {
        wt_ave += x / tot_wt;
        x = 0.0;
      }
      x += xo;
    }

    wt_ave += x / tot_wt;
  }
  else
  {
    for (; obj < obj_lim ; obj += s_obj)
    {
      double xo;

      get_x(obj, &xo);

      /* prevent overflow */
      if ( DBL_MAX - fabs(x) < fabs((double) xo) )
      {
        wt_ave += x / n_tot;
        x = 0.0;
      }
      x += xo;
    }

    wt_ave += x / n_tot;
  }

  return wt_ave;
}
