/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "lists/lists.h"

/*
  PROCEDURE
  partition - partition a list about an object

  DESCRIPTION
  partitions the list @code{obj[0 ... n]} into two sub-lists
  @code{obj[0 ... q]} and @code{obj[q+1 ... n]}, such that
  @code{obj[i] (0 <= i <= q) <= pobj <= obj[j] (q < j <= n)}
  where @code{pobj} is a supplied "pivot" point.
  returns the top object in the first sub-list, i.e. @code{obj[q]}.
  @code{pobj} need not be in the list.  in this case, if no object is less
  than @code{pobj}, @code{NULL} is returned.

  REFERENCES
  See "Introduction to Algorithms", T.H.Kormen, C.E.Leiserson, and
  R.L. Rivest sec. 8.1, p.154.

  AUTHOR
  Diab Jerius

*/

void *
partition(
  void *obj, 		/* list of objects to partition */
  void *p_obj, 		/* object about which to partition list.  need not be
			   in the list */
  void *work,		/* work area with size 2 * @code{s_obj} */
  size_t s_obj, 	/* size of an object in bytes */
  unsigned long n_obj, /* number of objects in list */
  int (*obj_comp)(const void *obj1, const void *obj2) /* function which
							 compares two objects and returns < 1,
							 0, > 1,
							 depending if
							 the first
							 object
							 is less than, equal to, or
							 greater 
							 than the
							 second */
  )
{
  char *lt_top, *gt_bot, *obj_top;
  char *x = (char *) work;
  char *t = (char *) work + 2 * s_obj;

  /* save pivot object */
  memcpy(x, p_obj, s_obj);

  /* top of $<$ region */
  lt_top = obj;

  /* bottom of $>$ region */
  obj_top = gt_bot = (char *) obj + (n_obj - 1) * s_obj;

  for (;;)
  {
    /* move downwards in $>$ region, looking for an object that's $\leq$  */
    while ( obj_comp(gt_bot, x) > 0 && gt_bot > (char *) obj ) gt_bot -= s_obj;

    /* move upwards in $<$ region, looking for an object that's $\geq$ */
    while ( obj_comp(lt_top, x) < 0 && lt_top < obj_top ) lt_top += s_obj;

    /* if two regions haven't merged, there are two out of place objects that should
       be swapped */
    if (lt_top < gt_bot)
    {
      memcpy(t, lt_top, s_obj);
      memcpy(lt_top, gt_bot, s_obj);
      memcpy(gt_bot, t, s_obj);
      gt_bot -= s_obj;
      lt_top += s_obj;
    }
    else        /* all done, clean up and return boundary object */
    {
      if (( void *) gt_bot == obj && obj_comp(gt_bot, x) > 0) gt_bot = NULL;
      return gt_bot;
    }
  }

  /* NOT REACHED */
}
