/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

#include "str/str.h"

#include "range/range.h"

#define DELIM " ,\t"

#define WHERE(ptr) (ptr - range_str)
#define ERROR(val) do { error = val; goto error; /* NOT REACHED */} while(0)

static int 		rangel_start_comp(const void *r1, const void *r2);

/*

  PROCEDURE
  rangel_parse - parse a list of integral ranges

  DESCRIPTION
  @code{rangel_parse} translates a list of ranges of the form
  @example
  r1[@{,| @}r2[@{,| @}r3]]...
  @end example
  into start-end pairs. The handling of the
  ranges depends upon the flags set in the @var{opts} arguement.
  @var{opts} is set to the logical OR of zero or more of the following values

  @table @code

  @vindex Range_SORT
  @item Range_SORT
  The resultant ranges will be sorted in ascending order by their lower
  bound

  @vindex Range_MERGE
  @item Range_MERGE
  Adjacent ranges will be merged, if possible.
  This implies @code{Range_SORT}.

  @vindex Range_INCOMPLETE
  @item Range_INCOMPLETE
  If set, incomplete ranges, i.e., those without an explicit start or end
  value, are accepted.  In this case, implied starts are set to the
  value of the passed @var{minval} argument, while implied ends are set to
  the value of the @var{maxval} argument.

  @vindex Range_UNSIGNED
  @item Range_UNSIGNED
  Only unsigned integers are allowed in the ranges.

  @end table

  A range has the format:

  @example
  [start]@{`:'@}[end]
  [start]@{`/'@}count
  @end example

  where @var{start} and @var{end} are optional limits on the range, and @var{count}
  specifies the number of elements in a range. 
  In the first form, if
  @var{start} or @var{end} is missing, @var{minval} and @var{maxval} are used appropriately
  (if the incomplete range flag is set). In the second form, if @var{start} is
  missing, the range starts at 1 if no previous range was specified,
  or directly follows the end of the previous range.

  @code{rangel_parse} creates a @code{RangeLList} object and returns it
  via the @var{rll} parameter.

  RETURNS
  It returns a code indicating whether an error ocurred. See
  @file{suplib/range.h} for the possible errors and codes.  It also
  returns the location of the error in the range specification via
  the passed @code{where} variable.

  AUTHOR
  Diab Jerius

*/


RangeErr
rangel_parse(
  RangeLList ** rll,	/* returned range list				*/
  char *range_spec,	/* string containing range to parse		*/
  RangeOpts opts,	/* range options				*/
  long minval,		/* minimum value for incomplete start		*/
  long maxval,		/* maximum value for incomplete end		*/
  long *where		/* position in string where error occurred	*/
)
{
  char *ptr, *tok, *range_str;
  size_t nr, lastpos;
  RangeLList *rl;
  RangeL *range;
  RangeErr error = RangeErr_OK;


  *rll = NULL;
  *where = 0;

  /* determine number of ranges */
  if ((nr = str_tokcnt(range_spec, DELIM)) == 0)
    return RangeErr_OK;

  if ( NULL == ( range_str = str_dup(range_spec) ) )
    return RangeErr_NOMEM;
  
  if ( NULL == (rl = rangel_new( nr )) )
  {
    free(range_str);
    return RangeErr_NOMEM;
  }

  lastpos = 0;
  range = rl->range;
  for (tok = strtok(range_str, DELIM) ;
       tok ;
       tok = strtok(NULL, DELIM), lastpos=range++->end)
  {
    size_t p, op, start = 0, end = 0 ;

    /* split token into start; op; end; */
    p = strcspn(tok, "/:");
    op = tok[p];

    /*
       determine if start or end limits have been specified.  if so, set flags to
       that effect; if not, preset limits to minval and maxval
    */

    if (p > 0)
      start = 1;
    else
      range->start = minval;

    if (p < strlen(tok) - 1)
      end = 1;
    else
      range->end = maxval;

    /*
       check that incomplete limits are allowed. the `/' operator is
       an exception: it will always take an incomplete start limit and
       will never take an incomplete end limit.
    */
    if (
        (op == ':' && ! (opts & Range_INCOMPLETE) && (!start || !end)) ||
        (op == '/' && !end)
       )
      ERROR( RangeErr_INCOMPLETE );

    /* split token into start and end limits */
    tok[p] = '\0';

    /* if start limit is present, get its value */
    if (start)
    {
      errno = 0;
      range->start = strtol(tok, &ptr, 10);

      if (errno)
	ERROR( RangeErr_ERANGE );

      if (*ptr != '\0')
	ERROR( RangeErr_ILLNUM );

      if ( opts & Range_UNSIGNED && range->start < 0)
	ERROR( RangeErr_NEGNUM );
    }

    /* if end limit is present, get its value */
    tok += p + 1;
    if (end)
    {
      errno = 0;
      range->end = strtol(tok, &ptr, 10);

      if (errno == ERANGE)
	ERROR( RangeErr_ERANGE );

      if (*ptr)		/* didn't find any number */
	ERROR( RangeErr_ILLNUM );

      if (opts & Range_UNSIGNED && range->end < 0)
	ERROR( RangeErr_NEGNUM );
    }

    /*
       determine range based upon range operand.  note that range
       operand `:' has already been implicitly taken care of
    */
    switch (op)
    {
      case '\0':	/* should be a single number */
        range->end = range->start;
	break;

      case '/':
        /* if no start, use end of last range as starting point */
        if ( !start )
        {
          if (lastpos == LONG_MAX)
	    ERROR( RangeErr_OFLOWSTART );
          else
            range->start = lastpos + 1;
        }
        if ( end )
        {
          if (range->end < 1)
	    ERROR( RangeErr_NONPOSCOUNT );
          else
          {
            --range->end;
            if (range->start > LONG_MAX - range->end)
	      ERROR( RangeErr_OFLOWEND );
            else
              range->end += range->start;
          }
        }
        break;

      case ':':
        if (opts & Range_UNSIGNED && range->start < 0)
	  ERROR( RangeErr_NEGNUM );
        break;

      default:
	ERROR( RangeErr_INTERNAL );
    }

    /* out of order limits are an error */
    if ( range->end < range->start )
      ERROR( RangeErr_ORDER );
  }

  /* sort ranges by starting value */
  if ( opts & Range_SORT || opts & Range_MERGE )
  {
    qsort(rl->range, nr, sizeof(RangeL), rangel_start_comp);
    if ( opts & Range_MERGE )
    {
      RangeL *r0 = rl->range,
             *r1 = rl->range + 1;

      for (; r1 < rl->range + nr ; r1++)
      {
        if (r1->start - r0->end > 1)
	{
	  r0++;
	  if ( r0 != r1 )
	    *r0 = *r1;
	}
        else
        {
          /* if the second range is longer, copy to first range */
          if (r1->end > r0->end)
            r0->end = r1->end;
        }
      }

      nr = (r0 - rl->range) + 1;
      if ((r0 = (RangeL *) realloc(rl->range, nr * sizeof(RangeL))) == NULL)
	ERROR( RangeErr_NOMEM );
      rl->range = r0;
    }
  }

  rl->n = nr;

  free( range_str );
  *rll = rl;
  return RangeErr_OK;

error:
  rangel_del( rl );
  free( range_str );
  *where = WHERE(tok);
  return error;
}


/*

  PROCEDURE
  rangel_new - create a new @code{RangeLList}

  DESCRIPTION
  @code{RangeLList} objects are used to contain a list of integral ranges.
  @code{rangel_new} creates such an object for a given number of ranges.
  
  RETURNS
  It returns a pointer to a @code{RangeLList} object upon success,
  @code{NULL} if it ran out of memory.

  AUTHOR
  Diab Jerius
 */
RangeLList *
rangel_new(
  size_t nr		/* the number of ranges to allocate space for */
  )
{
  RangeLList *rl = NULL;

  if ( (rl = (RangeLList *) malloc(sizeof(RangeLList))) )
  {
    if ( NULL ==
	 (rl->range = (RangeL *) malloc(nr * sizeof(RangeL))) )
    {
      free(rl);
      rl = NULL;
    }
  }

  return rl;
}

/*

  PROCEDURE
  rangel_del - destroy a @code{RangeLList} object.

  DESCRIPTION
  This routine frees all memory associated with a @code{RangeLList} object.
  
  RETURNS
  It returns @code{NULL}.

  AUTHOR
  Diab Jerius
 */
RangeLList *
rangel_del(
  RangeLList *rl		/* the @code{RangeLList} object to delete */
  )
{
  if ( rl )
  {
    if ( rl->range )
      free( rl->range );
    free( rl );
  }
  return NULL;
}


/*

  PROCEDURE
  rangel_reset - reset the @code{RangeLList} internal state

  DESCRIPTION
  The internal state of the passed @code{RangeLList} object is set so that
  the next call to @ref{rangel_next} will return the start of the first
  range.

  AUTHOR
  Diab Jerius
*/

void
rangel_reset(
  RangeLList *rl	/* the range list to reset */
  )
{
  rl->pos = NULL;
}

/*

  PROCEDURE
  rangel_next - get the next element in the range list

  DESCRIPTION
  @code{rangel_next} serves as an iterator over the passed list of ranges,
  returning the next element in the list.  To reset the iterator,
  @ref{rangel_reset}.  The element is returned via the @var{elem} argument.

  RETURNS
  It returns @samp{1} if there is an element left, @samp{0} if it has
  reached the end of the range list.

  AUTHOR
  Diab Jerius

*/

int
rangel_next(
  RangeLList *rl,		/* the range to iterate over */
  long *elem			/* the returned element */
  )
{
  if (rl->pos == NULL)
  {
    if (rl->n > 0)
    {
      rl->pos = &rl->range[0];
      *elem = rl->val = rl->pos->start;
      return 1;
    }
    else
    {
      return 0;
    }
  }

  if (rl->val < rl->pos->end)
  {
    *elem = ++rl->val;
    return 1;
  }

  if (++rl->pos < rl->range + rl->n)
  {
    *elem = rl->val = rl->pos->start;
    return 1;
  }

  return 0;
}

/*

  PROCEDURE
  rangel_minmax - get the minimum and maximum values in a range list

  DESCRIPTION

  @code{rangel_minmax} determines the minimum and maximum values in a
  range list, storing them in the passed @code{min} and @code{max}
  arguments.

  RETURNS
  It returns @code{1} upon success, @code{0} if the range was empty.

  AUTHOR
  Diab Jerius

*/
int
rangel_minmax(
  RangeLList *rl,		/* the range in question */
  long *min,			/* the minimum range value */
  long *max			/* the maximum range value */
  )
{
  RangeL* range = rl->range;
  RangeL* range_end = rl->range + rl->n;

  if ( 0 == rl->n )
    return 0;

  *min = range->start;
  *max = range->end;

  for ( range++ ; range < range_end ; range++ )
  {
    if ( *min > range->start ) *min = range->start;
    if ( *max < range->end   ) *max = range->end;
  }

  return 1;
}


/*

  PROCEDURE
  rangel_count - determine the number of elements in a range list

  DESCRIPTION
  @code{rangel_count} counts the number of elements in each range of
  a range list, returning the grand total.

  RETURNS
  It returns the number of elements in the range list.

  AUTHOR
  Diab Jerius

*/
unsigned long
rangel_count(
  RangeLList *rl		/* the range list to count */
  )
{
  unsigned long count = 0;
  RangeL *range, *range_lim;

  range = rl->range;
  range_lim = range + rl->n;

  for (; range < range_lim ; range++)
    count += range->end - range->start + 1;

  return count;
}


int
rangel_check(RangeLList *rl, long minval, long maxval)
{
  RangeL *range, *range_lim;

  range = rl->range;
  range_lim = range + rl->n;

  for (; range < range_lim ; range++)
    if (range->start < minval || range->end > maxval)
      return 1;

  return 0;
}

/*

  PROCEDURE
  rangel_rep_minval - replace matching range minimum values

  DESCRIPTION
  @code{rangel_rep_minval} replaces the minimum value for those
  ranges whose minimum values match the passed one with another
  value.

  AUTHOR
  Diab Jerius

*/
void
rangel_rep_minval(
  RangeLList *rl,	/* the list to scan */
  long minval,		/* the minimum value to match */
  long repval		/* the replacement minimum value */
  )
{
  RangeL *range, *range_lim;

  range = rl->range;
  range_lim = range + rl->n;
  for (; range < range_lim ; range++)
    if (range->start == minval) range->start = repval;
}

/*

  PROCEDURE
  rangel_rep_maxval - replace matching range maximum values

  DESCRIPTION
  @code{rangel_rep_maxval} replaces the maximum value for those
  ranges whose maximum values match the passed one with another
  value.

  AUTHOR
  Diab Jerius

*/
void
rangel_rep_maxval(
  RangeLList *rl,	/* the list to scan */
  long maxval,		/* the maximum value to match */
  long repval		/* the replacement maximum value */
  )
{
  RangeL *range, *range_lim;

  range = rl->range;
  range_lim = range + rl->n;
  for (; range < range_lim ; range++)
    if (range->end == maxval) range->end = repval;
}

/*

  PROCEDURE
  rangel_rep_val - replace matching range minimum and maximum values

  DESCRIPTION
  @code{rangel_rep_maxval} replaces the maximum value for those
  ranges whose maximum values match the passed one with another
  value. It also replaces the minimum value of the ranges which
  match the passed minimum value with another value.

  AUTHOR
  Diab Jerius

*/
void
rangel_rep_val(
  RangeLList *rl,	/* the list to edit */
  long minval,		/* the minimum value to match */
  long maxval,		/* the maximum value to match */
  long minrepval,	/* the replacement minimum value */
  long maxrepval	/* the replacement maximum value */
  )
{
  RangeL *range, *range_lim;

  range = rl->range;
  range_lim = range + rl->n;
  for (; range < range_lim ; range++)
  {
    if (range->start == minval) range->end = minrepval;
    if (range->end == maxval) range->end = maxrepval;
  }
}

static int
rangel_start_comp(const void *r1, const void *r2)
{
  return ((RangeL *)r1)->start > ((RangeL *)r2)->start ?  1 :
         ((RangeL *)r1)->start < ((RangeL *)r2)->start ? -1 : 0;
}

/*
  PROCEDURE
  rangel_dump - pretty print a @code{RangeLList}

  DESCRIPTION
  @code{rangel_dump} prints out a simple dump of a @code{RangeLList}
  to the passed output stream.

  AUTHOR
  Diab Jerius

*/
void
rangel_dump(
  FILE *fout,		/* where to print */
  RangeLList *rl	/* the list to print */
  )
{
  if ( rl && rl->range )
  {
    int i;

    for ( i = 0 ; i < rl->n ; i++ )
      fprintf( fout, "%ld %ld\n", rl->range[i].start, rl->range[i].end );
  }
}


#ifdef TEST
int
main( int argc, char *argv[] )
{
  RangeErr error;
  RangeLList *rl;
  long where;
  
  if ( argc < 2 )
  {
    fprintf( stdout, "usage: %s <range>\n", argv[0] );
    exit( 1 );
  }

  error = 
    rangel_parse( &rl, argv[1], Range_SORT, RANGEL_MIN, RANGEL_MAX, &where );

  if ( error )
    range_perror( stdout, error, argv[1], where );
  else
  {
    long min, max;

    if ( rangel_minmax( rl, &min, &max ) )
    {
      printf( "minimum: %ld maximum: %ld\n", min, max );
    }
    else
      printf( "empty range\n" );

    rangel_dump( stdout, rl );
  }

  rangel_del( rl );
}
#endif
