/*
 *
 *  The original routines upon which these are based are available
 *  here:
 *
 *    http://lib.stat.cmu.edu/apstat/
 *    http://www.ams.sunysb.edu/~nkfr/5PUBL.HTM
 *
 *  and were published in Applied Statistics, Vol. 43, No. 1 (1994), 265-270.
 *  "Algorithm AS 288: Exact Smirnov Two-Sample Tests
 *     for Arbitrary Distributions"
 *
 *  The http://lib.stat.cmu.edu/apstat/ code is copyrighted by the
 *  Royal Statistical Society. It may be distributed provided that no
 *  fee is charged.
 */


#include <stdlib.h>
#include <math.h>

#include "stats.h"
#include "local.h"

typedef struct
{
  double p;
  int m;
} stcalc_st;

static int cmp_stcalc_st (const void *st1, const void *st2 )
{
  return
      ((stcalc_st *) st1)->p < ((stcalc_st *) st2)->p  ? -1
    : ((stcalc_st *) st1)->p > ((stcalc_st *) st2)->p  ?  1
    :                                                     0;
}


/*

  PROCEDURE
  stcalc - calculate classical or weighted Smirnov statistic

  DESCRIPTION
  @code{stcalc} calculates statistics, number of categories
  (i.e. unique values in the sample) in the pooled
  sample and numbers of observations falling into each category
  for the two-sample Smirnov tests for arbitrary distributions.

  The following statistics are calculated.

  @enumerate

  @item
  sup | X - Y |
  
  @item
  sup ( X - Y )

  @item
  sup ( Y - X )

  @end enumerate

  RETURNS
  The number of categories is returned via the @code{k} argument.  The
  address of an integer array with @code{*k} elements holding the
  number of observations per category is returned via @code{m}.
  @code{stcalc} allocates the array; the calling routine must free it.

  The statistics are written to the @code{dstats} parameter, which
  is a caller-allocated array of minimum length 3.

  The return value of the subroutine may have the following possible values.

  @table @code

  @item 0
  no error

  @item 1
  the @code{nx} < 1 or @code{ny} < 1

  @item 2
  allocation of workspace failed

  @end table
  

  AUTHOR
  Original by Andrei M. Nikiforov
  C transcription by Diab Jerius

  COPYRIGHT
  This code is copyrighted by the Royal Statistical Society. It may be
  distributed provided that no fee is charged.

*/

int stcalc(
  int icw,	/* classical (1) or weighted (2) Smirnov statistic */
  int nx,	/* The number of observations in the first sample */
  int ny,	/* The number of observations in the second sample */
  double *x,	/* input: The first sample */
  double *y,	/* input: The second sample */
  int *k,	/* output: The number of categories in the pooled sample */
  int **m,	/* output: The number of observations in each
		   category. */
  double *dstats /* output: the calculated statistics */
  )
{
  /* Local variables */
  int ii, l, n, it;
  double zn, zp, slope;
  int retval = 0;
  int ki;

  stcalc_st *pm = (stcalc_st *) malloc( (nx + ny) * sizeof(stcalc_st) );
  if ( NULL == pm )
    return 2;




  /* Parameter adjustments */
  --dstats;
  --pm;
  --y;
  --x;

  /* Function Body */
  if (nx <= 0 || ny <= 0) 
    return 1;

  if ( NULL == m )
    return 3;

  n = nx + ny;

  for (l = 1; l <= nx; l++)
  {
    pm[l].p = x[l];
    pm[l].m = 1;
  }

  for (l = nx + 1; l <= n; l++)
  {
    pm[l].p = y[l - nx];
    pm[l].m = 2;
  }

  /*  Obtain ordered pooled sample P along with according */
  /*  ordering within the sample origin array M */

  qsort(pm+1, n, sizeof(*pm), cmp_stcalc_st );

  /*  Calculate the statistics and the number of observations falling */
  /*  into each category (in ascending order of category values) */

  slope = (double) ny / nx + 1;
  zn = 0.;
  zp = 0.;

  ii = 0;
  ki = 1;
  it = 0;

  for (l = 1; l <= n - 1 ; l++ )
  {
    if (pm[l].m == 1) 
      ii++;

    /*  Compare differencies between DF's (and define M) */
    /*  only when all points with current value are accumulated */

    if (pm[l].p < pm[l + 1].p)
    {
      double delta = ii * slope - l;
      if (icw == 2) {
	double weight = (double) l / n;
	delta /= sqrt( weight * (1 - weight));
      }
      if ( delta < zn ) zn = delta;
      if ( delta > zp ) zp = delta;

      pm[ki].m = l - it;
      it = l;
      ki++;
    }

  }

  dstats[3] = -zn / ny;
  dstats[2] =  zp / ny;
  dstats[1] = MAX(dstats[2],dstats[3]);
  pm[ki].m = n - it;

  *m = (int *) malloc( ki * sizeof(int) );
  if ( NULL == *m )
    retval = 2;

  else
  {
    int *ptr = *m - 1;
    for ( ii = 1 ; ii <= ki ; ii++ )
      ptr[ii] = pm[ii].m;
  }

  free( ++pm );

  *k = ki;

  return retval;
}

