/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <ctype.h>

#include <errno.h>

#include "str/str.h"




/*
  PROCEDURE
  str_varscan - scan a string for $VAR or $@{VAR@}

  DESCRIPTION
  @code{str_varscan} scans a strings for substrings of the form
  @code{$@{VAR@}} or @code{$VAR}.  @code{VAR}'s first
  character must be in the set @code{[a-zA-Z_]}, with the remaining
  characters being in the set @code{[a-zA-Z0-9_]}.

  When such a string is found, the passed callback function is called
  with the addresses of the first and last characters of the entire
  substring, as well as the first and last characters in @code{VAR}.

  RETURNS
  Upon success, it returns @code{0};

  Upon error, *errptr is set to point the the character that evoked
  the error, and one of the following is returned:

  @table @code

  @item EDOM
  There was an unbalanced @samp{@{}

  @end table

  AUTHOR
  Diab Jerius
  
 */
int
str_varscan(
  const char *string,	/* the string to process */
  void (*callback)(const char*, const char*, const char*, const char*, void*),	/* callback */
  void *udata,		/* extra data to pass to the callback routine */
  char **errptr		/* place to store a pointer to a bad spot in
			   the string */
  )
{
  const char *str = string;
  const char *bracket = NULL;
  int err = 0;

  for ( ; *str ; str++ )
  {
    /* these are the start and end of $VAR or ${VAR} and include
       the $ and optional }
     */
    const char* start;
    const char* end;

    /* these are the start and end of VAR */
    const char* vstart;
    const char* vend;
    
    /* if the loop is started and bracket != 0, that
       means there's an unbalanced opening bracket.
       we should do something about that */
    if ( bracket )
      break;

    /* all variables must start with $ */
    while( *str && *str != '$' )
      str++;

    /* but they might also be escaped */
    if ( str - 1 >= string && str[-1] == '\\' )
      continue;

    /* ran out of characters, bail */
    if ( *str == '\0' )
      break;
    
    /* start points at the $ */
    start = str;

    /* if the next character is a bracket, move on */
    if ( *++str == '{' )
    {
      bracket = str;
      str++;
    }

    /* variables start with [a-zA-Z_] */
    if ( ! ( isalpha( *str ) || *str == '_' ) )
      continue;

    /* vstart points at the first character of VAR */
    vstart = str;

    /* we're in! look for end of variable */
    while( *str && ( isalnum(*str) || *str == '_' ) )
      str++;

    /* vend points at the last character of VAR */
    vend = str - 1;
    
    /* substring started with a ${ */
    if ( bracket )
    {
      /* *str had better be another bracket */
      if ( '\0' == *str || *str != '}' )
      {
	/* nope.  back up str (as for loop will bump it) */
	str--;
	continue;
      }

      /* found our ket */
      bracket = NULL;

      /* the end of ${VAR} is the } */
      end = str;
    }

    /* substring started with $. */
    else
    {
      /* in this case the end of $VAR is the last char in VAR */
      end = vend;

      /* str points to the character after VAR. decrement it, as
	 the for(;;) will increment it, and that character will be
	 skipped */
      str--;
    }
    
    callback( start, end, vstart, vend, udata );
  }

  if ( bracket )
  {
    err = EDOM;
    if ( errptr ) *errptr = bracket;
  }

  return err;
}


typedef struct
{
  int nchar;
  char *buf;
} TallyData;

static void
tally_env_strlen(
  const char* start,
  const char* end,
  const char* vstart,
  const char* vend,
  void *udata
  )
{
  char *envvar;
  char c = vend[1];
  TallyData *tdata = (TallyData *) udata;

  strncpy( tdata->buf, vstart, (vend - vstart + 1) );
  tdata->buf[vend-vstart + 1] = '\0';

  if ( (envvar = getenv( tdata->buf )) )
    tdata->nchar += strlen(envvar);
}

typedef struct
{
  char* end;	/* the NULL at the end of the interpolated string buffer */
  char* next;	/* the first character after the last token in the
		   input string */

  /* if true, leave undefined variables in string */
  int   keep_undef;

  /* start and end of first undefined variable */
  char* ustart;
  char* uend;
		   
  /* work area */
  char *buf;
} InterpData;

static void
interpolate_string(
  const char *start,
  const char *end,
  const char* vstart,
  const char* vend,
  void *udata
  )
{
  InterpData *idata = (InterpData*) udata;
  size_t nchar;
  char *envvar;
  char c;

  /* first, concatenate the bits of the original text from the
     original string that lie between the end of the last token
     and the current token */

  nchar = start - idata->next;
  strncpy( idata->end, idata->next, nchar );
  idata->end += nchar;
  
  /* query value of environmental variable. copy name to
     work buffer first */
  strncpy( idata->buf, vstart, (vend - vstart + 1) );
  idata->buf[vend-vstart + 1] = '\0';

  envvar = getenv( idata->buf );

  /* if it's a real environmental variable, copy that */
  if ( envvar )
  {
    nchar = strlen(envvar);
    strncpy( idata->end, envvar, nchar );
    idata->end += nchar;
  }

  /* not an environmental variable */
  else
  {
    /* record it if its the first */
    if ( NULL == idata->ustart )
    {
      idata->ustart = vstart;
      idata->uend   = vend;
    }

    /* retain the ${VAR}/$VAR if requested */
    if ( idata->keep_undef )
    {
      nchar = end - start + 1;
      strncpy( idata->end, start, nchar );
      idata->end += nchar;
    }
  }

  /* and seal off the interpolated string */
  idata->end[0] = '\0';

  /* keep track of the character after the $... bit. that's
     the first to be copied next time around */
  idata->next = end + 1;
}



/*
  PROCEDURE
  str_interp - interpolates environmental variables into a string

  DESCRIPTION

  This routine scans a string for embedded environmental variables and
  returns a pointer to a copy of the string with the interpolated
  values.  The user is responsible for freeing the new string.

  The input string is scanned left to right for substrings of the
  form @code{$@{VAR@}} or @code{$VAR}.  @code{VAR} must begin with an
  alphabet character or the @samp{_} character.  The remaining
  characters must be alphabetic, numeric, or the @samp{_} character.


  RETURNS
  Upon success a pointer to the interpolated string is returned and
  @code{*error} is set to @code{0}.  The user is responsible for
  freeing the memory pointed to by the returned string.

  If @code{keep_undef} was set and there was an undefined variable,
  then @code{*errp} is set to @code{EFAULT} and @code{*ustart} and
  @code{*uend} point at the first and last characters in the variable
  name.  If multiple variables are undefined, they will point to the
  first undefined variable name.
  
  Upon error, @code{NULL} is returned and @code{*errp} is set to one
  of the following

  @table @code

  @item EDOM There was an unbalanced @samp{@{}.  @code{*ustart} is set
  to point the the character that evoked the error.  @code{*uend} is
  undefined.

  @item ENOMEM
  A memory allocation failed.

  @end table


  AUTHOR
  Diab Jerius
  
 */
char *str_interp(
  const char *text,		/* the text to be interpolated into */
  int  keep_undef,	/* if true, undefined variables are left
			   in the text.  if false, they are removed */
  char **ustart,	/* if not NULL, set to the start of the
			   undefined variable name */
  char **uend,		/* if not NULL, set to the end of the
			   undefined variable name */
  int *error		/* error code */
  ) {

  char* interpolated;

  TallyData tdata;
  InterpData idata;
  char* strptr;

  int err = 0;

  *error = 0;

  if ( NULL == text )
    return NULL;

  /* need buffer which will contain an environmental variable name.
     name can't be longer than passed string, but make sure buffer's 
     not overly long */
  tdata.nchar = strlen( text) + 1;
  idata.buf = tdata.buf = malloc( tdata.nchar > 1024
                                              ? 1024
					      : tdata.nchar );
  if ( NULL == tdata.buf )
  {
    error = ENOMEM;
    return NULL;
  }

  /*-----------------------------------------*/
  /* calculate length of final string */
  err = str_varscan( text, tally_env_strlen, &tdata, &strptr );

  if (err)
  {
    if ( ustart ) *ustart = strptr;
    if ( error )   *error = err;
    free( tdata.buf );
    return NULL;
  }

  interpolated = malloc( tdata.nchar );
  if ( NULL == interpolated )
  {
    if ( *error ) *error = ENOMEM;
    free( tdata.buf );
    return NULL;
  }

  /*-----------------------------------------*/
  /* generate interpolated string  */
  idata.end = interpolated;
  idata.next = text;
  idata.ustart = idata.uend = NULL;
  idata.keep_undef = keep_undef;

  /* this will generate everything up to the end of the
     last interpolated variable. no error checking required.  */
  str_varscan( text, interpolate_string, &idata,  NULL );

  /* copy the remainder of text after the last variable */
  strcpy( idata.end, idata.next );

  /* was there an undefined variable? */
  if ( keep_undef &&  NULL != idata.ustart )
  {
    if ( error  ) *error  = EFAULT;
    if ( ustart ) *ustart = idata.ustart;
    if ( uend   ) *uend   = idata.uend; 
  }


  free( tdata.buf );
  return interpolated;
}
