/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <strings.h>

#include "str/str.h"


/*
   PROCEDURE
   str_tokbq_init - initialize str_tokbq state information

   DESCRIPTION
   @code{str_tokbq_init} is called before the first call to
   @code{str_tokbq} to initialize its state information.

   RETURNS
   It returns the pased structure pointer

   AUTHOR
   Diab Jerius
*/
struct str_tokbqdata *
str_tokbq_init( 
  struct str_tokbqdata * tbqd
)
{
  tbqd->start_match = NULL;

  return tbqd;
}

/*
   PROCEDURE
   str_tokbq_free - free str_tokbq state information

   DESCRIPTION
   @code{str_tokbq_free} should be called if the full set
   of calls to @code{str_tokbq} (e.g. until it returns @code{NULL}
   is interrupted and will not be completed.

   AUTHOR
   Diab Jerius
*/
void
str_tokbq_free( 
  struct str_tokbqdata * tbqd
)
{
  if ( tbqd->start_match )
    free( tbqd->start_match );
  tbqd->start_match = NULL;
}

/*
   PROCEDURE
   str_tokbq - split a quoted string at specified delimiters

   DESCRIPTION
   This routine works like @code{strtok}, except that it
   understands escaped characters and quoted strings.  It handles
   balanced quote characters (i.e. @samp{[} and @samp{]}).  It does
   not handle nested quotes.
   
   It can collapse consecutive delimiter characters, or generate empty
   tokens.  For example, if the delimiter is @samp{,}, parsing of the
   string ",,a" would return three tokens, the first two empty
   strings.  Additionally @code{str_tokbq} can be instructed to
   restore the parsed string to its initial state (normally it sticks
   @samp{\0} where the delimiter characters are).

   @code{str_tokbq} is invoked multiple times, once per token.  Unlike
   @code{str_tok}, @code{str_tokbq} stores all of its state
   information externally, in a caller provided structure.  This must
   be initialized with @code{str_tokbq_init} before the first call to
   @code{str_tokbq}.  The contents are unique to each parsed string.

   Delimiters, quote characters, and the escape character may be
   changed between calls.

   Each invocation of @code{str_tokbq} returns a pointer to the next
   token.  If no more tokens are available, it returns @code{NULL}.
   If skip mode is on an empty string is considered to have no tokens.
   If skip mode is off, an empty string is considered to have a single
   token.

   The @code{actions} flag is used to specify how to handle
   consecutive delimiters, string restoration, and whether to
   recognize escape sequence.  The following flags (which may be
   combined with @code{|} are available:

   @table @code

   @item STRTOK_ESCAPE
   Recognize escape character sequences before quotes and delimiters.
   They are ignored elsewhere.  The escape character is given by the
   @code{escape_char} parameter.

   @item STRTOK_RESTORE @code{str_tokbq} replaces delimiter characters
   in the passed string with @samp{\0} as it parses it.  This flag
   directs @code{str_tokbq} to restore the character from the previous
   call.  To fully restore the string to its original state, the
   caller must either repeatedly invoke @code{str_tokbq} until it
   returns a @code{NULL} or invoke the @code{str_tokbq_restore}
   function.

   @item STRTOK_SKIP
   If set, this indicates that consecutive delimiters are collapsed
   ("skipped").

   @end table

   Quotes need not be the first character in a token (e.g., @samp{foo"
   "bar}) will be treated as one token, and can be either single (both
   forward and backwards) or double.  To embed quotes in like-quoted
   strings, escape them (e.g., @samp{foo"\""bar}).

   ERRORS
   On error, @code{str_tokbq} returns @code{NULL} and stores an error
   code in  @code{*error}.   The following errors are recognized:
   
   @table @code
   @item EINVAL
     There were unbalanced quotes.

   @end table

*/
char *
str_tokbq(
  char* string,			/* the string to parse */
  const char* delim,		/* the set of characters that delimit tokens */
  const char* open_quote,	/* the set of characters that delimit
				   opening quotes. e.g., @samp{"'@{[} */
  const char* close_quote,	/* the set of matching characters that delimit
				   closing quotes. These must be in
				   the same order as the opening
				   characters in @code{open_quote},
				   e.g., @samp{"'@}]} */
  char escape_char,		/* the escape character */
  int actions,			/* skip, restore, and escape modes */
  struct str_tokbqdata * tbqd, 	/* state information */
  int* error			/* returned error code */
  )
{
  char* start;
  char* cptr;

  *error = 0;

  /* first time around */
  if ( NULL == tbqd->start_match )
  {
    /* quotes, delim */
    tbqd->start_match = str_join( NULL, 3, delim, open_quote, close_quote );

    start = string;
    tbqd->dchar = '\0';
  }

  else
  {

    if (tbqd->ptr == NULL)
    {
      free( tbqd->start_match );
      tbqd->start_match = NULL;
      return NULL;
    }

    start = tbqd->ptr;

    if (actions & STRTOK_RESTORE)
      *start = tbqd->dchar;

    start++;
  }
  
  /* skip over any delimiters at the front of the string */
  if ( actions & STRTOK_SKIP)
    start += strspn(start, delim);

  
  cptr = start;
  for ( cptr = start; *cptr ; )
  {
    /* find the first quote or delimiter */
    char *fq = strpbrk( cptr, tbqd->start_match );

    /* ran off end of string; all done here */
    if ( NULL == fq )
      break;

    /* escaped character? */
    if (    (actions & STRTOK_ESCAPE)
	    && fq > start
	    && fq[-1]  == escape_char )
    {
      char *ep = fq;
      int nesc = 0;

      /* count the total number of escapes; must be odd for us to care */
      while( --ep >= start && *ep == escape_char )
	nesc++;

      if (nesc % 2 )
      {
	cptr = fq + 1;
	continue;
      }
    }
  
    /* delimiter? finished off another token */
    if ( strchr( delim, *fq ) )
    {
      tbqd->ptr   = fq;
      tbqd->dchar = *fq;
      *fq    = '\0';
      return start;
    }

    /* must be an open quote. find its companion */
    else if ( strchr( open_quote, *fq ) )
    {
      char quote = *fq;
      char* nptr = fq + 1;
      char matched_quote = close_quote[ index( open_quote, quote )
					- open_quote ];

      /* this loop is here to advance over escaped characters */
      while( 1 )
      {
	char* nq = index( nptr, matched_quote );

	/* no luck? unbalanced quote. that's bad */
	if ( NULL == nq )
	{
	  tbqd->ptr = nq;
	  tbqd->dchar = quote;

	  free( tbqd->start_match );
	  tbqd->start_match = NULL;

	  *error = EINVAL;
	  return NULL;
	}

	/* escaped character? skip it and try again */
	if (    (actions & STRTOK_ESCAPE)
		&& nq[-1]  == escape_char )
	{
	  char *ep = nq;
	  int nesc = 0;

	  /* count the total number of escapes; must be odd for us to care */
	  while( --ep >= nptr && *ep == escape_char )
	    nesc++;

	  if (nesc % 2)
	  {
	    nptr = nq + 1;
	    continue;
	  }
	}

	/* we found it, move on */
	cptr = nq + 1;
	break;
      }
    }

    /* must be a close quote. that's bad, it's out of sequence */
    else
    {
      tbqd->ptr   = fq;
      tbqd->dchar = *fq;

      free( tbqd->start_match );
      tbqd->start_match = NULL;

      *error = EINVAL;
      return NULL;
    }


  }

  /* this is reached upon hitting the end of the string */

  /*
    if skip is not turned on, even empty strings are treated
    as tokens.  however the default (strtok) behavior is to return a NULL.
    so, a little special processing for this case.  since
    estr is set to NULL, we'll only hit this code once. 
  */

  tbqd->ptr = NULL;
  return *start == '\0' && (actions & STRTOK_SKIP) ? NULL : start;
}


/*
   PROCEDURE
   str_tokbqcnt - count the number of tokens in a string

   DESCRIPTION
   This routine counts the number of tokens in a string as would be
   returned by @code{str_tokbq}.  It @emph{does} change the input string,
   but restores it to its initial state upon completion.  A constant
   string (e.g., a statically declared string) should not be passed.

   RETURNS
   Upon success it returns the number of tokens in the string.
   On error it returns @samp{-1}.

   AUTHOR
   Diab Jerius

*/
int
str_tokbqcnt(
  char* string,			/* the string to parse */
  const char* delim,		/* the set of characters that delimit tokens */
  const char* open_quote,	/* the set of characters that delimit
				   opening quotes. e.g., @samp{"'@{[} */
  const char* close_quote,	/* the set of matching characters that delimit
				   closing quotes. These must be in
				   the same order as the opening
				   characters in @code{open_quote},
				   e.g., @samp{"'@}]} */
  char escape_char,		/* the escape character */
  int actions,			/* skip, restore, and escape modes */
  struct str_tokbqdata * tbqd, 	/* state information */
  int* error			/* returned error code */
  )
{
  int cnt = 0;

  /* set STRTOK_RESTORE, as we shouldn't be whacking on the string */

  actions |= STRTOK_RESTORE;

  str_tokbq_init( tbqd );

  while( str_tokbq( string, delim, open_quote, close_quote, escape_char,
		    actions, tbqd, error ) )
    cnt++;

  return *error ? -1 : cnt;
}

/*
   PROCEDURE
   str_tokbq_restore - restore string passed to str_tokbq

   DESCRIPTION
   @code{str_tokbq} changes the scanned string as it processes it.  If
   the @code{restore} flag passed to @code{str_tokbq} is set then the
   string will be restored if the @code{str_tokbq} is called to
   completion. However, sometimes it's necessary to interrupt the
   parsing, and repeatedly calling the @code{str_tokbq} to restore the
   string is ridiculous.

   This function restores the string.

   RETURNS
   It returns a pointer to the next character that @code{str_tokbq}
   would have returned, except in the case where the end of string was
   reached, in which case it will return a pointer to the end of
   string character.  This allows one to pass the pointer to another
   independent invocation of @code{str_tokbq}, if,
   for example, continued parsing is to be done in another context.

   AUTHOR
   Diab Jerius

*/

char *
str_tokbq_restore(
  struct str_tokbqdata *tqbd
)
{
  if (tqbd->ptr == NULL)
    return NULL;
    
  *(tqbd->ptr) = tqbd->dchar;

  return tqbd->dchar ? ++(tqbd->ptr) : tqbd->ptr;
}
