/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "str/str.h"


/*   
  PROCEDURE
  str_tokbqenize - tokenize a string at specified delimiters

  DESCRIPTION
  This routine tokenizes the string using @code{str_tokbq}.  The 
  invoking routine calls @code{str_tokbqenize} once to parse all the tokens
  for a given delimiter.  The tokens are stored in @emph{string_argv}.  

  if the @code{restore} argument is false, @code{str_tokbqenize} will
  allocate memory for each of the tokens; otherwise it will return
  pointers to the locations in the original string.

  The routine @code{str_tokbqenize_free} is provided to free the memory
  allocated by @code{str_tokbqenize}

  RETURNS
  The number of tokens. The third argument, @code{string_argv}, 
  contains the parsed tokens.

  ERRORS
  On error, @code{str_tokbqenize} returns @code{0} and sets @code{errno}
  accordingly.   The following errors are recognized:

  @table @code

  @item EINVAL
    There were unbalanced quotes or the last character in the
    string was an escape character (i.e., no character to escape).

  @item ENOMEM
    A memory allocation failed

  @end table

  AUTHOR
  Diab Jerius
  Dan Nguyen
*/
int
str_tokbqenize(
  char* string,			/* the string to parse */
  char*** string_argv,		/* to store the parsed tokens */
  const char* delim,		/* the set of characters that delimit tokens */
  const char* open_quote,	/* the set of characters that delimit
				   opening quotes. e.g., @samp{"'@{[} */
  const char* close_quote,	/* the set of matching characters that delimit
				   closing quotes. These must be in
				   the same order as the opening
				   characters in @code{open_quote},
				   e.g., @samp{"'@}]} */
  char escape_char,		/* the escape character */
  int actions,			/* skip, restore, and escape modes */
  struct str_tokbqdata * tbqd, 	/* state information */
  int* error			/* returned error code */
  )
{
  char *tok;
  char **string_arr;
  int n_tokens;

  *string_argv = NULL;

  *error = 0;

  n_tokens =  str_tokbqcnt(string, delim, open_quote, close_quote,
			   escape_char, actions, tbqd, error);

  
  /* if there was an error, or no tokens, return */
  if ( n_tokens <= 0 ) 
    return 0;

  string_arr = *string_argv = (char**) malloc( (n_tokens + 1) * sizeof(char*));

  if ( NULL == string_arr )
  {
    *error = ENOMEM;
    return 0;
  }

  /* no need to check for errors in str_tokbq, as str_tokbqcnt has
     parsed the string once already */

  str_tokbq_init( tbqd );
  
  while( tok = str_tokbq(string, delim, open_quote, close_quote,
			 escape_char, actions, tbqd, error) )
  {
    if ( ! ( actions | STRTOK_RESTORE ) )
      *string_arr = tok;

    else
    {
      *string_arr = str_dup( tok );

      /* out of memory. must delete everything now */
      if ( NULL == *string_arr )
      {
	str_tokbqenize_free( *string_argv, actions );
	*error = ENOMEM;
	return 0;
      }
    }

    string_arr++;
  }

  /* last element is a sentinel */
  *string_arr = (char*) NULL;
  
  return n_tokens;

}

/*   
  PROCEDURE
  str_tokbqenize_free - free memory allocated by @code{str_tokbqenize}

  DESCRIPTION
  This routine frees the memory allocated by the @code{str_tokbqenize}.
  The user must call @code{str_tokbqenize_free} with the same value for
  restore as was used when @code{str_tokbqenize} was invoked.
  
  RETURNS
  void

  AUTHOR
  Diab Jerius
  Dan Nguyen

*/

void
str_tokbqenize_free(
  char** string_argv,	/* storage for the parsed tokens */
  int action		/* same as passed to str_tokbqenize */
)
{
  char **str_pptr = string_argv;

  if ( NULL == string_argv )
    return;
  
  if ( action & STRTOK_RESTORE )
    while ( *str_pptr )
      free( *str_pptr++ );

  free( string_argv );
}
