/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <testlib/testlib.h>


#include "str/str.h"


/* Solaris (2.8) doesn't have unsetenv, and putenv("foo=") doesn't
   cut it, so hope these aren't in existence */
#define VAR1 "WhasUP_UNDEFINED_ENVIRONMENTAL_VARIABLE1"
#define VAR2 "WhasDown_UNDEFINED_ENVIRONMENTAL_VARIABLE2"

#define len_VAR1 (sizeof(VAR1)-1)

#define DVAR1 "foo"
#define DVAR2 "goo"

typedef struct
{
  char* ifmt;
  char* ofmt;
  char* val1;
  char* val2;
  int   keep_undef;
  int   error;
  int   ustart;
  int   uend;
} Test;


Test tests[] =
{
  /*
    ifmt		ofmt		val1	val2	keep	error	ustart	uend */
  { "${%s",		NULL,		"a",	"b",	0,	EDOM,	 1,	-1 },
  { "${%s}/${%s}",	"%s/%s",	"a",	"b",	0,	0,	 -1,	-1 },
  { "$%s/$%s",		"%s/%s",	"a",	"b",	0,	0,	 -1,	-1 },
  { "a/${%s}/c",	"a/%s/c",	"b",	NULL,	0,	0,	 -1,	-1 },
  { "a/$%s/c",		"a/%s/c",	"b",	NULL,	0,	0,	 -1,	-1 },
  { "${%s}/b",		"%s/b",		"a",	NULL,	0,	0,	 -1,	-1 },
  { "$%s/b",		"%s/b",		"a",	NULL,	0,	0,	 -1,	-1 },
  { "${%s}/b",		"/b",		NULL,	NULL,	0,	0,	 -1,	-1 },
  { "${%s}/b",		"${%s}/b",	NULL,	NULL,	1,	EFAULT,	 2,	 1 + len_VAR1 },
  { "${%s}/${%s}",	"%s/%s",	NULL,	NULL,	0,	0,	 -1,	-1 },
  { "${%s}/${%s}",	"${%s}/${%s}",	NULL,	NULL,	1,	EFAULT,	  2,	 1 + len_VAR1 }
  
};

#define ntests (sizeof(tests) / sizeof(Test) )

void PutEnv( char *var, char *val )
{
  char envvar[1024];

  if ( NULL == val )
    return;

  sprintf(envvar, "var=%s", var, val );
  putenv(envvar);
}


int main( int argc, char **argv )
{
  int ii;
  char *istr;
  char label[1024];

  START( "str_interp" );


  for (ii = 0 ; ii < ntests ; ii++ )
  {
    char    buf[1024];
    char   save[1024];
    char  input[1024];
    char  label[1024];
    char  prefix[1024];
    char *result;
    int   error = -1;
    char *ustart = (char*) 0xDEADBEEF;
    char *uend   = (char*) 0xDEADBEEF;

    Test test = tests[ii];

    sprintf( input, test.ifmt,
	     test.val1 ? DVAR1 : VAR1,
	     test.val2 ? DVAR2 : VAR2 );

    strcpy( save, input );

    PutEnv( DVAR1, test.val1 );
    PutEnv( DVAR2, test.val2 );

#define LABEL(extra) (sprintf( label, "%-34s: %s", prefix, extra ), label)

    sprintf( buf, test.ifmt, DVAR1, DVAR2 );
    sprintf( prefix, "%-13s: keep_undef=%d", buf, test.keep_undef );
    if ( !test.val1 )
      strcat( prefix, " undef" );

    result = str_interp( input, test.keep_undef, &ustart, &uend, &error );

    TEST( LABEL("error"), test.error, error );

    /* result != NULL */
    if ( test.ofmt )
    {
      char output[1024];
      sprintf( output, test.ofmt,
	       test.keep_undef ? test.val1 ? test.val1 : VAR1
	                       : "",
	       test.keep_undef ? test.val2 ? test.val2 : VAR2
	                       : ""
	       );

      TESTSTRING( LABEL("result"), result, output );
    }

    /* expect null result */
    else
    {
      TEST( LABEL("null result"), result, NULL );
    }

    /* ensure that the original string is restored */
    TESTSTRING( LABEL("restore"), input, save );

    /* check ustart and uend */
    if ( test.ustart >= 0 )
      TEST( LABEL("ustart"), ustart, input + test.ustart );

    if ( test.uend >= 0 )
      TEST( LABEL("uend"),   uend,   input + test.uend );

    if ( result )
      free( result );

  }

  /******************************************************/
  SUMMARY();
  return FAILED( ) ? EXIT_FAILURE : EXIT_SUCCESS;
} 
