/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "str/str.h"

/*

  PROCEDURE
  tokmatch - match a string to a list of tokens

  DESCRIPTION
  @code{tokmatch} compares a string to a list of tokens in a user provided
  @code{TokList}.  The @code{TokList} should be sorted alphabetically by 
  name.

  RETURNS
  It returns the @code{id} of the matched token, @samp{-1} otherwise.
  Note that this restricts the values of @code{id}.

  AUTHOR
  Diab Jerius
*/
int
tokmatch(const char *token, const TokList *toklist )
{
 TokListToken tok, *tokf;

 tok.name = token;
 tokf = (TokListToken *) bsearch(&tok, toklist->toks, toklist->ntoks,
         sizeof(TokListToken), tokcmp);
 return (tokf == NULL) ? -1 : tokf->id;
}

/*

  PROCEDURE
  tokcmp - compare two tokens by name

  DESCRIPTION
  @code{tokcmp} compares two @code{TokListToken}'s by name.  It is
  suitable for use by @code{bsearch} or @code{qsort}.

  RETURNS
  It returns @samp{-1} if the first token's name is alphabetically first,
  @samp{0} if they have the same name, and @samp{1} if the first token's name
  follows the second, alphabetically.

  AUTHOR
  Diab Jerius

*/
int
tokcmp(const void *tok1, const void *tok2)
{
 return strcmp(((TokListToken *)tok1)->name,((TokListToken *)tok2)->name);
}


/*

  PROCEDURE
  tokcasematch - match a string to a list of tokens

  DESCRIPTION
  @code{tokmatch} compares a string to a list of tokens in a user provided
  @code{TokList}.  The @code{TokList} should be sorted alphabetically by 
  name. The string comparison is case insensitive.

  RETURNS
  It returns the @code{id} of the matched token, @samp{-1} otherwise.
  Note that this restricts the values of @code{id}.

  AUTHOR
  Diab Jerius
*/
int
tokcasematch(const char *token, const TokList *toklist )
{
 TokListToken tok, *tokf;

 tok.name = token;
 tokf = (TokListToken *) bsearch(&tok, toklist->toks, toklist->ntoks,
         sizeof(TokListToken), tokcasecmp);
 return (tokf == NULL) ? -1 : tokf->id;
}


/*

  PROCEDURE
  tokcasecmp - compare two tokens by name

  DESCRIPTION
  @code{tokcmp} compares two @code{TokListToken}'s by name.  The
  comparison is case insensitive. It is suitable for use by
  @code{bsearch} or @code{qsort}.

  RETURNS
  It returns @samp{-1} if the first token's name is alphabetically first,
  @samp{0} if they have the same name, and @samp{1} if the first token's name
  follows the second, alphabetically.

  AUTHOR
  Diab Jerius

*/
int
tokcasecmp(const void *tok1, const void *tok2)
{
 return strcasecmp(((TokListToken *)tok1)->name,((TokListToken *)tok2)->name);
}

