/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of suplib
 *
 * suplib is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * suplib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>

#include "str/str.h"

#include "units/units.h"

/*

  PROCEDURE
  units_parse - parse a string composed of a floating point number and
  a units specification.

  DESCRIPTION
  This routine parses a string composed of a floating point number and
  a units specification (e.g. @samp{1mm}, @samp{23 arcsec}).  White
  space between the number and the unit is ignored.  The caller must
  ensure that there is no white space following the unit specification.

  RETURNS
  The parsed number and the unit id are written into the passed
  @code{UnitVal} structure.  The @code{unit} element is only changed
  if a valid unit was found. @code{units_parse} returns a
  @code{UNITS_ParseErr} consistent with the results of the parse:

  @table @code

  @item UNITS_OK
    no errors were encountered

  @item UNITS_BADNUM
    there was an error while parsing the number

  @item UNITS_NOSPEC
    no units specification was present

  @item UNITS_BADSPEC
    the units specification wasn't recognized

  @end table

  AUTHOR
  Diab Jerius

*/

int
units_parse(
  const char *spec,		/* the specification to parse */
  UnitVal *uv,			/* the resultant number and unit id */
  const UnitsList *list		/* the list of units */
  )
{
  int unit;
  char *ptr;

  /* parse number in front of units specification */
  uv->value = strtod( spec, &ptr );

  /* check if strtod found a number */
  if ( ptr == spec )
    return UNITS_BADNUM;
  
  /* ignore spaces between the number and the units specification.
     spaces after the spec should be removed by the caller */
  while ( isspace( *ptr ) )
    ptr++;

  /* if nothing is left, there's no units specification */
  if ( '\0' == *ptr )
    return UNITS_NOSPEC;

  /* scan the lists looking for the first matching unit spec */
  unit = tokmatch( ptr, list->map );

  if ( -1 != unit )
    uv->unit = unit;

  return -1 == unit ? UNITS_BADSPEC : UNITS_OK;
}
