// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of suplibxx
//
// suplibxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// suplibxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#ifdef sgi
#include <ctype.h>
#else
#include <cctype>
#endif
#include <fstream>
#include <iostream>
#include <string>
using namespace std;

#include "io.h"
using namespace suplib;

#define WS " \t\n\r"

/**
 * \ingroup io
 *
 * \brief Reads physical and logical lines.
 * \param is the istream from which to read.
 * \param str the string into which to read.
 * \param opt the control options.
 * \param delim the character which indicates the end of a physical line.
 * \param continuation the character directly proceeding the delim
 * which  indicates a logical record continues on the following
 * physical line. 
 *
 * getrecord reads a line from the specified istream.  It always
 * reads a complete line, enlarging the string as necessary.  The opt
 * argument specifies a set of control flag, which are created by
 * logically OR'ing suplib::readopt values.
 *
 * \return It returns the input istream after attempting to read lines
 * from it.
 */
istream& suplib::getrecord(
  istream& is,
  string& str,
  int opt,
  char delim,
  char continuation
  ) {

  str.erase( );

  // split appart logical and physical reads here. in the latter
  // case, can optimize a bit by reading directly into target string

  if ( opt & suplib::READ_LOGICAL ) {
    
    string tmp;

    // are we there yet?
    bool done = false;

    while ( !done && getline( is, tmp, delim )  ) {

      // find last continuation char
      size_t last = tmp.find_last_of( continuation );

      // find last non whitespace char
      size_t eows = tmp.find_last_not_of( WS );

      // if there is nothing but whitespace after continuation...  a
      // logical line, and lo, a continuation character.  what shall
      // we do?
      if ( eows == last ) {

	// well, if this CLEAN thing is requested,
	// remove any whitespace after the continuation character
	// and insert the delimiter, since it's been scrubbed
	// by getline()

	if ( opt & suplib::CLEAN )
	{
	  tmp.erase( last + 1 );
	  tmp += delim;
	}

	// this really isn't raw, as we don't know if we
	// ran out of characters or if the delimiter was
	// removed by getline.  add the delimiter just to
	// make things peachy.
	else if (opt & suplib::RAW )
	  tmp += delim;

	// normally, we just zap the continuation character and all
	// trailing white space
	else
	  tmp.erase( last );

      }

      // ok, it's not a continuation character, just worry about
      // cleaning out the end of the line.  note that we
      // bail here, as we've hit the end of the record.
      else {

	// zap end of line (no white space)
	if ( opt & suplib::STRIP )
	  tmp.erase( eows + 1 );

	// same bogus argument as before
	else if (opt & suplib::RAW )
	  tmp += delim;
	
	// by default, just keep everything but the delimiter,
	// already removed by getline()

	// we're done here, break out of the getline() loop
	done = true;
      }

      // save the current line
      str.append( tmp );
    }

  }


  // read physical line
  else if ( getline( is, str, delim )  ) {

    // strip trailing white space
    if ( opt & suplib::STRIP )
    {
      size_t eows = str.find_last_not_of( WS );
      str.erase( eows + 1 );
    }

    // same silliness as above
    else if ( opt & suplib::RAW )
      str += delim;
  }


  return is;
}
