/*
//
// Copyright (C) 1991 Texas Instruments Incorporated.
//
// Permission is granted to any individual or institution to use, copy, modify,
// and distribute this software, provided that this complete copyright and
// permission notice is maintained, intact, in all copies and supporting
// documentation.
//
// Texas Instruments Incorporated provides this software "as is" without
// express or implied warranty.
//
// Created: LGO 11/27/89 -- Initial design
//
*/

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <math.h>

#include "testlib/testlib.h"

static int num_test;
static int tests_passed;
static int tests_failed;
static char* test_name;

#define LENGTH 128

static void change_header( char* name )
{
#ifdef NODEBUGGER
    time_t now = time( (time_t *) NULL );
    char* format = "%s%s%s";
#endif
    char the_time[LENGTH];
    int length;
    int l;
    char blank[LENGTH];
    char* title;

#ifdef NODEBUGGER
    /* the memory access/leak tool on Solaris thinks the 
       the following line has an access problem
    */
    strftime( the_time, LENGTH, "%x %X", localtime( &now ) );
#else
    strcpy( the_time, " " );
#endif


    length = strlen( name ) + strlen( the_time );
    l = 64 - length;

    if ( l > 0 ) {
      int i;
      for ( i = 0; i < l; i++ )
	blank[i] = ' ';
      blank[i] = '\0';
    } else {
      fprintf( stderr, "Please enter a shorter title `%s'\n",name );
      exit( 1 );
    }

    title = (char *) malloc( (size_t) LENGTH );
    #ifdef NODEBUGGER
      sprintf( title, format, name, blank, the_time );
    #else
      strcpy(title, name);
      strcat(title, blank);
      strcat(title, the_time);
    #endif

    length = strlen( title );
    /*
    //    title[ length-1 ] = '\0'; //get rid of carriage return
    */
    printf( title );
    free( title );
}


void Test_start(char* name) 
{
  num_test = 0;
  tests_passed = 0;
  tests_failed = 0;
  test_name = name;
  printf("-----------------------------------------------------------------------------\n" );
  printf( "Start Testing " );
  if ( test_name != NULL )
    change_header( test_name );
  printf( "\nTest   Line   Description\n" );
  printf("-----------------------------------------------------------------------------\n" );
  fflush(stdout);
}

void Test_begin(const char* msg, size_t line_no) 
{
  num_test++;
  printf( " %03d | %4lu | %-47s --> ", num_test, line_no, msg );
  fflush(stdout);
}

/*
// NOTE: We don't pass in the message (see test_begin) because
//       we want to ensure that the message is printed BEFORE
//       the test is executed.  This way when a test crashes
//       we can tell if it was during a test, or between tests.
*/
void Test_perform(int success) 
{
  if (success) 
  {
    tests_passed++;
    printf( "  PASSED\n" );
  }
  else 
  {
    tests_failed++;
    printf( "**FAILED**\n" );
  }
  fflush( stdout );
}

void Test_summary(void)
{
  printf ("-----------------------------------------------------------------------------\n");
  if (test_name != NULL) 
    printf( "%s ", test_name );
  printf( "Test Summary: ");
  printf( "%d passed, %d failed", tests_passed, tests_failed );
  if (tests_failed > 0) 
    printf( "\t\t\t*****" );
  printf("\n-----------------------------------------------------------------------------\n" );
  fflush(stdout);
}

int Test_failed( void )
{
  return tests_failed;
}

double safe_div( const double x, const double y ) {

  /*
    return x / y;
  */
  return ( 
    ( y < 1 && x > y * DBL_MAX )
    ?
    DBL_MAX
    :
    (
      ( (y > 1 && x < y * DBL_MIN) || x == 0 )
      ?
      0
      :
      x / y
      )
    );
}

/*
  Note that this test returns 0 when x and y compare to within the
  tolerance and 1 otherwise...

  That might be counterintuitive, 0 is success here and 1 is failure.
 */
int Test_fuzzy_compare( double x, double y, double tol )
{
  double diff = fabs( x - y );

  /*
    might as well start at the simplest comparison...
  */
  if ( x == y )
    return 0;

  /*
    alright, the comparison i'm hoping to use will always fail if one
    of the values is 0 so let's handle the special case first 
  */
  else if ( 0 == x || 0 == y )
    return diff < tol ? 0 : 1;

  else 
    return ( safe_div( diff, fabs( x ) ) <= tol && safe_div( diff, fabs( y ) ) <= tol ) ? 0 : 1;

}

int fuzzy_compare( double x, double y ) {
  return Test_fuzzy_compare( x, y, __TESTDFLT_TOL );
}


/*
  Note that this test returns 0 when x and y compare to within the
  tolerance and 1 otherwise...

  That might be counterintuitive, 0 is success here and 1 is failure.
 */
int Test_fuzzy_abs_compare( double x, double y, double tol )
{
  double diff = fabs( x - y );

  /*
    might as well start at the simplest comparison...
  */
  if ( x == y )
    return 0;

  else
    return diff < tol ? 0 : 1;

}

int fuzzy_abs_compare( double x, double y ) {
  return Test_fuzzy_abs_compare( x, y, __TESTDFLT_TOL );
}
