// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of tracefctxx
//
// tracefctxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// tracefctxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iostream>
#include <fstream>
#include <cstring>
#include <unistd.h>
using namespace std;

#include "TraceFct.h"

#define NUM_BLANK_SPACES "   "

bool	TraceFct::print_upon_enter_and_exit(false);
int	TraceFct::stack_level_to_print(-1);
string	TraceFct::progname("UNKNOWN PROGRAM");
string	TraceFct::prefix = TraceFct::init_prefix();
char	TraceFct::outbuf[8192];
ostream* TraceFct::ostr = &cerr;
list<string> TraceFct::function_stack;

char *
TraceFct::init_prefix( void )
{
  sprintf(outbuf, "# %d: %s: ", getpid(), progname.c_str() );
  return outbuf;
}

//====================================================================//


//! Class initializer

/*!
  This initializer should be used to initialize the class, and should
  be called once in the program, before any subsequent calls to the
  per-function initializers

  \param name	the name of the program, preferrably argv[0].  path
                information (if present) is not stripped.

  \param print_it	if true, a message will be
			   output upon entry and exit of a
			   function

  \param num_fct_to_print	if positive, only the requested number of
		     	   function names will be printed.  if
		     	   non-positive, all of the function names
		     	   will be printed.
  

  The calling routine can also set up some output options which
  determine the depth of the function stack to print and whether a
  diagnostic should be printed at every entry and return of a
  function.

*/

TraceFct::TraceFct( string prog_name, bool print_it, int num_fct_to_print ) 
{

  progname = prog_name;

  print_upon_enter_and_exit = print_it;

  stack_level_to_print = num_fct_to_print;

  function_stack.clear();

  ostr = &cerr;

  prefix = init_prefix( );
}

//@{

/*!
  \brief Function entry constructor
  \param name	the name of the function

  This constructor is used when entering a function which should be
  registered.
*/

TraceFct::TraceFct(string& name)
{
  function_stack.push_back( name );

  if (print_upon_enter_and_exit)
    message( "[%2d] %s: entering\n", function_stack.size(), 
	     function_stack.back().c_str() );
}

/*!
  \brief Function entry constructor
  \param name	the name of the function

  This constructor is used when entering a function which should be
  registered.
*/
TraceFct::TraceFct(const char* name)
{
  function_stack.push_back( name );

  if (print_upon_enter_and_exit)
    message( "[%2d] %s: entering\n", function_stack.size(), 
	     function_stack.back().c_str() );
}

//@}

/*!

  The destructor. 

*/
TraceFct::~TraceFct( )
{
  // if the function stack is empty, we must be at the top level;
  // nothing to do.
  if ( ! function_stack.empty() )
  {
    if (print_upon_enter_and_exit)
    {
      message( "[%2d] %s: leaving\n", function_stack.size(), 
	       function_stack.back().c_str() );
    }

    function_stack.pop_back();
  }

  else
  {
    // close the output stream
    if ( &cerr != ostr )
      delete ostr;
  }
}

//====================================================================//


/*!
  \brief print a message
  \param print_nl print a closing newline?
  \param format   a printf style format string
  \param args	  objects to print

  This (class) method prints a message to the TraceFct output stream,
  using a printf() style format string and a \c va_list.  If 
  \a print_nl is true, a final newline character is output if
  there isn't one already there.

  \c vprint is aware of embedded newlines and ensures that a
  message prefix is output after them.

 */
void
TraceFct::vprint(
  bool print_nl,
  const char *format,
  va_list args
  )
{
  char *start, *end;

  vsprintf(outbuf, format, args);

  for ( start = outbuf ; end = strchr( start, '\n' ) ; start = end )
  {
    char save;
    save = *++end;
    *end = '\0';
    
    if ( ostr == &cerr )
      *ostr << prefix;

    *ostr << start;

    *end = save;
  }

  if ( *start )
  {
    if ( ostr == &cerr )
      *ostr << prefix;

    *ostr << start;
  }

  if (  print_nl && '\n' != outbuf[strlen(outbuf)-1] )
    *ostr << '\n';
}


/*!
  \brief print a message
  \param print_nl print a closing newline?
  \param format   a printf style format string
  \param ...	  objects to print

  A convenience wrapper around vprint();

 */
void
TraceFct::print(
  bool print_nl,
  const char *format,
  ...
  )
{
  va_list args;
  va_start( args, format );
  vprint( print_nl, format, args );
  va_end(args );
}

/*!
  \brief print a message
  \param print_nl print a closing newline?
  \param format   a printf style format string
  \param ...	  objects to print

  A convenience wrapper around vprint();

 */
void
TraceFct::print(
  bool print_nl,
  Exception& ex
  )
{
  deque<string>::const_iterator begin = ex.begin();
  deque<string>::const_iterator end   = ex.end();

    while( begin != end )
      print( print_nl, (*begin++).c_str() );
}


//====================================================================//

/*!
  \brief Print the current function stack
  \param ostr	the output file stream

  This prints the current function stack to the TraceFct output stream

*/
void
TraceFct::dump_stack()
{

  message( "Stack:\n" );

  unsigned max_level = stack_level_to_print > -1 ?
              stack_level_to_print : function_stack.size();

  if ( function_stack.size() < max_level )
    max_level = function_stack.size();

  list<string>::iterator begin = function_stack.begin();
  list<string>::iterator   end = function_stack.end();
  advance( begin, function_stack.size() - max_level );

  int depth = function_stack.size();
  while( begin != end )
  {
    end--;
    message( " [%2d] %s\n", depth, end->c_str() );
    depth--;
  }

  if ( stack_level_to_print == -1 ||
       max_level > function_stack.size() )
    message( " [ 0] <TOP>\n");
}


//====================================================================//

/*!
  \brief handle exit output
  \param format   a printf style format string
  \param args	  objects to print

  This prints the passed message (using \c TraceFct::vprint()) and
  calls TraceFct::dump_stack();

 */

void
TraceFct::vexit_print(
  const char *format,
  va_list args
)
{
  vprint( true, format, args );

  dump_stack(  );
}


//====================================================================//


/*! \name Exit

    These functions print an error message to \c stderr as well as
    the \b tracefct output stream (if different from \c stderr) and
    then exits the program with the supplied error code.  The error
    message is passed in the same fashion as the arguments to \c
    vprintf or \c sprintf.  The message may contain multiple output
    lines (i.e., multiple newline characters).  If a trailing newline
    character is not specified, it will be appended.  

*/

//@{

// doxygen braindamage; can't pull inlines from header file into
// a member group
/*!
  \fn   static void TraceFct::exit (int exit_code, const string& msg)
  \brief Exit a program, dumping the function stack.
  \param exit_code  the exit code to be returned to the system
  \param msg	    a message to print
*/
  

/*!
  \brief Exit a program, dumping the function stack.
  \param exit_code  the exit code to be returned to the system
  \param format	    a printf style format string
  \param ...        additional arguments
*/

void
TraceFct::exit
(
  int exit_code,
  const char *format,		
  ...
)
{
  va_list args;

  if ( ostr != &cerr )
  {
    va_start(args, format);
    vexit_print( format, args );
    va_end(args);
    delete ostr;
  }

  ostr = &cerr;

  va_start(args, format);
  vexit_print( format, args );
  va_end(args);
  
  std::exit(exit_code);
}

/*!
  \brief Exit a program, dumping the function stack.
  \param exit_code  the exit code to be returned to the system
  \param ex	    The Exception Stack to print

  This version of the exit function will print out the passed
  \c Exception stack.

*/

void
TraceFct::exit
(
  int exit_code,
  Exception& ex
)
{
  deque<string>::const_iterator begin = ex.begin();
  deque<string>::const_iterator end   = ex.end();
  
  if ( ostr != &cerr ){
    print( true, ex );
    delete ostr;
  }

  ostr = &cerr;

  print( true, ex );
  
  dump_stack();
  std::exit(exit_code);
  
}

// doxygen braindamage; can't pull inlines from header file into a member group
/*!
  \fn  void TraceFct::die (const string& msg)
  \brief Exit a program with error EXIT_FAILURE, dumping the function stack.
  \param msg	    a message to print
*/

/*!
  \fn void TraceFct::die( Exception& ex )

  \brief Exit a program with error EXIT_FAILURE, dumping the function stack.
  \param ex	    The Exception Stack to print

  This version of the exit function will print out the passed
  \c Exception stack.
*/
  

/*!
  \brief Exit a program with error EXIT_FAILURE, dumping the function stack.
  \param format	    a printf style format string
  \param ...        additional arguments
*/

void
TraceFct::die
(
  const char *format,		
  ...
)
{
  va_list args;

  if ( ostr != &cerr )
  {
    va_start(args, format);
    vexit_print( format, args );
    va_end(args);
    delete ostr;
  }

  ostr = &cerr;

  va_start(args, format);
  vexit_print( format, args );
  va_end(args);
  
  std::exit(EXIT_FAILURE);
}


//@}

//====================================================================//

/*! \name Messages

  The \c message functions print a user supplied message to the TraceFct output
  stream, prefixed by the standard TraceFct style 
  \ref messageprefix "prefix"
  string.  Unlike TraceFct::exit(), a newline character is \e not
  appended to the message.

  The \c println functions are similar, except that they ensure that
  a newline character \e is appended to the message.

*/


//@{

// doxygen braindamage; can't pull inlines from header file into
// a member group
/*!
  \fn static void TraceFct::message (const string& msg)
  \brief Print a message to the \c tracefct output stream.
  \param msg	      a message to print
*/
/*!
  \fn static void TraceFct::message (Exception& ex)
  \brief Print a message to the \c tracefct output stream.
  \param ex The exception to print
*/

/*!
  \brief Print a formatted message to the \c tracefct output stream.
  \param format	    a printf style format string
  \param ...        additional arguments

  The message is passed in the same fashion as the arguments
  to \c printf, allowing formatted output.

*/
void
TraceFct::message
(
  const char *format,		/* a printf style string */
  ...			/* objects to print */
)
{
  va_list args;

  va_start(args, format);

  vprint( false, format, args );

  va_end(args);
}


/*!
  
  \brief Print a formatted message to the \c tracefct output stream using a \c stdargs argument list.
  \param format  a printf style string
  \param args
q
  The error message is passed in the same fashion as the arguments to
  \c vprintf.  Note that \c vmessage expects a \c va_list to be
  passed, rather than a variable argument list.

*/
void
TraceFct::vmessage
(
  const char *format,
  va_list args		/* objects to print */
)
{
  vprint( false, format, args );
}


// doxygen braindamage; can't pull inlines from header file into
// a member group
/*!
  \fn static void TraceFct::println (const string& msg)
  \brief Print a message to the \c tracefct output stream.
  \param msg	      a message to print
*/
/*!
  \fn static void TraceFct::println (const char* msg)
  \brief Print a message to the \c tracefct output stream.
  \param msg	      a message to print
*/
/*!
  \fn static void TraceFct::println (Exception &ex)
  \brief Print a message to the \c tracefct output stream.
  \param ex	      the exception to print
*/


//@}

//====================================================================//


/*! \name Output Stream manipulation */

//@{

/*!

  \brief redirect the \c tracefct output stream
  \param file the file to which messages are to be written

  \open serves to change the TraceFct output stream to the specified
  file.  Normally TraceFct writes to \c stderr.  TraceFct::exit() \e
  always writes to \c stderr.  To reset the output stream, see
  TraceFct::close().  If the current output stream is not \c stderr,
  it is automatically closed.

  If the passed filename is the string \c stderr, open() will use \c
  stderr.

  Upon error, it writes and exit to \c stderr and exits.

*/
void
TraceFct::open(
  const string& file	/* the file to which messages are to be written */
  )
{
  ostream* new_ostr;

  if ( file != "stderr" )
  {
    try 
    {
      new_ostr = new ofstream( file.c_str() );

      if ( ! *new_ostr )
      {
	cerr << prefix << "unable to open TraceFct output stream `" 
	     << file << '\'' << endl;
	std::exit(1);
      }
    }
    catch (...)
    {
      // dunno what happend but it wasn't good!
      cerr << prefix 
	   << "error creating TraceFct output stream object for `" 
	   << file << '\'' << endl;
      std::exit(1);
    }

  }
  else
    new_ostr = &cerr;

  if ( ostr != &cerr )
    delete ostr;

  ostr = new_ostr;
}

/*!

  \brief reset the \c tracefct output stream to \c stderr.

  close() resets the output stream to \c stderr, closing
  the file previously opened by TraceFct::open().

*/

void
TraceFct::close( void )
{
  if ( &cerr != ostr )
  {
    delete ostr;
    ostr = &cerr;
  }
}


//@}

//====================================================================//
