// File:   t_vm_m3math.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of vm_math.cvs
//
// vm_math.cvs is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// vm_math.cvs is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <vm_math/vm_m3math.h>            // vm_M3Math

#include <cstdlib>
#include <iostream>
#include <testlib/testlib.h>

using std::cerr;

static
int vm_M3Math_test01()    /* test at(i,j) */
{
  double   cm[]  = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0 };
  return cm[vm_M3Math<double>::at(1,2)] == 6.0;
}

static
int vm_M3Math_test02()    /* test copy(m,cm) */
{
  double   cm[]  = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0 };
  double   m[9];
  vm_M3Math<double>::copy( m, cm );
  return (m[0] == 1.0) && (m[1] == 2.0) && (m[2] == 3.0) &&
         (m[3] == 4.0) && (m[4] == 5.0) && (m[5] == 6.0) &&
         (m[6] == 7.0) && (m[7] == 8.0) && (m[8] == 9.0);
}

static
int vm_M3Math_test03()    /* test init_by_row(m,v0,v1,v2) */
{
  double   v0[] = { 1.0, 2.0, 3.0 };
  double   v1[] = { 4.0, 5.0, 6.0 };
  double   v2[] = { 7.0, 8.0, 9.0 };
  double   m[9];
  vm_M3Math<double>::init_by_row( m, v0, v1, v2 );
  return (m[0] == 1.0) && (m[1] == 2.0) && (m[2] == 3.0) &&
         (m[3] == 4.0) && (m[4] == 5.0) && (m[5] == 6.0) &&
         (m[6] == 7.0) && (m[7] == 8.0) && (m[8] == 9.0);
}

static
int vm_M3Math_test04()    /* test init_by_col(m,v0,v1,v2) */
{
  double   v0[] = { 1.0, 2.0, 3.0 };
  double   v1[] = { 4.0, 5.0, 6.0 };
  double   v2[] = { 7.0, 8.0, 9.0 };
  double   m[9];
  vm_M3Math<double>::init_by_col( m, v0, v1, v2 );
  return (m[0] == 1.0) && (m[1] == 4.0) && (m[2] == 7.0) &&
         (m[3] == 2.0) && (m[4] == 5.0) && (m[5] == 8.0) &&
         (m[6] == 3.0) && (m[7] == 6.0) && (m[8] == 9.0);
}

int vm_M3Math_test05()    /* test inject_row(m,v,2) */
{
  double   m[] = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0 };
  double   v[] = { 1.0, 2.0, 3.0 };
  vm_M3Math<double>::inject_row( m, v, 2 );
  return (m[0] == 1.0) && (m[1] == 2.0) && (m[2] == 3.0) &&
         (m[3] == 4.0) && (m[4] == 5.0) && (m[5] == 6.0) &&
         (m[6] == 1.0) && (m[7] == 2.0) && (m[8] == 3.0);
}

static
int vm_M3Math_test06()    /* test inject_col(m,v,1) */
{
  double   m[] = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0 };
  double   v[] = { 1.0, 2.0, 3.0 };
  vm_M3Math<double>::inject_col( m, v, 1 );
  return (m[0] == 1.0) && (m[1] == 1.0) && (m[2] == 3.0) &&
         (m[3] == 4.0) && (m[4] == 2.0) && (m[5] == 6.0) &&
         (m[6] == 7.0) && (m[7] == 3.0) && (m[8] == 9.0);
}

int vm_M3Math_test07()    /* test extract_row(m,v,2) */
{
  double   cm[] = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0 };
  double   v[3];
  vm_M3Math<double>::extract_row( cm, v, 2 );
  return (v[0] == 7.0) && (v[1] == 8.0) && (v[2] == 9.0);
}

static
int vm_M3Math_test08()    /* test extract_col(m,v,1) */
{
  double   cm[] = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0 };
  double   v[3];
  vm_M3Math<double>::extract_col( cm, v, 1 );
  return (v[0] == 2.0) && (v[1] == 5.0) && (v[2] == 8.0);
}

static
int vm_M3Math_test09()    /* test set(m,2) */
{
  double   m[9];
  vm_M3Math<double>::set( m, 2.0 );
  return (m[0] ==  2.0) && (m[1] ==  2.0) && (m[2] ==  2.0) &&
         (m[3] ==  2.0) && (m[4] ==  2.0) && (m[5] ==  2.0) &&
         (m[6] ==  2.0) && (m[7] ==  2.0) && (m[8] ==  2.0);
}

static
int vm_M3Math_test10()    /* test dyad_product(m,v1,v2) */
{
  double   v1[] = { 4.0, 5.0, 6.0 };
  double   v2[] = { 2.0, 4.0, 8.0 };
  double   m[9];
  vm_M3Math<double>::dyad_product( m, v1, v2 );
  return (m[0] ==  8.0) && (m[1] == 16.0) && (m[2] == 32.0) &&
         (m[3] == 10.0) && (m[4] == 20.0) && (m[5] == 40.0) &&
         (m[6] == 12.0) && (m[7] == 24.0) && (m[8] == 48.0);
}

static
int vm_M3Math_test11()    /* test {add,sub,mul,div}_eq(m,m) */
{
  double a[] = { 1.0, 2.0, 3.0 };
  double b[] = { 4.0, 5.0, 6.0 };
  double c[] = { 2.0, 4.0, 8.0 };

  double m[9];
  double m1[9];

  vm_M3Math<double>::init_by_row(m,  a, b, c);
  vm_M3Math<double>::init_by_row(m1, a, b, c);

  vm_M3Math<double>::add_eq( m, m1 );
  vm_M3Math<double>::sub_eq( m, m1 );
  vm_M3Math<double>::mul_eq( m, m1 );
  vm_M3Math<double>::div_eq( m, m1 );
  return (m[0] == 1.0) && (m[1] == 2.0) && (m[2] == 3.0) &&
         (m[3] == 4.0) && (m[4] == 5.0) && (m[5] == 6.0) &&
         (m[6] == 2.0) && (m[7] == 4.0) && (m[8] == 8.0);
}

static
int vm_M3Math_test12()    /* test {add,sub,mul,div}_eq(m,r) */
{
  double a[] = { 1.0, 2.0, 3.0 };
  double b[] = { 4.0, 5.0, 6.0 };
  double c[] = { 2.0, 4.0, 8.0 };

  double m[9];
  double m1[9];

  vm_M3Math<double>::init_by_row(m, a, b, c);
  vm_M3Math<double>::init_by_row(m1, a, b, c);

  vm_M3Math<double>::add_eq( m1, 2.0 );
  vm_M3Math<double>::sub_eq( m1, 2.0 );
  vm_M3Math<double>::mul_eq( m1, 2.0 );
  vm_M3Math<double>::div_eq( m1, 2.0 );
  return (m[0] == 1.0) && (m[1] == 2.0) && (m[2] == 3.0) &&
         (m[3] == 4.0) && (m[4] == 5.0) && (m[5] == 6.0) &&
         (m[6] == 2.0) && (m[7] == 4.0) && (m[8] == 8.0);
}

static
int vm_M3Math_test13()    /* test {add,sub,mul,div}(m,cm,cm) */
{
  double a[] = { 1.0, 2.0, 3.0 };
  double b[] = { 4.0, 5.0, 6.0 };
  double c[] = { 2.0, 4.0, 8.0 };

  double m[9];
  double m1[9];
  double m2[9];

  vm_M3Math<double>::init_by_row(m1, a, b, c);
  vm_M3Math<double>::init_by_row(m2, a, b, c);

  vm_M3Math<double>::add(  m,  m1, m2 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::sub(  m,  m1, m2 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::mul(  m,  m1, m2 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::div(  m,  m1, m2 );

  return !fuzzy_compare(m[0], 1.0) &&
         !fuzzy_compare(m[1], 2.0) &&
         !fuzzy_compare(m[2], 3.0) &&
         !fuzzy_compare(m[3], 4.0) &&
         !fuzzy_compare(m[4], 5.0) &&
         !fuzzy_compare(m[5], 6.0) &&
         !fuzzy_compare(m[6], 2.0) &&
         !fuzzy_compare(m[7], 4.0) &&
         !fuzzy_compare(m[8], 8.0);
}

static
int vm_M3Math_test14()    /* test {add,sub,mul,div}(m,cm,r) */
{
  double a[] = { 1.0, 2.0, 3.0 };
  double b[] = { 4.0, 5.0, 6.0 };
  double c[] = { 2.0, 4.0, 8.0 };

  double m[9];
  double m1[9];
  double m2[9];

  vm_M3Math<double>::init_by_row(m1, a, b, c);
  vm_M3Math<double>::init_by_row(m2, a, b, c);

  vm_M3Math<double>::add(  m,  m1, 2.0 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::sub(  m,  m1, 2.0 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::mul(  m,  m1, 2.0 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::div(  m,  m1, 2.0 );

  return !fuzzy_compare(m[0], 1.0) &&
         !fuzzy_compare(m[1], 2.0) &&
         !fuzzy_compare(m[2], 3.0) &&
         !fuzzy_compare(m[3], 4.0) &&
         !fuzzy_compare(m[4], 5.0) &&
         !fuzzy_compare(m[5], 6.0) &&
         !fuzzy_compare(m[6], 2.0) &&
         !fuzzy_compare(m[7], 4.0) &&
         !fuzzy_compare(m[8], 8.0);
}

static
int vm_M3Math_test15()    /* test {add,sub,mul,div}(m,r,cm) */
{
  double a[] = { 1.0, 2.0, 3.0 };
  double b[] = { 4.0, 5.0, 6.0 };
  double c[] = { 2.0, 4.0, 8.0 };

  double m[9];
  double m1[9];

  vm_M3Math<double>::init_by_row(m1, a, b, c);

  vm_M3Math<double>::add(  m,  2.0, m1 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::sub(  m,  2.0, m1 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::mul(  m,  2.0, m1 );
  vm_M3Math<double>::copy( m1, m );
  vm_M3Math<double>::div(  m,  2.0, m1 );

  return !fuzzy_compare(m[0], -1.0    ) &&
         !fuzzy_compare(m[1], -1.0/2.0) &&
         !fuzzy_compare(m[2], -1.0/3.0) &&
         !fuzzy_compare(m[3], -1.0/4.0) &&
         !fuzzy_compare(m[4], -1.0/5.0) &&
         !fuzzy_compare(m[5], -1.0/6.0) &&
         !fuzzy_compare(m[6], -1.0/2.0) &&
         !fuzzy_compare(m[7], -1.0/4.0) &&
         !fuzzy_compare(m[8], -1.0/8.0);
}

static
int vm_M3Math_test16()    /* test lincomb(m3,2,m1,3,m2) */
{
  double const a[] = { 1.0, 2.0, 3.0 };
  double const b[] = { 4.0, 5.0, 6.0 };
  double const c[] = { 2.0, 4.0, 8.0 };
  double const d[] = { 2.0, 1.0, 4.0 };
  double const e[] = { 6.0, 4.0, 5.0 };
  double const f[] = { 3.0, 2.0, 1.0 };
  double m1[9];  vm_M3Math<double>::init_by_row( m1, a, b, c );
  double m2[9];  vm_M3Math<double>::init_by_row( m2, d, e, f );
  double m3[9];
  vm_M3Math<double>::lincomb( m3, 2.0, m1, 3.0, m2 );
  return (m3[0] ==  8.0) && (m3[1] ==  7.0) && (m3[2] == 18.0) &&
         (m3[3] == 26.0) && (m3[4] == 22.0) && (m3[5] == 27.0) &&
         (m3[6] == 13.0) && (m3[7] == 14.0) && (m3[8] == 19.0);
}

static
int vm_M3Math_test17()    /* test mvmult(v,cm,cv) */
{
  double const m1[] = { 2.0, 1.0, 4.0, 6.0, 4.0, 5.0, 3.0, 2.0, 1.0 };
  double const v1[] = { 2.0, 3.0, 4.0 };
  double       v2[3];
  vm_M3Math<double>::mvmult( v2, m1, v1 );
  return
     (v1[0] ==  2.0) && (v1[1] ==  3.0) && (v1[2] ==  4.0) &&
     (v2[0] == 23.0) && (v2[1] == 44.0) && (v2[2] == 16.0);
}

static
int vm_M3Math_test18()    /* test mtvmult(v,cm,cv) */
{
  double const m1[] = { 2.0, 1.0, 4.0, 6.0, 4.0, 5.0, 3.0, 2.0, 1.0 };
  double const v1[] = { 2.0, 3.0, 4.0 };
  double       v2[3];
  vm_M3Math<double>::mtvmult( v2, m1, v1 );
  return
     (v1[0] ==  2.0) && (v1[1] ==  3.0) && (v1[2] ==  4.0) &&
     (v2[0] == 34.0) && (v2[1] == 22.0) && (v2[2] == 27.0);
}

static
int vm_M3Math_test19()    /* test mmult(m,cm,cm) */
{
  double const m1[] = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 2.0, 4.0, 8.0 };
  double const m2[] = { 2.0, 1.0, 4.0, 6.0, 4.0, 5.0, 3.0, 2.0, 1.0 };
  double       m3[9];
  vm_M3Math<double>::mmult( m3, m1, m2 );
  return (m3[0] == 23.0) && (m3[1] == 15.0) && (m3[2] == 17.0) &&
         (m3[3] == 56.0) && (m3[4] == 36.0) && (m3[5] == 47.0) &&
         (m3[6] == 52.0) && (m3[7] == 34.0) && (m3[8] == 36.0);
}

static
int vm_M3Math_test20()    /* test negate(m) */
{
  double m[] = { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0 };
  vm_M3Math<double>::negate( m );
  return (m[0] == -1.0) && (m[1] == -2.0) && (m[2] == -3.0) &&
         (m[3] == -4.0) && (m[4] == -5.0) && (m[5] == -6.0) &&
         (m[6] == -7.0) && (m[7] == -8.0) && (m[8] == -9.0);
}

void
test_vm_M3Math() 
{

  TEST( "test at(i,j)",               vm_M3Math_test01(), 1 );
  TEST( "copy(m,cm)",                 vm_M3Math_test02(), 1 );
  TEST( "init_by_row(m,v0,v1,v0)",    vm_M3Math_test03(), 1 );
  TEST( "init_by_col(m,v0,v1,v0)",    vm_M3Math_test04(), 1 );
  TEST( "inject_row(m,v,2)",          vm_M3Math_test05(), 1 );
  TEST( "inject_col(m,v,1)",          vm_M3Math_test06(), 1 );
  TEST( "extract_row(m,v,2)",         vm_M3Math_test07(), 1 );
  TEST( "extract_col(m,v,1)",         vm_M3Math_test08(), 1 );
  TEST( "set(m,2.0)",                 vm_M3Math_test09(), 1 );
  TEST( "dyad_product(m,v1,v2)",      vm_M3Math_test10(), 1 );
  TEST( "{add,sub,mul,div}_eq(m,m)",  vm_M3Math_test11(), 1 );
  TEST( "{add,sub,mul,div}_eq(m,r)",  vm_M3Math_test12(), 1 );
  TEST( "{add,sub,mul,div}(m,cm,cm)", vm_M3Math_test13(), 1 );
  TEST( "{add,sub,mul,div}(m,cm,r)",  vm_M3Math_test14(), 1 );
  TEST( "{add,sub,mul,div}(m,r,cm)",  vm_M3Math_test15(), 1 );
  TEST( "lincomb(m3,c1,cm1,c2,cm2)",  vm_M3Math_test16(), 1 );
  TEST( "mvmult(v,cm,cv)",            vm_M3Math_test17(), 1 );
  TEST( "mtvmult(v,cm,cv)",           vm_M3Math_test18(), 1 );
  TEST( "mmult(m,cm,cm)",             vm_M3Math_test19(), 1 );
  TEST( "negate(m)",                  vm_M3Math_test20(), 1 );

}
