// File:   t_vm_v3math.cc
// Author: Terry Gaetz

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of vm_math.cvs
//
// vm_math.cvs is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// vm_math.cvs is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <vm_math/vm_v3math.h>            // vm_V3Math

#include <cstdlib>
#include <iostream>
#include <cstdio>
#include <testlib/testlib.h>

using std::cerr;

static
int vm_V3Math_test01()    /* test copy(v,cv) */
{
  double   acv[3]  = { 1.0, 2.0, 3.0 };
  double   x[3];
  vm_V3Math<double>::copy( x, acv );
  return (x[0] == 1.0) && (x[1] == 2.0) && (x[2] == 3.0);
}

static
int vm_V3Math_test02()    /* test set(v,r,r,r) */
{
  double   x[3];
  vm_V3Math<double>::set( x, 4.0, 5.0, 6.0 );
  return (x[0] == 4.0) && (x[1] == 5.0) && (x[2] == 6.0);
}

static
int vm_V3Math_test03()    /* test set(v,r) */
{
  double   x[3];
  vm_V3Math<double>::set( x, 4.0 );
  return (x[0] == 4.0) && (x[1] == 4.0) && (x[2] == 4.0);
}

static
int vm_V3Math_test04()    /* test {add,sub,mul,div}_eq(v,cv) */
{
  double   acv[3]  = { 4.0, 5.0, 6.0 };
  double   x[3];
  vm_V3Math<double>::set( x, 2.0, 4.0, 8.0 );
  vm_V3Math<double>::add_eq( acv, x );
  vm_V3Math<double>::sub_eq( acv, x );
  vm_V3Math<double>::mul_eq( acv, x );
  vm_V3Math<double>::div_eq( acv, x );
  return !fuzzy_compare(acv[0], 4.0) && 
         !fuzzy_compare(acv[1], 5.0) && 
         !fuzzy_compare(acv[2], 6.0);
}

static
int vm_V3Math_test05()    /* test {add,sub,mul,div}_eq(v,r) */
{
  double   acv[3]  = { 4.0, 5.0, 6.0 };
  vm_V3Math<double>::add_eq( acv, 4.0 );
  vm_V3Math<double>::sub_eq( acv, 4.0 );
  vm_V3Math<double>::mul_eq( acv, 4.0 );
  vm_V3Math<double>::div_eq( acv, 4.0 );
  return !fuzzy_compare(acv[0], 4.0) && 
         !fuzzy_compare(acv[1], 5.0) && 
         !fuzzy_compare(acv[2], 6.0);
}

static
int vm_V3Math_test06()    /* test add(v,cv,cv) */
{
  double  a[3] = { 2.0, 3.0,  4.0 };
  double  b[3] = { 7.0, 2.0, -3.0 };
  double  c[3];
  vm_V3Math<double>::add( c, a, b );
  return !fuzzy_compare(c[0], 9.0) && 
         !fuzzy_compare(c[1], 5.0) && 
         !fuzzy_compare(c[2], 1.0);
}

static
int vm_V3Math_test07()    /* test sub(v,cv,cv) */
{
  double  a[3] = { 2.0, 3.0,  4.0 };
  double  b[3] = { 7.0, 2.0, -3.0 };
  double  c[3];
  vm_V3Math<double>::sub( c, a, b );
  return !fuzzy_compare(c[0], -5.0) && 
         !fuzzy_compare(c[1],  1.0) && 
         !fuzzy_compare(c[2],  7.0);
}

static
int vm_V3Math_test08()    /* test mul(v,cv,cv) */
{
  double  a[3] = { 2.0, 3.0,  4.0 };
  double  b[3] = { 7.0, 2.0, -3.0 };
  double  c[3];
  vm_V3Math<double>::mul( c, a, b );
  return !fuzzy_compare(c[0],  14.0) && 
         !fuzzy_compare(c[1],   6.0) && 
         !fuzzy_compare(c[2], -12.0);
}

static
int vm_V3Math_test09()    /* test mul(v,cv,cv) */
{
  double  a[3] = { 8.0, 16.0,  4.0 };
  double  b[3] = { 4.0,  2.0, -3.0 };
  double  c[3];
  vm_V3Math<double>::div( c, a, b );
  return !fuzzy_compare(c[0],   2.0) && 
         !fuzzy_compare(c[1],   8.0) && 
         !fuzzy_compare(c[2],   4.0 / -3.0);
}

static
int vm_V3Math_test10()    /* test negate(v) */
{
  double  a[3] = { 2.0, 3.0,  4.0 };
  vm_V3Math<double>::negate( a );
  return (a[0] == -2.0) && (a[1] == -3.0) && (a[2] == -4.0);
}


static
int vm_V3Math_test11()    /* test lincomb(x,3,a,2,b) */
{
  double  a[3];
  double  b[3];
  double  x[3];
  vm_V3Math<double>::set( a, 1.0, 2.0, 3.0 );
  vm_V3Math<double>::set( b, 4.0, 5.0, 6.0 );
  vm_V3Math<double>::lincomb(x, 3.0, a, 2.0, b);
  return !fuzzy_compare(x[0],  11.0) && 
         !fuzzy_compare(x[1],  16.0) && 
         !fuzzy_compare(x[2],  21.0);
}

static
int vm_V3Math_test12()    /* test dot(a,b) */
{
  double  a[3];
  double  b[3];
  vm_V3Math<double>::set( a, 2.0, 3.0, 4.0 );
  vm_V3Math<double>::set( b, 4.0, 5.0, 6.0 );
  return !fuzzy_compare( vm_V3Math<double>::dot(a,b), 47.0 );
}

static
int vm_V3Math_test13()    /* test cross(c,a,b) */
{
  double  a[3] = { 2.0, 3.0, 4.0 };
  double  b[3] = { 4.0, 5.0, 6.0 };
  double  c[3];
  vm_V3Math<double>::cross(c,a,b);
  return !fuzzy_compare(c[0],  -2.0) &&
         !fuzzy_compare(c[1],   4.0) &&
         !fuzzy_compare(c[2],  -2.0);
}

static
int vm_V3Math_test14()    /* test is_unit_vector(a) */
{
  double  a[3];
  double  b[3];
  vm_V3Math<double>::set( a,  cos(0.1), sin(0.1), 0.0 );
  vm_V3Math<double>::set( b, -sin(0.1), cos(0.1), 0.0 );
  return vm_V3Math<double>::is_unit_vector(a, 100.0 * DBL_EPSILON);
}

static
int vm_V3Math_test15()    /* test are_orthogonal(cv,cv) */
{
  double  a[3];
  double  b[3];
  vm_V3Math<double>::set( a,  cos(0.1), sin(0.1), 0.0 );
  vm_V3Math<double>::set( b, -sin(0.1), cos(0.1), 0.0 );
  return vm_V3Math<double>::are_orthogonal(a,b, 100.0 * DBL_EPSILON);
}

static
int vm_V3Math_test16()    /* test are_orthonormal(cv,cv) */
{
  double  a[3];
  double  b[3];
  vm_V3Math<double>::set( a,  cos(0.1), sin(0.1), 0.0 );
  vm_V3Math<double>::set( b, -sin(0.1), cos(0.1), 0.0 );
  return vm_V3Math<double>::are_orthonormal(a,b, 100.0 * DBL_EPSILON);
}

static
int vm_V3Math_test17()    /* test unitize(v) */
{
  double  a[3] = { 2.0, 3.0,  4.0 };
  double ra = sqrt( a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);
  double rb = vm_V3Math<double>::unitize( a );
  return !fuzzy_compare( ra, rb ) &&
         vm_V3Math<double>::is_unit_vector( a, 100.0 * DBL_EPSILON );
}

static
int vm_V3Math_test18()    /* test unitize(v,cv) */
{
  double a[3] = { 2.0, 3.0,  4.0 };
  double b[3];
  double rv = vm_V3Math<double>::unitize( b, a );
  return !fuzzy_compare(a[0],   2.0) &&
         !fuzzy_compare(a[1],   3.0) &&
         !fuzzy_compare(a[2],   4.0) &&
         !fuzzy_compare(rv, sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])) &&
         vm_V3Math<double>::is_unit_vector( b, 100.0 * DBL_EPSILON );
}

void
test_vm_V3Math() 
{

  TEST( "copy(v,cv)",                 vm_V3Math_test01(), 1 );
  TEST( "set(v,r,r,r)",               vm_V3Math_test02(), 1 );
  TEST( "set(v,r)",                   vm_V3Math_test03(), 1 );
  TEST( "{add,sub,mul,div}_eq(v,cv)", vm_V3Math_test04(), 1 );
  TEST( "{add,sub,mul,div}_eq(v,r)",  vm_V3Math_test05(), 1 );
  TEST( "add(v,cv,cv)",               vm_V3Math_test06(), 1 );
  TEST( "sub(v,cv,cv)",               vm_V3Math_test07(), 1 );
  TEST( "mul(v,cv,cv)",               vm_V3Math_test08(), 1 );
  TEST( "div(v,cv,cv)",               vm_V3Math_test09(), 1 );
  TEST( "negate(v)",                  vm_V3Math_test10(), 1 );
  TEST( "lincomb(x,3,a,2,b)",         vm_V3Math_test11(), 1 );
  TEST( "dot(a,b)",                   vm_V3Math_test12(), 1 );
  TEST( "cross(v,a,b)",               vm_V3Math_test13(), 1 );
  TEST( "is_unit_vector(a)",          vm_V3Math_test14(), 1 );
  TEST( "are_orthogonal(a,b)",        vm_V3Math_test15(), 1 );
  TEST( "are_orthonormal(a,b)",       vm_V3Math_test16(), 1 );
  TEST( "unitize(v)",                 vm_V3Math_test17(), 1 );
  TEST( "unitize(v,cv)",              vm_V3Math_test18(), 1 );

}
